/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

  /*****************************
    Status:
    Main elements are a display label, a dropdown button, a window with a List child class DropList in it
    As for today, the list is fixed visible under the label. No window implementation yet
    All functions are implemented, but refer to the fixed list instead of the window
    Complete ItemListener support, the Choice throws an ItemEvent every time an item is selected

    TO DO:
    => move the list into the dropdown window:
      -> calculate available frame from bottom of label to bottom of MAIN SCREEN
      -> on button:
        - if window doesn't exist yet, build it out of available frame and optimal list height (dropList.getOptimalHeight() )
        - if window exists and visible, hide it again
        - if available frame changed, recalculate window to new frame, recalculate list to new window size
        - if optimal list height changed (elements added to or removed from list), recalculate window to new list size, possibly cropped by frame
        - else (if nothing changed), just make the list visible
    => replace label and button by a more user-friendly display
      -> label framed, mouse border scrolling, key scrolling if label text bigger then label width
      -> button should show a down-arrow
    => keyboard support: drop and select commands with arrows and enter-key

  */


package java.awt;

import java.util.*;
import java.awt.peer.*;
import java.awt.event.*;

public class Choice extends Panel implements ItemSelectable, ActionListener {

  /******************************/
  /**
   * Variables
   */
  //layout label and button : replace by custom elements in later fase
  private Button dropButton;
  private Label display;

  // drop-down list in window
  /** our elements list*/
  private DropList dropList;
  /** Window to display the list in*/
  private Window dropWindow;
  /** frame bounds of the window by definition
  topleft point = bottom left of the label (or this.topleft + label width)
  width = this.width
  height: if possible the height all list elements (plus list frame) so the list is displayed without scrollbars,
  if not the maximum height between topleft and bottom of complete screen*/
  private Rectangle dropFrame;
  /** the height between window topleft and bottom of screen. This is NOT the window height for the window might be smaller if it
  has to display less elements then fit the screen */
  private int freeDropHeight;

  /** listener for ItemEvents */
  private ItemListener iListener;

  /******************************/
  /**
   * Constructor
   */
  public Choice() {
    // build screen
    setLayout(new BorderLayout());
      dropButton = new Button("#");
      dropButton.addActionListener(this);
    add(dropButton, BorderLayout.EAST);
      display = new Label();
    add(display, BorderLayout.CENTER);
      dropList = new DropList();
    add(dropList, BorderLayout.SOUTH);

    // no drop window yet
    dropWindow = null;
    dropFrame = new Rectangle();
    freeDropHeight = -1;

    // no listeners yet
    iListener = null;
  }

  /****************************************************************/
  /**
  * add item listener (either the listener itself or an AWTEventMulticaster ItemListener handling it
  */
  public void addItemListener(ItemListener newlistener) {
    iListener = AWTEventMulticaster.add(iListener,newlistener);
  }

  /****************************************************************/
  /**
  * remove item listener. If the listener was in the  AWTEventMulticaster ItemListener list, return a new AWTEventMulticaster without it
  * if the listener was the only listener remaining, let AWTEventMulticaster delete it and return a null
  */
  public void removeItemListener(ItemListener oldlistener) {
    iListener = AWTEventMulticaster.remove(iListener,oldlistener);
  }
  /****************************************************************/
  /**
  * On a call to Component.processEvent, use our own listeners to handle all Item and action listeners
  * ( not that the list itself never calls this function, yet we reservr the option to receive action and Item events
  * from either the component superclass or from a custom-build derived list
  */
  protected void processEvent(AWTEvent e) {
    if(e instanceof  ItemEvent) {
      if(iListener != null) {
        iListener.itemStateChanged((ItemEvent)e);
      }
    }
    else {
      super.processEvent(e);
    }
  }


  /****************************************************************/
  /**
  * Process ItemEvent as described in Itemselectable interface
  */
  protected void processItemEvent(ItemEvent iet){
    if(iListener != null) {
      iListener.itemStateChanged(iet);
    }
  }

  /** our own special form to be called from the droplist*/
  private void processDropList(String item) {
    processItemEvent(new ItemEvent(this, ItemEvent.ITEM_STATE_CHANGED, item, ItemEvent.SELECTED));
  }

  /**
   * @status  ??  (seems not to be implemented)
   * @remark  ?? (seems not to be implemented)
   */
  public void addNotify() {
  }

  /******************************/
  /** Event manipulation:  */
  /******************************/
  /******************************/
  /**
   * On button clicked: drop down list, reeve list up again
   */
  public void actionPerformed(ActionEvent event) {
    Rectangle available = getAvailableFrame();
    int optimal = dropList.getOptimalHeight();
    if(dropWindow == null) {
      //first time initialisation: build list
      if (buildWindow() ) {
        setWindowFrame(available, optimal);
        dropWindow.setVisible(true);
      }
    }
    else if(dropWindow.isVisible()){
      // window is visible => hide it again
      dropWindow.setVisible(false);
    }
    else if(available.width != dropFrame.width || available.height != freeDropHeight) {
      // display area changed => change window just as well ){
      setWindowFrame(available, optimal);
      dropWindow.setVisible(true);
    }
    else if(dropFrame.height != optimal) {
      // optimal height changed and window is big enough to display optimal=>
      setWindowFrame(available, optimal);
      dropWindow.setVisible(true);
    }
    else if(available.x!=dropFrame.x || available.y!= dropFrame.y){
      // available window space moved=> move window with it
      setWindowOffset(available);
      dropFrame.setBounds(available);
      dropWindow.setVisible(true);
    }
    else {
      // neither window nor frame isn't changed => simply make it visible
      dropWindow.setVisible(true);
    }

  }

  /******************************/
  /**
   * auxilliary from list:
   * On list item clicked: display the latest choice
   * and close the window
   */
  private void displaySelection(String selection) {
    display.setText(selection);
    if(dropWindow!= null && dropWindow.isVisible()) // drop window MUST exist and be visible in order to display the list in the first place
      dropWindow.setVisible(false);
  }

  /******************************/
  /**
   * auxilliary for window:
   * Build the window (return true if succeeded)
   * @status :  not implemented)
   */
   public boolean buildWindow(){
     // temporary no window may be constructed => always return false
     return false;
   }
  /**
   * change window size and location
   * @status :  not implemented)
   */
   public void setWindowFrame(Rectangle frame, int optimalheight){
   }
  /**
   * change window offset
   * @status :  not implemented)
   */
   public void setWindowOffset(Rectangle frame){
   }
  /**
   * Calculate available space
   * @status :  not implemented)
   */
   public Rectangle getAvailableFrame() {
     return new Rectangle();
   }

  /******************************/
  /** List manipulation functions :
  * Forward the calls to the according
  * calls of the dropdown list
  */
  /******************************/

  /******************************/
  /** adding   */
  public synchronized void add(String item) {
    dropList.add(item);
  }
  
  public synchronized void addItem(String item) {
    dropList.add(item);
  }

  /******************************/
  /** inserting  */
  public synchronized void insert(String item, int index) {
    dropList.add(item, index);
  }

  /******************************/
  /** removing  */
  public synchronized void remove(String item) {
    dropList.remove(item);
  }

  public synchronized void remove(int index) {
    dropList.remove(index);
  }

  public synchronized void removeAll() {
    dropList.removeAll();
  }


  /******************************/
  /** selecting   */
  public synchronized void select(int index) {
    dropList.select(index);
    displaySelection((index>=0)?dropList.getSelectedItem():"");
  }

///// still needs work
  public synchronized void select(String item) {
    displaySelection(item);
  }

  /******************************/
  /** get number of items  */
  public int getItemCount() {
    return dropList.getItemCount();
  }
  
  /******************************/
  /** get item by number   */
  public String getItem(int index) {
    return dropList.getItem(index);
  }

  /******************************/
  /** return (list) selection   */
  public int getSelectedIndex() {
    return dropList.getSelectedIndex();
  }
  
  public synchronized String getSelectedItem() {
    return dropList.getSelectedItem();
  }
  
  /******************************/
  /** return all selected objects
  * There is either ONE object selected : the list's selection
  * or no selection has been made yet. So check for list selection
  * and return the selection object accordingly */
  public synchronized Object[] getSelectedObjects() {
    if(dropList.getSelectedIndex()<0) {
      // no selection has been made yet
      return null;
    }
    else {
      // return the ONE selection
      Object[] selected = new Object[1];
      selected[0]= dropList.getSelectedItem();
      return selected;
    }
  }


  /******************************/
  /******************************/
  /******************************/
  /******************************/
  /**
   * Our own subclass of List:
   * => on 'popup' show up in its own window
   * => on selection, forward selection to display and close window again
   * => use event handler of Choice
   * => ALWAYS single selection
   */
   private class DropList extends List {
     /** special droplist variables */

     /** droplist constructor */
     public DropList() {
       super(4,false);
     }

    /**************************************************************/
    /**
    * Listeners
    * we don't use listeners in the inner list,
    * but forward the calls to the Choice listener
    */
    /**************************************************************/
    public void addItemListener(ItemListener newlistener) {
    }

    public void removeItemListener(ItemListener oldlistener) {
    }

    protected void processItemEvent(ItemEvent iet){
    }


    public void addActionListener(ActionListener newlistener) {
    }

    public void removeActionListener(ActionListener oldlistener) {
    }

    /** special case: when clicked an item the List throws an actionEvent.
    * Use this to close the List window, update the display and make Choice throw an item event*/
    protected void processActionEvent(ActionEvent aet) {
      processDropList(aet.getActionCommand());
      displaySelection(aet.getActionCommand());
    }

    /**************************************************************/
    /**
    * Special DropList function getOptimalHeight:
    * calculate the height needed to display the whole list without scrollbars
    */
    public int getOptimalHeight() {
      return RudolphListPeer.getHeight(this.getItemCount(), painterMetrics);
    }

    /**************************************************************/
    /**
    * Calculate list for new screen size
    */
    public void setViewportHeight(int w, int h) {
      screenViewport.setSize(w,h);
      calculateScrollbarSettings();
    }

    public int getViewportwidth() {
      return screenViewport.width;
    }

    public int getViewportHeight() {
      return screenViewport.height;
    }

  // end of inner class DropList
  }

  /******************************/
  /**
   * debug
   */
  protected String paramString() {
    return "java.awt.Choice ["+dropList.paramString();
  }



}
