/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

package java.awt;

import java.util.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.Color;

public abstract class Component implements java.awt.image.ImageObserver, MenuContainer, java.io.Serializable {

  // Tree lock:
  static private final Object lock;
  
  static {
    lock = new Object();
    initRudolph();
  }

  // Refresh types:
  static final int REFRESH_LOCAL = 1;
  static final int REFRESH_GLOBAL = 2;

  // Cache types:
  static final int CACHE_NONE = 0;
  static final int CACHE_FONT = 1;
  static final int CACHE_FGCOLOR = 2;
  static final int CACHE_BGCOLOR = 3;
  static final int CACHE_VISIBILITY = 4;

  // Default font on which we can fall-back:
  public static final Font DEFAULT_FONT = new Font("helvP08", Font.PLAIN, 8);

  // (text) alignment definitions
  public static final float BOTTOM_ALIGNMENT = 1.0f;
  public static final float CENTER_ALIGNMENT = 0.5f;
  public static final float LEFT_ALIGNMENT   = 0.0f;
  public static final float RIGHT_ALIGNMENT  = 1.0f;
  public static final float TOP_ALIGNMENT    = 0.0f;

  // Component dimensions:
  int x = 0;
  int y = 0;
  int height = 0;
  int width = 0;

  // Sizes:
  Dimension preferredSize;
  Dimension minimumSize;
  Dimension maximumSize;

  // Name:
  String name;

  // Visibility:
  boolean visible = true;

  // Validation:
  boolean valid = false;
  boolean validate = false;

  // Enabled:
  boolean enabled = true;

  // Notification:
  protected boolean notified = false;

  //Locale:
  private Locale locale;

  // Parent container:
  Container parent = null;

  // Background and foreground color:
  Color background = null;
  Color foreground = null;

  // Events:
  ComponentListener componentListener;
  MouseMotionListener mouseMotionListener;
  MouseListener mouseListener;
  FocusListener focusListener;
  ActionListener actionListener;
  KeyListener keyListener;

  // Font:
  Font font = null;
  static boolean renderer = false;

  /*
  ** Focus :
  **
  ** These fields are updated from Event_addFocusEvent (Event.c)
  ** and are used by the taskbar to revert the focus to the 
  ** original Component that had focus..
  ** The next 2 methods are used to revert focus and to get 
  ** the Component which currently has focus.
  */
  
  private static Component focusComponent = null;
  private static Component focusComponentPrev = null;

  public static void revertFocus() {
    focusComponent = focusComponentPrev;
    if(focusComponent != null) {
      focusComponent.dispatchEvent(new FocusEvent(focusComponent, FocusEvent.FOCUS_GAINED));
    }
  }

  public static Component getFocusComponent() {
    return focusComponent;
  }

  

  static protected void initRudolph() {
    if (!System.getProperty("com.acunia.wonka.awt", "true").equals("false")) {
      Component.awt(1);
    }
  }

  native private static synchronized void awt(int status);

  /****************************************************************/
  /**  Constructor / destructor                                   */
  /****************************************************************/

  protected Component() {
    createPeer();    
  }

  native private void createPeer();

  native protected void finalize();

  /****************************************************************/
  /**  Name                                                       */
  /****************************************************************/
  public void setName(String name) {
    this.name = name;
  }

  public String getName() {
    return (name != null) ? name : "";
  }


  /****************************************************************/
  /**  Colors and fonts                                           */
  /****************************************************************/
  
  public void setForeground(Color color) {
    synchronized(getTreeLock()) {
      this.foreground = color;
      if (valid) invalidate();
      refresh(Component.REFRESH_LOCAL, Component.CACHE_FGCOLOR);
    }
  }

  public Color getForeground() {
    if (this.foreground != null) return this.foreground;
    return (parent != null) ? parent.getForeground() : null;
  }

  /****************************************************************/
  /**  Background color
  */

  public void setBackground(Color color) {
    background = color;
    if (valid) {
      invalidate();
    }
    refresh(Component.REFRESH_LOCAL, Component.CACHE_BGCOLOR);
  }

  public Color getBackground() {
    if (this.background != null) return this.background;
    return (parent != null) ? parent.getBackground() : null;
  }

  /****************************************************************/
  /**  Color model
  */

  native public java.awt.image.ColorModel getColorModel();

  /****************************************************************/
  /**  Font
  */

  public synchronized void setFont(Font font) {
    synchronized(getTreeLock()) {
      this.font = font;
      if (valid) invalidate();
      refresh(Component.REFRESH_LOCAL, Component.CACHE_FONT);
    }
  }

  public Font getFont() {
    if (this.font != null) {
      return this.font;
    }
    else if (parent != null) {
      return parent.getFont();
    }
    else {
      return null;
    }
  }

  /****************************************************************/
  /**  Font metrics
  */

  public FontMetrics getFontMetrics(Font font) {
    return new FontMetrics((font == null) ? Component.DEFAULT_FONT : font);
  }

  /**
   * @status dummy implementation, avoids a linkage error
   * @remark not compliant with specifications
   */
  public synchronized void setCursor(Cursor cursor)
  {
    com.acunia.wonka.Etc.woempa(9, "UNIMPLEMENTED: java.awt.Component.setCursor");
  }

  native public Cursor getCursor();


  /****************************************************************/
  /**   Boundaries                                                */
  /****************************************************************/

  public void setBounds(int x, int y, int width, int height) {
   
    boolean l = (this.x != x || this.y != y) ? true : false;
    boolean s = (this.width != width || this.height != height) ? true : false;

    if(componentListener != null) {
      dispatchEvent(new ComponentEvent(this, ComponentEvent.COMPONENT_RESIZED));
      if(l) dispatchEvent(new ComponentEvent(this, ComponentEvent.COMPONENT_RESIZED));
    }
    
    if (s || l) {

      synchronized (getTreeLock()) {

        this.x = x;
        this.y = y;
        this.width = width;
        this.height = height;

        // Invalidate component:
        if (valid) {
          invalidate();
        }

        if (s && this instanceof Container) {
          ((Container)this).scale(width, height);
        }

        // Refresh();
        refresh(Component.REFRESH_GLOBAL, Component.CACHE_NONE);

        // TEMPORARY DEBUG OUTPUT:
        // System.out.println("[DB] * component resized:");
        // System.out.println("[DB]     - component : " + this);
        // System.out.println("[DB]     - new size  : x = "+ x +", y = "+ y +", w = "+ width +", h = "+ height);
      }
    }
  }

  public void setBounds(Rectangle rectangle) {
    setBounds(rectangle.x, rectangle.y, rectangle.width, rectangle.height);
  }

  public Rectangle getBounds() {
    return new Rectangle(x, y, width, height);
  }

  public void setSize(int w, int h) {
    setBounds(this.x, this.y, w, h);
  }

  public void setSize(Dimension dimension) {
    setBounds(this.x, this.y, dimension.width, dimension.height);
  }

  public Dimension getSize() {
    return new Dimension(width, height);
  }

  /*
  ** Deprecated...
  */

  public Dimension size() {
    return new Dimension(width, height);
  }
  
  public int getX() {
    return x;
  }

  public int getY() {
    return y;
  }

  public int getHeight() {
    return height;
  }

  public int getWidth() {
    return width;
  }

  /****************************************************************/
  /**  Position inside component?
  */

  native public boolean contains(int x, int y);

  native public boolean contains(Point point);

  /****************************************************************/
  /**  Dimensions                                                 */
  /****************************************************************/

  public Dimension getMinimumSize() {
    return minimumSize();
  }

  public Dimension getPreferredSize() {
    return preferredSize(); 
  }

  public Dimension getMaximumSize() {
    return maximumSize();
  }

  public Dimension minimumSize() {
    synchronized (getTreeLock()) {
      if (minimumSize == null || valid == false) {
        minimumSize = new Dimension(width, height);
      }
      return minimumSize;
     }
  }

  public Dimension preferredSize() {
    synchronized (getTreeLock()) {
      if (minimumSize == null || valid == false) {
        minimumSize = new Dimension(width, height);
      }
      return minimumSize;
    }
  }

  public Dimension maximumSize() {
    return new Dimension(Short.MAX_VALUE, Short.MAX_VALUE);
  }

  /****************************************************************/
  /**  Location relative to parent                                */
  /****************************************************************/

  public void setLocation(int x, int y) {
    setBounds(x, y, this.width, this.height);
  }

  public void setLocation(Point point) {
    setBounds(point.x, point.y, this.width, this.height);
  }

  public Point getLocation() {
    return new Point(x, y);
  }

  private native int getAbsX();
  private native int getAbsY();

  public Point getLocationOnScreen() {
    return new Point(getAbsX(), getAbsY());
  }

  /****************************************************************/
  /**  Child component at a given position
  */
  
  native public Component getComponentAt(int x, int y);

  native public Component getComponentAt(Point point);

  public void list(java.io.PrintStream out, int indent) {
  }

  public void list(java.io.PrintWriter out, int indent) {
  }

  public void list() {
    list(System.out,0);
  }

  public void list(java.io.PrintStream out) {
    list(out,0);
  }

  public void list(java.io.PrintWriter out) {
    list(out,0);
  }

  /****************************************************************/
  /**  Parent
  */
  
  public Container getParent() {
    return parent;
  }

  /****************************************************************/
  /**  Layout of components
  */
  
  public void doLayout() {
  }

  public float getAlignmentX() {
    return CENTER_ALIGNMENT;
  }

  public float getAlignmentY() {
    return CENTER_ALIGNMENT;
  }

  public void addNotify() {
    if (notified == false) {
      // createPeer();
      notified = true;
    }
  }

  public void removeNotify() {
  }

  /****************************************************************/
  /**  Popup menu                                                 */
  /****************************************************************/

  native public synchronized void add(PopupMenu popup);

  native public synchronized void remove(MenuComponent popup);

  /****************************************************************/
  /**  Tree lock                                                  */
  /****************************************************************/

  /**
   * @status implemented
   * @remark not compliant with specifications
   */
  public final Object getTreeLock() {
    return lock;
  }

  /****************************************************************/
  /**  Locale                                                     */
  /****************************************************************/

  public void setLocale(java.util.Locale locale){
    this.locale = locale;
  }

  public java.util.Locale getLocale() throws IllegalComponentStateException{
    if(locale != null){
      return locale;
    }
    if(parent != null){
      return parent.getLocale();
    }
    throw new IllegalComponentStateException("no locale set for this component or its parents");
  }

  /****************************************************************/
  /**  Traversable focus                                          */
  /****************************************************************/

  public void requestFocus() {
    /* Do nothing */
  }

  public void transferFocus() {
    /* Do nothing */
  }
  
  /**
   * @status not implemented
   * @remark not compliant with specifications
   */
  public boolean isFocusTraversable() {
    return true;
  }

  /****************************************************************/
  /**  Validation                                                 */
  /****************************************************************/

  public void validate() {
    valid = true;
  }

  public void invalidate() {
    if (valid) {
      synchronized(getTreeLock()) {
        minimumSize = null;
        preferredSize = null;
        valid = false;

        // if parent exists, invalidate parent component
        if ((parent != null) && (parent.valid)) {
          parent.invalidate();
        }
      }
    }
  }

  public boolean isValid() {
    return valid;
  }


  /****************************************************************/
  /**  Enable / disable component and listeners                   */
  /****************************************************************/

  public void setEnabled(boolean condition) {
    if (background != null) {
      // set background bright when selected, dark when not
      if (enabled == false && condition == true) {
        // background switched to enabled => background brighter
        setBackground(background.brighter());
      }
      else if (enabled == true && condition == false) {
        // background switched to disabled => background darker
        setBackground(background.darker());
      }
    }
    // in any case.. set the new condition
    enabled = condition;
  }

  public boolean isEnabled() {
    return enabled;
  }


  /****************************************************************/
  /**  Listeners                                                  */
  /****************************************************************/

  public synchronized void addComponentListener(ComponentListener listener) {
    componentListener = AWTEventMulticaster.add(componentListener, listener);
  }

  public synchronized void removeComponentListener(ComponentListener listener) {
    componentListener = AWTEventMulticaster.remove(componentListener, listener);
  }


  public synchronized void addFocusListener(java.awt.event.FocusListener listener) {
    focusListener = AWTEventMulticaster.add(focusListener, listener);
  }

  public synchronized void removeFocusListener(java.awt.event.FocusListener listener) {
    focusListener = AWTEventMulticaster.remove(focusListener, listener);
  }

  public synchronized void addKeyListener(java.awt.event.KeyListener listener) {
    keyListener = AWTEventMulticaster.add(keyListener, listener);
  }

  public synchronized void removeKeyListener(java.awt.event.KeyListener listener) {
    keyListener = AWTEventMulticaster.remove(keyListener, listener);
  }

  public synchronized void addMouseListener(MouseListener listener) {
    mouseListener = AWTEventMulticaster.add(mouseListener, listener);
  }


  public synchronized void removeMouseListener(MouseListener listener) {
    mouseListener = AWTEventMulticaster.remove(mouseListener, listener);
  }

  public synchronized void addMouseMotionListener(MouseMotionListener listener) {
    mouseMotionListener = AWTEventMulticaster.add(mouseMotionListener, listener);
  }

  public synchronized void removeMouseMotionListener(java.awt.event.MouseMotionListener listener) {
    mouseMotionListener = AWTEventMulticaster.remove(mouseMotionListener, listener);
  }

  /****************************************************************/
  /**  Listener events                                            */
  /****************************************************************/

  /**
   * @status dummy implementation
   * @remark not compliant with specifications: just prints a message
   */
  protected final void enableEvents(long eventTypes){}

  native protected final void disableEvents(long eventTypes);

  protected void dispatchEventImpl(AWTEvent event) {
    if (enabled == false) {
      // If a component is not enabled it can not generate events
      // nor respond to events (or user input for that matter).
      return;
    }
    
    if (event instanceof MouseEvent) {
      switch(event.getID()) {
        case MouseEvent.MOUSE_ENTERED:
        case MouseEvent.MOUSE_EXITED:
        case MouseEvent.MOUSE_PRESSED:
        case MouseEvent.MOUSE_RELEASED:
        case MouseEvent.MOUSE_CLICKED:
          processMouseEvent((MouseEvent) event);
          break;
        case MouseEvent.MOUSE_MOVED:
        case MouseEvent.MOUSE_DRAGGED:
          processMouseMotionEvent((MouseEvent) event);
          break;
      }
    }
    else if (event instanceof KeyEvent) {
      processKeyEvent((KeyEvent) event);
    }
    else if (event instanceof FocusEvent) {
      processFocusEvent((FocusEvent) event);
    }
    else if (event instanceof ItemEvent) {
      Object source = event.getSource();
      if (source instanceof Checkbox) {
        ((Checkbox) source).processEvent((ItemEvent) event);
      }
    }
    else if (event instanceof ComponentEvent) {
      processComponentEvent((ComponentEvent) event);
    }
    else if (event instanceof ActionEvent) {
      Object source = event.getSource();
      if (source instanceof Button) {
        ((Button) source).processEvent((ActionEvent) event);
      }
    }
  }

  public final void dispatchEvent(AWTEvent event) {
    dispatchEventImpl(event);
  }

  protected void processEvent(AWTEvent event) {
    if (event instanceof FocusEvent) {
      processFocusEvent((FocusEvent) event);
    }
    else if (event instanceof MouseEvent) {
      switch(event.getID()) {
        case MouseEvent.MOUSE_ENTERED:
        case MouseEvent.MOUSE_EXITED:
        case MouseEvent.MOUSE_PRESSED:
        case MouseEvent.MOUSE_RELEASED:
        case MouseEvent.MOUSE_CLICKED:
          processMouseEvent((MouseEvent) event);
          break;
        case MouseEvent.MOUSE_MOVED:
        case MouseEvent.MOUSE_DRAGGED:
          processMouseMotionEvent((MouseEvent) event);
          break;
      }
    }
    else if (event instanceof KeyEvent) {
      processKeyEvent((KeyEvent) event);
    }
    else if (event instanceof ComponentEvent) {
      processComponentEvent((ComponentEvent) event);
    }
  }

  /****************************************************************/
  /**  Listener events type by type:                              */
  /****************************************************************/

  /**  Component event:  */
  protected void processComponentEvent(ComponentEvent event) {
    if (componentListener != null) {
      switch(event.getID()) {
        case ComponentEvent.COMPONENT_RESIZED:
          componentListener.componentResized(event);
          break;
        case ComponentEvent.COMPONENT_MOVED:
          componentListener.componentMoved(event);
          break;
        case ComponentEvent.COMPONENT_SHOWN:
          componentListener.componentShown(event);
          break;
        case ComponentEvent.COMPONENT_HIDDEN:
          componentListener.componentHidden(event);
          break;
      }
    }
  }

  /**  focus event:  */
  protected void processFocusEvent(FocusEvent event) {
    if (focusListener != null) {
      switch(event.getID()) {
      case FocusEvent.FOCUS_GAINED:
        focusListener.focusGained(event);
        break;
      case FocusEvent.FOCUS_LOST:
        focusListener.focusLost(event);
        break;
      }
    }
  }

  /**  key event:  */
  protected void processKeyEvent(java.awt.event.KeyEvent event) {
    if (keyListener != null) {
      switch(event.getID()) {
        case KeyEvent.KEY_TYPED:
          keyListener.keyTyped(event);
          break;
        case KeyEvent.KEY_PRESSED:
          keyListener.keyPressed(event);
          break;
        case KeyEvent.KEY_RELEASED:
          keyListener.keyReleased(event);
          break;
      }
    }
  }

  /**  mouse event:  */
  protected void processMouseEvent(MouseEvent event) {
    if (mouseListener != null) {
      switch(event.getID()) {
        case MouseEvent.MOUSE_EXITED:
          mouseListener.mouseExited(event);
          break;
        case MouseEvent.MOUSE_ENTERED:
          mouseListener.mouseEntered(event);
          break;
        case MouseEvent.MOUSE_PRESSED:
          mouseListener.mousePressed(event);
          break;
        case MouseEvent.MOUSE_RELEASED:
          mouseListener.mouseReleased(event);
          break;
        case MouseEvent.MOUSE_CLICKED:
          mouseListener.mouseClicked(event);
          break;
      }
    }
  }
  
  /**  mouse motion event:  */
  protected void processMouseMotionEvent(java.awt.event.MouseEvent event) {
    if (mouseMotionListener != null) {
      switch(event.getID()) {
        case MouseEvent.MOUSE_MOVED:
          mouseMotionListener.mouseMoved(event);
          break;
        case MouseEvent.MOUSE_DRAGGED:
          mouseMotionListener.mouseDragged(event);
          break;
      }
    }
  }
  
  /****************************************************************/
  /**  Printing functionality                                     */
  /****************************************************************/

  /**
   * @status not implemented
   * @remark not compliant with specifications
   */
  native public void print(Graphics context);

  /**
   * @status not implemented
   * @remark not compliant with specifications
   */
  native public void printAll(Graphics context);

  /****************************************************************/
  /**  Diagnostics                                                */
  /****************************************************************/

  public String toString() {
    String display = (name == null)? getClass().getName() : name;
    return display +" - bounds: x = "+ x +", y = "+ y +", w = "+ width +", h = "+ height;
  }

  /**
   * @status not implemented
   * @remark not compliant with specifications
   */
  protected String paramString() {
    return getClass().getName() +"<"+name+"> - bounds("+ x +", "+ y +", "+ width +", "+ height+")";
  }

  /****************************************************************/
  /**  Visibility                                                 */
  /****************************************************************/

  public boolean isVisible() {
    return visible;
  }

  public void setVisible(boolean condition) {
    synchronized(getTreeLock()) {
      if (condition) {
        // Show component:
        if (!visible) {
          visible = true;

          // Component listener:
          if (componentListener != null) {
            System.out.println("[DB] TODO: java.awt.Component: throw ContainerEvent.COMPONENT_SHOWN");
          }

          // Invalidate parent container (if any):
          if (this instanceof Container) {
            invalidate();
            refresh(Component.REFRESH_GLOBAL, Component.CACHE_VISIBILITY);
          }
          else if (parent != null) {
            parent.invalidate();
            parent.refresh(Component.REFRESH_GLOBAL, Component.CACHE_VISIBILITY);
          }

        }
      }
      else {
        // Hide component:
        if (visible) {
          visible = false;

          // Component listener:
          if (componentListener != null) {
            System.out.println("[DB] TODO: java.awt.Component: throw ContainerEvent.COMPONENT_HIDDEN");
          }

          // Invalidate parent container (if any):
          if (this instanceof Container) {
            invalidate();
            refresh(Component.REFRESH_GLOBAL, Component.CACHE_VISIBILITY);
          }
          else if (parent != null) {
            parent.invalidate();
            parent.refresh(Component.REFRESH_GLOBAL, Component.CACHE_VISIBILITY);
          }
        }
      }
    }
  }

  public void show() {
    setVisible(true);
  }

  public boolean isDisplayable() {
    return true;
  }

  /**
   * @status not implemented
   * @remark not compliant with specifications
   */
  public boolean isShowing() {
    return isVisible();
  }

  /****************************************************************/
  /**  Graphics                                                   */
  /****************************************************************/

  /**
   * @status implemented
   * @remark not compliant with specifications: a Graphics object is 
   *         always constructed and returned, even if the component 
   *         is not visible or its peer component does not exist.
   */
  native public Graphics getGraphics();

  public Toolkit getToolkit() {
    return Toolkit.getDefaultToolkit();
  }

  /**creates an Image of desired size, the image is directly available*/
  native public Image createImage(int w, int h);

  public Image createImage(java.awt.image.ImageProducer producer) {
    return (java.awt.Image)(new com.acunia.wonka.rudolph.Image(producer));
  }

  /**If the desired image is not available yet, starts building it*/
  /**
   * @status implemented
   * @remark implemented
   */
  public boolean prepareImage(Image image, ImageObserver observer) {
    return prepareImage(image, -1, -1, observer);
  }

  /**
   * @status implemented
   * @remark implemented
   */
  public boolean prepareImage(Image image, int w, int h, ImageObserver observer) {
    return Toolkit.getDefaultToolkit().prepareImage(image, w, h, observer);
  }

  /**returns the status of the image as a series of ImageObserver flags*/
  /**
   * @status implemented
   * @remark implemented
   */
  public int checkImage(Image image, ImageObserver observer) {
    return checkImage(image, -1, -1, observer);
  }

  /**
   * @status implemented
   * @remark implemented 
   */
  public int checkImage(Image image, int w, int h, ImageObserver observer) {
    return Toolkit.getDefaultToolkit().checkImage(image, w, h, observer);
  }


  /**
   * Called by Image.getWidth(ImageObserver) and getHeight(ImageObserver). Graphics.drawImage(..... ImageObserver)
   * and Component prepareImage(ImageObserver) / CheckImage(ImageObserver) to send diagnostic data to the
   * <this>-ImageObserver-interface Component.
   * designed to be overridden by subclasses to get the diagnostics
   * returns true = <yes, there are still more of these messages to come> unless the flags contain ImageObserver.ALLBITS
   */
  public boolean imageUpdate(Image image, int flags, int x, int y, int w, int h) {
    return((flags & ImageObserver.ALLBITS) == 0);
  }


  public void repaint() {
    repaint(0, 0, 0, width, height);
  }

  public void repaint(long ms) {
    repaint(ms, 0, 0, width, height);
  }

  public void repaint(int x, int y, int w, int h) {
    repaint(0, x, y, w, h);
  }

  public void repaint(long ms, int x, int y, int w, int h) {
    Graphics g = getGraphics();
    if (g != null) {
      update(g);
    }
  }

  public void paint(Graphics context) {
  }

  public void update(Graphics g) {
    if (valid == true) {
      refresh(REFRESH_LOCAL, Component.CACHE_NONE);
    }
  }

  public void paintAll(Graphics context) {
    // Validate container:
    validate();
  }

  /****************************************************************/
  /** Protected screen-refresh / update
  */

  protected void refresh(int type, int cache) {
    /* Refresh component if required: */
    if ((parent != null) && (parent.validate == true)) {
      // prune - if the parent container is invalid, we are likely to
      //         be in the middle of updating/validating an entire
      //         container in which case we postpone to update the
      //         individual components until the parent container is
      //         entirely validated/layouted.
      synchronized(getTreeLock()) {
        tag(type, cache, false);
      }
    }
    else if ((parent == null) && !(this instanceof Window)) {
      // prune - ignore refresh as the component is still floating
      //         in the void (not attached to a parent container).
      synchronized(getTreeLock()) {
        tag(type, cache, false);
      }
    }
    else if (this.isVisible()) {
      /* Validate the current component: */
      if (valid == false) {
        this.validate();
      }

      /* Re-render the off-screen buffer: */
      if (renderer == false) {
        renderer = true;
        synchronized(getTreeLock()) {
          tag(type, cache, true);
        }
        renderer = false;
      }
      else {
        synchronized(getTreeLock()) {
          tag(type, cache, false);
        }
      }
    }
  }

  native private void tag(int status, int cache, boolean render);
}
