/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

package java.awt;

import java.awt.event.*;
import java.awt.peer.*;


public class ScrollPane extends Container implements AdjustmentListener {

  public final static int SCROLLBARS_AS_NEEDED = 0;
  public final static int SCROLLBARS_ALWAYS = 1;
  public final static int SCROLLBARS_NEVER = 2;

  private Scrollbar hsb;
  private Scrollbar vsb;

  private int vx;
  private int vy;
  private int vwidth;
  private int vheight;

  private int scrollbarDisplayPolicy;

  private Component comp = null;
  private Panel viewport = new Panel(null);
  private Panel componentpanel;

  private native void setViewport(Component viewport, int vx, int vy, int vwidth, int vheight);

  private void changeViewport(Component viewport, int vx, int vy, int vwidth, int vheight) {
    if(vx + vwidth > viewport.getWidth()) vx = viewport.getWidth() - vwidth;
    if(vy + vheight > viewport.getHeight()) vy = viewport.getHeight() - vheight;
  
    setViewport(viewport, vx, vy, vwidth, vheight);
  }

  public ScrollPane() {
    this(SCROLLBARS_AS_NEEDED);
  }
	
  public ScrollPane(int policy) {
    if (policy < 0 || policy > 2) {
      throw new IllegalArgumentException("invalid policy specified : " + policy);
    } 
  
    hsb = new Scrollbar(Scrollbar.HORIZONTAL, 0, 5, 0, 5);
    vsb = new Scrollbar(Scrollbar.VERTICAL, 0, 5, 0, 5);
    
    hsb.setVisible(false);
    vsb.setVisible(false);
    
    hsb.addAdjustmentListener(this);
    vsb.addAdjustmentListener(this);
    
    scrollbarDisplayPolicy = policy;

    setLayout(new ScrollPaneLayout());

    add(hsb, ScrollPaneLayout.HSB);
    add(vsb, ScrollPaneLayout.VSB);
    add(viewport, ScrollPaneLayout.VIEW);

  }

  protected final void addImpl(Component component, Object constraints, int position) {

    /*
    ** A ScrollPane can only have one child component so we remove
    ** existing child components, if any.
    */
   
    if(component != hsb && component != vsb && component != viewport) {
      
      if(comp != null) {
        if(comp instanceof Container) {
          viewport.remove(comp);
        } else {
          viewport.remove(componentpanel);
        }
      }

      comp = component;
      
      if(component instanceof Container) {
        viewport.add(comp);
      } else {
        componentpanel = new Panel();
        componentpanel.add(comp);
        viewport.add(componentpanel);
      }

      doLayout();
      
    } else {
      super.addImpl(component, constraints, position);
    }
  }

  public void doLayout() {
    super.doLayout();

    /*
    ** Get the size of the scrollpane
    */
    
    vwidth = getSize().width;
    vheight = getSize().height;

//    System.out.println(" -+ (1)  " + getSize().width + ", " + getSize().height);
   
    if(comp != null) {
      
      Dimension ps = comp.getPreferredSize();
    
      /*
      ** Decide wether to show the scrollbars.
      */

      switch(scrollbarDisplayPolicy) {
      
        case SCROLLBARS_NEVER:   
          break;

        case SCROLLBARS_ALWAYS: 
          hsb.setVisible(true);
          vheight -= getHScrollbarHeight();
          vsb.setVisible(true);
          vwidth -= getVScrollbarWidth();
          break;

        case SCROLLBARS_AS_NEEDED:
          if(ps.width > vwidth) {
            hsb.setVisible(true); 
            vheight -= getHScrollbarHeight();
          } 
          else {
            hsb.setVisible(false);
          }
          if(ps.getSize().height > vheight) {
            vsb.setVisible(true); 
            vwidth -= getVScrollbarWidth();
          } 
          else {
            vsb.setVisible(false);
          }
          break;
      }

      /*
      ** Calculate the scrollbar properties.
      */

      hsb.setMaximum(Math.max(0, ps.width - vwidth) + vwidth);
      vsb.setMaximum(Math.max(0, ps.height - vheight) + vheight);
      hsb.setVisibleAmount(vwidth);
      vsb.setVisibleAmount(vheight);

      /*
      ** Make the component grow larger if it's not as wide or
      ** not as high as the viewport
      */

      if(ps.width < vwidth) ps.width = vwidth;
      if(ps.height < vheight) ps.height = vheight;

      /*
      ** Change the size of the component (or the panel which
      ** contains the component)
      */
      
      synchronized(getTreeLock()) {
        viewport.setBounds(0, 0, vwidth, vheight);
        
        if(comp instanceof Container) {
          comp.setBounds(0, 0, ps.width, ps.height);
          changeViewport(comp, vx, vy, vwidth, vheight);
        } else {
          componentpanel.setBounds(0, 0, ps.width, ps.height);
          comp.setBounds(0, 0, ps.width, ps.height);
          changeViewport(componentpanel, vx, vy, vwidth, vheight);
        }
      }

      /*
      ** Refresh.
      */

      comp.refresh(Component.REFRESH_GLOBAL, CACHE_NONE);  
      
    }
  }

  public int getScrollbarDisplayPolicy() {
    return scrollbarDisplayPolicy;
  }

  public Adjustable getHAdjustable() {
    return hsb;
  }

  public Adjustable getVAdjustable() {
    return vsb;
  }

  public int getHScrollbarHeight() {
    return hsb.getHeight();
  }

  public Point getScrollPosition() {
    return new Point(hsb.getValue(), vsb.getValue());
  }

  public int getVScrollbarWidth() {
    return vsb.getWidth();
  }
 
  public Dimension getViewportSize() {
    return new Dimension(vwidth, vheight);
  }

  public void setScrollPosition(Point p) {
    if(p.x > comp.getSize().width - vwidth) {
      p.x = Math.max(0, comp.getSize().width - vwidth);
    }
    if(p.y > comp.getSize().height - vheight) {
      p.y = Math.max(0, comp.getSize().height - vheight);
    }
    
    setScrollPosition(p.x, p.y);
  }
 
  public void setScrollPosition(int x, int y) {
    hsb.setValue(x);
    vsb.setValue(y);

    vx = x;
    vy = y;

    if(comp instanceof Container) {
      changeViewport(comp, vx, vy, vwidth, vheight);
    } else {
      changeViewport(componentpanel, vx, vy, vwidth, vheight);
    }

    comp.refresh(Component.REFRESH_GLOBAL, CACHE_NONE);
  }


  public void adjustmentValueChanged(AdjustmentEvent event) {
    vx = hsb.getValue();
    vy = vsb.getValue();
    
    if(comp instanceof Container) {
      changeViewport(comp, vx, vy, vwidth, vheight);
    } else {
      changeViewport(componentpanel, vx, vy, vwidth, vheight);
    }

    comp.refresh(Component.REFRESH_GLOBAL, CACHE_NONE);
  }

  public String paramString() {
    return "";
  }

  public void printComponents(Graphics g) {
  }

  public Dimension getMinimumSize() {
    return new Dimension(0, 0);
  }
  
  public Dimension getMaximumSize() {
    return getPreferredSize();
  }
  
  public Dimension getPreferredSize() {
    return new Dimension(comp.getPreferredSize().width, comp.getPreferredSize().height);
  }
  
  public void setSize(int w, int h) {
    super.setSize(w, h);
    doLayout();
  }

  public void setSize(Dimension dimension) {
    super.setSize(dimension);
    doLayout();
  }

  public void setBounds(int x, int y, int width, int height) {
    super.setBounds(x, y, width, height);
    doLayout();
  }

  public void setBounds(Rectangle rectangle) {
    super.setBounds(rectangle);
    doLayout();
  }

  public Component getComponent(int index) {
    if(comp != null && index == 0) {
      return comp;
    } else {
      throw new ArrayIndexOutOfBoundsException();
    }
  }
  
  public int getComponentCount() {
    if(comp != null) {
      return 1;
    } else {
      return 0;
    }
  }
  
  public Component[] getComponents() {
    Component[] list = new Component[0];
    if(comp != null) {
      list = new Component[1];
      list[0] = comp;
    }
    return list;
  }

}
