/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: canvas.c,v 1.37 2002/03/22 12:56:43 dumon Exp $ 
*/

#include <string.h>

#include "interpreter.h"  // runThread()
#include "hashtable.h"    // ht_read
#include "strings.h"      // cstring2String()
#include "methods.h"	  // w_string_equals
#include "realms.h"
#include "threads.h"       // currentWonkaThread
#include "ts-mem.h"       // allocMem(), releaseMem()
#include "stack.h"        // allocBottomFrame(), releaseBottomFrame()
#include "clazz.h"        // w_clazz

#include "awt-classes.h"  // F_Component_*
#include "canvas.h"       // includes "rudolph.h" 
#include "platform.h"

#include "Event.h"
#include "Button.h"
#include "Checkbox.h"
#include "Component.h"
#include "Graphics.h"
#include "Label.h"
#include "TextField.h"

r_canvas rootCanvas;
int native_paint;

r_canvas canvas_constructor(w_instance canvasInstance) {

  w_thread thread = currentWonkaThread;
  r_buffer buffer;
  r_canvas canvas;
     
  /*
  ** Initialize the off-screen buffer.  We use lazy allocation or
  ** JIT buffer allocation.  That is, we don't allocate memomy for
  ** the buffer now and from start, we mark it as being released.
  ** With Java's layout managers in mind, the container is likely
  ** going to be rescaled anyway.
  */
 
  buffer = allocClearedMem(thread, sizeof(r_Buffer));
    
  /*
  ** Allocate memory for the canvas:
  */
  
  canvas = allocMem(thread, sizeof(r_Canvas)); 
  
  /*
  ** Initialize the canvas:
  */
  
  canvas->buffer = buffer;
  
  /*
  ** Initialize the canvas specific parts of the component:
  */
  
  canvas->component = (r_component)canvasInstance[F_Component_wotsit];
  canvas->component->tag = Z_CONTAINER;
  canvas->component->object = (unsigned long)canvas;
  canvas->component->refresh = 2;
               
  return canvas;

}

inline char *canvas_spaces(const int depth) {

  const char *spaces = "                                	  ";  // 40 spaces
  return (char *)(spaces + strlen(spaces) - (2 * depth));

}

static inline void canvas_blitter(w_word *dst, w_word *src, w_size size) {

  /*
  ** Dump screen data from source to destination.  Don't use 'memcpy';
  ** it copies only one byte at a time.  A loop copying one w_word at 
  ** a time is recommended ...
  **
  **   while (size) {
  **     *dst++ = *src++;
  **     size -= 4;
  **   }
  **
  ** ... though a faster implementation (proposed by "Duffs"), with 
  ** less iterations, exists: it copies 8 words per iteration.
  */

  w_int num_words = size / sizeof(w_word);
  w_int duffs = (num_words + 7) / 8;

  switch (num_words % 8) {
    case 0: do { *dst++ = *src++;
    case 7:      *dst++ = *src++;
    case 6:      *dst++ = *src++;
    case 5:      *dst++ = *src++;
    case 4:      *dst++ = *src++;
    case 3:      *dst++ = *src++;
    case 2:      *dst++ = *src++;
    case 1:      *dst++ = *src++;
            } while (--duffs > 0);
  }
}

static inline w_word *canvas_transpose(w_ubyte *src, w_int width, w_int height) {
  w_ubyte *dst = allocMem(currentWonkaThread, pixels2bytes(width * height));
  w_int  x, y;
  
  for (x=0; x < width; x++) {
    for (y=0; y < height; y++) {
      pixelset(pixelget(src, width, x, (height - (y + 1))), dst, height, y, x);
    }
  }
  
  return (w_word *)dst;
}

w_int canvas_verifyBounds(r_component dst, r_component src) {

  if (dst && dst->instance && src && src->instance) {
    return Component_verifyBounds(dst) && Component_verifyBounds(src) && Container_verifyBounds(dst, src);
  }
  else if (src && src->instance) {
    return Component_verifyBounds(src);
  }
  else if (dst && dst->instance) {
    return Component_verifyBounds(dst);
  }
  else {
    return 0;
  }

  /*
  ** TODO: for performance's sake, most of these verification results 
  **       can be cached.
  */
  
}

void canvas_copyAll(r_canvas dstCanvas, r_canvas srcCanvas) {

  w_instance instance = srcCanvas->component->instance;
  w_ubyte * src;
  w_ubyte * dst;
  int x;
  int y;
  int w;
  int h;
  int i;
  
  /*
  ** Debug output:
  */
  
  #ifdef DEBUG

    if (dstCanvas != rootCanvas && dstCanvas->buffer->status == 0) {
      wabort(ABORT_WONKA, "the destination canvas' buffer has been released.\n");
    }

    if (srcCanvas != rootCanvas && srcCanvas->buffer->status == 0) {
      wabort(ABORT_WONKA, "the source canvas' buffer has been released.\n");
    }

  #endif

  /*
  ** Copy buffer to the parent buffer:
  */

  x = (int)instance[F_Component_x];
  y = (int)instance[F_Component_y];

  Component_intersect(x, y, srcCanvas->buffer->vw, srcCanvas->buffer->vh, 0, 0, dstCanvas->buffer->fw, dstCanvas->buffer->fh, &x, &y, &w, &h);

  src = srcCanvas->buffer->buffer + pixels2bytes(srcCanvas->buffer->fw * srcCanvas->buffer->oy);
  dst = dstCanvas->buffer->buffer + pixels2bytes(dstCanvas->buffer->fw * y);
  
  woempa(5, "%d x %d  ->  %d x %d\n", w, h, dstCanvas->buffer->fw, dstCanvas->buffer->fh);
   
  for (i = 0; i < h; i++) {
    pixelcopyline(dst, src, x, srcCanvas->buffer->ox, w);
    src += pixels2bytes(srcCanvas->buffer->fw);
    dst += pixels2bytes(dstCanvas->buffer->fw);
  }
  
  /* 
  ** TODO: we tried optimizing this as much as possible as this
  **       function is frequently called according to gprof.  
  **       But still, in case the memory is aligned we could use
  **       a faster algorithm to blit a canvas' buffer.
  */

}
    
void buffer_realloc(r_canvas canvas) {
  void *buffer;
  int width;
  int height;

  if (canvas->component && canvas->component->instance) {
   
    width = canvas->component->instance[F_Component_width];
    height = canvas->component->instance[F_Component_height];
    
    if (width < 1 || height < 1 || width * height < 1 || width * height > MAX_SIZE) { 
      width = 0;
      height = 0;
    }
  
    /*
    ** Allocate the canvas' buffer:
    */
      
    buffer = reallocMem(canvas->buffer, sizeof(r_Buffer) + pixels2bytes(width * height));
    if (!buffer) {
      woempa(9, "No memory to allocate a canvas buffer!\n");

      return;
    }

    canvas->buffer = buffer;
    canvas->buffer->fw = width;
    canvas->buffer->fh = height;
    canvas->buffer->ox = 0;
    canvas->buffer->oy = 0;
    canvas->buffer->vw = width;
    canvas->buffer->vh = height;
    
    /*
    ** Debug output:
    */
      
    woempa(5, "cache: realloc buffer %p (%dx%d, allocated %d bytes, %d bytes in use) of container '%w'\n", canvas, width, height, pixels2bytes(width * height), x_mem_total() - x_mem_avail(), (instance2clazz(canvas->component->instance))->dotified);

    /*
    ** Mark the buffer as being allocated:
    */
      
    canvas->buffer->status = 1;

    /*
    ** Mark the component as being invalid:
    */
    
    canvas->component->refresh = 2;
  }
}

void buffer_release(r_canvas canvas) {

  r_component component;

  /*
  ** Recursivly do the same thing for all child component:
  */

  if (rootCanvas != canvas) {

    /*
    ** Release the canvas' buffer:
    */
    
    canvas->buffer = reallocMem(canvas->buffer, sizeof(r_Buffer));
    
    /*
    ** Debug output:
    */

    woempa(5, "cache: release buffer %p (%dx%d, released %d bytes, %d bytes in use) of container '%k'\n", canvas, canvas->buffer->fw, canvas->buffer->fh, canvas->buffer->fw * canvas->buffer->fh * sizeof(r_pixel), x_mem_total() - x_mem_avail(), (instance2clazz(canvas->component->instance)));

    /*
    ** Mark the buffer as being released:
    */
    
    canvas->buffer->status = 0;

    /*
    ** Mark the component as being invalid:
    */
    
    canvas->component->refresh = 2;

  }
      
  for (component = canvas->component->headChild; component != NULL; component = component->next) {

    if (component->tag == Z_CONTAINER) {
     
      if (canvas->buffer->status == 1 && Component_isVisible(canvas->component) == 0) {

        woempa(5, "canvas %p, %d, %d\n", canvas, canvas->buffer->status, Component_isVisible(component)); 

        buffer_release((r_canvas)component->object);
      }
    }
  }

}

void canvas_drawCanvas(r_canvas canvas, w_int depth) {

  r_canvas childCanvas;
  r_component c;

  if (canvas != rootCanvas && canvas->component && canvas->component->instance && canvas->component->parent && canvas->component->parent->instance) {
    if (!canvas_verifyBounds(canvas->component->parent, canvas->component)) {
      return;
      // wabort(ABORT_WONKA, "Oops. canvas_drawCanvas failed due to invalid bounds!\n");
    }
  }
  else if (canvas != rootCanvas && canvas->component && canvas->component->instance) {
    if (!Component_verifyBounds(canvas->component)) {
      return;
      // wabort(ABORT_WONKA, "Oops: canvas_drawCanvas failed due to invalid bounds!\n");
    }
  }

  if ((canvas == rootCanvas) || (canvas->component && canvas->component->parent && canvas->component->instance && canvas->component->instance[F_Component_visible])) {
    
    /*
    ** The off-screen buffer is tagged invalid:
    **   not able to use the cached off-screen buffer - rerender the
    **   canvas instead
    */
      
    if (canvas->component->refresh == 2) {

      /*
      ** Fix the damaged area:
      */
        
      canvas_repare(canvas->component);
     
      /*
      ** Call the Containers's paint() method in Java:
      */
      
      canvas_paint(canvas->component);

      /*
      ** Iterate through the component tree to redraw components (if necessary):
      */
      
      for (c = canvas->component->headChild; c != NULL; c = c->next) {
        
        if (c->tag == Z_CONTAINER) {

          if (canvas_verifyBounds(canvas->component, c)) {

            /*
            ** Resolve the child canvas:
            */
            
            childCanvas = (r_canvas) c->object;
            
            /*
            ** Debug output:
            */
            
            #ifdef DEBUG
              if (childCanvas && childCanvas->component && childCanvas->component->instance[F_Component_visible] && childCanvas->component->refresh == 2) { 
                woempa(5, "%scanvas '%k' (canvas = %p, x = %i, y = %i, w = %i, h = %i) [invalid cache - render]\n", canvas_spaces(depth), (instance2clazz(childCanvas->component->instance)), childCanvas, childCanvas->component->instance[F_Component_x], childCanvas->component->instance[F_Component_y], childCanvas->component->instance[F_Component_width], childCanvas->component->instance[F_Component_height]);
              }
            #endif
              
            /*
            ** Draw child canvas:
            */
            
            canvas_drawCanvas(childCanvas, depth + 1);
          }
        }
        else {
          
          canvas_drawComponent(c, 0);
        }     

      }        
      
      /*
      ** Mark the Container as being refreshed:
      */
      
      canvas->component->refresh = 0;
    } 
    
    /*
    ** The off-screen buffer is still valid:
    **   no need to rerender the canvas - use the cached off-screen 
    **   canvas instead
    */
    
    else if (canvas->component->refresh == 1) {
      
      /*
      ** Iterate through the component tree to redraw components (if necessary):
      */
       
      for (c = canvas->component->headChild; c != NULL; c = c->next) {
        
        if (c->tag == Z_CONTAINER) {
        
          childCanvas = (r_canvas) c->object;
      
          #ifdef DEBUG 
            if (childCanvas && childCanvas->component && childCanvas->component->refresh == 2) {
              woempa(5, "%scanvas '%p' (x = %i, y = %i, w = %i, h = %i) [invalid cache - render]\n", canvas_spaces(depth), childCanvas, childCanvas->component->instance[F_Component_x], childCanvas->component->instance[F_Component_y], childCanvas->component->instance[F_Component_width], childCanvas->component->instance[F_Component_height]);
            }
            else { 
              woempa(5, "%scanvas '%p' (x = %i, y = %i, w = %i, h = %i) [valid cache - prune]\n", canvas_spaces(depth), childCanvas, childCanvas->component->instance[F_Component_x], childCanvas->component->instance[F_Component_y], childCanvas->component->instance[F_Component_width], childCanvas->component->instance[F_Component_height]); 
            }
          #endif
          
          if (childCanvas->component->refresh != 0) {  
            canvas_drawCanvas(childCanvas, depth + 1);
          }
       
        } 
     
      } 
                
      /*
      ** Mark the Container as being refreshed:
      */
      
      canvas->component->refresh = 0;

    }

    /*
    ** Visualize the Container's canvas:
    */
    
    if (canvas != rootCanvas && depth != 0) {
      r_canvas parentCanvas = (r_canvas)canvas->component->parent->object;
        
      /*
      ** Debug output:
      */
     
      #ifdef DEBUG
        if (parentCanvas == rootCanvas) { 
          woempa(5, "calling canvas_copyAll(%p = ROOTCANVAS, %p): buffer 1: %p, buffer 2: %p\n", parentCanvas, canvas, parentCanvas->buffer, canvas->buffer); 
        }
        else { 
          woempa(5, "calling canvas_copyAll(%p, %p): buffer 1: %p, buffer 2: %p\n", parentCanvas, canvas, parentCanvas->buffer, canvas->buffer); 
        }
      #endif
      
      /*
      ** Validate the parent buffer:
      */
      
      if (parentCanvas != rootCanvas) {
        if (parentCanvas->buffer->status == 0) {
          buffer_realloc(parentCanvas);
        }
      }
      
      canvas_copyAll(parentCanvas, canvas);
      
    }
    else {
      
      canvas_canvas2screen(canvas);
    }       
  }
}

void canvas_drawComponent(r_component c, w_int display) {

  r_canvas canvas;
 
  if (c && c->parent) {

    if (c->tag == Z_CONTAINER) {
      canvas = (r_canvas)c->object;
    }
    else {
      canvas = (r_canvas)c->parent->object;
    }
    
    if (c->instance[F_Component_visible] && canvas_verifyBounds(canvas->component, c)) {
    
      /*
      ** Redraw the component:
      */
      
      if (c->tag == Z_CONTAINER) {
        canvas->component->parent->refresh = 2;
        canvas->component->refresh = 2;
        canvas_drawCanvas(canvas, 0);
      }
      else {

        // registry_dump(rootCanvas, 0);
        
        /*
        ** This is mandatory: leave this out and you'll face update/refresh
        ** problems ... if the canvas you are about to draw in is not up to
        ** date, make sure to update it first.
        */
               
        if (display && canvas->component->refresh) {
          canvas_drawCanvas(canvas, 1);
        }
        
        /*
        ** Fix the damaged area:
        */
   
        canvas_repare(c);

        switch (c->tag) {
          case Z_BUTTON:
            woempa(6, "button '%w'\n", String2string((w_instance)c->instance[F_Button_label]));
            Button_drawButtonDefault(canvas->buffer, c);
            break;
          case Z_CHECKBOX:
            woempa(6, "checkbox '%w'\n", String2string((w_instance)c->instance[F_Checkbox_label]));
            Checkbox_drawCheckboxDefault(canvas->buffer, c);
            break;
          case Z_LABEL:
            woempa(6, "label '%w'\n", String2string((w_instance)c->instance[F_Label_text]));
            Label_drawLabel(canvas->buffer, c);
            break;
          case Z_TEXTFIELD:
            woempa(6, "textfield '%w'\n", String2string((w_instance)c->instance[F_TextComponent_text]));
            TextField_drawTextField(canvas->buffer, c);
            break;
          case Z_COMPONENT:
            woempa(6, "component '%k'\n", instance2clazz(c->instance));
            break;
          default:
            wabort(ABORT_WONKA, "ERROR: unknown or unsupported component (type = %i)\n", c->tag);
        }
        
        /*
        ** Call the Component's paint() method in Java:
        */
        
        canvas_paint(c);

        /*
        ** Update the screen when needed:
        */
     
        if (display) {
          canvas_canvas2screen(canvas);
        }
 
        /*
        ** This is mandatory: leave this out and you'll face update/refresh
        ** problems ... if the canvas you are about to draw in is not up to
        ** date, make sure to update it first.
        */

        canvas->component->refresh = 1;
      }    
    }
  }
}

void canvas_repare(r_component component) {

  w_instance instance; 
  r_canvas canvas;
  w_instance col;
  r_pixel pixel;
  w_int x;
  w_int y;
  w_int w;
  w_int h;
  w_int i;

  if (component->instance == NULL) {
    /*
    ** Do nothing, we are trying to fix the rootCanvas.
    */

    return;
  }
  
  if (component->tag == Z_CONTAINER) {
    
    /*
    ** Clear entire canvas:
    */
    
    instance = component->instance;
    x = 0;
    y = 0;
    w = (w_int)instance[F_Component_width];
    h = (w_int)instance[F_Component_height];
    canvas = (r_canvas)component->object;
  }
  else {

    /*
    ** Clear part of parent canvas:
    */

    canvas = (r_canvas)component->parent->object;
    instance = component->instance;
    x = (w_int)instance[F_Component_x];
    y = (w_int)instance[F_Component_y];
    w = (w_int)instance[F_Component_width];
    h = (w_int)instance[F_Component_height];
  }
  
  /*
  ** Validate and reallocate the canvas buffer if necessary:
  */
      
  if (canvas->buffer->status == 0) {
    buffer_realloc(canvas);
  }
    
  /*
  ** Resolve background color:
  */
    
  Component_getBackground(component, &col, Component_getColor(F_SystemColor_window));
    
  /*
  ** Debug output:
  */
    
  woempa(5, "fixed damaged area: %d, %d, %d, %d, %d, %d\n", x, y, w, h, canvas->buffer->fw, canvas->buffer->fh);
      
  /*
  ** Redraw background:
  */
    
  pixel = color2pixel(col);
  for (i = 0; i < h; i++) {
    drawHLine(canvas->buffer, x, y + i, w, pixel);
  }
}

inline static w_size round_up(w_size value, w_size rounding) {
  return (value + (rounding - 1)) & ~(rounding - 1);
}

inline static w_size round_down(w_size value, w_size rounding) {
  return (value) & (unsigned int)(0 - rounding);
}

/*#include <stdio.h>
static int drawCount = 0;*/

void canvas_canvas2screen(r_canvas canvas) {
 
  static int count = 0;
  
  r_component component;
  w_ubyte *src;
  w_ubyte *dst;

  int i;
  int x;
  int y;
  int w;
  int h;
  int dx;
  int dy;

  w_word *swapped_buffer;
    
  if (canvas == rootCanvas) {    
    /*
    ** Debug output:
    */
    
    woempa(5, "called canvas_canvas2screen(%p = ROOTCANVAS)\n", rootCanvas);

    /*
    ** If canvas is tagged as being invalid, make sure to refresh it first:
    */
    
    if (rootCanvas->component->refresh) {
      canvas_drawCanvas(rootCanvas, 0);
    }
    
    /*
    ** Copy off-screen buffer to video buffer:
    */

    if (swap_display) {
      swapped_buffer = canvas_transpose(rootCanvas->buffer->buffer, screen->width, screen->height);
      canvas_blitter((w_word *)screen->video, (w_word *)swapped_buffer, pixels2bytes(screen->width * screen->height));
      releaseMem(swapped_buffer);
    } 
    else {
      canvas_blitter((w_word *)screen->video, (w_word *)rootCanvas->buffer->buffer, pixels2bytes(screen->width * screen->height));
    }
    
    /*
    ** Periodically try to release invisible buffers.
    */
    
    count = count + 1;
    if (count > 10) {
      buffer_release(rootCanvas);
      count = 0;
    }
  }
  else if (Component_isVisible(canvas->component)) {

    component = canvas->component;
    
    #ifdef DEBUG
      if (canvas->buffer->status == 0) {
        wabort(ABORT_WONKA, "Oops!  This buffer %p of canvas %p of component %k has been released!\n", canvas->buffer, canvas, (instance2clazz(component->instance)));
      }
    #endif
         
    /*
    ** Resolve canvas coordinates:
    */
    
    getAbsoluteCoordinates(component, &x, &y, &w, &h, &dx, &dy, 0);
    
    /*
    ** Debug output:
    */
   
    woempa(5, "called canvas_canvas2screen(%k = %p): x = %i, y = %i, w = %i, h = %i, dx = %i, dy = %i, refresh = %d\n", instance2clazz(component->instance), canvas, x, y, w, h, dx, dy, component->refresh);

    /*
    ** Validate canvas if required:
    */

    if (component->refresh) {
      canvas_drawCanvas(canvas, 0);
    }

    /*
    ** Copy canvas to root canvas:
    */

    src = canvas->buffer->buffer + pixels2bytes(canvas->buffer->fw * dy);
    dst = rootCanvas->buffer->buffer + pixels2bytes(y * rootCanvas->buffer->fw);

    for (i = 0; i < h; i++) {
      pixelcopyline(dst, src, x, dx, w);
      dst += pixels2bytes(rootCanvas->buffer->fw);
      src += pixels2bytes(canvas->buffer->fw);
    }

    /* 
    ** Resolve the parent stand-alone frame (java.awt.Window, 
    ** java.awt.Frame or direct subclasses):
    */
   
    while (component->parent && component->parent->parent) {
      component = component->parent;
    }

    /*
    ** Display the next canvas in the row; that is, the one with a
    ** higher z-depth/order.  Note that this function calls itself
    ** (recursion) to render the entire list of canvases that have
    ** a higher z-order.
    */
    
    if (component->next) {
      canvas_canvas2screen((r_canvas)component->next->object);
    }
 
    /*
    ** Video memory requires memory alignment on 32-bit words:
    */
    
    if (swap_display) {
      swapped_buffer = canvas_transpose(rootCanvas->buffer->buffer, screen->width, screen->height);
      canvas_blitter((w_word *)screen->video, (w_word *)swapped_buffer, pixels2bytes(screen->width * screen->height));
      releaseMem(swapped_buffer);
    } 
    else {
      /*w = w + (x - (w_int)round_down((w_size)x, 4));
      x = (w_int)round_down((w_size)x, 4);
      w = (w_int)round_up((w_size)w, 4);
      x = 0; y = 0; w = screen->width; h = screen->height;*/

      dst = screen->video + pixels2bytes((y * screen->width) + x);
      src = rootCanvas->buffer->buffer + pixels2bytes((y * screen->width) + x);
    
      /*
      ** Copy root canvas to video memory:
      */
        
      /*for (i = 0; i < h; i++) {
        canvas_blitter((w_word *)dst, (w_word *)src, pixels2bytes(w));
      
        dst += pixels2bytes(screen->width);
        src += pixels2bytes(screen->width-1);
      }*/
      
      canvas_blitter(
	(w_word*) screen->video, 
	(w_word*) rootCanvas->buffer->buffer,
	pixels2bytes(screen->width*screen->height)
      );
    }
  }
  else {

    component = canvas->component;

    /* 
    ** Resolve the parent stand-alone frame (java.awt.Window, 
    ** java.awt.Frame or direct subclasses):
    */
   
    while (component->parent && component->parent->parent) {
      component = component->parent;
    }

    /*
    ** Display the next canvas in the row; that is, the one with a
    ** higher z-depth/order.  Note that this function calls itself
    ** (recursion) to render the entire list of canvases that have
    ** a higher z-order.
    */
    
    if (component->next) {
      canvas_canvas2screen((r_canvas)component->next->object);
    }

  }
}

void canvas_paint(r_component component) {

  static w_string name = NULL;
  static w_string sig = NULL;
  w_thread thread = currentWonkaThread;
  w_clazz clazz = instance2clazz(component->instance);
  w_instance graphics = NULL;
  w_frame bottom = NULL; 

  if (name == NULL) {
    // we only need to create this w_string once
    name = cstring2String(thread, "paint", 5);
    sig = cstring2String(thread, "(Ljava/awt/Graphics;)", strlen("(Ljava/awt/Graphics;)"));
  }
  
  if(!component->paint_method) {
    component->paint_method = getMethod_name_sig(clazz, name, sig);
  }

  /*
  ** Check whether the method is non-empty:
  */
  
  if (component->paint_method == NULL || component->paint_method->code_length <= 1) {
    return;
  }

  /*
  ** Check whether the component is visible:
  */
  
  if (!(component->instance && component->instance[F_Component_visible])) {
    return;
  }
 
  #ifdef DEBUG
    {
      r_canvas canvas = (r_canvas)component->object;
      if (component->instance && component->tag == Z_CONTAINER && canvas->buffer->status == 0) {
        wabort(ABORT_WONKA, "Oops! The buffer of canvas %p has been released!\n", canvas);
      }
    }
  #endif
    
  graphics = Component_getGraphics(NULL, component->instance);

  if (graphics) {           
    bottom = allocBottomFrame(thread, component->paint_method);
    if (bottom) {
      bottom->locals[0] = (w_word)component->instance;
      bottom->locals[1] = (w_word)graphics;

      native_paint = 1;
    
      setFlag(bottom->flags, FRAME_JNI);
      runThread(thread, bottom);

      native_paint = 0;

      releaseBottomFrame(bottom);
    }
    removeLocalReference(thread->top, graphics);
  }
}

/*
** Description:
**   A line algorithm using shift operators instead of
**   multiplications and divisions!
** Algorithm:
**   Given two line endpoints, we are trying to find the
**   'in-between' points on a pixel grid.  Briefly, our
**   algorithm determines subsequent points from the start
**   point by making a decision between the two next
**   available points by determining which is closer to the
**   ideal point.
** Remark:
**   Once we are in the for-loop the calculation of a pixel
**   requires 2.5 additions (average)! :)
*/

enum {
  r_top = 0x1, 
  r_bottom = 0x2, 
  r_right = 0x4, 
  r_left = 0x8
};

int cohen_sutherland_code(int x, int y, r_buffer buffer) {
  int c = 0;
  
  if (y > buffer->fh - 1) {
    c |= r_top;
  }
  else if (y < 0) {
    c |= r_bottom;
  }
  if (x > buffer->fw - 1) {
    c |= r_right;
  }
  else if (x < 0) {
    c |= r_left;
  }
  
  return c;
}

void drawLine(r_buffer buffer, int x1, int y1, int x2, int y2, r_pixel color) {
 
  int dx;
  int dy;
  int bx1;
  int by1;
  int bx2;
  int by2;
  int x;
  int y;
  int stop;
  int code;
  int code0;
  int code1; 
  int xIncr;
  int yIncr;
  
  /*
  ** Take a handy short-cut when we have to draw a horizontal 
  ** line:
  */
    
  if (y1 == y2) {
    drawHLineClipped(buffer, (x1 < x2 ? x1 : x2), y1, abs(x2 - x1), color);
    return;
  }
  
  /*
  ** Initialize the temporaries:
  */
  
  bx1 = 0;
  by1 = 0;
  bx2 = buffer->fw - 1;
  by2 = buffer->fh - 1;

  x = 0;
  y = 0;
  
  stop = 0;
  
  /*
  ** Clip the line:
  */
  
  code0 = cohen_sutherland_code(x1, y1, buffer);
  code1 = cohen_sutherland_code(x2, y2, buffer);

  while (!stop) {
    
    /*
    ** The line is completly inside our buffer.
    */
    
    if (!(code0 | code1)) {
      stop = 1;
      continue;
    }


    /*
    ** The line is completly outside our buffer.
    */
    
    if (code0 & code1) {
      stop = 2;
      continue;
    }

    /*
    ** We have to clip the line:
    */
   
    code = code0 ? code0 : code1;
    
    if (code & r_top) {
      x = x1 + (x2 - x1) * (by2 - y1) / (y2 - y1);
      y = by2;
    } 
    else if (code & r_bottom) {
      x = x1 + (x2 - x1) * (by1 - y1) / (y2 - y1);
      y = by1;
    } 
    else if (code & r_right) {
      x = bx2;
      y = y1 + (y2 - y1) * (bx2 - x1) / (x2 - x1);
    }
    else {
      x = bx1;
      y = y1 + (y2 - y1) * (bx1 - x1) / (x2 - x1);
    }

    /*
    ** Update our coordinates and iterate:
    */

    if (code == code0) {
      x1 = x;
      y1 = y;
      code0 = cohen_sutherland_code(x1, y1, buffer);
    } 
    else {
      x2 = x;
      y2 = y;
      code1 = cohen_sutherland_code(x2, y2, buffer);
    }
  }

  /*
  ** Render the line:
  */
 
  if (stop == 1) {
 
    /*
    ** Initialize the components of the algorithm that are not
    ** affected by the slope or direction of the line.
    */
    
    dx = abs(x2 - x1);
    dy = abs(y2 - y1);

    woempa(5, "called r_drawLine(%p, %i, %i, %i, %i, %d)\n", buffer, x1, y1, x2, y2, color);

    /*
    ** Determine 'directions' to increment x and y (regardless of
    ** decision):
    */
    
    if (x1 > x2) { 
      xIncr = -1;
    } 
    else { 
      xIncr = 1; 
    }

    if (y1 > y2) { 
      yIncr = -1; 
    }
    else {
      yIncr = 1;
    }

    /* 
    ** Determine variable and initiate appropriate line drawing
    ** routine:
    */
    
    if (dx >= dy) {
      int dPr = dy << 1;
      int dPru = dPr - (dx << 1);
      int p = dPr - dx;
  
      for (; dx >= 0; dx--) {
        drawPixel(buffer, x1, y1, color);
        if (p > 0) {
          x1 += xIncr;
          y1 += yIncr;
          p += dPru;
        }
        else {
          x1 += xIncr;
          p += dPr;
        }
      }
    }
    else {
      int dPr = dx << 1;
      int dPru = dPr - (dy << 1);
      int p = dPr - dy;
  
      for (; dy >= 0; dy--) {
        drawPixel(buffer, x1, y1, color);
        if (p > 0) {
          x1 += xIncr;
          y1 += yIncr;
          p += dPru;
        }
        else {
          y1 += yIncr;
          p += dPr;
        }
      }
    }
  }
}

