
#include "stdarg.h"
#include "oswald.h"  // memory allocation

#include "j-spot.h"
#include "ir.h"
#include "assembler.h"
#include "code-generator.h"

j_asm asm_create_block(void) {

  j_asm block;
 
  /*
  ** Allocate a chunk of memory that can be used to emit the generated
  ** machine code to.  Remark that we assume that the pointer pointing 
  ** to this memory block is word aligned.
  */
  
  block = (j_asm)x_mem_calloc(sizeof(j_Asm));
  block->size = 5000;
  block->base = (void *)x_mem_calloc(block->size);
  block->step = block->base;
  
  return block;
}

void asm_free_block(j_asm block) {

  j_ref ref_curr;
  j_ref ref_swap;
 
  ref_curr = block->refs;
  while (ref_curr) {
    ref_swap = ref_curr;
    ref_curr = ref_curr->next;
    x_mem_free(ref_swap);
  } 
  
  x_mem_free(block);

  /*
  ** NOTE: remark that we don't free the memory chunk that holds the 
  **       generated machine code.
  */

}

int asm_emit(j_asm block, int endian, int count, ...) {

  va_list ap;

  int shift = 0;
  int bytes = 0;
  int data = 0;
  int insn = 0;
  int bits = 0;
  int loop = 0;
  int size = 0;
  
  size = count / 8; 
  loop = count;
  
  if (endian) {
    block->step = block->step + size - 1;
  }
  
  va_start(ap, count);

  while (loop > 0) {
    bits = va_arg(ap, int);
    data = va_arg(ap, int);

    if (bits < 0 || bits > 32) {
      wabort(ABORT_WONKA, "invalid instruction: bits = %d\n", bits);
    }
   
    if (block->step - 32 > block->base + block->size) {
      wabort(ABORT_WONKA, "code buffer is full: size = %d\n", block->size);
    }
    
    shift = shift + bits;
    insn |= data << (32 - shift);

    while (shift > 7) {
      if (endian) {
        *block->step-- = (char)(insn >> 24);
      }
      else {
        *block->step++ = (char)(insn >> 24);
      }
      bytes = bytes + 1;
      shift = shift - 8;
      insn = insn << 8;
    }
  
    loop = loop - bits;
  }
  
  va_end(ap);
 
  if (shift != 0 || bytes != size) {
    wabort(ABORT_WONKA, "invalid instruction: shift = %d, size = %d, bytes = %d\n", shift, size, bytes);
  }

  if (endian) {
    block->step = block->step + size + 1;
  }
 
  return size;

}

char * asm_export(j_asm block) {
 
  /* 
  ** Reallocate the memory chunk that holds the generated assembly code
  ** such that it is contracted to minimal size.
  */
  
  x_mem_realloc(block->base, (x_size)(block->step - block->base));
 
  asm_patch_references(block);
  
  return block->base;
}

void asm_dump(j_asm block) {
 
  #ifdef DEBUG
  
    char * i;
    char * j;
  
    fprintf(stderr, "----------------\n");
    fprintf(stderr, "raw compiled method:\n");
  
    for (i = block->base; i < block->step; i = i + 4) {
      fprintf(stderr, "   %10p  |  %s  (%8x hex)\n", i, asm_insn(*(j_word *)i), *(j_word *)i);
    }

    fprintf(stderr, "disassembled method:\n");

    for (j = block->base; j < block->step; j = j) {
      fprintf(stderr, "   %10p  |  ", j);
      j = (char *)cg_disassemble((j_word *)j);
      fprintf(stderr, "\n");
    }
  
    fprintf(stderr, "----------------\n");

  #endif

}

char * asm_insn(j_word inst) {

  /*
  ** Return a string with the bit-wise representation of the
  ** specified word.
  */
 
  unsigned int size = 31; 
  char * buffer = (char *)x_mem_alloc(size + 8);
  unsigned int mask = (unsigned)(1 << (size));
  int i = 0;

  memset(buffer, 0x00, (unsigned)size);

  while (mask) {
    if (i % 9 == 8) {
      *(buffer + i++) = ' ';
    }
    if (inst & mask) {
      *(buffer + i++) = '1';
    }
    else {
      *(buffer + i++)  = '0';
    }
    mask >>= 1;
  }
  
  *(buffer + i++) = '\0';
                                                      
  return (char *)buffer;
                                                               
}

void asm_add_reference(j_asm block, char * address, j_node dst) {

  j_ref ref = x_mem_calloc(sizeof(j_Ref));
 
  woempa(9, "add reference: src at address %p, dst node %p at address %p\n", address, dst, dst->address); 

  ref->address = address;
  ref->dst = dst;

  if (block->refs) {
    ref->next = block->refs;
    block->refs = ref;
  }
  else {
    block->refs = ref;
  }

  // NOTE:
  //   We don't care about duplicate entries.
}

void asm_patch_references(j_asm block) {

  j_ref ref = block->refs;
 
  while (ref) {
    *((int *)(ref->address - 4)) = (int)(ref->dst->address - ref->address);
    ref = ref->next;
  }
}

