/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

#include "oswald.h"        /* x_thread_... */
#include "wonka.h"         /* the w_ types */
#include "ts-mem.h"        /* allocMem_system */
#include "realms.h"        /* allocMem_system */
#include "jdwp.h"          /* all the jdwp stuff */
#include "core-classes.h"  /* F_... stuff */
#include "arrays.h"        /* instance2arrayof */
#include "strings.h"       /* string2UTF8 */
#include <string.h>        /* strncmp */

/*
** threadgrp_name : Returns the name of the thread with given ID.
*/

w_void jdwp_threadgrp_name(jdwp_command_packet cmd) {
  w_instance  threadgroup;
  w_group     group;
  w_ubyte     *name;
  w_int       length;
  
  /*
  ** Get the address of the group out of the packet.
  */
  
  memcpy(&threadgroup, &cmd->data[0], 4);
  group = (w_group)threadgroup[F_ThreadGroup_wotsit];
  
  /* 
  ** sanity check : See if this is indeed a threadgroup.
  */
  
  if (strncmp(group->label, "group", 5) == 0) {

    /*
    ** Everything checks out. This is indeed a threadgroup. 
    ** Now all that's left is converting the name from a wonka string in a 
    ** UTF8 string, pack it and send it back to the debugger.
    */
    
    name = jdwp_string2UTF8(group->name, &length);
    
    jdwp_send_reply(cmd->id, name, length + 4, jdwp_err_none);

    x_mem_free(name);
    
  } else {

    /*
    ** This is not (or no longer) a threadgroup... Let the debugger know about 
    ** this little fact.
    */
    
    jdwp_send_reply(cmd->id, NULL, 0, jdwp_err_invalid_object);

  }
  
  return;
}


/*
** threadgrp_parent: Returns the parent of the threadgroup with given ID.
*/

w_void jdwp_threadgrp_parent(jdwp_command_packet cmd) {
  w_instance  threadgroup;
  w_group     parent;
  w_group     group;
  w_int       groupID;
  
  /*
  ** Get the address of the group out of the packet.
  */
  
  memcpy(&threadgroup, &cmd->data[0], 4);
  group = (w_group)threadgroup[F_ThreadGroup_wotsit];
  
  /* 
  ** sanity check : See if this is indeed a threadgroup.
  */
  
  if (strncmp(group->label, "group", 5) == 0) {

    /*
    ** Everything checks out. This is indeed a threadgroup. 
    ** Now all that's left is converting the name from a wonka string in a 
    ** UTF8 string, pack it and send it back to the debugger.
    */
    
    parent = group->parent;

    if((parent == group) || (parent == NULL)) {
      
      /*
      ** If 'parent' equals the given threadgroup or the parent is NULL, then this 
      ** threadgroup has no parent and is a toplevel threadgroup.
      */

      jdwp_send_reply(cmd->id, NULL, 0, jdwp_err_invalid_object);
      
    } else {

      /*
      ** There actually is a parent, so get the instance and send it to the debugger.
      */
      
      groupID = (w_int)(parent->ThreadGroup);
     
      jdwp_send_reply(cmd->id, &groupID, 4, jdwp_err_none);  
      
    }
    
  } else {

    /*
    ** This is not (or no longer) a threadgroup... Let the debugger know about 
    ** this little fact.
    */
    
    jdwp_send_reply(cmd->id, NULL, 0, jdwp_err_invalid_object);

  }
  
  return;
}

/*
** threadgrp_children: Returns the direct children of this threadgroup. 'Children' are the 
**                     different threads and threadgroups that have this threadgroup as their
**                     parent. Threads/threadgroups in child threadgroups are not reported.
*/

w_void jdwp_threadgrp_children(jdwp_command_packet cmd) {
  w_instance  group;
  w_group     threadgroup;
  w_instance  threads_vector;
  w_instance  groups_vector;
  w_instance  threads_array;
  w_instance  groups_array;
  w_instance  *threads_data = NULL;
  w_instance  *groups_data = NULL;
  w_int       threads_count = 0;
  w_int       groups_count = 0;
  w_ubyte     *reply;
  w_int       i;
  w_int       position = 0;
  w_int       ID;
  
  /* 
  ** Get the threadgroup out of the packet
  */

  memcpy(&group, &cmd->data[0], 4);
  threadgroup = (w_group)group[F_ThreadGroup_wotsit];

  /*
  ** TODO: Add a sanity check to see if 1) this is a threadgroup, 2) it still exists...
  */
  
  /*
  ** Now that we have a reference to the threadgroup, we need to get the list of its childeren.
  ** This can be done with some JNI magic : The threadgroup instance has 2 vectors; one contains
  ** the child threadgroups, the other one contains the flock of threads.
  */

  threads_vector = (w_instance)threadgroup->ThreadGroup[F_ThreadGroup_flock];
  groups_vector =  (w_instance)threadgroup->ThreadGroup[F_ThreadGroup_children];

  /*
  ** Get the number of elements in the 2 vectors.
  */

  if(threads_vector) {
    threads_count = threads_vector[F_Vector_elementCount];
    threads_array = (w_instance)threads_vector[F_Vector_elementData];
    threads_data = instance2Array_instance(threads_array);
  } 

  if(groups_vector) {
    groups_count =  groups_vector[F_Vector_elementCount];
    groups_array = (w_instance)groups_vector[F_Vector_elementData];
    groups_data = instance2Array_instance(groups_array);
  }

  /*
  ** Allocate memory for the reply. We need a word for every thread/threadgroup and also 
  ** 2 words for the 2 counters.
  */
  
  reply = x_mem_alloc((w_word)((2 + threads_count + groups_count) * 4));

  /*
  ** Iterate over the elements and store them in the reply.
  ** First the numbers of threads and all the threads.
  */

  threads_count = swap_int(threads_count);
  memcpy(&reply[position], &threads_count, 4);
  position += 4;
  threads_count = swap_int(threads_count);

  for(i = 0; i < threads_count; i++) {
    ID = (w_int)((w_instance)(threads_data[i]));
    memcpy(&reply[position], &ID, 4);
    position += 4;
  }

  /*
  ** Then the number of groups and all the groups.
  */
  
  groups_count = swap_int(groups_count);
  memcpy(&reply[position], &groups_count, 4);
  position += 4;
  groups_count = swap_int(groups_count);
  
  for(i = 0; i < groups_count; i++) {
    ID = (w_int)((w_instance)(groups_data[i]));
    memcpy(&reply[position], &ID, 4);
    position += 4;
  }

  /*
  ** Send this bunch to the debugger.
  */
  
  jdwp_send_reply(cmd->id, reply, (2 + threads_count + groups_count) * 4, jdwp_err_none);

  /* 
  ** Clean up
  */

  x_mem_free(reply);
  
  return;
}


/*
** The dispatcher for the 'ThreadGroup reference' command set.
*/

w_void dispatch_threadgrpref(jdwp_command_packet cmd) {

  switch((jdwp_threadgrpref_cmd)cmd->command) {
    case jdwp_threadgrpref_name:      jdwp_threadgrp_name(cmd);        break;
    case jdwp_threadgrpref_children:  jdwp_threadgrp_children(cmd);    break;
    case jdwp_threadgrpref_parent:    jdwp_threadgrp_parent(cmd);      break;
    default:
      jdwp_send_reply(cmd->id, NULL, 0, jdwp_err_not_implemented);
  }
  
}

