###########################################################################
# Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 #
#                                                                         #
# This software is copyrighted by and is the sole property of Acunia N.V. #
# and its licensors, if any. All rights, title, ownership, or other       #
# interests in the software remain the property of Acunia N.V. and its    #
# licensors, if any.                                                      #
#                                                                         #
# This software may only be used in accordance with the corresponding     #
# license agreement. Any unauthorized use, duplication, transmission,     #
#  distribution or disclosure of this software is expressly forbidden.    #
#                                                                         #
# This Copyright notice may not be removed or modified without prior      #
# written consent of Acunia N.V.                                          #
#                                                                         #
# Acunia N.V. reserves the right to modify this software without notice.  #
#                                                                         #
#   Acunia N.V.                                                           #
#   Vanden Tymplestraat 35      info@acunia.com                           #
#   3000 Leuven                 http://www.acunia.com                     #
#   Belgium - EUROPE                                                      #
###########################################################################

.file "atomic.s"

.text
.align 4
.global compare_swap

#
# w_boolean compare_swap(void *address, w_word old, w_word new);
#
# Atomically compare and swap; when the contents of 'address' are the 
# same as 'old', we write 'new' in there, otherwise, we keep the contents
# unchanged. When we swap the contents, we return 1, otherwise 0.
#
# Note: we use a spinlock and a signal that indicates that a 'LOCK IN
# PROGRESS' is taking place. This identifier is set by using the atomic
# xchg opcode. As long we get the 'LOCK IN PROGRESS' identifier from the
# xchg opcode, we keep on spinning since somebody else is in the
# critical section, indicated by (CS). It is up to the caller to implement
# a priority inversion prevention mechanism! We do call x_spinlock_backoff with the
# failed count to implement exponential backoff...
#

compare_swap:
      pushl %ebp                 # save %ebp
      movl %esp, %ebp            # and overwrite it with the new stack pointer
      subl $4, %esp              # make room for 1 local argument
      movl $0, -4(%ebp)          # set count to 0

cs_spinlock:
      movl $0xffffffff, %eax     # load %eax with our 'LOCK IN PROGRESS' identifier
      movl 8(%ebp), %edx         # load %edx with our 'address' argument
      xchgl %eax, (%edx)         # atomically exchange contents of %eax with '*address'
      cmpl $0xffffffff, %eax     # see if we have a 'LOCK IN PROGRESS' and if
      jne cs_critical            # that is not the case, we are in the critical section
      pushl -4(%ebp)             # push current count on stack
      call x_spinlock_backoff    # call our custom backoff function
      addl $4, %esp              # restore stack pointer to previous state
      incl -4(%ebp)              # increment our lock count
      jmp cs_spinlock            # and try again...

cs_critical:                     # here we have entered the critical section
      movl 12(%ebp), %ecx        # (CS) load %ecx with our 'old' argument
      cmpl %eax, %ecx            # (CS) compare our 'old' argument with the contents of %eax
      je cs_swap                 # (CS) if they are equal, do the swap
      movl %eax, (%edx)          # (CS) replace *addres with the 'old' argument, leave the critical section
      xorl %eax, %eax            # set %eax to 0 to return as result
      jmp cs_epilogue            # go back to our caller without swapping

cs_swap:
      movl 16(%ebp), %ecx        # (CS) load %ecx with our 'new' argument
      movl %ecx, (%edx)          # (CS) and write it back to '*address', leave the critical section
      movl $1, %eax              # write 1 as compare_swap result

cs_epilogue:
      movl %ebp, %esp            # restore previous stack pointer
      popl %ebp                  # restore %ebp saved at the entry
      ret                        # and return to our caller

# -----------------------------------------------------------------------------------------------------------------------

.text
.align 4
.global peek_swap

peek_swap:
      pushl %ebp                 # save %ebp
      movl %esp, %ebp            # and overwrite it with the new stack pointer
      subl $4, %esp              # make room for 1 local argument
      movl $0, -4(%ebp)          # set count to 0

ps_spinlock:
      movl $0xffffffff, %eax     # load %eax with our 'LOCK IN PROGRESS' identifier
      movl 8(%ebp), %edx         # load %edx with our 'address' argument
      xchgl %eax, (%edx)         # atomically exchange contents of %eax with '*address'
      cmpl $0xffffffff, %eax     # see if we have a 'LOCK IN PROGRESS' and if
      jne ps_critical            # that is not the case, we are in the critical section
      pushl -4(%ebp)             # push current count on stack
      call x_spinlock_backoff    # call our custom backoff function
      addl $4, %esp              # restore stack pointer to previous state
      incl -4(%ebp)              # increment our lock count
      jmp ps_spinlock            # and try again...

ps_critical:                     # here we have entered the critical section
      movl %eax, (%edx)          # (CS) write back old value, leaving the critical section and result in %eax

ps_epilogue:
      movl %ebp, %esp            # restore previous stack pointer
      popl %ebp                  # restore %ebp saved at the entry
      ret                        # and return to our caller

# -----------------------------------------------------------------------------------------------------------------------

.text
.align 4
.global lock_flags

#
# w_boolean lock_flags(void *address, w_flags flags)
#
# Lock a set of flags at the address. The flags that can be used are the
# lower 31 bits of the w_flags, the highest bit is used as a LOCK IN PROGRESS bit
# and will always be reset to 0 when entering this routine. When at least
# one of the passed bit flags is allready set at the address, the lock does not
# succeed and we will return a 0, otherwise we return a 1.
#

lock_flags:
      pushl %ebp                 # save %ebp
      movl %esp, %ebp            # and overwrite it with the new stack pointer
      subl $4, %esp              # make room for 1 local argument
      movl $0, -4(%ebp)          # set count to 0
      movl $0x7fffffff, %ecx     # load %ecx with our mask
      movl 12(%ebp), %eax        # load %eax with our flags word
      andl %ecx, %eax            # set highest bit to zero to be on the safe side, leave result in %eax
      movl %eax, 12(%ebp)        # and write it back

lf_spinlock:
      movl $0x80000000, %eax     # load %eax with our 'LOCK IN PROGRESS' identifier
      movl 8(%ebp), %edx         # load %edx with our 'address' argument
      xchgl %eax, (%edx)         # atomically exchange contents of %eax with '*address'
      testl $0x80000000, %eax    # see if our 'LOCK IN PROGRESS' bit is set
      je lf_critical             # that is not the case, we are in the critical section, %eax contains the current flags
      pushl -4(%ebp)             # push current count on stack
      call x_spinlock_backoff    # call our custom backoff function
      addl $4, %esp              # restore stack pointer to previous state
      incl -4(%ebp)              # increment our lock count
      jmp lf_spinlock            # and try again...

lf_critical:                     # here we have entered the critical section
      movl 12(%ebp), %ecx        # (CS) load %ecx with our flags words
      testl %eax, %ecx           # (CS) test our old flags against the new flags
      je lf_set                  # (CS) if none are set, set the flags
      movl %eax, (%edx)          # (CS) clash, replace *addres with the 'old' flags, leaving critical section
      xorl %eax, %eax            # set %eax to 0 to return as result
      jmp lf_epilogue            # go back to our caller without swapping

lf_set:
      orl %ecx, %eax             # (CS) or flags together, leave result in %eax
      movl %eax, (%edx)          # (CS) write back result, thereby leaving the critical section
      movl $1, %eax              # set result code to 1

lf_epilogue:
      movl %ebp, %esp            # restore previous stack pointer
      popl %ebp                  # restore %ebp saved at the entry
      ret                        # and return to our caller

# -----------------------------------------------------------------------------------------------------------------------

.text
.align 4
.global unlock_flags

#
# Reverse, unlock the flags; this operation always succeeds...
#

unlock_flags:
      pushl %ebp                 # save %ebp
      movl %esp, %ebp            # and overwrite it with the new stack pointer
      subl $4, %esp              # make room for 1 local argument
      movl $0, -4(%ebp)          # set count to 0

uf_spinlock:
      movl $0x80000000, %eax     # load %eax with our 'LOCK IN PROGRESS' identifier
      movl 8(%ebp), %edx         # load %edx with our 'address' argument
      xchgl %eax, (%edx)         # atomically exchange contents of %eax with '*address'
      testl $0x80000000, %eax    # see if our 'LOCK IN PROGRESS' bit is set
      je uf_critical             # that is not the case, we are in the critical section, %eax contains the current flags
      pushl -4(%ebp)             # push current count on stack
      call x_spinlock_backoff    # call our custom backoff function
      addl $4, %esp              # restore stack pointer to previous state
      incl -4(%ebp)              # increment our lock count
      jmp uf_spinlock            # and try again...

uf_critical:                     # here we have entered the critical section
      movl 12(%ebp), %ecx        # (CS) load %ecx with our flags words
      notl %ecx                  # (CS) invert flags in %ecx
      andl $0x7fffffff, %ecx     # (CS) reset highest bit
      andl %ecx, %eax            # (CS) AND the passed flags with its inverted version, leave result in %eax
      movl %eax, (%edx)          # (CS) write back the result, leaving the critical section

uf_epilogue:
      movl %ebp, %esp            # restore previous stack pointer
      popl %ebp                  # restore %ebp saved at the entry
      ret                        # and return to our caller

# -----------------------------------------------------------------------------------------------------------------------

.text
.align 4
.global peek_flags

#
# w_flags peek_flags(void *address);
#
# Lock the flags, read the current status of the flags, unlock them and
# return the the 'current' status of the flags as the result. Note that at the
# time that the result is back at the caller function, the flags could have
# changed allready, so *DONT* use this to check a flag and then try to set
# it without going through the lock_flags function call.
#

peek_flags:
      pushl %ebp                 # save %ebp
      movl %esp, %ebp            # and overwrite it with the new stack pointer
      subl $4, %esp              # make room for 1 local argument
      movl $0, -4(%ebp)          # set count to 0

pf_spinlock:
      movl $0x80000000, %eax     # load %eax with our 'LOCK IN PROGRESS' identifier
      movl 8(%ebp), %edx         # load %edx with our 'address' argument
      xchgl %eax, (%edx)         # atomically exchange contents of %eax with '*address'
      testl $0x80000000, %eax    # see if our 'LOCK IN PROGRESS' bit is set
      je pf_critical             # that is not the case, we are in the critical section, %eax contains the current flags
      pushl -4(%ebp)             # push current count on stack
      call x_spinlock_backoff    # call our custom backoff function
      addl $4, %esp              # restore stack pointer to previous state
      incl -4(%ebp)              # increment our lock count
      jmp pf_spinlock            # and try again...

pf_critical:                     # here we have entered the critical section
      movl %eax, (%edx)          # (CS) write back old value, leaving the critical section and result in %eax

pf_epilogue:
      movl %ebp, %esp            # restore previous stack pointer
      popl %ebp                  # restore %ebp saved at the entry
      ret                        # and return to our caller

#
# x_boolean x_enter_atomic(void * address, void * contents) {
#     %eax                    8(%ebp)          12(%ebp)
#

.text
.align 4
.global xslurp_enter_atomic

# is not being used
xslurp_enter_atomic:
      pushl %ebp                 # save %ebp
      movl %esp, %ebp            # and overwrite it with the new stack pointer
      movl $1, %eax              # Load the 'looking value' in %eax
      movl 8(%ebp), %edx         # load %edx with our 'address' argument
      xchgl %eax, (%edx)         # Atomically exchange the contents and lock others out
      cmpl $1, %eax              # Is anyone else trying to look ?
      je ea_false                # Yes, jump to return false
      cmpl $0, %eax              # No one was looking, is the content 0 ?
      jne ea_restore             # No, jump to replace old contents
      movl 12(%ebp), %eax        # Yes, so we put our new contents in
      movl %eax, (%edx)          # and write them back
      movl $1, %eax              # Prepare 'true' return value
      jmp ea_exit                # and exit

  ea_restore:
      movl %eax, (%edx)          # Restore old contents

  ea_false:
      xorl %eax, %eax            # Prepare 'false' return value

  ea_exit:
      movl %ebp, %esp            # restore previous stack pointer
      popl %ebp                  # restore %ebp saved at the entry
      ret                        # and return to our caller
