/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: monitor.c,v 1.5 2002/01/10 10:28:06 dumon Exp $
*/

//#include <monitor.h>
#include <oswald.h>

x_status x_monitor_create(x_monitor monitor) {

  monitor->owner = NULL;
  monitor->n_waiting = 0;
  monitor->l_waiting = NULL;
  monitor->count = 0;
  return x_event_init(&monitor->Event, xe_monitor);

}

x_status x_monitor_delete(x_monitor monitor) {

  x_status status;

  x_enter_critical;
    
  if (monitor->owner != NULL && monitor->owner != thread_current) {
    status = xs_not_owner;
  }
  else {
    if (monitor->owner == thread_current) {  
      /*
      ** Get this event from our list of owned events, undo any possible priority inversion
      ** and destroy the monitor event.
      */
    
      xi_remove_owned_event(thread_current, &monitor->Event);
      monitor->owner = NULL;
      if (thread_current->c_prio != thread_current->a_prio) {
        xi_thread_priority_set(thread_current, xi_find_safe_priority(thread_current));
      }
    }
    status = xi_event_destroy(&monitor->Event);
  }

  x_exit_critical;
  
  return status;

}

/*
** Internal routine to try enter a monitor.
*/

inline static x_status x_monitor_try_enter(x_monitor monitor, x_boolean decrement_competing) {

  
  if (x_event_is_deleted(monitor)) {
    if (decrement_competing) {
      monitor->Event.n_competing--;
    }
    return xs_deleted;
  }

  if (x_event_type_bad(monitor, xe_monitor)) {
    return xs_bad_element;
  }
  
  if (monitor->owner == NULL) {
    xi_add_owned_event(thread_current, &monitor->Event);
    monitor->count = 1;
    monitor->owner = thread_current;
    return xs_success;
  }
  else if (monitor->owner == thread_current) {
    monitor->count += 1;
    return xs_success;
  }
  else {
    return xs_no_instance;
  }

}

x_status x_monitor_enter(x_monitor monitor, x_sleep timeout) {

  x_status status;

  if (x_in_context_critical(timeout)) {
    return xs_bad_context;
  }

  x_enter_critical;

  status = x_monitor_try_enter(monitor, false);
  
  if (status == xs_no_instance) {
    if (timeout) {
      if (thread_current->c_prio < monitor->owner->c_prio) {
        xi_thread_priority_set(monitor->owner, thread_current->c_prio);
      }
      while (timeout && status == xs_no_instance) {
        timeout = x_event_compete_for(&monitor->Event, timeout);
        status = x_monitor_try_enter(monitor, true);
      }
    }
  }
  
  x_exit_critical;

  return status;

}

/*
** An optimized function that is the same as x_monitor_enter(monitor, x_eternal) since
** this is used a lot in the memory implementation.
*/

x_status x_monitor_eternal(x_monitor monitor) {

  x_status status;

  x_enter_critical;

  status = x_monitor_try_enter(monitor, false);
  
  if (status == xs_no_instance) {
    if (thread_current->c_prio < monitor->owner->c_prio) {
      xi_thread_priority_set(monitor->owner, thread_current->c_prio);
    }
    while (status == xs_no_instance) {
      x_event_compete_for(&monitor->Event, x_eternal);
      status = x_monitor_try_enter(monitor, true);
    }
  }
  
  x_exit_critical;

  return status;

}

/*
** Exit a monitor. The current thread yields after exiting a monitor to enforce fairness...
*/

x_status x_monitor_exit(x_monitor monitor) {

  x_enter_critical;

  if (x_event_is_deleted(monitor)) {
    return xs_deleted;
  }

  if (x_event_type_bad(monitor, xe_monitor)) {
    return xs_bad_element;
  }

  if (monitor->owner == NULL || monitor->owner != thread_current) {
    return xs_not_owner;
  }
  
  monitor->count -= 1;

  if (monitor->count == 0) {
    xi_remove_owned_event(thread_current, &monitor->Event);
    monitor->owner = NULL;
    if (thread_current->c_prio != thread_current->a_prio) {
      xi_thread_priority_set(thread_current, xi_find_safe_priority(thread_current));
    }
    x_event_signal(&monitor->Event);
  }

  x_exit_critical;
    
  return xs_success;

}

/*
** Force release of a monitor by another thread. For callbacks.
*/

x_status x_monitor_release(x_monitor monitor) {

  x_enter_critical;

  if (x_event_is_deleted(monitor)) {
    return xs_deleted;
  }

  if (x_event_type_bad(monitor, xe_monitor)) {
    return xs_bad_element;
  }

  if (monitor->owner) {
    xi_remove_owned_event(monitor->owner, &monitor->Event);
    monitor->owner = NULL;
    x_event_signal(&monitor->Event);
  }

  x_exit_critical;
    
  return xs_success;

}

/*
** Add a thread to the list of waiting threads for a monitor ...
**
** Implementation note: if waiting would become a time critical element in Oswald, we could
** implement the walk over the priorities of the thread into a skiplist like structure. Each thread
** would then have not a single list by means of 'l_waiting' but it would have an array of priority
** ranges so that we can skip a number of priority comparisons.
*/

inline static void xi_add_waiting_thread(x_monitor monitor, x_thread thread) {

  x_thread current;
  volatile x_thread * update;
  x_ubyte c_prio;

  thread->waiting_for = monitor;
  monitor->n_waiting += 1;

#ifdef RUNTIME_CHECKS

  if (thread->l_waiting != NULL) {
    loempa(9, "waiting is not null for thread %d.\n", thread->id);
    abort();
  }

#endif
  
  if (monitor->l_waiting == NULL) {
    monitor->l_waiting = thread;
  }
  else {
    update = &monitor->l_waiting;
    c_prio = thread->c_prio;
    for (current = monitor->l_waiting; current; current = current->l_waiting) {
      if (current->c_prio > c_prio) {
        thread->l_waiting = *update;
        *update = thread;
        return;
      }
      update = &current->l_waiting;
    }
    
    /*
    ** If we get here, the return in the for loop was not taken, but there where threads in the 
    ** list. Update the 'update' value, which means, add the thread at the end of the list.
    */
    
    *update = thread;

  }

}

/*
** ... and the reverse; remove a thread from the waiting list of a monitor.
*/

void xi_remove_waiting_thread(x_monitor monitor, x_thread thread) {

  x_thread current;
  volatile x_thread * update = &monitor->l_waiting;

  for (current = monitor->l_waiting; current; current = current->l_waiting) {
    if (current == thread) {
      *update = thread->l_waiting;
      thread->l_waiting = NULL;
      thread->waiting_for = NULL;
      monitor->n_waiting -= 1;
      return;
    }
    update = &current->l_waiting;
  }

#ifdef RUNTIME_CHECKS

  loempa(9, "Thread %d was not on event waiting list !!\n", thread->id);
  abort();

#endif
    
}

/*
** The special compete_for function we use for re-acquiring a monitor event.
** It is special since we don't add the re-acquiring thread in priority order
** in the list, but just insert it at the front. We also don't return nor take
** a timeout value since it will always be x_eternal...
*/

inline static void xi_monitor_compete_for(x_monitor monitor) {

  x_event event = &monitor->Event;
  
  thread_current->competing_for = event;

  thread_current->l_competing = event->l_competing;
  event->l_competing = thread_current;

  event->n_competing++;
  xi_thread_becomes_pending(xi_event_timeout_action, x_eternal, xe_monitor);
  xi_thread_reschedule();

}

/*
** The 'wait' implementation of a monitor.
*/

x_status x_monitor_wait(x_monitor monitor, x_sleep timeout) {

  x_status status;
  
  x_enter_critical;

  /*
  ** Necessary checks ...
  */

  if (x_event_is_deleted(monitor)) {
    return xs_deleted;
  }

  if (x_event_type_bad(monitor, xe_monitor)) {
    return xs_bad_element;
  }

  if (monitor->owner == NULL || monitor->owner != thread_current) {
    return xs_not_owner;
  }

  /*
  ** Record count and release claims ...
  */
  
  thread_current->m_count = monitor->count;
  xi_remove_owned_event(thread_current, &monitor->Event);
  monitor->count = 0;
  monitor->owner = NULL;

  /*
  ** Put ourselves in the waiting list and signal the availability of the monitor to the next
  ** competing thread, if there is one...
  */
  
  xi_thread_becomes_pending(xi_wait_timeout_action, timeout, xt_waiting);
  xi_add_waiting_thread(monitor, thread_current);
  x_event_signal(&monitor->Event);

  /*
  ** OK, we are back from waiting. We were either notified, or the wait count was over or we
  ** where interrupted on this monitor. We reset the waiting list since when we would be the
  ** only thread waiting, we didn't pass through 'remove from waiting list'.
  */

  thread_current->waiting_for = NULL;
  thread_current->l_waiting = NULL;

  status = x_monitor_try_enter(monitor, false);
  
  if (status == xs_no_instance) {
    if (thread_current->c_prio < monitor->owner->c_prio) {
      xi_thread_priority_set(monitor->owner, thread_current->c_prio);
    }
    while (status == xs_no_instance) {
      xi_monitor_compete_for(monitor);
      status = x_monitor_try_enter(monitor, true);
    }
  }

  if (status == xs_success) {
    monitor->count = thread_current->m_count;
  }
  
  x_exit_critical;
  
  return status;
  
}

/*
** The single thread notification function. Make the first thread waiting, ready to run. It will
** be the highest priority thread.
*/

x_status x_monitor_notify(x_monitor monitor) {

  x_thread next;
  
  x_enter_critical;

  if (x_event_is_deleted(monitor)) {
    return xs_deleted;
  }

  if (x_event_type_bad(monitor, xe_monitor)) {
    return xs_bad_element;
  }

  if (monitor->owner == NULL || monitor->owner != thread_current) {
    return xs_not_owner;
  }

  /*
  ** Get the first waiting thread from the list and reschedule...
  */
  
  next = monitor->l_waiting;
  if (next) {
    xi_thread_remove_pending(next);
    monitor->l_waiting = next->l_waiting;
    next->l_waiting = NULL;
    monitor->n_waiting -= 1;
  }

  /*
  ** If the thread removed from the list is the highest ready to run, make it run.
  */

  xi_thread_reschedule();

  x_exit_critical;

  return xs_success;

}

/*
** Notify all waiting threads and empty the waiting list.
*/

x_status x_monitor_notify_all(x_monitor monitor) {

  x_thread waiter;
  x_thread next;

  x_enter_critical;

  if (x_event_type_bad(monitor, xe_monitor)) {
    return xs_bad_element;
  }

  if (x_event_is_deleted(monitor)) {
    return xs_deleted;
  }

  if (monitor->owner == NULL || monitor->owner != thread_current) {
    return xs_not_owner;
  }

  /*
  ** Get all waiting threads from the list in the monitor, out of their pending
  ** status and reset the monitor waiting list and count to NULL ...
  */
    
  for (waiter = monitor->l_waiting; waiter; waiter = next) {
    xi_thread_remove_pending(waiter);
    next = waiter->l_waiting;
    waiter->l_waiting = NULL;
  }

  monitor->l_waiting = NULL;
  monitor->n_waiting = 0;

  /*
  ** ... now see if any of these threads is the highest priority ready thread and if
  ** so, let it run.
  */
  
  xi_thread_reschedule();

  x_exit_critical;
  
  return xs_success;

}

/*
** Remove a waiting thread from a monitor.
**
** Returns: 
**
** xs_success     | when the thread was successfully found in the list of
**                | waiting threads and was removed from the list.
** ---------------+-----------------------------------------------------------
** xs_no_instance | when the thread was not found in the list of waiting
**                | threads.
*/

x_status x_thread_stop_waiting(x_thread thread) {

  x_status status = xs_no_instance;
  
  x_enter_critical;
  
  if (thread->waiting_for) {
    xi_thread_remove_pending(thread);
    xi_remove_waiting_thread(thread->waiting_for, thread);
    status = xs_success;
  }

  x_exit_critical;
  
  return status;
    
}
