###########################################################################
# Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 #
#                                                                         #
# This software is copyrighted by and is the sole property of Acunia N.V. #
# and its licensors, if any. All rights, title, ownership, or other       #
# interests in the software remain the property of Acunia N.V. and its    #
# licensors, if any.                                                      #
#                                                                         #
# This software may only be used in accordance with the corresponding     #
# license agreement. Any unauthorized use, duplication, transmission,     #
#  distribution or disclosure of this software is expressly forbidden.    #
#                                                                         #
# This Copyright notice may not be removed or modified without prior      #
# written consent of Acunia N.V.                                          #
#                                                                         #
# Acunia N.V. reserves the right to modify this software without notice.  #
#                                                                         #
#   Acunia N.V.                                                           #
#   Vanden Tymplestraat 35      info@acunia.com                           #
#   3000 Leuven                 http://www.acunia.com                     #
#   Belgium - EUROPE                                                      #
###########################################################################

#! /usr/bin/awk -f

#
# $Id: unicode.awk,v 1.12 2002/03/19 15:41:31 gray Exp $
#

# This script converts UnicodeData.txt into a set of lookup tables 
# which are convenient to use when implementing the methods of
# java.util.Character.  
#
# The tables used to implement the JDK 1.1 methods isLetter(), isDigit(),
# etc. are output as C declarations:
#
#   Each array 'const w_byte character_class_row_<xy>[]' contains
#   256 bytes, each of which holds information concerning one
#   unicode character.  The high nibble contains the character
#   'class':
#     0   Character not defined (is unassigned)
#     1   Character is lower-case letter
#     2   Character is upper-case letter
#     3   Character is title-case letter
#     4   Character is some other kind of "letter"
#     5   Character is a digit
#     6   Character is whitespace
#     7   Character is formatting
#     8   Character is none of the above (but is defined)
#     >8  reserved, will not be output by this program.
#   For classes 1 and 2 the low nibble contains information
#   to enable upper case to be derived from lower or vice versa:
#     0   There is no corresponding other-case letter
#     i   (0<i<F) Use upper_to_lower_delta[i] : add it to the
#         upper-case to get the lower, or subtract it from the
#         lower-case to get the upper.
#     F   Use to_lower_case[][] or to_upper_case[][] (see below).
#
#   'const int upper_to_lower_delta []'
#     contains the "deltas" which cover most upper-to-lower case
#     conversions (and vice versa).  upper_to_lower_delta[0]
#     holds the number 'n' of deltas, upper_to_lower_delta[1..n] 
#     holds the deltas themselves.  
#
#   'const char *character_class[]' holds 256 pointers,
#   each pointing to 'character_class_row_<xy>[]' for some <xy>.
#   character_class[charcode/256][charcode%256] yields the
#   information byte corresponding to unicode character 'charcode'.
#
#   The two arrays 'const w_char to_lower_case[][]' and
#   'const w_char to_upper_case[][]' each hold a number
#   of pairs of characters, being respectively {upper,lower}
#   and {lower,upper}.  The pairs are stored in ascending
#   order, and are terminated by {0,0}.
#
#   The array 'const w_char to_title_case[][]' hold a number
#   of pairs of characters, each being {upper or lower case,title case}.
#   The pairs are stored in ascending order, and are terminated by {0,0}.
#
#   The methods introduced in JDK 1.2 and later provide for a more 
#   detailed classification of characters, which is not a straightforward
#   refinement of the JDK 1.1 classification.  Therefore we build
#   further tables which enable the general category and numeric 
#   value to be determined.
#
#   Class RuleBasedCollator needs to be able to look up a character's
#   canonical combining class and decomposition (canonical or not),
#   so we provide functions for this too.
#
#   Although the functions to perform these lookups all appear very
#   similar (they are all based on a binary search algorithm), there
#   are differences in the type of data they access (char, int, w_string
#   pointer) and in the result to be returned in the event of a `miss'.
#   Caveat optimisator!
#
# TODO:
#   Enable the UnicodeData.txt file to be filtered according to a list
#   of ``Subsets'': if a Subset is not included, all its characters are
#   treated as unassigned.  The Subsets marked with a `+' below will
#   always be included, even if they are not requested.
#
#   See Annex A of ISO 10646-1:2000.
#    1+  BASIC LATIN                        0020-007E
#    2+  LATIN-1 SUPPLEMENT                 00A0-00FF
#    3   LATIN EXTENDED-A                   0100-017F
#    4   LATIN EXTENDED-B                   0180-024F
#    5   IPA EXTENSIONS                     0250-02AF
#    6   SPACING MODIFIER LETTERS           02B0-02FF
#    7   COMBINING DIACRITICAL MARKS        0300-036F
#    8   BASIC GREEK                        0370-03CF
#    9   GREEK SYMBOLS AND COPTIC           03D0-03FF
#   10   CYRILLIC                           0400-04FF
#   11   ARMENIAN                           0530-058F
#   12   BASIC HEBREW                       05D0-05EA
#   13   HEBREW EXTENDED                    0590-05CF,05EB-05FF
#   14   BASIC ARABIC                       0600-065F
#   15   ARABIC EXTENDED                    0660-06FF
#   16   DEVANGARI                          0900-097F,200C,200D
#   17   BENGALI                            0980-09FF,200C,200D
#   18   GURMUKHI                           0A00-0A7F,200C,200D
#   19   GUJARATI                           0A80-0AFF,200C,200D
#   20   ORIYA                              0B00-0B7F,200C,200D
#   21   TAMIL                              0B80-0BFF,200C,200D
#   22   TELUGU                             0C00-0C7F,200C,200D
#   23   KANNADA                            0C80-0CFF,200C,200D
#   24   MALAYALAM                          0D00-0D7F,200C,200D
#   25   THAI                               0E00-0E7F
#   26   LAO                                0E80-0EFF
#   27   BASIC GEORGIAN                     10D0-10FF
#   28   GEORGIAN EXTENDED                  10A0-10CF
#   29   HANGUL JAMO                        1100-11FF
#   30   LATIN EXTENDED ADDITIONAL          1E00-1EFF
#   31   GREEK EXTENDED                     1F00-1FFF
#   32+  GENERAL PUNCTUATION                2000-206F
#   33+  SUPERSCRIPTS AND SUBSCRIPTS        2070-209F
#   34+  CURRENCY SYMBOLS                   20A0-20CF
#   35   COMBINIG DIACRITICAL MARKS FOR SYMBOLS
#                                           20D0-20FF
#   36   LETTERLIKE SYMBOLS                 2100-214F
#   37   NUMBER FORMS                       2150-218F
#   38   ARROWS                             2190-21FF
#   39   MATHEMATICAL OPERATORS             2200-2FF
#   40   MISCELLANEOUS TECHNICAL            2300-23FF
#   41   CONTROL PICTURES                   2400-243F
#   42   OPTICAL CHARACTER RECOGNITION      2440-245F
#   43   ENCLOSED ALPHANUMERICS             2460-24FF
#   44   BOX DRAWING                        2500-257F
#   45   BLOCK ELEMENTS                     2580-259F
#   46   GEOMETRIC SHAPES                   25A0-25FF
#   47   MISCELLANEOUS SYMBOLS              2600-26FF
#   48   DINGBATS                           2700-27BF
#   49   CJK SYMBOLS AND PUNCTUATION        3000-303F
#   50   HIRAGANA                           3040-309F
#   51   KATAKANA                           30A0-30FF
#   52   BOPOMOFO                           3100-312F,31A0-31BF
#   53   HANGUL COMPATIBILTY JAMO           3130-318F
#   54   CJK MISCELLANEOUS                  3190-319F
#   55   ENCLOSED CJK LETTERS AND MONTHS    3200-32FF
#   56   CJK COMPATIBILTY                   3300-33FF
#   60   CJK UNIFIED IDEOGRAPHS             4E00-9FFF
#   61   PRIVATE USE AREA                   E000-F8FF
#   62   CJK COMPATIBILTY IDEOGRAPHS        F900-FAFF
#   63   ALPHABETIC PRESENTATION FORMS      FB00-FB4F
#   64   ARABIC PRESENTATION FORMS-A        FB50-FDFF
#   65   COMBINING HALF-MARKS               FE20-FE2F
#   66   CJK COMPATIBILTY FORMS             FE30-FE4F
#   67   SMALL FORM VARIANTS                FE50-FE6F
#   68   ARABIC PRESENTATION FORMS-B        FE70-FEFE
#   69   HALFWIDTH AND FULLWIDTH FORMS      FF00-FFEF
#   70   SPECIALS                           FFF0-FFFD
#   71   HANGUL SYLLABLES                   AC00-D7A3
#   72   BASIC TIBETAN                      0F00-0FBF
#   73   ETHIOPIC                           1200-137F
#   74   UNIFIED CANADIAN ABORIGINAL SYLLABICS
#                                           1400-167F
#   75   CHEROKEE                           13A0-13FF
#   76   YI SYLLABLES                       A000-A48F
#   77   YI RADICALS                        A490-A4CF
#   78   KANGXI RADICALS                    2F00-2FDF
#   79   CJK RADICALS SUPPLEMENT            2E80-2EFF
#   80   BRAILLE PATTERNS                   2800-28FF
#   81   CJK UNIFIED IDEOGRAPHS EXTENSION A 3400-4DBF
#   82   OGHAM                              1680-169F
#   83   RUNIC                              16A0-16FF
#   84   SINHALA                            0D80-0DFF
#   85   SYRIAC                             0700-074F
#   86   THAANA                             0780-07BF
#   87   BASIC MYANMAR                      1000-104F,200C,200D
#   88   KHMER                              1780-17FF,200C,200D
#   89   MONGOLIAN                          1800-18AF
#   90   EXTENDED MYANMAR                   1050-109F
#   91   TIBETAN                            0F00-0FFF
#  200   ZERO-WIDTH BOUNDARY INDICATORS     200B-200D,FEFF
#  201   FORMAT SEPARATORS                  2028-2029
#  202   BI-DIRECVTIONAL FORMAT MARKS       200E-200F
#  203   BI-DIRECTIONAL FORMAT EMBEDDINGS   202A-202E
#  204   HANGUL FILL CHARACTERS             3164,FFA0
#  205   CHARACTER SHAPING SELECTORS        206A-206D
#  206   NUMERIC SHAPE SELECTORS            206E-206F
#  207   IDEOGRAPHIC DESCRIPTION CHARACTERS 2FF0-2FFF
# To this list we add one special value:
#  999   (Wildcard)                         0000-FFFD

# Edit the following function to select the subsets which will be recognised.
# E.g. for Japanese set subsets[49] = subsets[50] = subsets[51] = 1 .

function select_subsets() {
  subsets[999] = 1
}

function supported(codenum) {
  if (codenum < 256 || subsets[999] || codenum >= 0x2000 && codenum < 0x20d0) {
    return 1;
  }

  if (codenum < 0x2000) {
    if (codenum < 0x0900) {
      if (codenum < 0x0400) {
        if (codenum < 0x0300) {
          # Subsets 1, 2 (0x0000..0x0100) are always supported.
          if (codenum < 0x0180) {
            return subsets[3]
          }
          else if (codenum < 0x0250) {
            return subsets[4]
          }
          else if (codenum < 0x02b0) {
            return subsets[5]
          }
          else {
            return subsets[6]
          }
        }
        else { # 0x0300..0x03ff
          if (codenum < 0x0370) {
            return subsets[7]
          }
          else if (codenum < 0x03d0) {
            return subsets[8]
          }
          else {
            return subsets[9]
          }
        }
      }
      else { # 0x0400..0x8ff
        if (codenum < 0x05d0) {
          if (codenum < 0x0500) {
            return subsets[10]
          }
          else if (codenum < 0x0590) {
            return subsets[11]
          }
          else {
            return subsets[13]
          }
        }
        else if (codenum < 0x0600) {
          if (codenum < 0x05eb) {
            return subsets[12]
          }
          else {
            return subsets[13]
          }
        }
        else { # 0x0600..0x08ff
          if (codenum < 0x0660) {
            return subsets[14]
          }
          else if (codenum < 0x0700) {
            return subsets[15]
          }
          else if (codenum < 0x0750) {
            return subsets[85]
          }
          else if (codenum < 0x07c0) {
            return subsets[86]
          }
          else {
            return 0
          }
        }
      }
    }
    else { # 0x0900..0x1fff
      if (codenum < 0x0e00) {
        if (codenum < 0x0a00) {
          if (codenum < 0x0980) {
            return subsets[16]
          }
          else {
            return subsets[17]
          }
        }
        else if (codenum < 0x0b00) {
          if (codenum < 0x0a80) {
            return subsets[18]
          }
          else {
            return subsets[19]
          }
        }
        else if (codenum < 0x0c00) {
          if (codenum < 0x0b80) {
            return subsets[20]
          }
          else {
            return subsets[21]
          }
        }
        else if (codenum < 0x0d00) {
          if (codenum < 0x0c80) {
            return subsets[22]
          }
          else {
            return subsets[23]
          }
        }
        else { # 0x0d00..0x0dff
          if (codenum < 0x0d80) {
            return subsets[24]
          }
          else {
            return subsets[84]
          }
        }
      }
      else { # 0x0e00..0x1fff
        if (codenum < 0x1000) {
          if (codenum < 0x0e80) {
            return subsets[25]
          }
          else if (codenum < 0x0f00) {
            return subsets[26]
          }
          else {
            return subsets[91] || ((codenum < 0x0fc0) && subsets[72])
          }
        }
        else if (codenum < 0x1800) {
          if (codenum < 0x1050) {
            return subsets[87]
          }
          else if (codenum < 0x10a0) {
            return subsets[90]
          }
          else if (codenum < 0x10d0) {
            return subsets[28]
          }
          else if (codenum < 0x1100) {
            return subsets[27]
          }
          else if (codenum < 0x1200) {
            return subsets[29]
          }
          else if (codenum < 0x13a0) {
            return subsets[73]
          }
          else if (codenum < 0x1400) {
            return subsets[75]
          }
          else if (codenum < 0x1680) {
            return subsets[74]
          }
          else if (codenum < 0x16a0) {
            return subsets[82]
          }
          else if (codenum < 0x1700) {
            return subsets[83]
          }
          else if (codenum < 0x1780) {
            return 0;
          }
          else { # 0x1780..0x17ff
            return subsets[88]
          }
        }
        else { # 0x1800..0x1fff
          if (codenum < 0x1e00) {
            return 0
          }
          else if (codenum < 0x1f00) {
            return subsets[30]
          }
          else { # 0x1f00..0x1fff
            return subsets[31]
          }
        }
      }
    }
  }
  else { # 0x2000..
    if (codenum < 0x3000) {
      if (codenum < 0x2400) {
        # Subsets 32--34 (0x2000..0x20cf) are always supported.
        if (codenum < 0x2100) {
          return subsets[35]
        }
        else if (codenum < 0x2150) {
          return subsets[36]
        }
        else if (codenum < 0x2190) {
          return subsets[37]
        }
        else if (codenum < 0x2200) {
          return subsets[38]
        }
        else if (codenum < 0x2300) {
          return subsets[39]
        }
        else {
          return subsets[40]
        }
      }
      else { # 0x2400..0x2fff
        if (codenum < 0x2440) {
          return subsets[41]
        }
        else if (codenum < 0x2460) {
          return subsets[42]
        }
        else if (codenum < 0x2500) {
          return subsets[43]
        }
        else if (codenum < 0x2580) {
          return subsets[44]
        }
        else if (codenum < 0x25a0) {
          return subsets[45]
        }
        else if (codenum < 0x2600) {
          return subsets[46]
        }
        else if (codenum < 0x2700) {
          return subsets[47]
        }
        else if (codenum < 0x27c0) {
          return subsets[48]
        }
        else if (codenum < 0x2800) {
          return 0
        }
        else if (codenum < 0x2900) {
          return subsets[80]
        }
        else if (codenum < 0x2e80) {
          return 0
        }
        else if (codenum < 0x2f00) {
          return subsets[79]
        }
        else if (codenum < 0x2fe0) {
          return subsets[78]
        }
        else if (codenum < 0x2ff0) {
          return 0
        }
        else {
          return subsets[207]
        }
      }
    }
    else { # 0x3000 ..
      if (codenum < 0xe000) {
        if (codenum < 0x3040) {
          return subsets[49]
        }
        else if (codenum < 0x30a0) {
          return subsets[50]
        }
        else if (codenum < 0x3100) {
          return subsets[51]
        }
        else if (codenum < 0x3130) {
          return subsets[52] || ((codenum == 0x3164) && subsets[204])
        }
        else if (codenum < 0x3190) {
          return subsets[53]
        }
        else if (codenum < 0x31a0) {
          return subsets[54]
        }
        else if (codenum < 0x31c0) {
          return subsets[52]
        }
        else if (codenum < 0x3200) {
          return 0
        }
        else if (codenum < 0x3300) {
          return subsets[55]
        }
        else if (codenum < 0x3400) {
          return subsets[56]
        }
        else if (codenum < 0x4e00) {
          return subsets[81]
        }
        else if (codenum < 0xa000) {
          return subsets[60]
        }
        else if (codenum < 0xa490) {
          return subsets[76]
        }
        else if (codenum < 0xa4d0) {
          return subsets[77]
        }
        else {
          return 0
        }
      }
      else { # 0xe000..
        if (codenum < 0xfe00) {
          if (codenum < 0xf900) {
            return subsets[61]
          }
          else if (codenum < 0xfb00) {
            return subsets[62]
          }
          else if (codenum < 0xfb50) {
            return subsets[63]
          }
          else {
            return subsets[64]
          }
        }
        else { #0xfe00..
          if (codenum < 0xfe20) {
            return 0
          }
          else if (codenum < 0xfe30) {
            return subsets[65]
          }
          else if (codenum < 0xfe50) {
            return subsets[66]
          }
          else if (codenum < 0xfe70) {
            return subsets[67]
          }
          else if (codenum < 0xff00) {
            return subsets[68]
          }
          else if (codenum < 0xfff0) {
            return subsets[69] || ((codenum == 0xffa0) && subsets[204])
          }
          else if (codenum < 0xfffe) {
            return subsets[70]
          }
        }
      }
    }
  }

  return 0;
}

# Append an item to a list.  All items except the first are preceded by ``, ''.
# We add tabs and newlines in an attempt to produce something human-readable.
function appendtolist(list,new) {
  itemlen=length(new)+3
  itemsperline=int(80/itemlen)
  if(list=="") return "\t" new
  else if(int((length(list)+itemlen)/80) > int(length(list)/80)) return list",\n\t"new
  else return list",  "new
}

# Optional information provided if 'debug' is defined
# (e.g. run with -v debug=1).

function debugoutput(number,name,class,uce,lce,tce) {
      printf "/* %04X %s */\n",number,name
      debugline=sprintf("class %s (%s) subclass %X ",class,classname[class],subclass)
      if(uce) debugline=debugline sprintf("upper is %s ",uce)
      if(lce) debugline=debugline sprintf("lower is %s ",lce)
      if(tce) debugline=debugline sprintf("title is %s ",tce)
      printf "/* %s */\n",debugline
    }

# Build up the classtable as rows of 256 bytes each, where each byte holds
# a class (high nibble) and subclass (low nibble).  On completing each
# row, check to see if it's a duplicate of one we made earlier; if it
# is, re-use the earlier row.

function buildtableentry(n) {
    currentrow=appendtolist(currentrow,sprintf("0x%x%x",class,subclass))

    m=n%256
    if(m==255) {
      r=(n-m)/256;

      dup=0
      for(s=0;s<r;++s) if(row[s]==currentrow) { dup=1; break }

      if(dup) {
        table[r]=s
      } else {
        row[r]=currentrow
        table[r]=r
      }
      currentrow=""
    }
}

BEGIN {
  FS=";"
  select_subsets()

# debug=1;

# Initialise numcompats to 0, otherwise it we end up generating a subscript`[]'.
# But we initialize numcats to 1, so that we can tell that the first entry to
# be added to catindex is there (otherwise it would be 0, which looks like
# no entry).  Later we will use category 0 to mean `unassigned character code'.
# Hack hack hack ...
  numcompats = 0
  numcats = 1

# Fields of UnicodeData.txt
# Note that these are 1 greater than in the Unicode Standard,
# because awk numbers fields starting from 1.

  codefield=1
  namefield=2
  catfield=3
  combifield=4
  decompfield=6
  digitvalfield=7
  numericvalfield=9
  ucefield=13  
  lcefield=14
  tcefield=15

# Our character classes 

  class_undefined = 0
  class_lowercase = 1
  class_uppercase = 2
  class_titlecase = 3
  class_other_letter = 4
  class_digit = 5
  class_whitespace = 6
  class_format = 7
  class_miscellaneous = 8

# Names of our character classes, for debugging purposes

  classname[class_undefined]="undefined"
  classname[class_lowercase]="lowercase"
  classname[class_uppercase]="uppercase"
  classname[class_titlecase]="titlecase"
  classname[class_mod_letter]="mod.lettr"
  classname[class_other_letter]="misclettr"
  classname[class_digit]="digit    "
  classname[class_whitespace]="whitespce"
  classname[class_miscellaneous]="other    "

# Some common differences between uppercase and lowercase forms
# (lowercase = uppercase + delta)

  delta[1]=1
  delta[2]=2
  delta[3]=16
  delta[4]=32
  delta[5]=48
  delta[6]=26
  delta[7]=80
  delta[8]=-112
  delta[9]=-8
  numdeltas=9
  
# Emit our #include's

  print "#include \"wonka.h\""
  print "#include \"hashtable.h\""
  print "#include \"strings.h\""
  print "#include \"unicode.h\""
  print " "

  print "/* " ENVIRON["UNICODE_SUBSETS"] " */"
}

{
# We force 'code' to be treated as a string, by prefixing "" 
# (otherwise e.g. 2000 would be treated as a decimal number).

  code=""$codefield

# Handle a gap in the sequence of codes.  There are two kinds of gap:
# empty ones and full ones.  Empty gaps represent unassigned codes,
# full ones represent a block of characters with the same name and 
# properties.  

  if(blockname) {
    name=blockname # class and subclass remain same throughout block
  }
  else {
    name=""
    class=class_undefined; subclass=0
  }

  while(sprintf("%04X",number)!=code) {

    if(debug&&class) debugoutput(number,name,class);

    buildtableentry(number);

    ++number
  }

# If not in a supported subset, build a (null) table entry and skip the rest.

  if(!supported(number)) {
    buildtableentry(number);

    next;

  }

# First challenge: deal with the blocks.  A block appears in the table as 
# two entries named <foo, First> and <foo, Last>.  We treat every entry 
# from first to last as if it were named <foo>.

  startblock=index($namefield,", First");
  stopblock=index($namefield,", Last");

  if(startblock) blockname=substr($namefield,1,startblock-1) substr(name,startblock+7)

  if(blockname) name=blockname
  else name=$namefield

  if(stopblock) blockname=""

# Some hacks apparently required by the Java spec (but not by ISO 10646).

  if(code=="0100") forceLetterOrDigit=1
  else if(code=="2000") {
                        forceLetterOrDigit=0
  }
  else if(code=="3000") forbidLetterOrDigit=0

# Get the digit and/or numerical values and the combining class, if nonzero. 
# Build lists in ascending order of character code.

  digitval=""$digitvalfield;
  if (digitval != "") {
    digitvalues=appendtolist(digitvalues,"{0x"code","digitval"}")
    ++numdigitvalues
  }

  numericval=""$numericvalfield;
  if (numericval != "") {
    if (index(numericval,"/") || index(numericval,"-")) {
      numericval = "-2"
    }
    numericvalues=appendtolist(numericvalues,"{0x"code","numericval"}")
    ++numnumericvalues
  }

  if ($combifield) {
    combiclass[number] = $combifield
    combilist=appendtolist(combilist,"{0x"code","$combifield"}")
    ++numcombis
  }

  category = $catfield;
  uce=""$ucefield
  lce=""$lcefield
  tce=""$tcefield

  if (!catindex[category]) {
    catindex[category] = numcats;
    ++numcats
  }

  if (category != prevcat) {
    catlist = appendtolist(catlist,"{0x"code","catindex[category]"}")
    prevcat = category;
    ++numcatpairs
  }

  class=class_undefined; subclass=0;

# Let's collect some useful evidence from the unicode name of the character.

  smallName=index(name,"SMALL LETTER") || index(name,"SMALL LIGATURE");
  capsName=index(name,"CAPITAL LETTER") || index(name,"CAPITAL LIGATURE");
  digitName=index(name,"DIGIT");

# More useful evidence from the general category code of the character.

  letterCat = index(category,"L") == 1;
  digitCat = index(category,"Nd") == 1;
  formatCat = index(category,"Cf") == 1;
  spaceCat = index(category,"Z") == 1;

# Even the decomposition field can be of interest ..
  nobreakDecomp = index($decompfield,"<noBreak>") == 1;

# This is where the classification proper begins.

# If no lower-case equivalent, but has upper-case equivalent or has a small-
# sounding name, classify as "lower-case letter".  The subclass indicates
# how the upper-case equivalent is to be derived:
#   0   don't bother, there ain't none
#   n,0<n<F
#       subtract delta[n]
#   F   look it up in a table wot we will output later

  if(!lce && (uce || smallName)) {
    class=class_lowercase
    for(d in delta) if(sprintf("%04X",number-delta[d])==uce) subclass=d

    if(subclass==0 && uce) {
      subclass=15;
      ucespecial=appendtolist(ucespecial,"{0x"code",0x"uce"}")
    }
  }

# If no upper-case equivalent, but has lower-case equivalent or has a large-
# sounding name, classify as "upper-case letter".  The subclass indicates
# how the lower-case equivalent is to be derived:
#   0   don't bother, there ain't none
#   n,0<n<F
#       add delta[n]
#   F   look it up in a table what we will output later

  else if(!uce && (lce || capsName)) {
    class=class_uppercase
    for(d in delta) if(sprintf("%04X",number+delta[d])==lce) subclass=d

    if(subclass==0 && lce) {
      subclass=15;
      lcespecial=appendtolist(lcespecial,"{0x"code",0x"lce"}")
    }
  }

# If both lower- and  upper-case equivalents exist, this must be "title case".
# Converting to lower- or upper-case is always a table lookup job.

  else if(uce && lce) {
    class=class_titlecase;
    subclass=15;
    lcespecial=appendtolist(lcespecial,"{0x"code",0x"lce"}")
    ucespecial=appendtolist(ucespecial,"{0x"code",0x"uce"}")
  }

# In JDK 1.2 and 1.3, a digit is a digit if it has "DIGIT" in its name.
# In JDK 1.4 it's a digit if its general category is "Nd".
# 1.2/1.3:
  else if (digitName) {
# 1.4:
# else if(digitCat) {
	class=class_digit;
        subclass=digitval
  }

  else if (formatCat) {
	class=class_format;
  }

  else if ((number >= 9 && number <=13) || (number >= 28 && number <= 31) || (spaceCat && !nobreakDecomp)) {
        class=class_whitespace;
  }

# Anything not yet classified is either a "miscellaneous letter" or "other".

  if(class==class_undefined && name) {
    if(letterCat || forceLetterOrDigit ) {
        class=class_other_letter;
    } else class=class_miscellaneous;
  }

# If a title-case equivalent is specified and it is different to the
# upper case, add it to our little collection.

  if(tce && tce!=uce) {
      tcespecial=appendtolist(tcespecial,"{0x"code",0x"tce"}")
    }

  if(debug&&class) debugoutput(number,name,class,uce,lce,tce);

  buildtableentry(number);

# If this character has a decomposition, make a note of this.

  if ($decompfield) {
    raw = $decompfield
    rawdecomp[number] = raw
    n = split(raw, decomp, " ")
    i = 1
    j = index(decomp[1],"<")
    if (j) {
      k = index(decomp[1],">")
      comptype = substr(decomp[1], j+1, k-2)
      compatnames[comptype] = "string_" comptype
      compatlist = compatlist "\n  compatible[" numcompats "].from = 0x" code ";"
      compatlist = compatlist "\n  compatible[" numcompats "].to = string_" comptype ";"
      ++numcompats
      ++i
    }

    thisdecomp = sprintf("static w_char decomp_%s[] = {%d",code,n-i+1)
    for (; i<=n; ++i) {
      thisdecomp = thisdecomp sprintf(", 0x%s",""decomp[i])
    }
    thisdecomp = thisdecomp "};"
    decomplist = decomplist "\n" thisdecomp
    decopairlist=appendtolist(decopairlist,"{0x"code", decomp_"code"}")
  }

# Proceed with the next character

  ++number

}

END{

# There is a gap at the end, for the two non-characters 0xFFFE and 0xFFFF.

  while(number!=65536) {

    if(debug&&class) debugoutput(number,name,0);

    buildtableentry(number);

    ++number
  }

  for (category in catindex) {
    print "#define CATEGORY_" category " " catindex[category]
  }
  print "#define NUMBER_OF_CATEGORIES " numcats
  print "static w_string category_name[NUMBER_OF_CATEGORIES];"
  print ""

# Create the list of mappings from code to digit value, and a function
# to search the list.  The function takes an int rather than a w_char
# just to avoid spurious warnings from gcc.
#
# Note: For the time being we don't do this, because the old 1.1 code
# seems to work fine.
#
#  printf "const w_char_conversion to_digit_value[] = {\n\t"
#  printf digitvalues
#  printf ", {0xffff, 0xffff},{0xffff, 0xffff}\n};\n"
#  print  ""
#  print  "w_int charToDigitValue(w_char ch) {"
#  print  "  w_int min = 0;"
#  print  "  w_int max = " numdigitvalues - 1 ";"
#  print  "  w_int try = " numdigitvalues " / 2;"
#  print  "  w_int next;"
#  print  "  while (try != min && try != max) {"
#  print  "  if (to_digit_value[try].from > ch) {"
#  print  "      next = (try + min) / 2;"
#  print  "      max = try;"
#  print  "      try = next;"
#  print  "    }"
#  print  "  else if (to_digit_value[try].from < ch) {"
#  print  "      next = (try + max + 1) / 2;"
#  print  "      min = try;"
#  print  "      try = next;"
#  print  "    }"
#  print  "    else {"
#  print  "      break;"
#  print  "    }"
#  print  "  }"
#  print  "  if (to_digit_value[try].from == ch) {"
#  print  "    woempa(1, \"Looking for %04X, found it at %d\\n\", ch, try);"
#  print  "    return to_digit_value[try].to;"
#  print  "  }"
#  print  "  woempa(1, \"Looking for %04X, didn't find it\\n\", ch);"
#  print  "  return -1;"
#  print  "}"
#  print  ""

# Create the list of mappings from code to numeric value, and a function
# to search the list.  The function takes an int rather than a w_char
# just to avoid spurious warnings from gcc.

  printf "const w_char_to_number to_numeric_value[] = {\n\t"
  printf numericvalues
  printf ", {0xffff, 0},{0xffff, 0}\n};\n"
  print  ""
  print  "w_int charToNumericValue(w_char ch) {"
  print  "  w_int min = 0;"
  print  "  w_int max = " numnumericvalues - 1 ";"
  print  "  w_int try = " numnumericvalues " / 2;"
  print  "  w_int next;"
  print  "  while (try != min && try != max) {"
  print  "  if (to_numeric_value[try].from > ch) {"
  print  "      next = (try + min) / 2;"
  print  "      max = try;"
  print  "      try = next;"
  print  "    }"
  print  "  else if (to_numeric_value[try].from < ch) {"
  print  "      next = (try + max + 1) / 2;"
  print  "      min = try;"
  print  "      try = next;"
  print  "    }"
  print  "    else {"
  print  "      break;"
  print  "    }"
  print  "  }"
  print  "  if (to_numeric_value[try].from == ch) {"
  print  "    woempa(1, \"Looking for %04X, found it at %d\\n\", ch, try);"
  print  "    return to_numeric_value[try].to;"
  print  "  }"
  print  "  woempa(1, \"Looking for %04X, didn't find it\\n\", ch);"
  print  "  return -1;"
  print  "}"
  print  ""

# Create a list of mappings from code value onto canonical combining class, 
# and a function to search the list.

#  combilist=appendtolist(combilist,"{0,0}")
  printf "const w_char_to_number to_combining_class[] = {\n\t"
  printf combilist
  printf "\n};\n\n"
  print  ""
  print  "w_int charToCombiningClass(w_char ch) {"
  print  "  w_int min = 0;"
  print  "  w_int max = " numcombis - 1 ";"
  print  "  w_int try = (" numcombis " + 0) / 2;"
  print  "  w_int next;"
  print  "  while (try != min && try != max) {"
  print  "  if (to_combining_class[try].from > ch) {"
  print  "      next = (try + min) / 2;"
  print  "      max = try;"
  print  "      try = next;"
  print  "    }"
  print  "  else if (to_combining_class[try].from < ch) {"
  print  "      next = (try + max + 1) / 2;"
  print  "      min = try;"
  print  "      try = next;"
  print  "    }"
  print  "    else {"
  print  "      break;"
  print  "    }"
  print  "  }"
  print  "  if (to_combining_class[try].from == ch) {"
  print  "    woempa(1, \"Looking for %04X, found it at %d\\n\", ch, try);"
  print  "    return to_combining_class[try].to;"
  print  "  }"
  print  "  woempa(1, \"Looking for %04X, didn't find it\\n\", ch);"
  print  "  return 0;"
  print  "}"
  print  ""

# Create the list of mappings from code to decomposition, and a function
# to search the list.  Both canonical and compatibility decompositions
# are included.

  print  decomplist
  print  ""
  print  "const w_char_decomposition decomposition[] = {"
  print  "\t" decopairlist
  print  "};"
  print  ""
  print  "w_char* charToDecomposition(w_char ch) {"
  print  "  w_int min = 0;"
  print  "  w_int max = " numdigitvalues - 1 ";"
  print  "  w_int try = " numdigitvalues " / 2;"
  print  "  w_int next;"
  print  "  while (try != min && try != max) {"
  print  "  if (decomposition[try].from > ch) {"
  print  "      next = (try + min) / 2;"
  print  "      max = try;"
  print  "      try = next;"
  print  "    }"
  print  "  else if (decomposition[try].from < ch) {"
  print  "      next = (try + max + 1) / 2;"
  print  "      min = try;"
  print  "      try = next;"
  print  "    }"
  print  "    else {"
  print  "      break;"
  print  "    }"
  print  "  }"
  print  "  if (decomposition[try].from == ch) {"
  print  "    woempa(1, \"Looking for %04X, found it at %d\\n\", ch, try);"
  print  "    return decomposition[try].to;"
  print  "  }"
  print  "  woempa(1, \"Looking for %04X, didn't find it\\n\", ch);"
  print  "  return NULL;"
  print  "}"
  print  ""

# Define the array which maps character codes onto compatibility types,
# and a function to search it.

  print "static w_char_to_string compatible[" numcompats "];"
  print  ""
  print  "void createDecompositionTables() {"
  for (cname in compatnames) {
    print "  w_string string_" cname " = cstring2String(W_Thread_system, \"" cname "\", " length(cname) ");"
  }
  print  compatlist
  print  ""
  print  "}"
  print  ""
  print  "void createCharacterTables() {"
  print "  category_name[0] = cstring2String(W_Thread_system, \"--\", 2);"
  for (category in catindex) {
    print "  category_name[CATEGORY_" category "] = cstring2String(W_Thread_system, \"" category "\", " length(category) ");"
  }
  print  "}"
  print  ""

  print  "w_string charToCompatibilityType(w_char ch) {"
  print  "  w_int min = 0;"
  print  "  w_int max = " numcompats - 1 ";"
  print  "  w_int try = " numcompats " / 2;"
  print  "  w_int next;"
  print  "  while (try != min && try != max) {"
  print  "  if (compatible[try].from > ch) {"
  print  "      next = (try + min) / 2;"
  print  "      max = try;"
  print  "      try = next;"
  print  "    }"
  print  "  else if (compatible[try].from < ch) {"
  print  "      next = (try + max + 1) / 2;"
  print  "      min = try;"
  print  "      try = next;"
  print  "    }"
  print  "    else {"
  print  "      break;"
  print  "    }"
  print  "  }"
  print  "  if (compatible[try].from == ch) {"
  print  "    woempa(1, \"Looking for %04X, found it at %d\\n\", ch, try);"
  print  "    return compatible[try].to;"
  print  "  }"
  print  "  woempa(1, \"Looking for %04X, didn't find it\\n\", ch);"
  print  "  return NULL;"
  print  "}"
  print  ""

# Create a list of mappings from code value onto general category, 
# and a function to search the list.  We list only the characters
# which have a different category from their predecessor.

  printf "const w_char_conversion to_category[] = {\n\t"
  printf catlist
  printf ", {0xFFFE,0}\n};\n\n"
  print  ""
  print  "w_string charToCategory(w_char ch) {"
  print  "  w_int min = 0;"
  print  "  w_int max = " numcatpairs + 1";"
  print  "  w_int try = " numcatpairs " / 2;"
  print  "  w_int next;"
  print  ""
  print  "  if (!CHAR_CLASS(ch)) {"
  print  ""
  print  "  return category_name[0];"
  print  ""
  print  "  }"
  print  ""
  print  "  while (try != min && try != max) {"
  print  "  if (to_category[try].from > ch) {"
  print  "      next = (try + min) / 2;"
  print  "      max = try;"
  print  "      try = next;"
  print  "    }"
  print  "  else if (to_category[try].from < ch) {"
  print  "      next = (try + max + 1) / 2;"
  print  "      min = try;"
  print  "      try = next;"
  print  "    }"
  print  "    else {"
  print  "      break;"
  print  "    }"
  print  "  }"
  print  ""
  print  "  if (to_category[try].from == ch) {"
  print  "    woempa(7, \"Looking for %04X, found it (category %d) at %d\\n\", ch, to_category[try].to, try);"
  print  "    woempa(7, \"Returning ``%w''\n\", category_name[to_category[try].to]);"
  print  "  }"
  print  "  else {"
  print  "    woempa(7, \"Looking for %04X, found %04X (%d) at %d\\n\", ch, to_category[min].from, to_category[min].to, min);"
  print  "    woempa(7, \"Returning ``%w''\n\", category_name[to_category[min].to]);"
  print  "  }"
  print  ""
  print  "  return category_name[to_category[to_category[try].from == ch ? try : min].to];"
  print  "}"
  print  ""

# Print out all the (non-duplicate) rows of the classification table:

  for(r=0;r<256;++r) if(row[r]) {
      printf "static const w_byte character_class_row_%02x[] = {\n",r
      print "\t"row[r]
      print "};\n\n"
  }

# Now the table of rows:

  printf "const w_byte * character_class[] = {\n"
  for(r=0;r<256;++r) printf "\tcharacter_class_row_%02x, \n",table[r] 
  printf "};\n\n"

# Now for the little table of deltas (preceded by a count and 
# terminated by a zero):

  printf "const int upper_to_lower_delta[] = { %d",numdeltas
  for(i=0;i<numdeltas;++i) printf ", %d",delta[i+1];
  printf ", 0};\n\n"

# Finally the three tables of special case conversions.

  ucespecialt=appendtolist(ucespecialt,"{0,0}")
  printf "const w_char_conversion to_upper_case[] = {\n\t"
  printf ucespecial
  printf "\n};\n"

  lcespecialt=appendtolist(lcespecialt,"{0,0}")
  printf "const w_char_conversion to_lower_case[] = {\n\t"
  printf lcespecial
  printf "\n};\n"

  tcespecialt=appendtolist(tcespecialt,"{0,0}")
  printf "const w_char_conversion to_title_case[] = {\n\t"
  printf tcespecial
  printf "\n};\n"

}
