#ifndef _FIFO_H
#define _FIFO_H

/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: fifo.h,v 1.3 2002/02/20 08:49:05 dfrederickx Exp $
*/

#include "realms.h"
#include "wonka.h"

/** Source file fifo.h
**
** A FIFO is a first-in, first-out queue of pointers (\texttt{void*}).
** New ``leaves'' are automativally added as existing ones become full:
** when a leaf is exhausted (number of puts = number of gets), it is
** moved from the beginning of the FIFO to the end and hence recycled.
*/
//@{

/** Default size for a FIFO leaf.
** Please use a number that is one less then a true binary number
** because we loose a pointer size in a fifoLeaf. e.g. 511, 1023, ... and
** this would give nice sized FifoLeafs.
*/

#define FIFO_DEFAULT_LEAF_SIZE   511

/// FIFO leaf structure.  The leaves of a FIFO form a simple singly-linked list.
typedef struct FifoLeaf {
  struct FifoLeaf *next;
  void *data[1];
} FifoLeaf;

/** FIFO header structure.
*/
//@{
typedef struct w_Fifo {
/// index into putFifoLeaf where the next put will occur (postincrementingly).
  w_size putFifoIndex;
/// index into getFifoLeaf where the next get will occur (postincrementingly).
  w_size getFifoIndex;
/// size with which each new leaf will be allocated.
  w_size leafSize;
/// capacity of each leaf (in words)
  w_size leafElements;
// number of elements currently in use (put and not yet got)
  w_size numElements;
// the currenty leaf for getting
  FifoLeaf *getFifoLeaf;
// the currenty leaf for putting
  FifoLeaf *putFifoLeaf;
/// the realm which ``owns'' this FIFO (all memory allocation will be carried out in the name of this realm).
  w_realm realm;
} w_Fifo;
//@}

#define FIFO_SUCCESS   0
#define FIFO_NO_MEMORY 1

/// Allocate a new FIFO
// Each leaf will contain \texttt{leafElements} elements.
// For speed, choose \texttt{leafElements} to be one less than a power of two.
w_fifo allocFifo(w_realm realm, w_size leafElements);

/// Release a FIFO
void releaseFifo(w_fifo f);

/** Contract a FIFO (release exhausted leaves).
** @return the number of bytes of memory which were freed.
*/
int contractFifo(w_fifo f);

/** Append an element e to FIFO f
 ** Returns 0 if succeeded, <0 if not (e.g. no memory for new leaf).
 */
w_int putFifo(void *e, w_fifo f);

/// Remove the front element of FIFO f
void *getFifo(w_fifo f);

//@}
#endif /* _FIFO_H */
