#ifndef _HEAP_H
#define _HEAP_H

#include "jni.h"
#include "oswald.h"

/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: heap.h,v 1.11 2002/02/14 09:58:16 gray Exp $
*/

/**
 
  This file defines the API towards the object store.

  The object store is divided into two regions: a transient object store
  which exists only in RAM, and a persistent object store which is backed
  up onto some persistent medium (flash, disk, drum, mercury delay line...).
  The two regions have slightly different object headers and use different
  low-level storage allocation strategies, but these subtleties are hidden
  from the user by the magic herein contained.
 */

/**
  Object flags word

  These flags are present at the same position (1 word before the start of
  the instance fields) in both transient and persistent objects.

  1. Object "colours" used by the mark-and-sweep algorithm.  Note that the
  position of the first four flags is essential to the correct working
  of the isMarked() funtion of collector.c .
*/
// A FINALIZE_BLACK object has been marked as reachable only becuse it is referred to by objects which are finalizable.
#define O_FINALIZE_BLACK     0x00000001
// A PHANTOM_BLACK object has been marked as reachable by phantom references only.
#define O_PHANTOM_BLACK      0x00000002
// A WEAK_BLACK object has been marked as weakly reachable, but not softly or strongly reachable.
#define O_WEAK_BLACK         0x00000004
// A SOFT_BLACK object has been marked as softly reachable, but not strongly reachable.
#define O_SOFT_BLACK         0x00000008
// A BLACK object has been marked as strongly reachable.
#define O_BLACK              0x00000010
// A GRAY object is one to which a new reference has been created, so it needs re-marking
#define O_GRAY               0x00000020 

/**
 2. Immortal objects which should never become garbage 
*/
// A GREEN object has just been created. It is too young to die.
#define O_GREEN              0x00000040 
// A SYSTEM object belongs to the ``system'' (e.g. is needed by a system class).
#define O_SYSTEM             0x00000080 

/**
 3. Finalization etc.
*/
// It is forbidden to create new references to a CONDEMNED object.  Existing references should be nulled. */
#define O_CONDEMNED          0x00000400 
// An object is FINALIZABLE if it has a finalizer which has not yet been called
#define O_FINALIZABLE        0x00001000 
// An object is FINALIZING if it is on the finalizer_fifo or its finalize() methodnqueue() method has been called
#define O_FINALIZING         0x00002000 
// An object is ENQUEUEABLE if it has an enqueue() which has not yet been called
#define O_ENQUEUEABLE        0x00004000 
// An object is ENQUEUEING if it is on the enqueue_fifo or its enqueue() method has been called
#define O_ENQUEUEING         0x00008000 
// An object has WINDOW set if it is subject to analysis by the garbage collector.
#define O_WINDOW             0x00010000
// A TARGET object is eligible for finalization
#define O_TARGET             0x00020000
// A GARBAGE object has become garbage and can be freed. 
#define O_GARBAGE            0x00040000

#define O_CACHEABLE          0x00100000 /* Set when instance can be put in the cache of a clazz. */

/**
 4. Persistent object store support 
*/
// A PERSISTENT object resides in persistent memory
#define O_PERSISTENT         0x00200000
// A LOCAL_COPY is a copy in transient memory of a PERSISTENT object
#define O_LOCAL_COPY         0x00400000
// A DIRTY object is a LOCAL_COPY which has been updated since it was last copied
#define O_DIRTY              0x00800000

/**
 5. Object locking
  If O_HAS_LOCK is set, then the instance is linked to a monitor via lock_hashtable.
*/
#define O_HAS_LOCK           0x80000000

/**
 ** Collective names for related roups of lags
*/

/**
 Inferior versions of BLACKness.
*/
#define O_NEAR_BLACK (O_WEAK_BLACK | O_SOFT_BLACK | O_PHANTOM_BLACK | O_FINALIZE_BLACK)

/**
** An object is considered WHITE if none of these are set.  
*/
#define O_NON_WHITE (O_BLACK | O_NEAR_BLACK | O_GRAY | O_GREEN | O_ENQUEUEING | O_PERSISTENT | O_LOCAL_COPY | O_SYSTEM)


/** ``Magic number'' which identifies an object as an object
*/
#define OBJECT_MADGIC 0xbeafbabe

/**
  Object header structure w_Object
 
  Defines the header of a transient object.
  An instance pointer (w_instance) will point at the first word
  of the `fields' array).
*/
typedef struct w_Object {
// \texttt{w_Clazz} structure describing the class of this object
  w_clazz clazz;
// Object flags as defined above
  w_flags flags;
// the instance data for this object, this is a variable sized array.
  w_word fields[0];
} w_Object;

/** Instance allocation

  Macro for the most common type of _allocInstance.
  If called during system startup, just allocates the instance:
  if it is not made reachable from the roots before the garbage
  collector is started up, bye bye instance.
  If called later, calling thread must be part of the VM (either
  created from Java, or a native thread which attached itself);
  a local reference will be created in the calling Java frame.
  Normally this local reference is destroyed when returning from
  native code to Java, but if that will happen too late (or never),
  you should call releaseLocalReference manually.
*/
#define allocInstance(t, c) _xxxx_allocInstance(t, c)

/** Macro for oddball cases where you need to set extra flags.
*/
#define allocInstanceWithFlags(t, c, f) _xxxx_allocInstanceWithFlags(t, c, f)

extern w_instance (*_xxxx_allocInstance)(w_thread thread, w_clazz clazz);
extern w_instance (*_xxxx_allocInstanceWithFlags)(w_thread thread, w_clazz clazz, w_flags flags);
extern w_instance (*_xxxx_allocArrayInstance)(w_thread thread, w_clazz clazz, w_size size, w_flags flags);

/**
 ** Allocation of array instances
*/
// Base function: allocate using source filename and line number, clazz, dimansions, flags.
w_instance _allocateArrayInstance(w_thread thread, w_clazz clazz, w_int fixedDimensions, w_word lengths[], w_flags flags);
// Convenience macro: get source file and line number automatically
#define allocArrayInstanceWithFlags(t, c, d, l, f) _allocateArrayInstance(t, c, d, l, f)
// Even more convienient (no flags) -- also obeys convention of starting with `alloc'
#define allocArrayInstance(t, c, d, l) _allocateArrayInstance(t, c, d, l, 0)

#ifndef offsetof
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

/** Convert an instance pointer to a pointer to the object header.
*/
inline static w_object _instance2object(const char *f, int l, w_instance instance) {
  w_object object = (w_object)(((unsigned char *)(instance)) - offsetof(w_Object, fields));
  return object;
}

/**
 Macro for the most common type of _instance2object
*/
#define instance2object(i) _instance2object(__FILE__, __LINE__, i)

/**
 Convert an instance pointer to a pointer to the corresponding w_Clazz structure.
*/
w_clazz instance2clazz(w_instance ins);

/**
** instance2flagsptr doesn't check OBJECT_MADGIC, because it also needs
** to work for persistent objects (which have a different header, see below). 
** The only thing we can be sure of is that the flags word is located 8 
** bytes before the instance fields.
*/
inline static w_word* instance2flagsptr(w_instance instance) {
  w_word *flags = (w_word*)(((unsigned char *)(instance)) - offsetof(w_Object, fields) + offsetof(w_Object, flags));
  return flags;
}

/** Convenience macro
*/
#define instance2flags(ins) (*instance2flagsptr(ins))

/** We distinguish four types of references:
 1. The address of a plain ordinary transient instance.
 2. The address of a "local copy" of a persistent instance.
    Such a local copy lives in the same heap as normal transient instances.
 3. The address of (the master copy of) a persistent instance.
    These live in a special area of memory which is periodically backed up
    onto the persistent medium.
 4. A persistent reference.  This is not an absolute address but an offset
    within the persistent memory zone, plus a constant PS_OFFSET which
    enables a persistent reference to be easily distinuished from a real
    memory address.  A persistent reference is therefore independent of
    the location of the persistent memory zone.
*/
#define INSTANCE_IS_TRANSIENT  0
#define INSTANCE_IS_LOCAL_COPY 1
#define INSTANCE_IS_PERSISTENT 2
#define PERSISTENT_REFERENCE   3

#ifdef ENABLE_PS
int typeOfInstance(w_instance);
#else
#define typeOfInstance(instance) INSTANCE_IS_TRANSIENT
#endif

/** PS_OFFSET is the ammount which is added to the offset of a persistent
 ** instance relative to the start of the persistent area in order to make
 ** a persistent reference.  Choose a value which will clearly distinguish
 ** persistent references from normal memory addresses.  Any reference
 ** higher than (or equal to) PS_OFFSET will be treated as a persistent
 ** reference.
 */
#ifdef ENABLE_PS
#define PS_OFFSET 0xe0000000ULL

#define isPersistentRef(ref) (((w_word)ref) >= PS_OFFSET)

#define isLocalCopy(instance) (typeOfInstance(instance) == INSTANCE_IS_LOCAL_COPY)

w_instance createPersistentInstance(w_instance ref);
#else
#define isPersistentRef(ref)  WONKA_FALSE
#define isLocalCopy(instance) WONKA_FALSE
#endif

/** Start of the persistent memory zone.  (The area of memory which is backed
 ** up on non-volatile memory).
 */
extern char *PS_start;
/** End of the persistent memory zone.  
 ** The persistent memory zone includes *PS_start but not *PS_end.
 */
extern char *PS_end;

#ifdef ENABLE_PS
/** Convert a persistent reference to the address of a persistent instance.
*/
static inline w_instance ref2pinstance(void *ref) {
  return ref ? (w_instance)(PS_start + ((char*)ref - (char*)PS_OFFSET)) : NULL;
}

/** Convert the address of a persistent instance to a persistent reference.
*/
static inline w_instance pinstance2ref(void *ref) {
  return ref ? (w_instance)((char*)PS_OFFSET + ((char*)ref - PS_start)) : NULL;
}
#endif

extern w_int F_Array_data;

void newReferenceTo(w_instance child);

/*
** markLocalCopy and markPersistentRef are called by the transient garbage
** collector (TGC) when it encounters a reference to a persistent object
** or a local copy of one, respectively, during its "mark" phase.  For a 
** persistent instance there is no further processing for the TGC to do;
** for a local copy, if markLocalCopy returns WONKA_TRUE then the TGC
** should continue to process the object normally.
**
** releaseLocalCopy is called when the TGC would like to release the
** memory used by a local copy.  If it returns WONKA_TRUE then TGC
** will go right ahead, otherwise it will wait and (probably) try
** again later.
**
** besmirchLocalCopy marks a local copy as `dirty'.
*/

void markPersistentRef(w_instance);
void markPersistentInstance(w_instance);
w_boolean markLocalCopy(w_instance);
w_boolean releaseLocalCopy(w_instance);
void besmirchLocalCopy(w_instance tinstance);

/** To set a reference field of any object, always use this function (acts as a ``write barrier'')
*/
static inline void setReferenceField(w_instance parent, w_instance child, w_int slot) {
  if (child == NULL) {
    parent[slot] = 0;

    return;

  }

#ifdef ENABLE_PS
  {
    w_flags flags = instance2flags(parent);

    if (isSet(flags, O_PERSISTENT)) {
      wabort(ABORT_WONKA, "Instance %p is persistent instance, not allowed to call setReferenceField\n", parent);
    }

    if (isLocalCopy(child)) {
      woempa(1,"  --> markLocalCopy(child)\n");
      markLocalCopy(child);
    }

    if (isLocalCopy(parent)) {
      parent[slot] = (w_word)createPersistentInstance(child);
    }
    else {
      parent[slot] = (w_word)(child);
    }
  }
#else
  parent[slot] = (w_word)(child);
#endif

  newReferenceTo(child);
}


/** Similarly, to set an element of a reference array, always use this function.
*/
static inline void setArrayReferenceField(w_instance parent, w_instance child, w_int slot) {
  if (child == NULL) {
    parent[F_Array_data + slot] = 0;

    return;

  }

#ifdef ENABLE_PS
  {
    w_flags flags = instance2flags(parent);

    if (isSet(flags, O_PERSISTENT)) {
      wabort(ABORT_WONKA, "Array %p is persistent instance, not allowed to call setArrayReferenceField\n", parent);
    }

    if (isLocalCopy(child)) {
      woempa(1,"  --> markLocalCopy(child)\n");
      markLocalCopy(child);
    }

    if (isLocalCopy(parent)) {
      parent[F_Array_data + slot] = (w_word)createPersistentInstance(child);
    }
    else {
      parent[F_Array_data + slot] = (w_word)(child);
    }
  }
#else
  parent[F_Array_data + slot] = (w_word)(child);
#endif

  newReferenceTo(child);
}


/** Create a new global reference to an instance.
*/
void newGlobalReference(w_instance instance);

/** Get the "local copy" of a persistent instance. All accesses, including 
** lock operations, should take place in the local copy.
** \texttt{pinstance} must be non-null and really persistent.
** If the local copy does not already exist then it will be created.
*/
w_instance pinstance2local_copy(w_instance pinstance);

/**
** Ditto, but starting from a persistent reference.
*/
w_instance ref2local_copy(w_instance pinstance);

/**
 ** Starting with a local copy, get the address of the persistent
 ** instance of which it is a copy.
 */
w_instance local_copy2pinstance(w_instance pinstance);

/*
** Overall structure of a Pobject:
**
**  +----------------+
**  | to be defined  |
**  +----------------+
**  ! NULL or ptr --------> local copy (w_object)
**  +----------------+
**  ! hashcode       ! (copied from the transient object at the moment it 
**  +----------------+                         becomes persistent)          
**  ! class --------------> persistent instance representing the class of
**  +----------------+                                          the object
**  ! flags          !
**  +----------------+
**  ! object         !
**  : contents,      :
**  ! see below      !
**  +----------------+
**
** If the object is a String:                 If the object is a Class:
**  +----------------+                        +----------------+
**  | to be defined  |                        | to be defined  |
**  +----------------+                        +----------------+
**  ! NULL or ptr --------> local copy        ! NULL or ptr --------> local copy
**  +----------------+                        +----------------+
**  ! hashcode       !                        ! hashcode       !
**  +----------------+                        +----------------+
**  ! class --------------> Pinstance         ! class --------------> Pinstance
**  +----------------+  for class String      +----------------+ for class Class
**  ! flags          !                        ! flags          !
**  +----------------+                        +----------------+
**  ! length         !                        ! loader (see Note)
**  +----------------+                        +----------------+
**  : char1  ! char2 :                        ! Prot. domain ----> Pobject
**  +---    etc.     +                        +----------------+
**  :                :                        ! dotified name ---> Pobject
**  +----------------+                        +----------------+
**                                            ! static fields  !
**                                            : of the class   :
** For all other instances:                   +----------------+
**  +----------------+
**  | to be defined  |
**  +----------------+
**  ! NULL or ptr --------> local copy
**  +----------------+
**  ! hashcode       !
**  +----------------+
**  ! class --------------> Pinstance representing the class of the object
**  +----------------+
**  ! flags          !
**  +----------------+
**  ! instance       !
**  : fields         :
**  +----------------+
*/

typedef struct w_Pobject {
  w_word     spare;
  w_object   local_copy;
  w_int      hashcode;
  w_instance class;
  w_flags    flags;
  w_word     fields[1];
} w_Pobject;

#define instance2pobject(pinstance) ((w_pobject)(((unsigned char *)(pinstance)) - offsetof(w_Pobject, fields)))

/** The hashtable in which global references are stored. (See newGlobalRef).
*/
extern w_hashtable globals_hashtable;

void initHeap(void);
/** Start up the GC thread
*/
void startHeap(void);

void readStatics (w_clazz clazz);

/** Heap variables - treat as read-only!
    Number of instances currently in use
*/
extern w_size instance_use;
/**
    Number of instances allocated so far
*/
extern w_size instance_allocated;
/**
    Number of instances returned (garbage-collected) so far
*/
extern w_size instance_returned;

/*
** Create and initialise the garbage collector. Only call this once!
*/
void gc_create(JNIEnv *env, w_instance theGarbageCollector);

/*
** Run one pass of the garbage collector.  Returns the number of instances collected.
** Caller must won gc_monitor.
*/
w_int gc_collect(w_instance theGarbageCollector);

/*
** Ask the garbage collector to free up some momory, and return the amount
** which was really freed.  Caller must own gc_monitor.  Includes a built-in
** wait(), so may take a second or so to return.
*/
w_int gc_request(w_int bytes);

/*
** Run the finalizers of all the finalizable objects discovered by gc_collect().
*/
void gc_finalization(void);

/*
** Enqueue all the `dead' references discovered by gc_collect().
*/
void gc_references(void);

/*
** Perform miscellaneous tasks (e.g. clearing the instance caches, if any).
*/
void gc_housekeeping(void);

/**
    GC monitor
    Acquire and perform a wait on this lock in order to know when memory
    may have become available.
*/
extern x_monitor gc_monitor;

/**
   Used for synchronization with the garbage collector.
   Use \texttt{kickGC} to ask the garbage collector to run again soon.
 */
void kickGC(void);

/**
  createRoot() creates the root of persistency, either by reading data
  from the persistent medium or (if this fails) by creating an empty set.
*/
void createRoot(w_instance);

/*
void _swizzleRef(const char*, int, w_instance*);
#define swizzleRef(inst) _swizzleRef(inst)
*/

void markTransientRef(w_instance instance);

/*
** A w_reclaim_callback is called by the garbage collector to ask the callee
** to be so kind as to free up some memory.  The first parameter is the number
** of bytes that GC would like to reclaim, and the second is a w_instance
** which somehow identifies the subsystem which is asking for memory to be
** reclaimed (for example, the AWT might use the unique instance of 
** java.awt.Toolkit for this purpose).  The value NULL is reserved for use
** by the heap itself.  This parameter can be used to implement a kind of
** priority among subsystems, e.g. subsystem X might be more willing to
** give up memory to subsystem Y than to Z.
**
** The result returned is the number of bytes that the 
** callee was able or willing to release.  (This may only be an estimate, 
** but it needs to be a good one if the mechanism is to work  properly).
*/
typedef w_size(*w_reclaim_callback)(w_int requested, w_instance caller);

/*
** Listeners are registered and deregistered using registerReclaimCallback
** and deregisterReclaimCallback respectively.  In the current implementation,
** so long as no callbacks are deregistered the callbacks will always be 
** used in the order in which they were registered.
** A w_reclaim_callback can be called by any thread, at any priority!
** In particular, it is perfectly normal for a call to gc_reclaim() to
** cause the caller's own callback to be invoked.
** Callbacks should be written with this in mind.
*/
void registerReclaimCallback(w_reclaim_callback);
void deregisterReclaimCallback(w_reclaim_callback);

/*
** To force a reclaim, gc_reclaim(bytes) is called.
*/

w_size gc_reclaim(w_int bytes, w_instance caller);

/*
** The finalizer_fifo is used by the garbage collector thread to inform the
** finalizer thread of objects which need finalizing.  Access is protected
** by a mutex.
*/
extern w_fifo  finalizer_fifo;
extern x_mutex finalizer_fifo_mutex;

/*
** The enqueue_fifo is used by the garbage collector thread to inform the
** finalizer thread of objects which need enqueueg.  Access is protected
** by a mutex.
*/
extern w_fifo  enqueue_fifo;
extern x_mutex enqueue_fifo_mutex;

/*
** min_heap_free is the amount of heap which will be reserved for native
** code (cannot be used by Java instances).
*/
extern w_size     min_heap_free;

/*
** gc_instance and gc_thread are the instance of com.acunia.wonka.GarbageCollector
** and the w_Thread associated with the Garbage Collector singleton.
*/
extern w_instance gc_instance;
extern w_thread   gc_thread;

void startCollector(void);

/*
** Check whether the amount of memory requested is within the realm's quota.
** Returns WONKA_TRUE if so, WONKA_FALSE otherwise.
*/
extern w_boolean  checkRealmMemoryQuota(w_realm realm, w_int bytes);

static inline w_instance makeLocal(w_instance objectref) {
#ifdef ENABLE_PS
  if (isPersistentRef(objectref)) {
    return pinstance2local_copy(ref2pinstance(objectref));
  }
  else if (typeOfInstance(objectref) == INSTANCE_IS_PERSISTENT) {
    return pinstance2local_copy(objectref);
  }
  else {
    return objectref;
  }
#else
  return objectref;
#endif
}

#endif /* _HEAP_H */
