#ifndef _PS_MEM_H
#define _PS_MEM_H

/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: ps-mem.h,v 1.2 2001/11/28 15:00:31 gray Exp $
*/

/** File wonka/include/ps-mem.h
** Manage the persistent storage area.
*/
//@{

#ifndef offsetof
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

#include "wonka.h"

/**@name struct w_Chunk
** The structure of a persistent memory chunk for Wonka.  We use "boundary
** tagging", with a pointer to the characters "free" or "busy" as our tag:
** \begin{verbatim}
**          free block                        busy block
**     +------------------+               +-------------------+
**     !0! size in bytes  !               !1!  size in bytes  !     
**     +------------------+               +-------------------+     
**     !   not used       !               ! allocating realm  !         
**     +------------------+               +-------------------+         
**     !   not used       !               !    not used       !         
**     +------------------+               +-------------------+         
**     :                  :               :                   :         
**     :    contents      :               :    contents       :         
**     :                  :               :                   :         
**     +------------------+               +-------------------+         
**     !0! size in bytes  !               !1!  size in bytes  !     
**     +------------------+               +-------------------+         
** \end{verbatim}
**                                                                      
** The total size of each chunk (including the header and tailer shown above)
** is always a power of two and always at least 64 bytes.  We use a "buddy"
** algorithm for allocation, thereby guaranteeing that a chunk will only
** straddle a disk sector or memory page boundary when it is too big to fit
** into a single disk sector or memory page.  Initially the whole memory
** consists of a single free chunk.  The size indicated in the first and 
** last words is the total size, not the size of the contents.
**
*/
//@{
typedef struct w_PSchunk *w_pschunk;
///
typedef struct w_PSchunk {
  w_word     front_tag;             /*  0 */
  w_word     realm;                 /*  4 */
  w_word     flagz;                 /*  8 */
  /* Contents begins at offset ...     12 */
  /* We arrange that sizeof(w_PSchunk) gives */
  /* the total overhead, i.e. 20 bytes       */
  unsigned char contents[sizeof(w_word)];
} w_PSchunk;

///
#define THIS_BACK_TAG_PTR(chunkptr) \
  ((w_word*)&chunkptr->contents[((chunkptr->front_tag) & 0x7fffffff) - sizeof(w_PSchunk)])
///
#define PREV_BACK_TAG_PTR(chunkptr) (((w_word*)chunkptr)-1)
///
#define PS_MIN_SIZE 64
//@}

/**
** Macro to get back a w_pschunk pointer from a memory block that was 
** allocated through the _allocPSmem routine.
*/

#define PSblock2chunk(block) ((w_pschunk)((unsigned char *)block - offsetof(w_PSchunk, contents)))

/** Lock used to protect access to the PS-mem structures.
*/
//@{
/// The lock itself.
extern x_monitor PSmemory_monitor;
///
void _lockPSmem(const char*,int);
///
void _unlockPSmem(const char*,int);
/// Macro to lock the semaphore.
#define lockPSmem() _lockPSmem(__FILE__,__LINE__);
///
#define unlockPSmem() _unlockPSmem(__FILE__,__LINE__)
//@}

/// Allocate a memory block.
#define allocPSmem(t, n, s)           _allocPSmem(__FILE__, __LINE__, t, (w_size)n, (w_size)s)
// Release a memory block.
#define releasePSmem(b)               _releasePSmem(__FILE__, __LINE__, b)

void *_allocPSmem(const char *file, const int, w_thread thread, w_size num, w_size size);
void _releasePSmem(const char *file, const int, void *block);

/**
** Initialization: called during tx_application_define
*/

void initPSmem(void);
void clearPSmem(void);

/**
** Startup function called from PersistentRoot.c
*/

void startPSmem(void);

void *firstPSblock(void);
void *nextPSblock(void *block);

void updatePSblock(void *block);
void updatePSchunk(w_pschunk chunk, w_size size);
void syncPSmem(void);
//@}

#endif /* _PS_MEM_H */
