#ifndef _STACK_H
#define _STACK_H

/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: stack.h,v 1.4 2002/02/20 08:49:05 dfrederickx Exp $
*/

/**@name File wonka/include/stack.h
*/
//@{

#include "wonka.h"
#include "grobag.h"

#ifdef RUNTIME_CHECKS
#include "methods.h"
#include "clazz.h"
#endif

/*
** Maximum number of stack frames we allow a thread to accumulate.
** When running the Mauve tests and the VTE we never see more than
** 30 frames ...
*/
#define MAX_STACK_DEPTH 300

typedef struct w_Trace *w_trace;

/** One ``frame'' of a stack trace
*/
//@{
typedef struct w_Trace {
/// pointer to the trace of the calling frame
  w_trace previous;
/// the method we were calling
  w_method method;
/// the code pointer of our \textbf{caller} when this frame was called
  w_code caller_current;
/// frame flags
  w_int flags;
} w_Trace;

//@}

/** A stack ``leaf''
** Our Java stack is built out of a series of ``leaves'': whenever the
** next frame to be built will not fit into the current leaf, we append
** a new leaf.  Once added, leaves are never removed: probably the stack
** will grow to the same size again during the lifetime of the thread.
** (When the thread terminates, all leaves are of course released).
*/
//@{
typedef struct w_Leaf {
/// The number of \texttt{w_Item}s for which there is still space after the current frames.
  w_size remain;
/// The total size of the leaf (in \texttt{w_Item}s) and the number of
  w_size size;
/// The stack leaf proper
  w_word items[0];
} w_Leaf;

/** Fiddle factor
** When deciding whether a new frame will fit into the current leaf, we
** overestimate the number of items we need by this amount.  Is hack.
*/
#define STACK_SLACK 7
//@}

/** Stack frame
*/
//@{
/** Stack frame flags
*/
//@{
/// Frame is the bottom frame of a thread
#define FRAME_ROCK_BOTTOM                 0x00000001   
/// Frame is a startframe for runThread() 
#define FRAME_BOTTOM                      0x00000002   
/// Frame is used to run a <clinit> method 
#define FRAME_CLINIT                      0x00000004   
/// Frame is a host frame for a native method
#define FRAME_NATIVE                      0x00000010   
/// Frame is used in a JNI call
#define FRAME_JNI                         0x00000020   
/// Frame is JIT-compiled 
#define FRAME_JIT                         0x00000040   
/// Frame is used in reflection invocation
#define FRAME_REFLECTION                  0x00000080   
/// Frame was built using doPrivileged/
#define FRAME_PRIVILEGED                  0x00000100   
//@}

/** The stack frame structure
*/
//@{
typedef struct w_Frame {
/// Top-of-stack, i.e. the location where the next item will be pushed.
  w_word * o_top;
/// Stack base: pop no further.
  w_word * o_base;
/// Local variables (address of local variable 0).
  w_word * locals;
/// Java method currently executing (iff native==NULL).
  w_method method;
/// Native method currently executing, or NULL.
  w_method native;
/// Pointer to the bytecode (+1) we were executing when this frame was pushed.
  w_code saved_code;
/// Pointer to the bytecode (+1) we were executing when the native method (if any) was called.
  w_code native_code;
/// The stack leaf in which this frame resides.
  w_leaf leaf;
/// The amount of space which remained in the leaf when another frame was pushed on top of this one.
  w_size saved_remain;
/// The stack frame ``under'' this one (or NULL).
  w_frame previous;
/// The stack frame ``on top of'' this one (or NULL).
  w_frame next;
/// TODO NOT in use yet. 
//  It is there to get rid of the passed pointer in popJavaFrame. The compiler
// can not make the c_code in the interpreter loop a register since it is 
// passed via a pointer to the popJavaFrame function. When we would store the 
// next to execute opcode in this variable it can be picked up by the 
// interpreter loop here and wouldn't have to be passed as a pointer
// argument. This could be wrapped in a union with 'native'
  w_code next_code; 
/// Pointer to an instance monitor which is held by this frame (or NULL).
  x_monitor monitor;
/// The thread to which this frame belongs.
  w_thread thread;
/// Miscellaneous flags, see above.
  w_flags flags;

  /*
  ** The fields for managing local references.
  */

/// The local references created by a native method running in this frame.
  w_grobag localRefs;
  w_instance firstRef;
/// The Java program counter: only valid during exception handling, unless DEBUG is enabled.
//  Even when DEBUG is enabled, this may not always be precise: it will however
//  be updated at every transfer of control.  Caveat emptor!
  w_short pc;

  w_short depth; /* Number of frames already on stack when this one was created */
/// The set of ProtectionDomains applicable to this frame (created on demand).
  w_instance cached_context;
/// Another set of ProtectionDomains which must also be checked (or NULL).
  w_instance inherited_context;
} w_Frame;

/// The default local references capacity of a frame. 
#define DEFAULT_REFERENCES_CAPACITY   16
//@}
//@}

/** Local reference stuff
*/
//@{
/** Add a local reference from a stack frame to an instance.
** @param frame w_frame The frame where the reference is created.
** @param instance w_instance The instance referred to.
** @doc
**  For each local reference we actually create a global reference (which
**  is visible to the garbage collector), and in \texttt{frame->localRefs} 
**  we just note the local reference so that when we leave the native function
**  we can delete the global one.  Geddit?
*/
void addLocalReference(w_thread, w_instance instance);
/** Remove a local reference from a stack frame to an instance.
** @param thread w_thread The thread on which the reference was created.
** @param instance w_instance The instance referred to.
** @doc
**  This actually removes both the entry  in \texttt{frame->localRefs} and
**  the global reference.
*/
void removeLocalReference(w_frame, w_instance instance);
/** Remove all local reference from a stack frame.
** @param frame w_frame The frame where the references were created.
** @doc
**  This function is called when a native function exits.
*/
void removeAllLocalReferences(w_frame);
//@}

/** Stack frame manipulations
*/
//@{
/** Push a new Java frame on top of a given frame.
** @param parent w_frame The frame on top of which the new frame is to be pushed.
** @param c_code w_code* Bytecode pointer on entry / in new frame (in/out parameter).
** @param n_method w_method The method to be invoked in the new frame.
** @doc
** The descriptor of n_method is used to calculate the stack requirements,
** and to ensure that the parameters passed by the caller are installed
** as local variables of the callee.
**
** This function will allocate a new w_Frame structure unless it can re-use
** one left behind by a previous \texttt{popJavaFrame} (q.v.).
*/
w_frame pushJavaFrame(w_frame parent, w_code *c_code, w_method n_method);

/** Pop a java frame from the stack.
** @param parent w_frame The frame which is to be popped.
** @param c_code w_code* Bytecode pointer on entry / in new frame (in/out parameter).
** @doc
** The descriptor of n_method is used to determine which stack items if
** any should be passed back to the calling frame as a return value.
**
** The w_Frame structure is not released: it is left for possible re-use by
** a subsequent \texttt{pushJavaFrame}.
*/
w_frame popJavaFrame(w_frame current, w_code *c_code);

/** Reset the operand stack of a frame to the initial position.
** @param parent w_frame The frame which is to be reset.
** @doc
**  Resets \texttt{o_top} to \texttt{o_base}.
*/
void clearJavaOperands(w_frame);

/** Discard a frame without copying any return values. 
** @param parent w_frame The frame which is to be discarded.
** @param c_code w_code* The bytecode pointer in the underlying frame (output parameter).
** @doc
**  Used inside exception handling code.
*/
w_frame clearJavaFrame(w_frame, w_code *c_code);

/// Allocate a frame without inheriting any parameters from a previous frame.
//  Used during thread startup, reflection, and JNI.
w_frame allocBottomFrame(w_thread thread, w_method method);

/// Release a frame without copying any return values.
//  Used after thread startup, reflection, and JNI.
void releaseBottomFrame(w_frame frame);

/// Dump a stack frame in a welter of \texttt{woempa}s.
void dumpFrame(const char *s, w_frame frame);
//@}

/// Get the security domain associated with a frame.
// Note that for the time being we simply ignore native code.
// Steven, you may need to adapt this.
#define frame2domain(f) ((w_instance)(clazz2Class((f)->method->clazz)[F_Class_domain]))

void handleSynchronization(w_frame frame, w_method method);
/** TODO: docify the rest of this stuff
** getCurrentMethod and getCurrentClazz return the Java method currently
** being executed and the class in which it was defined, respectively.
** getCallingMethod and getCallingClazz return the method which called
** the current Java method and the class in which it was defined.
** Note that we say the -Java- method: if called from within a native
** method, these functions will return information for the Java method
** which invoked the native code.
** getCurrentInstance returns the 'this' of the instance method currently
** being executed (or null if a static method), and getCallingInstance
** does the same for the method which called the current Java method.
*/
w_method getCurrentMethod(w_thread thread);
w_clazz getCurrentClazz(w_thread thread);
w_instance getCurrentInstance(w_thread thread);
w_method getCallingMethod(w_thread thread);
w_clazz getCallingClazz(w_thread thread);
w_instance getCallingInstance(w_thread thread);
w_int callHeight(w_frame frame);


w_frame allocFrame(w_frame parent);
w_leaf allocLeaf(w_thread thread, w_size size);
void releaseStack(w_frame bottom);
w_trace createStackTrace(w_frame top);
void destroyStackTrace(w_trace);

#ifdef RUNTIME_CHECKS
#define stack_pops_pushes(f, pops, pushes) _stack_pops_pushes(__FILE__, __LINE__, f, pops, pushes)
inline static void _stack_pops_pushes(const char *f, int l, w_frame frame, w_int pops, w_int pushes) {

  w_int avail = frame->o_top - frame->o_base;
  w_int space = (frame->o_base + frame->method->stack_i) - frame->o_top;

  if (space + pops < pushes) {
    wabort(ABORT_WONKA, "%s:%d Stack overflow, pushes %d words, only %d left.\n", f, l, pushes, space + pops);
  }

  if (avail < pops) {
    wabort(ABORT_WONKA, "%s:%d Stack underflow, pops %d words, only %d available.\n", f, l, pops, avail);
  }

}
#else
#define stack_pops_pushes(f, pops, pushes) 
#endif

#endif /* _STACK_H */
