#ifndef _TS_MEM_H
#define _TS_MEM_H

/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: ts-mem.h,v 1.3 2002/02/20 08:49:05 dfrederickx Exp $
*/

#include <stddef.h>

#include "oswald.h"
#include "realms.h"
#include "wonka.h"

/*
** The internal structure of a memory chunk for Wonka when DEBUG is defined. 
*/

#ifdef DEBUG // --------------------------------------------------------------

typedef struct w_Chunk {          // offset of  ends at
  w_word       madgic;            //     0         3  madgic word to identify a debug memory chunk
  const char * file;              //     4         7  in which file the chunk was allocated
  w_int        line;              //     8        11  at which line the chunk was allocated
  w_int        back;              //    12        15  an index to the backfence byte
  w_size       size;              //    16        19  the size as was requested by the user for datax
  w_word       front;             //    20        23  the front control word to check for out of bound writes
  w_ubyte      data[8];           //    24       ...  pre-allocate 4 bytes for backfence
} w_Chunk;

#else

typedef struct w_Chunk {
  w_ubyte      data[0];
} w_Chunk;

#endif // --- DEBUG ----------------------------------------------------------

typedef struct w_Chunk *          w_chunk;

/*
** Macros for allocating and freeing memory. Note that when DEBUG is defined, the routines
** get a _d_ prepended and have file and line number information as last parameters. For
** non DEBUG compiles, the macros refer to the normal _x memory functions.
*/

#ifdef DEBUG

#define allocMem(t, n)              _d_allocMem_realm(thread2realm(t), n, __FILE__, __LINE__)
#define allocClearedMem(t, n)       _d_allocClearedMem_realm(thread2realm(t), n, __FILE__, __LINE__)
#define allocMem_realm(r, n)        _d_allocMem_realm(r, n, __FILE__, __LINE__)
#define allocClearedMem_realm(r, n) _d_allocClearedMem_realm(r, n, __FILE__, __LINE__)
#define allocMem_system(n)          _d_allocMem_realm(W_Realm_system, n, __FILE__, __LINE__)
#define allocClearedMem_system(n)   _d_allocClearedMem_realm(W_Realm_system, n, __FILE__, __LINE__)
#define reallocMem(b, s)            _d_reallocMem(b, s, __FILE__, __LINE__)
#define releaseMem(b)               _d_releaseMem(b, __FILE__, __LINE__)
#define discardMem(b)               _d_discardMem(b, __FILE__, __LINE__)

#else

#define allocMem(t, n)              _allocMem_realm(thread2realm(t), n)
#define allocClearedMem(t, n)       _allocClearedMem_realm(thread2realm(t), n)
#define allocMem_realm(r, n)        _allocMem_realm(r, n)
#define allocClearedMem_realm(r, n) _allocClearedMem_realm(r, n)
#define allocMem_system(n)          _allocMem_realm(W_Realm_system, n)
#define allocClearedMem_system(n)   _allocClearedMem_realm(W_Realm_system, n)
#define reallocMem(b, s)            _reallocMem(b, s)
#define releaseMem(b)               _releaseMem(b)
#define discardMem(b)               _discardMem(b)

#endif /* DEBUG */

/*
** The ID for anonymous memory; please use static const w_size to assign ID's.
*/

static const w_size ID_anon = 32;

/*
** Get the chunk reference from a void * block.
*/

inline static w_chunk block2chunk(void * block) {
  return ((w_chunk)((unsigned char *)block - offsetof(w_Chunk, data)));
}

extern w_realm realm_table[];

/*
** Mask to extract the realm index from a memory tag.  Max 7 bits
** (OSwald memory manager restriction).
*/
#define REALM_INDEX_MASK 0x7f

/*
** Return the owning realm of a memory block. The prototypes to change the owner
** realm of a memory block are given below.
*/

inline static w_realm getMemOwner(void * block) {
  return realm_table[x_mem_tag_get(block2chunk(block)) & REALM_INDEX_MASK];
}

void resetMemOwner(void * block, w_thread thread);
void resetMemOwnerRealm(void * block, w_realm realm);

inline static x_word getMemTag(void * block) {
  return x_mem_tag_get(block2chunk(block));
}

inline static void setMemTag(void * block, x_word tag) {
  x_mem_tag_set(block2chunk(block), tag);
}

/*
** The function prototypes for the allocation and release routines.
*/

#ifdef DEBUG

void * _d_allocMem_realm(w_realm realm, w_size size, const char * file, const int line);
void * _d_allocClearedMem_realm(w_realm realm, w_size size, const char * file, const int line);
void * _d_reallocMem(void * block, w_size size, const char * file, const int line);
void   _d_releaseMem(void * block, const char * file, const int line);
void   _d_discardMem(void * block, const char * file, const int line);

#else

void * _allocMem_realm(w_realm realm, w_size size);
void * _allocClearedMem_realm(w_realm realm, w_size size);
void * _reallocMem(void * block, w_size size);
void _releaseMem(void * block);
void _discardMem(void * block);

#endif /* DEBUG */

/*
** For chasing memory leaks, the reportMemStat function can be used. It dumps a sorted
** text representation of the memory that is in use and where it is allocated. Sorting
** can be done on a number of equal blocks or filename basis.
*/

#ifdef DEBUG

#define SORT_BY_HITS                  1
#define SORT_BY_FILE                  2

void reportMemStat(w_int type);
#define heapCheck _heapCheck(__FUNCTION__, __LINE__)
void _heapCheck(const char * funtion, const int line);

inline static const char * allocedInFile(void * block) {
  return block2chunk(block)->file;
}

inline static const int allocedAtLine(void * block) {
  return block2chunk(block)->line;
}

#endif /* DEBUG */

#endif /* _TS_MEM_H */
