#ifndef _WORDSET_H
#define _WORDSET_H

/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: wordset.h,v 1.2 2002/02/12 17:25:44 gray Exp $
*/

#include "grobag.h"

/** A Wordset is an expandable array of words, in no particular order.
** @doc
** By "in no particular order" we mean in fact that the functions below
** may result in a re-ordering of the wordset outside of the user's
** control, so the user should never assume that a particular word
** will remain at a particular offset.
**
** A \texttt{w_Wordset} is really just a \texttt{w_Grobag} used in a
** particular way.  Like the \texttt{\w_Grobag} operations, all operations
** on a wordset take a *w_wordset (i.e. a **w_Wordset) as parameter;
** this is because the w_wordset pointer will be updated when new space is
** allocated.
*/

#define w_Wordset w_Grobag
typedef struct w_Wordset *w_wordset;

/** We always allocate a multiple of WORDSET_SIZE_GRANULARITY words.
*/
#define WORDSET_SIZE_GRANULARITY 16

/** Reserve space for at least 'newsize' words of contents.
**
** If \texttt{wordsetptr} is currently \texttt{NULL}, allocates a new Wordset.
*/
#define ensureWordsetCapacity(thread,wordsetptr,newsize) ensureGrobagCapacity(thread,wordsetptr,((((newsize)+WORDSET_SIZE_GRANULARITY-1) & (-WORDSET_SIZE_GRANULARITY)) * sizeof(w_word)))

/** Add a word to a wordset, after expanding it if necessary.
    Duplicates may be created.
*/
static inline void addToWordset(w_thread thread, w_wordset* wordset, w_word what) {
  register w_word *w;
  w_int   i;

  if (*wordset) {
    i = ((*wordset)->occupancy) / sizeof(w_word);
    ensureWordsetCapacity(thread, wordset, i+1);
  }
  else {
    i = 0;
    ensureWordsetCapacity(thread, wordset, 1);
  }
  w = (w_word*)((*wordset)->contents);
  w[i] = what;
  (*wordset)->occupancy = (i+1) * sizeof(w_word);
}

/** Remove first word from a waitset.  Returns the word or NULL.
*/
static inline w_word takeFirstFromWordset(w_wordset* wordset) {
  if (*wordset && (*wordset)->occupancy) {
    w_word *w = (w_word*)((*wordset)->contents);
    w_int   n = ((*wordset)->occupancy) / sizeof(w_word);
    w_word result;

    woempa(1,"Wordset %p contains %d items\n",*wordset,(*wordset)->occupancy / sizeof(w_word));
    result = w[0];
    woempa(1,"Removed front element 0x%08x\n",result);
    if (n > 1) {
      w[0] = w[n-1];
      woempa(1,"  -- replaced with 0x%08x\n",w[0]);
    }
    (*wordset)->occupancy -= sizeof(w_word);

    return result;

  }
  else {

    return 0;

  }
}

/** Look for a word in a wordset.  Returns WONKA_TRUE iff the word was found.
*/
static inline w_boolean isInWordset(w_wordset* wordset, w_word what) {
  w_word *w;
  w_size  i;

  if (*wordset == NULL) {

    return WONKA_FALSE;

  }

  w = (w_word*)((*wordset)->contents);
  for (i = 0; i < ((*wordset)->occupancy) / sizeof(w_word); ++i) {
    if (w[i] == what) {

      return WONKA_TRUE;

    }
  }

  return WONKA_FALSE;
}

/** Remove a word from a wordset.  Returns WONKA_TRUE iff the word was found.
*/
static inline w_boolean removeFromWordset(w_wordset* wordset, w_word what) {
  w_word *w = (w_word*)((*wordset)->contents);
  w_int   n = ((*wordset)->occupancy) / sizeof(w_word);
  w_int   i;

  for (i=0; i < n; ++i) {
    if (w[i] == what) {
      w[i] = w[n-1];
      (*wordset)->occupancy -= sizeof(w_word);

      return WONKA_TRUE;

    }
  }

  return WONKA_FALSE;
}

/**
 ** Returns the n'th element of a wordset.
 */
inline static w_word elementOfWordset(w_wordset* wordset, w_int n) {
  w_word *w = (w_word*)((*wordset)->contents);
  return w[n];
}

/** Sort a Wordset. Will need to be re-sorted if anything is added or removed.
*/
inline static void sortWordset(w_wordset* wordset) {
  w_word *w = (w_word*)((*wordset)->contents);
  w_int   n = (*wordset)->occupancy;
  w_int   i;
  w_int   j;

  for (i = 0; i < n; ++i) {
    for (j = i+1; j < n; ++j) {
      if (w[j] < w[i]) {
        w_word x = w[i];
        w[i] = w[j];
        w[j] = x;
      }
    }
  }
}

/** Release the memory used by a Wordset
** (Sets \texttt{*wordset} to \texttt{NULL}.
*/
inline static void releaseWordset(w_wordset* wordset) {
  releaseMem(*wordset);
  *wordset = NULL;
}

/*
** Find out how many items are in a wordset.
*/
inline static w_size sizeOfWordset(w_wordset *wordset) {
  return (*wordset) ? (*wordset)->occupancy / sizeof(w_word) : 0;
}

#endif /* _WORDSET_H */
