/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: DeviceInputStream.java,v 1.1.1.1 2001/11/21 10:41:34 pollock Exp $
*/

package com.acunia.device;

import java.io.*;

/**
 * DeviceInputStream implements InputStream for a Device of the SM platform.
 * It supports the setting of thresholds and timeouts for reading data.
 */
public class DeviceInputStream extends InputStream implements Runnable {

  /** The internal device ID for this device.
  */
  private int devid;

  /** True iff the thresholding mechanism is active.
  */
  private boolean thresholdActive = false;

  /** True iff the timeout mechanism is active.
  */
  private boolean timeoutActive = false;

  /**
   ** The currently applicable threshold value, as supplied by the user of 
   ** the DeviceInputStream.  Note that an rxThreshold of zero
   ** really means one (read returns as soon as *some* data is available).
   */
  private int rxThreshold = 0;

  /**
   ** We call the effective threshold (the minimum number of bytes which 
   ** must be available before read() will return) "desired".
   */
  private int desired;

  /**
   ** The currently applicable timeout value, as supplied by the user of 
   ** the InputStream.  Note that an rxTimeout of zero really means infinity. 
   */
  private int rxTimeout = 0; // milliseconds

  /**
   ** The pause built into the spinlock when we
   ** are waiting for some data to become available.
   */
  private final long micronap =  100; // milliseconds

  /** Used in the monitor thread where theoretically we
   ** should execute an open-ended wait(), but we haven't the
   ** balls.
   */
  private final long meganap = 10000; // milliseconds

  /**
   ** A one-byte buffer used to implement read() 
   ** as a special case of read(array,offset,length).
   */
  private byte [] microbuffer = new byte[1];

  /**
   ** Reference to the thread which monitors the number of input
   ** bytes available and wakes up the read method as required.
   ** It is started when the timeout or threshold mechanism becomes
   ** active, and terminates when they both become inactive or the
   ** InputStream is closed ("open" becomes false).
   */
  private Thread inputMonitor = null;

  /**
   ** Set true when the InputStream is created, and false when
   ** close() is called.
   */
  private boolean open;

  /**
   ** The name of the underlying device.
   */
  private String name;

  /**
   ** createFromString initialises this DeviceInputStream object using a
   ** lookup based on "name".  Returns the internal device ID.
   */
  private native int createFromString(String name)
    throws SecurityException;

  /**
   ** Constructor DeviceInputStream(String) calls createFromString
   ** and starts up the monitor thread.
   */
  public DeviceInputStream(String name) {
    // TODO: security check?

    devid = createFromString(name);
    this.name = name;
    this.open = true;
  }
  
  /**
   ** Method readIntoBuffer performs the "guts" of a read,
   ** shifting bytes from the i/o driver into the buffer.
   */
  private native int readIntoBuffer(int devid, byte[] b, int off, int len)
    throws IOException;
  
  /** Method read() (single-character read) is implemented as a
   ** special case of read(array,offset,length).  This could be
   ** done more efficiently: feel free if you have the time.
   */
  public synchronized int read()
    throws IOException
  {
    int readlen = read(microbuffer,0,1);

    if (readlen>0) {

      return microbuffer[0] & 0xff;

    }
    else return readlen;

  }
  
  /**
   ** Method read(array) is a convenience method, a shorthand
   ** for read(array, 0, b.length).
   */
  public synchronized int read(byte[] b)
    throws IOException, NullPointerException 
  {
    return read(b, 0, b.length);
  }
  
  /**
   ** OK, now for the real business.
   ** Method read(byte[] b, int off, int len) first (after sanity checks)
   ** determines the minimum number of characters to be read ("desired")
   ** and then wakes up the monitor thread.  If a Timeout was specified
   ** we then wait to be notified by the monitor thread, until at least
   ** "desired" characters are available.  If no Timeout was specified
   ** then we just spinlock (using micronap) until at least "desired" 
   ** characters are available.  In both cases the wait loop is aborted
   ** if an interrupt is received.
   ** Finally we either copy the lesser of (available(),length) bytes
   ** into the buffer and return the number of bytes copied, or (if no
   ** data is available, i.e. we timed out or were interrupted) return -1.
   */
  public synchronized int read(byte[] b, int off, int len)
    throws IOException, NullPointerException, ArrayIndexOutOfBoundsException 
  {
    long deadline;  // absolute time when timeout will expire
    long remaining; // elapsed time from now until deadline

    if(b==null) throw new NullPointerException();

    if(off<0 || len<0 || off+len > b.length) {
      throw new ArrayIndexOutOfBoundsException();
    }

    if (!thresholdActive) {
      desired = len;
    }
    else if (rxThreshold>0) {
      desired = rxThreshold;
    } 
    else {
      desired = 1;
    }

    if (timeoutActive && rxTimeout>0) {
      deadline = System.currentTimeMillis()+rxTimeout;
      remaining = rxTimeout;
    }
    else {
      deadline = Long.MAX_VALUE;
      remaining = Long.MAX_VALUE;
    }

    this.notify();

    try {
      while (remaining > 0 && available() < desired) {
        if (rxTimeout > 0) {
          this.wait(remaining);
          remaining = deadline - System.currentTimeMillis();
        }
        else { // no rxTimeout
          this.wait(micronap);
        }
      }
    } catch (InterruptedException e) {}

    int bytesToRead = available();

    desired = 0;

    if (bytesToRead==0) {

      return -1;

    }
    else if (bytesToRead>len) {

	return readIntoBuffer(devid, b, off, len);

    }
    else {

      return readIntoBuffer(devid, b, off, bytesToRead);

    }
  }
  
  /**
   ** Method skip0 removes and discards exactly n characters from the
   ** input stream (unless an IOException happens first).
   */
  public native long skip0(int devid, long n)
    throws IOException;

  /**
   ** Method skip calls skip0 to skip either n characters or available()
   ** characters, whichever is the less.
   */
  public long skip(long n)
    throws IOException
  {
    int bytesToRead = available();

    if (bytesToRead==0) {

      return 0;

    }
    else if (bytesToRead>n) {

      return skip0(devid, n);

    }
    else return skip0(devid, bytesToRead);

  }

  /**
   ** Method available() returns the maximum number of bytes which could
   ** now be read from the input stream without blocking.
   */
  public native synchronized int available()
    throws IOException;
  
  /**
   ** Method close() sets "open" false.
   */
  public void close()
    throws IOException
  {
    open = false;
  }
  
  
  /**
   ** Method setRxThreshold(t) adjusts the value of rxThreshold and
   ** then gives the monitor a "kick".
   */
  public synchronized void setRxThreshold(int t) {
// TODO: does set(0) really mean clear()?
    if (t==0) {
      clearRxThreshold();
    }
    else {
      thresholdActive = true;
      if (inputMonitor == null) {
        inputMonitor = new Thread(this,name+"-monitor");
        inputMonitor.start();
      }
      rxThreshold = t;
      this.notify();
    }
  }

  /**
   ** Method clearRxThreshold(t) sets rxThreshold to zero (and hence
   ** desired to 1) and then gives the monitor a "kick".
   */
  public synchronized void clearRxThreshold() {
    thresholdActive = false;
    rxThreshold = 0;
    desired = 1;
    this.notify();
  }

  /**
   ** Method setRxTimeout(t) adjusts the value of rxTimeout and
   ** then gives the monitor a "kick".
   */
  public synchronized void setRxTimeout(int t) {
// TODO: does set(0) really mean clear()?
    if (t==0) {
      clearRxTimeout();
    }
    else {
      timeoutActive = true;
      if (inputMonitor == null) {
        inputMonitor = new Thread(this,name+"-monitor");
        inputMonitor.start();
      }
      rxTimeout = t;
      this.notify();
    }
  }

  /**
   ** Method clearRxTimeout(t) sets rxTimeout to zero and
   ** then gives the monitor a "kick".
   */
  public synchronized void clearRxTimeout() {
    timeoutActive = false;
    rxTimeout = 0;
    this.notify();
  }

  /**
   ** Method run() implements the monitor thread.  This sits in an endless
   ** loop until the InputStream is closed:
   **   - wait until at least "desired" bytes are available;
   **   - notify the reading thread;
   **   - wait to be notified in return (e.g. because of a new read).
   ** The last wait should really be open-ended, but that would be too
   ** scary.
   */
  public synchronized void run() {
    while (open && (timeoutActive || thresholdActive)) {
      try {
        while (available()<desired) {
          this.wait(micronap);
        }
        this.notify();
        this.wait(meganap);
      } catch (InterruptedException e) {
      } catch (IOException e) {
      }
    }
  }

}
