
/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: GarbageCollector.java,v 1.4 2002/01/16 08:27:34 gray Exp $
*/

package com.acunia.wonka;

public final class GarbageCollector implements Runnable {

  private static GarbageCollector theGarbageCollector;

  /**
   ** The priority at which the garbage collection thread is started.
   */
  private static final int INITIAL_PRIORITY = 10;

  /**
   ** The minimum priority at which the GC thread will run.
   */
  private static final int MIN_PRIORITY = 1;

  /**
   ** The maximum priority at which the GC thread will run.
   */
  private static final int MAX_PRIORITY = 10;

  /**
   ** RATE_FALLING_THRESHOLD: if the current round of gc collects fewer items
   ** than this then we will always consider the collection rate to be `falling'.
   ** When the collection rate is falling we lengthen the sleep between gc passes.
   */
  private static final int RATE_FALLING_THRESHOLD = 200;

  /**
   ** RATE_RISING_THRESHOLD: if the current round of gc collects fewer items
   ** than this then we will never consider the collection rate to be `rising'.
   ** When the collection rate is rising we shorten the sleep between gc passes.
   */
  private static final int RATE_RISING_THRESHOLD = 400;

  /**
   ** HIGHER_PRIORITY_THRESHOLD: when the occupancy of the heap rises above this
   ** percentage, we start to increase the priority of the gc thread.
   */
  private static final float HIGHER_PRIORITY_THRESHOLD = 0.80f;

  /**
   ** LOWER_PRIORITY_THRESHOLD: when the occupancy of the heap falls below this
   ** percentage, we start to gradually reduce the priority of the gc thread.
   */
  private static final float LOWER_PRIORITY_THRESHOLD = 0.50f;

  /**
   ** Priority of the GC thread.  Updated after each cycle, depending
   ** on the occupancy of the cache.  Booset to MAX_PRIORITY each time
   ** we are `kicked'.  (N.B. a `kick' can also come from native code).
   */
  private int priority = INITIAL_PRIORITY;

  /**
   ** Numver of GC cycles completed so far.
   */
  private int passes;

  /**
   ** Number of times we have been `kicked'.
   */
  private int kicks;

  /**
   ** The number of objects which were collected in the most recent pass.
   */
  private int eaten;

  /**
   ** The number of objects collected since system startup.
   */
  private int eaten_total;

  /**
   ** The value of eaten_total prior to the most recent pass.
   */
  private int eaten_previous;

  /**
   ** The number of objects which were collected in some previous pass
   ** (not necessarily the most recent); if we collect significantly
   ** more or less then we will adjust our sleep time accordingly.
   */
  private int eaten_baseline;

  /**
   ** Time to sleep between GC cycles.  Updated after each cycle, depending
   ** on whether our collection rate is rising or falling.
   */
  private int sleep_millis = 100;

  /**
   ** The minimum time for which GC will sleep between cycles.
   ** Can be adjusted using setMinSleep().
   */
   private int min_sleep_millis = 100;

  /**
   ** The maximum time for which GC will sleep between cycles.
   ** Can be adjusted using setMaxSleep().
   */
   private int max_sleep_millis = 10000;


  /**
   ** The identity of the thread which performs GC.
   */
  private Thread thread;

  /**
   ** Create the native code structures used by GC.
   */
  private native void create();

  /**
   ** Our very private constructor.
   */
  private GarbageCollector() {
    thread = new Thread(this, "Undertaker");
    thread.setPriority(INITIAL_PRIORITY);
    thread.setDaemon(true);
    thread.start();
  }

  /**
   ** Return the solitary instance of GarbageCollector.
   */
  public static synchronized GarbageCollector getInstance() {
    if (theGarbageCollector == null) {
      theGarbageCollector = new GarbageCollector();
    }

    return theGarbageCollector;
  }

  /**
   ** Run the native code which performs a GC cycle.
   */
  private native int collect();

  /**
   ** We say the collection rate is rising if it exceeds both an
   ** absolute minimum and 25% more than the `baseline'.  Each
   ** time this is so, we update the baseline.
   */
  private boolean collectionRateIsRising () {
    if (eaten > RATE_RISING_THRESHOLD && eaten > eaten_baseline * 5 / 4) {
      eaten_baseline = eaten;

      return true;

    }

    return false;
  }

  /**
   ** We say the collection rate is rising if it is less than both an
   ** absolute minimum and 80% of the `baseline'.  Each time this is so, 
   ** we update the baseline.
   */
  private boolean collectionRateIsFalling () {
    if (eaten_baseline < RATE_FALLING_THRESHOLD || eaten < eaten_baseline * 4 / 5) {
      eaten_baseline = eaten;

      return true;

    }

    return false;
  }

  /**
   ** We limit the sleep time to be between min_sleep_millis and max_sleep_millis.
   */
  private int limit_sleep_time(int time) {
    if (time < min_sleep_millis) {

      return min_sleep_millis;

    }
    if (time > max_sleep_millis) {

      return max_sleep_millis;

    }

    return time;
  }

  /**
   ** Kick the GarbageCollector, make it do some work instead of sleeping.
   ** Will throw NullPointerException if GC thread is not yet started.
   */
  public void kick() {
    kicks += 3;
    thread.setPriority(MAX_PRIORITY);
    thread.interrupt();
  }

  /**
   ** Wait for the GarbageCollector to do a certain number of passes.
   ** Will throw NullPointerException if GC thread is not yet started.
   */
  public synchronized void work(int numPasses) {
    int wait_until = passes + numPasses;
    while (passes - wait_until > 0) {
      try {
        wait();
      }
      catch (InterruptedException ie) {
      }
    }
  }

  /**
   ** Ask GC to free up a certain number of bytes of memory.
   ** Returns the number of bytes which were really freed 
   ** (can be more or less).
   */
  public native synchronized int request(int bytes);

  public void setMinSleep(int millis) {
    min_sleep_millis = millis;
  }

  public void setMaxSleep(int millis) {
    max_sleep_millis = millis;
  }

  public void run() {
    create();

    try {
      Thread.sleep(sleep_millis);
    }
    catch (InterruptedException e) {}

    while (true) {
      try {
        priority = thread.getPriority();
        float heap_total = memTotal();
        float heap_avail = memAvail();
        float heap_occupancy = (heap_total - heap_avail) / heap_total;
        if (heap_occupancy > HIGHER_PRIORITY_THRESHOLD && priority < MAX_PRIORITY) {
          priority = (priority + MAX_PRIORITY + 1) / 2;
          thread.setPriority(priority);
        }
        else if (heap_occupancy < LOWER_PRIORITY_THRESHOLD && priority > MIN_PRIORITY) {
          priority -= 1;
          thread.setPriority(priority);
        }
 
        synchronized(this) {
          ++passes;
          eaten_previous = eaten_total;
          eaten_total = collect();
          eaten = eaten_total - eaten_previous;
          this.notifyAll();
        }

        housekeeping();

        if (collectionRateIsRising()) {
          sleep_millis = limit_sleep_time(sleep_millis / 2);
        }
        else if (collectionRateIsFalling()) {
          sleep_millis = limit_sleep_time(sleep_millis * 5 / 3);
        }
  
        try {
          synchronized(this) {
            this.wait(sleep_millis);
          }
        }
        catch (InterruptedException ie) {
        }
 
        if (kicks > 0) {
          sleep_millis = limit_sleep_time(sleep_millis / kicks);
          --kicks;
          synchronized(this) {
            ++passes;
            collect();
            this.notifyAll();
          }
        }
 
      }
      catch (Throwable t) {
      }
    }
  }

  native void housekeeping();

  static native long memTotal();
  static native long memAvail();

}

