/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/**
 * $Id: IntHashtable.java,v 1.1 2001/12/07 15:52:22 ruelens Exp $
 */

package com.acunia.wonka;

public class IntHashtable {

  private transient int capacity;
  private float loadFactor;
  private transient int occupancy;

  private final static float DEFAULT_LOADFACTOR = (float).75;
  private final static int DEFAULT_CAPACITY = 101;

  private int[] keys;
  private transient Object[] value;

  public IntHashtable() {
    this(DEFAULT_CAPACITY, DEFAULT_LOADFACTOR);
  }

  public IntHashtable(int initialCapacity) {
    this(initialCapacity, DEFAULT_LOADFACTOR);
  }

  public IntHashtable(int initialCapacity, float loadFactor) {
    if ( initialCapacity < 0 || loadFactor <= 0.0f )
    	throw new IllegalArgumentException("IntHashtable needs positive numbers");
    this.capacity = (initialCapacity < 5 ? 5 :initialCapacity);
    this.keys = new int[capacity];
    this.value = new Object[capacity];
    this.loadFactor = (loadFactor > 1.0f ? DEFAULT_LOADFACTOR : loadFactor );
  }

  public int size() {
    return this.occupancy;
  }

  public Object get(int key)  {
    if (this.keys == null) {
      return null;
    }
    int i;
    int j = 0;
    while(true) {
      i = probe(key, j);
      if(this.keys[i] == 0) {
        return null;
      } else if(key == keys[i]) {
        return this.value[i];
      } else ++j;
    }
  }          

  public Object put(int key, Object newvalue)  {
    int i;
    int j = 0;
    int newsize;
    if (this.keys==null) {
      this.keys = new int[DEFAULT_CAPACITY];
      this.value = new Object[DEFAULT_CAPACITY];
      this.capacity = DEFAULT_CAPACITY;
    }
    while(true) {
      i = probe(key, j);
      if(this.keys[i] == 0) {
        this.keys[i] = key;
        this.value[i] = newvalue;
        ++this.occupancy;
        newsize = this.checkSize(+1);
        if(newsize!=0) {
          resize(newsize);
        }
        return null;
      } 
      else if(key == keys[i]) {
        Object oldvalue = this.value[i];
        this.value[i] = newvalue;
        return oldvalue;
      }
      ++j;
    }
  }

  public Object remove(int key)  {
    int   i;
    int   j = 0;
    int   newsize;

    if (this.keys == null) {
      return null;
    }

    while(true) {
      i = probe(key,j);
      if(this.keys[i] == 0) {
        return null; 
      } else {
        if(key == keys[i]) {
          Object oldvalue = this.value[i];
          deleteSlot(i);
//          --this.occupancy;
          newsize = this.checkSize(-1);
          if(newsize!=0) {
            resize(newsize);
          }
          return oldvalue;
        }
      }
      ++j;
    }
  }

  public void clear() {
     if (occupancy > 0) {
     keys = new int[capacity];
     value = new Object[capacity];
     occupancy = 0;
     }
  }

  private int firstBusySlot(int i) {
    int j;
    if (i<0 || i>=capacity || this.keys==null) {
      return -1;
    }
    for(j=i;j<capacity;++j)
      if (this.keys[j] != 0)  {
        return j;
      }
    return -1;
  } 

  private int probe(int hashcode, int sequence) {
    int rehash = (hashcode-sequence) % this.capacity;
    if (rehash<0) rehash = rehash + this.capacity;
    return rehash;
  }

 private int checkSize(int direction) {
    float targetsquared, current, currentsquared;

    targetsquared = loadFactor * loadFactor;
    current = (float)occupancy / (float)capacity;
    currentsquared = current * current;

    if (direction>=0 && currentsquared<loadFactor
     || direction<=0 && targetsquared>current
    ) {
      return 0;
    }

    int newsize = (int)((float)occupancy/loadFactor);

    if (newsize<occupancy+2) newsize = occupancy+2;

    if (direction>=0 && newsize<=capacity) newsize = 0;
    if (direction<=0 && newsize>=capacity) newsize = 0;

    return newsize;

  } 

  private void resize(int newsize) {
    int oldsize = this.capacity;
    int[] oldkeys = this.keys;
    Object[] oldvalues = this.value;
    int[] newkeys = new int[newsize];
    Object[] newvalues = new Object[newsize];

    this.capacity = newsize;
    this.occupancy = 0;
    this.keys = newkeys;
    this.value = newvalues;

    int oldindex, newindex;

    for (oldindex=0;oldindex<oldsize;++oldindex) {
      int key = oldkeys[oldindex];
      if (key != 0)
        this.put(key,oldvalues[oldindex]);
    }
  } 

  private void deleteSlot(int slotIndex) {
    int vacant, current, home, distance1, distance2;
    boolean happy;

    --occupancy;
    current = slotIndex;

    while(true) {
    // R1
      keys[current] = 0;
      vacant = current;

    // R2
      happy = true;
      while(happy) {
        current = current==0 ? capacity-1 : current-1;

    // R3
        if (keys[current] == 0) {

          return; // normal termination

        }
        home = probe(keys[current],0);
        distance1 = current<=home ? home - current : capacity + home - current;
        distance2 = current<=vacant ? vacant - current : capacity + vacant - current;

        happy = distance1<distance2;
      // back to R2
      }

    // R4
      keys[vacant] = keys[current];
      value[vacant] = value[current];

    // repeat from R1 ...
    }
  }
}
