/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: BufferedInputStream.java,v 1.1.1.1 2001/11/21 10:41:34 pollock Exp $
*/

package java.io;

public class BufferedInputStream extends FilterInputStream {

  protected byte [] buf;
  protected int count = 0;
  protected int pos = 0;
  protected int marklimit = 0;
  protected int markpos = -1;

  public BufferedInputStream(InputStream in){
    super(in);
    buf = new byte[2048];
  }

  public BufferedInputStream(InputStream in, int size) {
    super(in);
    if(size <= 0){
      throw new IllegalArgumentException("size should be > 0");
    }
    buf = new byte[size];
  }

  public synchronized int read() throws IOException {
    if(buf == null){
      throw new IOException("Stream is closed");
    }
    if(count <= pos){
      if(!updateBuffer()){
        return -1;
     }
    }
    pos++;
    validateMark();
    return buf[pos-1];
  }

  public synchronized int read(byte[] b, int off, int len) throws IOException {
    //System.out.println("reading bytes -- off = "+off+", len = "+len+", pos = "+pos+", count ="+count);
    if(buf == null){
      throw new IOException("Stream is closed");
    }
    if(off < 0 || len < 0 || (off + len) > b.length){
      throw new IndexOutOfBoundsException();
    }
    if(len == 0){
      return 0;
    }
    int rd = 0;
    //first take the bytes out of buf
    if(count > pos){
      int cp = count - pos;
      if(len > cp){
        System.arraycopy(buf, pos, b, off, cp);
        off += cp;
        len -= cp;
        pos += cp;
        rd +=cp;
      }else{
        System.arraycopy(buf, pos, b, off, len);
        pos += len;
        validateMark();
        return len;
      }
    }
    //we need more bytes ...
    //System.out.println("we need more bytes -- pos = "+pos+", count ="+count);
    if(len > buf.length && (markpos == -1 || pos > (markpos+marklimit))){
      len = in.read(b, off, len);
      //System.out.println("bytes read from stream -- pos = "+pos+", count ="+count);
    }else{
      //System.out.println("updated buffer -- pos = "+pos+", count ="+count);
      len = updateBuffer() ? read(b, off, len) : -1;
    }
    if(len == -1){
      rd += (rd == 0 ? 0 : 1);
    }
    validateMark();
    return len + rd;
  }

  public synchronized long skip(long n) throws IOException {
    if(buf == null){
      throw new IOException("Stream is closed");
    }
    if(n <= 0){
      return 0;
    }
    long ret = n;
    if((n - count + pos) > 0){
      //not enough bytes in the buffer
      ret = count - pos;
      pos = count;
      if((n - ret) > buf.length && (markpos == -1 || pos > (markpos+marklimit))) {
        ret = in.skip(n-ret)+ret;
      }else if(updateBuffer()){
        ret = skip(n-ret)+ret;
      }
    }else{
      //take bytes from the buffer
      pos += (int)n;
    }
    validateMark();
    return ret;
  }

  public boolean markSupported() {
     return true;
  }

  public synchronized void mark(int markLimit){
     marklimit = (markLimit > 0 ? markLimit : 0);
     markpos = pos;
  }

  public synchronized void reset() throws IOException {
    //System.out.println("markpos = "+markpos+", pos = "+pos+", marklimit "+marklimit);
    if(buf == null || markpos == -1 || pos > markpos+marklimit){
      throw new IOException();
    }
    pos = markpos;
  }

  public synchronized void close() throws IOException {
    if(buf != null){
      markpos = -1;
      marklimit = 0;
      in.close();
      buf = null;
      in = null;
    }
  }

  public synchronized int available() throws IOException {
    if(buf == null){
      throw new IOException("Stream is closed");
    }
    return (count - pos) + in.available();
  }

  /**
  ** this method will will read bytes from in and place them in the buffer
  ** it will update count and pos and should take care of mark/reset stuff
  */
  private boolean updateBuffer() throws IOException {
    if(pos >= (markpos + marklimit)){
      count = in.read(buf, 0, buf.length);
      marklimit = 0;
      pos = 0;
      markpos = -1;
      if(count == -1){
        count = 0;
        return false;
      }
      return true;
    }else{
      //a position is marked and still valid
      byte[] bytes = buf;
      if (marklimit > buf.length){
         //let the buffer grow
         bytes = new byte[marklimit];
      }
      int cp = count - markpos;
      System.arraycopy(buf, markpos, bytes, 0 ,cp);
      count = in.read(bytes, cp, bytes.length - cp);
      pos = cp;
      markpos = 0;
      buf = bytes;
      if(count == -1){
        count = cp;
        return false;
      }
      count += cp;
      return true;
    }
  }

  private void validateMark(){
    //System.out.println("validating mark "+markpos+", "+marklimit+", "+pos);
    if (markpos != -1 && pos > (markpos + marklimit)){
      markpos = -1;
      marklimit = 0;
    }

  }
}
