/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: ObjectInputStream.java,v 1.4 2002/03/27 08:57:17 ruelens Exp $
*/

package java.io;


import java.lang.reflect.Array;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

public class ObjectInputStream extends InputStream implements ObjectInput, ObjectStreamConstants {

  private static final boolean DEBUG = false;

  /** Stream itself. */
  private DataInputStream dis = null;


  // ----------------------------------------------------------------
  //   Buffering blockdata

  /** Maximum buffer size for blockdata is 1024. */
  private static final int MAX_BUFFERSIZE = 1024;

  /** Buffer for blockdata. */
  private byte[] blockDataBuffer = new byte[MAX_BUFFERSIZE];

  /** Current position to read data into the blockdata buffer. */
  private int bufferPointer = 0;
  private int bufferSize    = 0;

  /** Bytes read in stream. */
  private Vector bytesRead = new Vector();

  /** Used in available() and readObject(). */
  private boolean inReadObject = false;

  /**
   * Reading ahead is sometimes necessary.
   * Used in available() and readObject().
   */
  private boolean hasReadAhead = false;
  private byte    readAhead    = 0x00;

  // ----------------------------------------------------------------

  private Object targetObject = null;
  private ObjectStreamClass targetOSClass = null;

  /** Current free handle id. */
  private int handleCounter = baseWireHandle;

  private Hashtable handles = new Hashtable();

  private boolean resolveObjectEnabled = false;

  private ClassLoader myClassLoader = getCallingClass().getClassLoader();

  protected ObjectInputStream()
      throws StreamCorruptedException, IOException {
    //TODO: IOException - Thrown if not called by a subclass.
    //      SecurityException - if a security manager exists
    //      and its checkPermission method denies enabling subclassing.

    initialize();

    //TODO: ObjectInputStream()
    throw new RuntimeException("Implementation does not support subclassing.");
  }

  //OK
  public ObjectInputStream (InputStream in)
      throws StreamCorruptedException, IOException {

    initialize();

    this.dis = new DataInputStream(in);
    readStreamHeader();
  }

  //OK
  /**
   * Huge side-effects!
   * If available data is blockdata, this function will fetch
   * new block if necessary.
   */
  public int available() throws IOException {
    // Trivial buffer check. If code is perfect, this check is not needed.
    if (bufferPointer > bufferSize) {
      error ( "Invalid bufferstatus: "
                + "bufferPointer=" + bufferPointer
                + ", bufferSize="  + bufferSize );
    }

    // First handle if method is called from within readObject.
    if (inReadObject) {
      if ( isEmptyBuffer() ) {
        return 0;
      }
      else {
        OptionalDataException exc = new OptionalDataException();
        exc.length = bufferSize - bufferPointer;
        exc.eof = false;
        throw exc;
      }
    }

    // Now we are sure method is called from a BLOCK_DATA method.

    // If buffer is empty, take a refill.
    if (isEmptyBuffer()) {
      if (DEBUG) print ("DSER: Buffer is empty. Trying to load new BLOCK_DATA.");
      if ( (readAhead() == TC_BLOCKDATA) || (readAhead() == TC_BLOCKDATALONG) ) {
        if (DEBUG) print ("DSER: TC_BLOCKDATA found. Loading...");
        fillBuffer();
        readAheadDone();
      }
    }

    if (DEBUG) print ("DSER: bufferSize - bufferPointer: " + (bufferSize - bufferPointer) );

    // Return available bytes.
    return bufferSize - bufferPointer;
  }

  //OK
  public void close() throws IOException {
    dis.close();
  }

  //OK
  protected final boolean enableResolveObject (boolean enable) {
    boolean oldValue = this.resolveObjectEnabled;

    //TODO: SecurityException - if a security manager exists
    //      and its checkPermission method denies enabling the stream to
    //      do replacement of objects in the stream.

    resolveObjectEnabled = enable;
    return oldValue;
  }

  public int read() throws IOException {
    return readByte();
  }

  public int read (byte[] buffer, int offset, int count) throws IOException {
    int available = available();
    if (available == 0) {
      return -1;
    }

    int totalBytesToRead = Math.min (count, available);

    for (int i = 0; i < totalBytesToRead; i++) {
      buffer[offset+i] = (byte)readByte();
    }

/*  System.arraycopy ( this.blockDataBuffer, this.bufferPointer,
                       buffer, offset,
                       totalBytesToRead );

    bufferPointer += totalBytesToRead;
*/
    return totalBytesToRead;
  }

  public boolean readBoolean() throws IOException {
    return (readByte() == 0x01) ? true : false;
  }

  public byte readByte() throws IOException {
    if (available() > 0) {
      byte result = blockDataBuffer[bufferPointer];
      bufferPointer++;

      if (DEBUG) {
        addByte(result);
      }

      return result;
    }
    else {
      throw new EOFException ("No BLOCK_DATA available.");
    }
  }

  public char readChar() throws IOException {
    int result = 0x00;

    for (int i = 0; i < 2; i++) {
      result = ( result << 8 ) | ( readByte() & 0xFF );
    }

    return (char)result;
  }

  public double readDouble() throws IOException {
    return Double.longBitsToDouble (readLong());
  }

  public float readFloat() throws IOException {
    return Float.intBitsToFloat (readInt());
  }

  public void readFully(byte[] data) throws IOException {
    readFully(data, 0, data.length);
  }

  public void readFully(byte[] data, int offset, int size) throws IOException {
    int done = 0;
    while (done < size-offset) {
     int bytesRead = read(data, offset+done, size-done);
     if (bytesRead == -1) {
       throw new EOFException();
     }
     done += bytesRead;
    }
  }

  public int readInt() throws IOException {
    int result = 0x00;

    for (int i = 0; i < 4; i++) {
      result = ( result << 8 ) | ( readByte() & 0xFF );
    }

    return result;
  }

  public String readLine() throws IOException {
    throw new RuntimeException("Deprecated method.");
  }

  public long readLong() throws IOException {
    long result = 0x00;

    for (int i = 0; i < 8; i++) {
      result = ( result << 8 ) | ( readByte() & 0xFF );
    }

    return result;
  }

  public short readShort() throws IOException {
    int result = 0x00;

    for (int i = 0; i < 2; i++) {
      result = ( result << 8 ) | ( readByte() & 0xFF );
    }

    return (short)result;
  }

  public int readUnsignedByte() throws IOException {
    return readByte();
  }

  public int readUnsignedShort() throws IOException {
    return readShort();
  }

  public String readUTF() throws IOException {
    byte[] string = new byte[readShort()];
    readFully(string);
    return new String(string, "UTF8");
  }

  //OK
  protected void readStreamHeader()
      throws IOException,StreamCorruptedException {

    short stream_magic = readShortNow();
    if (stream_magic != ObjectStreamConstants.STREAM_MAGIC) {
      throw new StreamCorruptedException("Magic number incorrect.");
    }

    short stream_version = readShortNow();
    if (stream_version != ObjectStreamConstants.STREAM_VERSION) {
      throw new StreamCorruptedException("Version number incorrect.");
    }
  }

  public int skipBytes (int count) throws IOException {
    return read (new byte[count]);
  }

  // ----------------------------------------------------------------
  //   Helper methods for class.
  // ----------------------------------------------------------------

  private String toHex (byte b) {
    return Integer.toHexString(((char)b & 0xff));
  }

  private void addByte(byte b) {
    System.out.println(bytesRead.size() + ") " + toHex(b) + "  (" + ((char)b) + ")");
    bytesRead.add (new Byte(b));
  }

  private void print (String s) {
    System.out.println(s);
  }

  private void error (String message) {
    System.out.println(message);
    try {
      this.close();
    }
    catch(IOException ioe){}
    throw new InternalError(message);
  }

  private void error (String message, Throwable error) {
    System.out.println(message);
    error.printStackTrace();
    try {
      this.close();
    }
    catch(IOException ioe){}
    throw new InternalError(message);
  }

  private void initialize() {
    targetObject = null;
    targetOSClass = null;

    resolveObjectEnabled = false;

    handleCounter = baseWireHandle;
    handles       = new Hashtable();

    bytesRead     = new Vector();
  }

  private boolean isEmptyBuffer() {
    return bufferSize == bufferPointer;
  }

  private void fillBuffer() throws IOException {
    // TODO: This constraint can be relaxed. Bit to difficult, for now.
    if ( !isEmptyBuffer() ) {
      error ("Buffer must be empty before next buffer can be read.");
    }

    this.bufferPointer = 0;
    if (readAhead() == TC_BLOCKDATA) {
      this.bufferSize = ((char)readByteNow()) & 0x0ff;
    }
    else if (readAhead() == TC_BLOCKDATALONG) {
      this.bufferSize = readIntNow();
    }
    else {
      throw new EOFException("Expected TC_BLOCKDATA.");
    }

    if (DEBUG) print ("BufferSize: " + bufferSize);


    //TODO: wait for fix
    //dis.readFully(blockDataBuffer, 0, bufferSize);
    for (int i = 0; i < bufferSize; i++) {
      blockDataBuffer[i] = dis.readByte();
    }

    if (DEBUG) {
      for (int i = 0; i < bufferSize; i++) {
        System.out.println(i + ") " + toHex(blockDataBuffer[i]) + "  (" + ((char)blockDataBuffer[i]) + ")");
      }
    }
  }

  /** Checks if a class is an array of primitives. */
  private boolean isPrimitiveArray(Class clazz) {
    String classname = clazz.getName();
    if (classname.length() >= 2) {
      boolean isArray = (classname.charAt(0) == '[');
      boolean isArrayOfArrays  = (classname.charAt(1) == '[');
      boolean isArrayOfObjects = (classname.charAt(1) == 'L');

      return isArray && !isArrayOfArrays && !isArrayOfObjects;
    }

    return false;
  }

  /** Returns the typecode of primitive array. */
  private char getPrimitiveArrayTypeCode(Class clazz) {
    if (!isPrimitiveArray(clazz)) {
      error ("This is not a primitive array, please do check first.");
    }

    return clazz.getName().charAt(1);
  }

  /** Gets you the next byte, can be a cached one. */
  private synchronized byte readAhead() throws IOException {
    if (!hasReadAhead) {
      readAhead    = readByteNow();
      hasReadAhead = true;
    }

    return readAhead;
  }

  /**
   * Flushes the next byte from the cache.
   * Must be called at a certain moment after the readAhead().
   */
  private synchronized void readAheadDone() {
    readAhead    = 0x00;
    hasReadAhead = false;
  }

  boolean readBooleanNow() throws IOException {
    if (DEBUG) {
      byte b = readByteNow();
      return (b == 0x01) ? true : false;
    }

    return dis.readBoolean();
  }

  byte readByteNow() throws IOException {
    if (DEBUG) {
      byte result = dis.readByte();
      addByte(result);
      return result;
    }

    return dis.readByte();
  }

  char readCharNow() throws IOException {
    if (DEBUG) {
      int result = 0x00;

      for (int i = 0; i < 2; i++) {
        result = ( result << 8 ) | ( readByteNow() & 0xFF );
      }

      return (char)result;
    }

    return dis.readChar();
  }

  double readDoubleNow() throws IOException {
    if (DEBUG) {
      return Double.longBitsToDouble (readLongNow());
    }

    return dis.readDouble();
  }

  float readFloatNow() throws IOException {
    if (DEBUG) {
      int data = readIntNow();
      return Float.intBitsToFloat (data);
    }

    return dis.readFloat();
  }

  int readIntNow() throws IOException {
    if (DEBUG) {
      int result = 0x00;

      for (int i = 0; i < 4; i++) {
        result = ( result << 8 ) | ( ((char)readByteNow()) & 0xFF );
      }

      return result;
    }

    return dis.readInt();
  }

  long readLongNow() throws IOException {
    if (DEBUG) {
      long result = 0x00;

      for (int i = 0; i < 8; i++) {
        result = ( result << 8 ) | ( readByteNow() & 0xFF );
      }

      return result;
    }

    return dis.readLong();
  }

  short readShortNow() throws IOException {
    if (DEBUG) {
      int result = 0x00;

      for (int i = 0; i < 2; i++) {
        result = ( result << 8 ) | ( readByteNow() & 0xFF );
      }
      return (short)result;
    }

    return dis.readShort();
  }

  String readUTFNow() throws IOException {
    if (DEBUG) {
      short size = readShortNow();
      char[] string = new char[size];
      for (int i = 0; i < size; i++) {
        string[i] = (char)readByteNow();
      }
      return new String(string);
    }

    return dis.readUTF();
  }

  private synchronized void setInReadObject (boolean value) {
    this.inReadObject = value;
  }

  // ----------------------------------------------------------------
  //   Methods implementing the grammar.
  // ----------------------------------------------------------------

  //OK
  /**
   * classDesc: newClassDesc
   *            nullReference
   *            (ClassDesc)prevObject
   */
  private ObjectStreamClass readClassDesc()
      throws IOException, ClassNotFoundException {
    byte b = readByteNow();

    switch (b) {
      case TC_NULL:
        return null;
      case TC_CLASSDESC:
        return readNewClassDesc();
      case TC_REFERENCE:
        int handle = readIntNow();
        return (ObjectStreamClass)handles.get(new Integer(handle));
    }

    throw new StreamCorruptedException ("Problem in readClassDesc, read [" + b + "]");
  }

  //OK
  /**
   * newArray : TC_ARRAY classDesc newHandle (int)<size> values[size]
   */
  private Object readNewArray() throws IOException, ClassNotFoundException {
    // TC_ARRAY, already done

    // classDesc
    this.targetOSClass = readClassDesc();

    // (int)<size>
    int length = readIntNow();
    if (DEBUG) print ("DSER: Reading new array, size: " + length);

    // values[size] and newHandle
    Class clazz = this.targetOSClass.forClass();
    if (isPrimitiveArray(clazz)) {
      if (DEBUG) print ("DSER: Primitive array, size: " + length);
      switch (getPrimitiveArrayTypeCode(clazz)) {
        case 'B': {
          byte[] result = new byte[length];
          assignNewHandle(result);
          dis.readFully (result);

          return result;
        }
        case 'C': {
          char[] result = new char[length];
          assignNewHandle(result);
          for (int i = 0; i < length; i++) {
            result[i] = readCharNow();
          }

          return result;
        }
        case 'D': {
          double[] result = new double[length];
          assignNewHandle(result);
          for (int i = 0; i < length; i++) {
            result[i] = readDoubleNow();
          }

          return result;
        }
        case 'F': {
          float[] result = new float[length];
          assignNewHandle(result);
          for (int i = 0; i < length; i++) {
            result[i] = readFloatNow();
          }

          return result;
        }
        case 'I': {
          int[] result = new int[length];
          assignNewHandle(result);
          for (int i = 0; i < length; i++) {
            result[i] = readIntNow();
          }

          return result;
        }
        case 'J': {
          long[] result = new long[length];
          assignNewHandle(result);
          for (int i = 0; i < length; i++) {
            result[i] = readLongNow();
          }

          return result;
        }
        case 'S': {
          short[] result = new short[length];
          assignNewHandle(result);
          for (int i = 0; i < length; i++) {
            result[i] = readShortNow();
          }

          return result;
        }
        case 'Z': {
          boolean[] result = new boolean[length];
          assignNewHandle(result);
          for (int i = 0; i < length; i++) {
            result[i] = readBooleanNow();
          }

          return result;
        }
        default: {
          throw new StreamCorruptedException("Invalid primitive typecode for array: " + getPrimitiveArrayTypeCode(clazz));
        }
      }
    }
    else {
      if (DEBUG) print ("DSER: Object array, size: " + length);

      Object[] result = (Object[]) Array.newInstance(clazz.getComponentType(), length);
      assignNewHandle(result);

      Object[] objectarray = new Object[length];
      for (int i = 0; i < length; i++) {
        objectarray[i] = readObject();
      }

      if (DEBUG) print ("DSER: Copying array into correct arraytype.");
      System.arraycopy (objectarray, 0, result, 0, length);

      return result;
    }
  }

  //OK
  /** newClass: TC_CLASS classDesc newHandle */
  private Class readNewClass() throws IOException, ClassNotFoundException {
    //TC_CLASS, already done

    // classDesc
    ObjectStreamClass osc = readClassDesc();

    // newHandle
    Class clazz = osc.forClass();
    assignNewHandle(clazz);

    return clazz;
  }


  private ObjectStreamClass getObjectStreamClass (String classname, long serialVersionUID)
      throws IOException, ClassNotFoundException {

/*
    Class clazz = null;
    try {
      clazz = Class.forName(classnamer);
    }
    catch (ClassNotFoundException e) {
      if (DEBUG) print ("DSER: Couldn't find class " + classname);
    }

    if (clazz == null) {
      SecurityManager sm = new SecurityManager();
      if (DEBUG) print ("DSER: SecurityManager: " + sm);

      ClassLoader classloader = sm.currentClassLoader();
      if (classloader != null) {
        if (DEBUG) print ("DSER: ClassLoader: " + classloader);
        clazz = classloader.loadClass (classname);
        if (DEBUG) print ("DSER: Class: " + clazz);
      }
    }

    if (clazz == null) {
      throw new ClassNotFoundException();
    }
*/
    System.out.println("CG: using "+myClassLoader+" to load "+classname);
    Class clazz = Class.forName(classname,false,myClassLoader);
    // if Class.forName() throws ClassNotFoundException, so be it

    ObjectStreamClass result = ObjectStreamClass.lookup (clazz);

    if (result.getSerialVersionUID() != serialVersionUID) {
      //TODO: Check if correct exception is thrown.
      if (DEBUG) print ( "DSER: "
               + "SerialVersionUIDs for class " + clazz.getName() + " do not match. "
               + "Local version: " + result.getSerialVersionUID() + ". "
               + "Serialized version: " + serialVersionUID + ".");

      throw new InvalidClassException ( "SerialVersionUIDs for class " + clazz.getName() + " do not match. ");
    }
    return result;
  }

  //OK
  /**
   * newClassDesc:
   *   TC_CLASSDESC className serialVersionUID newHandle classDescInfo
   */
  private ObjectStreamClass readNewClassDesc()
      throws IOException, ClassNotFoundException {
    // TC_CLASSDESC, already done

    // className
    String className = readUTFNow();

    // serialVersionUID
    //TODO: Where do we check its serialVersionUID?
    long serialVersionUID = readLongNow();

    // newHandle
//    ObjectStreamClass result = new ObjectStreamClass(className,serialVersionUID);
    ObjectStreamClass result = getObjectStreamClass(className,serialVersionUID);
    assignNewHandle(result);

    // classDescInfo
    result.readClassDescInfo(this);
    if (DEBUG) print ("DSER: Done reading classdescription.");

    return result;
  }

  //OK
  /** newObject: TC_OBJECT classDesc newHandle classData[] */
  public Object readNewObject() throws IOException, ClassNotFoundException {
    // TC_OBJECT already done

    // classDesc
    this.targetOSClass = readClassDesc();
    if (DEBUG) print ("DSER: readNewObject, after readClassDesc " + this.targetOSClass);

    // allocating object, newHandle
    try {
      // Creates an instance without running the initializers
      // Is done because some classes do not have default
      // constructors to create a new instance.
      this.targetObject = allocNewInstance(targetOSClass.forClass());

      //TODO: if (targetObject == null) => no assignNewHandle?
      //TODO: is (targetObject == null) allowed at all?
      assignNewHandle(targetObject);
    }
    catch (IllegalAccessException e) {
      //TODO: what exception should be thrown?
      throw new ClassNotFoundException("JVDBG: illegal access");
    }
    catch (InstantiationException e) {
      //TODO: what exception should be thrown?
      throw new ClassNotFoundException("JVDBG: instantiationeexception");
    }


    // classData, depending on classDescFlag
    ObjectStreamClass[] supers = getSuperOSClasses(this.targetOSClass);


    for (int i = 0; i < supers.length; i++) {
      if (DEBUG) print ("DSER:" + i + ". targetobject: " + this.targetObject
                         + " and super: " + supers[i]);
      byte targetClassDescFlag = supers[i].getClassDescFlag();
      if ((SC_SERIALIZABLE & targetClassDescFlag) != 0 ) {
        // nowrclass
        Object remember = this.targetObject;
        this.targetOSClass = supers[i];
        if(this.targetOSClass.hasReadObjectMethod()){
          //readObject available ...
          boolean rememberRead = this.inReadObject;
          setInReadObject(false);
          if (DEBUG) print ("DSER: I will invoke readObject on " + targetObject + ", type " + supers[i]);
          supers[i].invokeReadMethod("readObject",targetObject,this);
          setInReadObject(rememberRead);
          // TC_ENDBLOCKDATA
          dis.readByte();
        }
        else {
          defaultReadObject();
        }
        this.targetObject = remember;
      }
      else if ( (SC_EXTERNALIZABLE & targetClassDescFlag) != 0 ) {
        if ( (SC_BLOCK_DATA & targetClassDescFlag) == 0 ) {
          // externalContents
          //TODO: Finish implementation.
          throw new RuntimeException("Not yet implemented.");
        }
        else {
          // objectAnnotation
          //TODO: Finish implementation. Call readExternal?
          throw new RuntimeException("Not yet implemented.");
        }
      }
    }

    if (DEBUG) print ("DSER: Returning " + targetObject + " of type " + targetObject.getClass());
    return targetObject;
  }

  //OK
  private Object readPreviousObject() throws IOException {
    int handle = readIntNow();
    if (DEBUG) print ("DSER: readPreviousObject, handle: " + handle);
    return (handles.get(new Integer(handle)));
  }

  //OK
  ObjectStreamClass readSuperClassDesc()
      throws IOException, ClassNotFoundException {
    return readClassDesc();
  }

  // ----------------------------------------------------------------

  public final Object readObject()
    throws OptionalDataException, ClassNotFoundException, IOException {

    // available() should return 0, if in readObject().
    setInReadObject (true);

    try {
      if (DEBUG) print ("DSER: In readObject.");

      // available() could have read 1 byte to much.
      // Use that byte.
      byte b = readAhead();

      // Invalidate byte.
      readAheadDone();

      if (DEBUG) print ("DSER: First byte: "+b);

      switch (b) {
        case TC_OBJECT:
          return readNewObject();
        case TC_CLASS:
          return readNewClass();
        case TC_ARRAY:
          return readNewArray();
        case TC_STRING:
          return readNewString();
        case TC_LONGSTRING:
          return readNewLongString();
        case TC_CLASSDESC:
          return readNewClassDesc();
        case TC_NULL:
          return null;
        case TC_REFERENCE:
          return readPreviousObject();
        case TC_EXCEPTION:
          return readException();
        case TC_RESET:
          return handleReset();
      }

      throw new StreamCorruptedException ("Problem reading object type, read [" + b + "]");
    }
    finally {
      setInReadObject (false);
    }
  }


  public void defaultReadObject() throws IOException, ClassNotFoundException  {
    if (DEBUG) print ("DSER: defaultReadObject of " + targetObject + " with osc " + targetOSClass);
    readClassData(targetOSClass,targetObject);
  }

  private void readClassData (ObjectStreamClass osc, Object obj)
    throws IOException, ClassNotFoundException {

/*
    ObjectStreamClass parent = osc.getParentObjectStreamClass();
    if (parent != null) {
      readClassData(parent,obj);
    }
*/
    if (DEBUG) print ("DSER: Reading field data for " + osc);
    osc.readFieldData(this,obj);
    if (DEBUG) print ("DSER: Reading fielddata done.");
  }

  private ObjectStreamClass[] getSuperOSClasses(ObjectStreamClass base) {
    int c = 0;
    ObjectStreamClass move = base;
    while (move != null) {
      c++;
      move = move.getParentObjectStreamClass();
    }
    move = base;
    ObjectStreamClass[] answer = new ObjectStreamClass[c];
    for (int i = c; i>0; i--) {
      answer[i-1] = move;
      move = move.getParentObjectStreamClass();
    }
    return answer;
  }

  //TODO: Is this correct?
  private Object readException() {
    return new Object();
  }

  //TODO: Is this correct?
  private Object handleReset() {
    return new Object();
  }

  private synchronized void assignNewHandle(Object object) {
    this.handles.put (new Integer(handleCounter), object);
    handleCounter++;
  }


  protected String readStringHandle() throws IOException  {
    int handle = readIntNow();
    return (String)handles.get(new Integer(handle));
  }

  protected String readNewString() throws IOException {
    String result = readUTFNow();
    assignNewHandle(result);
    return result;
  }

  protected String readNewLongString() throws IOException {
    byte[] string = new byte[(int)readLongNow()];
    dis.readFully(string);

    return new String (string);
  }


  /** Creates an instance of that clazz without running the initializers. */
  //TODO: Must run initializer of the non-serializable superclass
  //      by calling its default constructor. Exception if there is none.
  private native Object allocNewInstance(Class clazz) throws IllegalAccessException, InstantiationException;

  private native Class getCallingClass();

}


