/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: PipedInputStream.java,v 1.2 2002/03/14 16:37:29 gray Exp $
*/

package java.io;

/** An input stream which reads whatever is written to some output stream.
*/
public class PipedInputStream extends InputStream {

  /** The (default) size of the pipe buffer.
  */
  protected static final int PIPE_SIZE = 1024;

  /** The buffer in which bytes are stored.
  */
  protected byte[] buffer;

  /** The index where the next byte will be stored.
  */
  protected int in;

  /** The index where the next byte will be retrieved.
  */
  protected int out;
  
  /** The PipedOutputStream to which we are connected.
  */
  private PipedOutputStream peer;

  /** The current state of this pipe:
   **    UNOPENED = the pipe has never been connected
   **    HALFOPEN = waiting for the peer to call connect()
   **    OPEN     = everything hunkydory
   **    CLOSED   = closed, finished. It's all over.
   **    BROKEN   = the thread that created us is no longer alive.
   */
  private int state;
  private static final int UNOPENED = 0;
  private static final int HALFOPEN = 1;
  private static final int OPEN = 2;
  private static final int CLOSED = 3;
  private static final int BROKEN = 4;

  /**
   ** The thread which cunstructed this PipedInputStream.
   */
  private Thread reader;

  /** Create a PipedInputStream and connect it to a PipedOutputStream.
  */
  public PipedInputStream(PipedOutputStream src)
    throws IOException
  {
    create();
    connect(src);
  }

  /** Create a PipedInputStream which is not yet connected to a PipedOutputStream.
  */
  public PipedInputStream()
  {
    create();
  }

  /**
   ** Code shared by both constructors.
   */
  private void create() {
    state = UNOPENED;
    buffer = new byte[PIPE_SIZE];
    in = -1;
    out = 0;
    reader = Thread.currentThread();
  }

  /**
   ** Set the peer, and call its connect if it didn't call ours.
   */
  public synchronized void connect(PipedOutputStream src)
    throws IOException
  {
    switch (state) {
    case UNOPENED:
      peer = src;
      state = HALFOPEN;
      src.connect(this);
      /* fall through */
 
    case HALFOPEN:
      state = OPEN;
      break;

    case OPEN:
      throw new IOException("already connected to "+peer);

    case CLOSED:
      throw new IOException("previously closed");

    case BROKEN:
      throw new IOException("thread has terminated");

    }
  }

  /** Get the number of bytes which could be read from this stream without blocking.
  */
  public synchronized int available()
    throws IOException
  {
    if (state != OPEN || peer == null) {
      throw new IOException("stream not open");
    }

    return in < 0 ? 0 : in == out ? PIPE_SIZE : (in - out + PIPE_SIZE) % PIPE_SIZE;
  }

  /**
   ** Find out if our peer is in state BROKEN (throws IOException) or CLOSED
   ** (returns -1).  Else return a nonnegative result.
   */
  private int testPeer() throws IOException {
    if (peer.isBroken()) {
      throw new IOException("broken pipe");
    }

    return peer.isClosed() ? -1 : available();

  }

  /**
   ** Read one byte.  If no byte is available and the peer is still OPEN,
   ** wait for the peer to write something.
   */
  public synchronized int read() throws IOException {
    if (state != OPEN || peer == null) {
      throw new IOException("not connected");
    }

    while (in < 0) {
      if (testPeer() < 0) {

        return -1;

      }

      try {
        wait();
      } catch (InterruptedException ie) {
      }
    }

    byte result = buffer[out];
    out = (out+1) % PIPE_SIZE;
    if (out == in) {
      in = -1;
    }
    notify();

    return result;
  }

  /** 
   ** Read several bytes.  If nothing is available and the peer is still OPEN,
   ** wait for the peer to write at least one byte.  Return the number of
   ** bytes read.
   */
  public synchronized int read(byte[]b, int off, int len)
    throws IOException, NullPointerException, ArrayIndexOutOfBoundsException
  {
    if (b==null) 
      throw new NullPointerException();

    if(off<0 || len<0 || off+len > b.length)
      throw new ArrayIndexOutOfBoundsException();

    if (state != OPEN || peer == null) {
      throw new IOException("not connected");
    }

    while (in < 0) {
      if (testPeer() < 0) {

        return -1;

      }

      try {
        wait();
      } catch (InterruptedException ie) {
      }
    }

    int remaining = len;
    while (remaining-- > 0) {
      b[off++] = buffer[out];
      out = (out+1) % PIPE_SIZE;
      if (out == in) {
        in = -1;
        break;
      }
    }
    notify();

    return len - remaining;
  }

  /** Append one byte to the buffer.  Used by PipedOutputStream.
  */
  protected synchronized void receive(int ib)
    throws IOException
  {
    if (state != OPEN) {

      throw new IOException("closed by listener");

    }

    while (in == out) {
      try {
        wait();
      } catch (InterruptedException ie) {
        throw new InterruptedIOException();
      }
    }

    byte b = (byte)ib;
    if (in == -1) {
      buffer[0] = b;
      in = 1;
      out = 0;
    }
    else {
      buffer[in] = b;
      in = (in+1) % PIPE_SIZE;
    }
    notify();
  }

  /** Close the pipe.
  */
  public void close() throws IOException {
    buffer = null;
    in = -1;
    state = CLOSED;
  }

  /**
   ** Package-visible method used by PipedOutputStream to find out whether
   ** the thread which created the pipe has died.
   */
  boolean isBroken() {
    if (state == BROKEN) {

      return true;

    }

    if (!reader.isAlive()) {
      state = BROKEN;

      return true;

    }

    return false;
  }
}
