/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: PipedOutputStream.java,v 1.2 2002/03/14 16:37:29 gray Exp $
*/

package java.io;

/** An output stream which can be used as input by some input stream.
*/
public class PipedOutputStream extends OutputStream {

  /** The PipedInputStream to which we are connected.
  */
  private PipedInputStream peer;

  /** The current state of this pipe:
   **    UNOPENED = the pipe has never been connected
   **    HALFOPEN = waiting for the peer to call connect()
   **    OPEN     = everything hunkydory
   **    CLOSED   = closed, finished. It's all over.
   **    BROKEN   = the thread that created us is no longer alive.
   */
  private int state;
  private static final int UNOPENED = 0;
  private static final int HALFOPEN = 1;
  private static final int OPEN = 2;
  private static final int CLOSED = 3;
  private static final int BROKEN = 4;

  /**
   ** The thread which cunstructed this PipedOutputStream.
   */
  private Thread writer;

  /** Create a PipedOutputStream and connect it to a PipedInputStream.
  */
  public PipedOutputStream(PipedInputStream snk)
    throws IOException
  {
    writer = Thread.currentThread();
    state = UNOPENED;
    connect(snk);
  }

  /** Create a PipedOutputStream not yet connected to a PipedInputStream.
  */
  public PipedOutputStream()
  {
    writer = Thread.currentThread();
    state = UNOPENED;
  }

  /** Connnect this PipedOutputStream to a PipedInputStream.
  */
  public void connect(PipedInputStream snk)
    throws IOException
  {
    switch (state) {
    case UNOPENED:
      peer = snk;
      state = HALFOPEN;
      snk.connect(this);
      /* fall through */
 
    case HALFOPEN:
      state = OPEN;
      break;

    case OPEN:
      throw new IOException("already connected to "+peer);

    case CLOSED:
      throw new IOException("previously closed");

    case BROKEN:
      throw new IOException("thread has terminated");

    }
  }

  /**
   ** Write one byte, as soon as the buffer is non-full.
   */
  public void write(int ib) throws IOException {
    if (state != OPEN || peer == null) {
      throw new IOException("not connected");
    }

    synchronized(peer) {
      if (peer.isBroken()) {
        throw new IOException("broken pipe");
      }
      peer.receive(ib);
      flush();
    }
  }

  /**
   ** Write several bytes.
   */
  public void write(byte[] b, int off, int len)
    throws IOException, NullPointerException, ArrayIndexOutOfBoundsException 
  {
    if (state != OPEN || peer == null) {

      throw new IOException("not connected");

    }

    if(off<0 || len<0 || off+len>b.length)
      throw new ArrayIndexOutOfBoundsException();

    if (peer.isBroken()) {
      throw new IOException("broken pipe");
    }

    for (int i = 0; i < len; ++i) {
      peer.receive(b[i]);
    }
  }

  /** Tell anyone listening that there's data in the pipe, just in case
   ** they didn't notice.
   */
  public synchronized void flush()
    throws IOException
  {
    if (state != OPEN || peer == null) {

      throw new IOException("not connected");

    }

    synchronized(peer) {
      peer.notify();
    }
  }

  /** Close the pipe.
  */
  public void close()
    throws IOException
  {
    state = CLOSED;
    peer = null;
  }

  /**
   ** Package-visible method used by PipedInputStream to find out whether
   ** we closed the pipe yet.
   */
  boolean isClosed() {
    return state == CLOSED;
  }

  /**
   ** Package-visible method used by PipedInputStream to find out whether
   ** the thread which created the pipe has died.
   */
  boolean isBroken() {
    if (state == BROKEN) {

      return true;

    }

    if (!writer.isAlive()) {
      state = BROKEN;

      return true;

    }

    return false;
  }

}
