/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: PipedReader.java,v 1.1.1.1 2001/11/21 10:41:34 pollock Exp $
*/

package java.io;


public class PipedReader extends Reader {


	private static final int pSize = 1024;

/*
**  pipe and pointers are used internally to synchronize datatransfer ...
**
**  pipe is a char[] of size pSize. The written data is stored in this array
**
**  pointers is an int[] of size 2. pointers[0] is de read position, where pointers[1] is the write position.
**  Only the PipedWriter is allowed to change the write pointer(WP). THe read pointer (RP)can only be moved by the reader
**  the pipe is empty if the pointers are equal (same position in pipe) and the buffer is full if RP == WP+1 (this means the
**  effectieve buffer size is pSize -1) ...
**  NOTE: WP points to an empty cell, while RP points to an unread char (unless RP == WP)
**  There is no need to synchronize access to the pipe since the RP pointer is only moved after char are reads
**  so there is no chance of overwritting unread chars	
*/	
	private char[] pipe=null;
	private int [] pointers=null;
	
	private boolean closed = false;
	private Thread thread;
        private PipedWriter srcW;
	
	public PipedReader() {
		thread = Thread.currentThread();
	}
	
	public PipedReader(PipedWriter src) throws IOException {
		thread = Thread.currentThread();
	        src.connect(this);
	}
	
	
	public void close() throws IOException{
	 	synchronized (lock) {
		   if (!closed) {
		 	closed = true;
		 	pipe = null;
		   }
		}
	}

	public void connect(PipedWriter src) throws IOException {
	 	synchronized (lock) {
	 		src.connect(this);
	 	}
	}
	
	public int read() throws IOException {
    		synchronized (lock) {
	        	//System.out.println("thread is alive ="+thread.isAlive());
			if (closed || pipe == null || !thread.isAlive()) {
		 		throw new IOException();	
			}
    		        while(pointers[0] == pointers[1]) {
    		         	Thread.yield();
    		         	if(!thread.isAlive() || srcW.closed()) {
    		         	 	throw new IOException();
    		         	}
    		        }
	                int i = (int) pipe[pointers[0]];
	                pointers[0] = (pointers[0]+1) % pSize;
	                return i;
	        }	
	}
	
	public int read(char buff[], int offset, int count) throws IOException {
   		if (offset < 0 || count < 0 || offset+count > buff.length) throw new ArrayIndexOutOfBoundsException();
		int returnvalue = count;
    		synchronized (lock) {
		   if (closed || pipe == null || !thread.isAlive()) {
		 	throw new IOException();	
		    }
     		    while(count > 0) { //if count > 0 we still have bytes to write ...
	                //System.out.println("READING count ="+count+", R pointer "+pointers[0]+", W pointer "+pointers[1]);
    		        while(pointers[0] == pointers[1]) {
    		         	Thread.yield();
    		         	if(!thread.isAlive() || srcW.closed()) {
    		         	 	throw new IOException();
    		         	}
    		        }   		
    		        int available = (pSize + pointers[1] - pointers[0])%pSize;
	                //System.out.println("READING 2 count ="+count+", available ="+available+", R pointer "+pointers[0]+", W pointer "+pointers[1]);
    		        available = (available > count ? count : available);
    		        internalread(buff, offset, available);
    		        offset += available;
    		        count -= available;
	            }
	        }
		return returnvalue;
	}

	private void internalread(char [] buff, int offset, int count) {
                //System.out.println("internal READ count ="+count+", R pointer "+pointers[0]+", W pointer "+pointers[1]);
		if (count+pointers[0] <= pSize) {
		        System.arraycopy(pipe, pointers[0], buff, offset, count);
		        pointers[0] += count;
		}
		else {
			int nr = pSize - pointers[0];
		        System.arraycopy(pipe, pointers[0], buff, offset, nr);
		        pointers[0]=0;
		        internalread(buff, offset+nr, count-nr);				 	
		} 		
	 	
	
	}
		
	public boolean ready() throws IOException{
		synchronized (lock) {
			if (closed || pipe == null || !thread.isAlive()) {
		 		throw new IOException();	
			}
		  	return (pointers[1] != pointers[0]);
		}
	
	}
	
	// default acces methods : for communication between PipedReader and Writer ...
	
	void setPipe(char[] pp) {
	 	pipe = pp;
	}
	
	void setPointers(int[] ptrs) throws IOException {
	 	if (closed || pointers != null) {
	 		throw new IOException();
	 	}
	 	pointers = ptrs;
	}
	
	Thread switchThreads(Thread trd, PipedWriter src) {
	 	srcW = src;
	 	Thread t = this.thread;
	 	this.thread = trd;
	 	return t;
	}
	
	boolean closed() {
	 	return closed;
	}
	
}