/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: ClassLoader.java,v 1.10 2002/03/27 08:09:43 gray Exp $
*/

package java.lang;

import com.acunia.wonka.Etc;
import com.acunia.wonka.SystemClassLoader;
import java.io.InputStream;
import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.security.CodeSource;
import java.security.Permissions;
import java.security.ProtectionDomain;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.HashMap;
import java.util.Properties;
import java.util.Vector;

public abstract class ClassLoader {

  /**
   ** The application class loader, used to load the starting class and
   ** (confusingly) the one returned by getSystemClassLoader().
   */
  private static ClassLoader applicationClassLoader;

  /**
   ** The real system class loader, i.e. the one which loads from bootclasspath
   ** those classes (and resources) not loaded by the bootstrap class loader.
   */
  private static ClassLoader systemClassLoader;

  /**
   ** The default protection domain.  Has a CodeSource of (null,null)
   ** and default Permissions.
   */
  private static ProtectionDomain defaultProtectionDomain;

  /** The list of packages known to the bootstrap class loader.
   */
  private static Vector bootstrap_packages;

  /** The packages which have been defined by this ClassLoader.
   ** Only includes those for which a definePackage() was done.
   ** The key is the package name, with as associated data the 
   ** relevant Package object.
   */
  private Hashtable packages = new Hashtable();

  /** The code sources of the packages which have been defined by 
   ** this ClassLoader.
   ** Only includes those for which a definePackage() was done.
   ** The key is the Package object, with as associated data the 
   ** relevant CodeSource.
   */
  private HashMap package_sources;

  /**
   ** This lock controls access to the class hashtable, which lurks 
   ** in native code.
   */
  private Object lock = new Object();

  /**
   ** Every ClassLoader has a parent, except the system ClassLoader,
   ** which has (would have) no parent - or rather its parent is 
   ** the bootstrap class loader, represented as 'null'.
   */
  private ClassLoader parent;

  /**
   ** Get a reference to the default protection domain.
   ** A cheap substitute for having a static initialiser.
   */
  private static synchronized ProtectionDomain get_defaultProtectionDomain() {
    if (defaultProtectionDomain == null) {
      defaultProtectionDomain = new ProtectionDomain(new CodeSource(null,null),new Permissions());
    }
    return defaultProtectionDomain;
  }

  /**
   ** Identify the most useful classloader availble - the applicationClassLoader
   ** if already defined, but failing that the systemClassLoader.  Used to load
   ** resources (as opposed to classes) to avoid an infinite regress involving
   ** system.properties (we use systemClassLoader to load system.properties, so
   ** this needs to be in bootclasspath).
   ** Walks like a hack and quux like hack, so it probably is a hack.
   */
  private static ClassLoader getPertinentClassLoader() {
    ClassLoader loader = applicationClassLoader;
    if (loader == null) {
      systemClassLoader = SystemClassLoader.getInstance();
      loader = systemClassLoader;
    }

    return loader;
  }

  private static void permissionCheck(String permission) {
    if (com.acunia.wonka.SecurityConfiguration.USE_ACCESS_CONTROLLER) {
      java.security.AccessController.checkPermission(new RuntimePermission(permission));
    }
    else if (com.acunia.wonka.SecurityConfiguration.USE_SECURITY_MANAGER) {
      SecurityManager sm = System.getSecurityManager();
      if (sm != null) {
        sm.checkPermission(new RuntimePermission(permission));
      }
    }
  }

  /** Any ClassLoader created using the default constructor will
   ** have as its parent the system ClassLoader.
   */
  protected ClassLoader() throws SecurityException {
    this(getSystemClassLoader());
  }

  /**
   ** This new-to-Java-1.2 constructor allows the parent to be
   ** specified by the caller.
   */
  protected ClassLoader(ClassLoader parent) throws SecurityException {
    if (getCallingClassLoader() != null || applicationClassLoader != null) {
      if (com.acunia.wonka.SecurityConfiguration.USE_ACCESS_CONTROLLER) {
        java.security.AccessController.checkPermission(new RuntimePermission("createClassLoader"));
      }
      else if (com.acunia.wonka.SecurityConfiguration.USE_SECURITY_MANAGER) {
        SecurityManager sm = System.getSecurityManager();

        if (sm!=null) {
          sm.checkCreateClassLoader();
        }
      }
    }

    this.parent = parent;
    this.create();
  }

  private native void create();

  /**
   ** As of Java 1.2 this method is no longer abstract.  Instead it
   ** implements a default policy which can be be described as 
   ** "responsible delegation".
   */
  protected synchronized Class loadClass(String name, boolean resolve) 
    throws ClassNotFoundException 
  {
    Class loaded = findLoadedClass(name);
    if (loaded == null) {
      try {
        if (parent == null) {
          loaded = findClass(name);
        }
        else {
          loaded = parent.loadClass(name, false);
        }
      }
      catch (ClassNotFoundException e) {
        loaded = findClass(name);
        // If that throws a ClassNotFoundException then so be it.
      }
    }

    if (resolve) {
      try {
        resolveClass(loaded);
      }
      catch (NullPointerException e) {}
    }

    return loaded;
  }

  /**
   ** If no 'resolve' parameter is supplied then the default is 'false'
   ** (N.B. in Java 1.1 it was 'true'!)
   */
  public Class loadClass(String name) throws ClassNotFoundException {
    return loadClass(name, false);
  }

  /**
   ** The default implementation of findClass() is just a mugtrap: any
   ** real ClassLoader must override it.
   */
  protected Class findClass(String name) 
    throws ClassNotFoundException
  {
    String message = this.getClass()+" failed to override method findClass(String) of ClassLoader";
    System.err.println("----------------- Lame ClassLoader Warning --------------");
    System.err.println(message);
    System.err.println("------------------ This sucks!  Fix it !  ---------------");

    throw new ClassNotFoundException(message);
  }

  /** Turns a bag of bytecodes into a Class and register it in the namespace of this ClassLoader. 
   ** The class is not yet resolved.
   **
   ** If the name of the class begins with ``java.'' or some other
   ** bootstrap-only package, or (TODO) if the package implied by the class name
   ** already exists and has other signers, throw a SecurityException.
   ** If the ProtectionDomain is null, use the default protection domain.
   */
  protected final Class defineClass(String name, byte data[], int offset, int length, ProtectionDomain pd)
    throws NullPointerException, ArrayIndexOutOfBoundsException, ClassFormatError
  {
    if (offset < 0 || length < 0 || offset + length > data.length) {
      throw new ArrayIndexOutOfBoundsException();
    }

    int dot = name.lastIndexOf('.');
    String package_name = dot < 0 ? "" : name.substring(0, dot);

    // TODO: parameterise the list of forbidden packages
    //  --> use code of checkPackageDefinition() in SecurityManager?
    if ((package_name.startsWith("java.") || package_name.equals("com.acunia.wonka")) && !(this instanceof com.acunia.wonka.SystemClassLoader)) {
      throw new SecurityException("Classes in package "+name.substring(0,name.lastIndexOf('.'))+" may only be loaded by the bootstrap class loader!");
    }

    ProtectionDomain actual_pd;
    CodeSource actual_source;
    Package    existing_package;
    CodeSource existing_source;

    if (pd == null) {
      actual_pd = get_defaultProtectionDomain();
      actual_source = null;
    }
    else {
      actual_pd = pd;
      actual_source = pd.getCodeSource();
    }

// TODO: if destination package has other signers, throw SecurityException
//       Note that the list of signers is defined by the first class to be added to a package.

    synchronized(this) {
      if (package_sources == null) {
        package_sources = new HashMap();
        existing_package = null;
      }
      else {
        existing_package = getPackage(package_name);
      }

      if (existing_package == null) {
        existing_source = null;
      }
      else {
        existing_source = (CodeSource)package_sources.get(existing_package);
        if (existing_source == null) {
          package_sources.put(existing_package, actual_source);
        }
        else {
          if (!existing_source.equals(actual_source)) {
            throw new SecurityException("attempt to create class "+name+" from "+actual_source+" when package "+package_name+" was already defined from "+existing_source);
          }
        }
      }
    }

    return _defineClass(name,data,offset,length,actual_pd);
  }

  /**
   ** If no ProtectionDomain is specified, the default domain is used.
   */
  protected final Class defineClass(String name, byte data[], int offset, int length) 
    throws ClassFormatError 
  {
    return defineClass(name,data,offset,length,get_defaultProtectionDomain());
  } 

  /**
   ** The deprecated method defineClass(byte data[], int offset, int length) 
   ** is not provided. :-P
   */

  private native Class _defineClass(String name, byte data[], int offset, int length, ProtectionDomain pd)
    throws ClassFormatError;

  /**
   ** resolveClass forces all constants in the constant pool of c to be resolved.
   */
  native protected final void resolveClass(Class c) 
    throws NullPointerException;

  /**
   ** findSystemClass tries to load the named class using the system ClassLoader
   */
  protected final Class findSystemClass(String name) 
    throws ClassNotFoundException
  {
    ClassLoader loader = getPertinentClassLoader();
    if (loader == null) {
      throw new ClassNotFoundException("SystemClassLoader not yet defined");
    }
    return loader.findClass(name);
  }

  native private final Class findBootstrapClass(String name) 
    throws ClassNotFoundException;

  /**
   ** findLoadedClass looks for a class in the cache of classes for which
   ** this ClassLoader was the defining ClassLoader.
   */
  native protected final Class findLoadedClass(String name) ;

  /**
   ** findResource returns null: real resource-aware ClassLoaders
   ** override this method.
   */
  protected URL findResource(String name) {
    return null;
  }

  /**
   ** findResources returns an empty Enumeration: real resource-aware 
   ** ClassLoaders override this method.
   */
  protected Enumeration findResources(String name) 
    throws IOException
  {
    return new Enumeration() {
      public boolean hasMoreElements() { return false; }
      public Object nextElement() { return null; }
    };
  }

  /**
   ** getResource retrieves the named resource using "responsible
   ** delegation", rather like loadClass (except that there is no
   ** equivalent to "findLoadedClass").
   */
  public URL getResource(String name) {
    URL loaded;
    if (parent==null) {
      // TBD
      // loaded = getBootstrapResource(name);
     loaded = null;
    }
    else {
      loaded = parent.getResource(name);
    }
    if (loaded==null) {
      loaded = findResource(name);
    }
    return loaded;
  }
  
  /**
   ** getResources retrieves all the resource matching "name"  in a 
   ** responsibly delegating manner.  We first call getResources on
   ** our parent ClassLoader (or getBootstrapResources if parent is
   ** null), and then use a Vector to merge this with our own resource
   ** (if any).  TODO: make this more efficient.
   */
  public final Enumeration getResources(String name) {
    Vector merged = new Vector();
    if (parent==null) {
      // TBD
      // return getBootstrapResources(name);
    }
    else {
      Enumeration enum = parent.getResources(name);
      while (enum.hasMoreElements()) {
        merged.addElement(enum.nextElement());
      }
      URL ours = findResource(name);
      if (ours != null) {
        merged.addElement(ours);
      }
    }
    return merged.elements();
  }
  
  /** Get all resources using the system class loader.
   */
  public static Enumeration getSystemResources(String name) {
    ClassLoader loader = getPertinentClassLoader();
    if (loader == null) {
      return null; // Hm, should really be an empty enumeration???
    }
    return loader.getResources(name);
  }

  /**
   ** Set the signers of a class.
   */
  protected final void setSigners(Class cl, Object[] signers) {
    cl.signers = signers;
  }
  
  /**
   ** Get the identity of the system ClassLoader.
   ** Don't tell this to everybody!
   */
  public static synchronized ClassLoader getSystemClassLoader() {
    if (applicationClassLoader == null) {
      createApplicationClassLoader();
    }
    else {
      ClassLoader caller = getCallingClassLoader();

      if (caller != null && caller != SystemClassLoader.getInstance() && caller != applicationClassLoader) {
        permissionCheck("getClassLoader");
      }
    }

    return applicationClassLoader;
  }

  /** Get the identity of the parent ClassLoader.
   ** This too is sensitive information: only classes loaded by the
   ** bootstrap class loader or an ancestor of this ClassLoader have
   ** an automatic right to know, anyone else needs special permission.
   */
  public final ClassLoader getParent() {
    ClassLoader caller = getCallingClassLoader();
    if (!isDelegationAncestor(caller)) {
      permissionCheck("getClassLoader");
    }

    return parent;
  }

  static native ClassLoader getCallingClassLoader();

  protected boolean isDelegationAncestor(ClassLoader putative_ancestor) {
    return putative_ancestor == null || putative_ancestor == this || parent != null && parent.isDelegationAncestor(putative_ancestor);
  }

  // No longer final in 1.2
  /** Get a system resource in the form of an InputStream.
   */
  public static InputStream getSystemResourceAsStream(String name) {
    ClassLoader systemClassLoader = getPertinentClassLoader();
    if (systemClassLoader == null) {
      return null;
    }
    return systemClassLoader.getResourceAsStream(name);
  }

  public static native byte [] get_system_resource_as_bytes(String name);

  /** Get a resource in the form of an InputStream.
   */
  public InputStream getResourceAsStream(String name) {
    com.acunia.wonka.Etc.woempa(7, "ClassLoader/getResourceAsStream: "+this+" is looking for "+name);
    URL resource = getResource(name);
    com.acunia.wonka.Etc.woempa(7, "ClassLoader/getResourceAsStream: "+this+" found "+name+" at "+resource);
    if (resource == null) {

      return null;

    }
    else {
      try {

        return resource.openStream();

      }
      catch (IOException ioe) {

        return null;

      }
    }
  }

  /** Get a resource using the system class loader.
   */
  public static URL getSystemResource(String name) {
    ClassLoader systemClassLoader = getPertinentClassLoader();
    if (systemClassLoader == null) {
      return null;
    }
    return systemClassLoader.getResource(name);
  }

  /** Define a package by name.
   ** The name must be unique.
   */
  protected synchronized Package definePackage (String name, 
    String spectitle, String specversion, String specvendor,
    String impltitle, String implversion, String implvendor,
    URL sealbase)
    throws IllegalArgumentException
  {
    Package p = getPackage(name);
    if (p != null) {
      throw new IllegalArgumentException("duplicate package name: "+name);
    }

    p = new Package(name, spectitle, specversion, specvendor, impltitle, implversion, implvendor, sealbase);
    packages.put(name,p);

    return p;
  }

  /** Get the Package object associated with the given package name.
   ** If none found, returns null.
   */
  protected Package getPackage(String name) {
    Package p = packages == null ? null : (Package)packages.get(name);
    if (p == null) {
      if (parent == null) {
        p = getBootstrapPackage(name);
      }
      else {
        p = parent.getPackage(name);
      }
    }
    return p;
  }

  /** Get all Package objects known to this ClassLoader and its ancestors,
   ** as an array of all things.
   */
  protected Package[] getPackages() {
    Vector v = new Vector();
    Package[] ancestorpackages;
    if (parent == null) {
      ancestorpackages = getBootstrapPackages();
    }
    else {
      ancestorpackages = parent.getPackages();
      for (int i = 0; i < ancestorpackages.length; ++i) {
        v.addElement(ancestorpackages[i]);
      }
    }

    if (packages != null) {
      Enumeration ourpackages = packages.elements();
      while (ourpackages.hasMoreElements()) {
        v.addElement(ourpackages.nextElement());
      }
    }

    Object[] packobjs = v.toArray();
    Package[] packages = new Package[packobjs.length];
    System.arraycopy(packobjs, 0, packages, 0, packobjs.length);

    return packages;
  }

  /** Returns null, cos I haven't sussed out what it's supposed to do yet.
   */
  protected String findLibrary(String name) {
    return null;
  }

  /** Secret method to get a list of packages defined by the bootstrap
   ** class loader.  TODO: get this info from wre.jar.
   */
  synchronized static Package[] getBootstrapPackages() {
    if (bootstrap_packages == null) {
      bootstrap_packages = new Vector(40);
      bootstrap_packages.addElement(new Package("com.acunia.device"));
      bootstrap_packages.addElement(new Package("com.acunia.device.uart"));
      bootstrap_packages.addElement(new Package("com.acunia.device.serial"));
      bootstrap_packages.addElement(new Package("com.acunia.resource"));
      bootstrap_packages.addElement(new Package("com.acunia.wonka"));
      bootstrap_packages.addElement(new Package("com.acunia.wonka.security"));
      bootstrap_packages.addElement(new Package("java.lang"));
      bootstrap_packages.addElement(new Package("java.lang.ref"));
      bootstrap_packages.addElement(new Package("java.lang.reflect"));
      bootstrap_packages.addElement(new Package("java.io"));
      bootstrap_packages.addElement(new Package("java.util"));
      bootstrap_packages.addElement(new Package("java.net"));
      bootstrap_packages.addElement(new Package("java.awt.event"));
      bootstrap_packages.addElement(new Package("java.awt"));
      bootstrap_packages.addElement(new Package("java.security"));
      bootstrap_packages.addElement(new Package("java.math"));
      bootstrap_packages.addElement(new Package("java.text"));
      bootstrap_packages.addElement(new Package("java.rmi"));
      bootstrap_packages.addElement(new Package("javax.comm"));
    }

    Object[] packobjs = bootstrap_packages.toArray();
    Package[] packages = new Package[packobjs.length];
    System.arraycopy(packobjs, 0, packages, 0, packobjs.length);

    return packages;
  }

  /** Secret method to find a Package defined by the bootstrap
   ** class loader.  TODO: get this info from classes.zip.
   */
  synchronized static Package getBootstrapPackage(String name) {
    Package[] packages = getBootstrapPackages();

    for (int i = 0; i < packages.length; ++i) {
      if (packages[i].getName().equals(name)) {

        return packages[i];

      }
    }

    return null;
  }

  private static native String getCommandLineClasspath();

  static private void createApplicationClassLoader() {
    URL[] urls;
    Vector v = new Vector();
    int i;
    int j;

    String classpath = getCommandLineClasspath();
    i = classpath.indexOf(':');
    while(i != -1){
      v.add(classpath.substring(0,i));
      classpath = classpath.substring(i+1);
      i = classpath.indexOf(':');
    }
    v.add(classpath);

    String jarfiles = System.getProperty("com.acunia.wonka.jarfiles");

    i = jarfiles.indexOf(':');
    while(i != -1){
      v.add(jarfiles.substring(0,i));
      jarfiles = jarfiles.substring(i+1);
      i = jarfiles.indexOf(':');
    }
    v.add(jarfiles);

    int sz = v.size();
    urls = new URL[sz];

    j = 0;
    for(i=0 ; i < sz ; i++){
      String name = (String)v.get(i);
      if (name.indexOf(':') < 0) {
        name = "file:" + name;
      if (!name.endsWith(".jar") && !name.endsWith(".JAR") && !name.endsWith(".zip") && !name.endsWith(".ZIP") && !name.endsWith("/")) {
        name = name + "/";
      }}
      try {
        Etc.woempa(7,"url["+j+"] = "+name);
        urls[j++] = new URL(name);
      }
      catch(java.net.MalformedURLException mue) {
        mue.printStackTrace();
      }
    }

    applicationClassLoader = new URLClassLoader(urls, SystemClassLoader.getInstance());
    Properties props = System.getProperties();
    if (props != null) {
      props.put("java.class.path", classpath);
    }
  }

}
