/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: Runtime.java,v 1.6 2002/02/01 17:55:48 dumon Exp $
*/

package java.lang;

import java.io.IOException;
import java.io.OutputStream;
import java.io.InputStream;
import java.util.Vector;
import com.acunia.wonka.GarbageCollector;

/**
 ** A bunch of methods used to manipulate the runtime environment.
*/
public class Runtime {

  /**
   ** The single unique instance of Runtime.
   */
  private static Runtime theRuntime;

  /**
   ** The single unique instance of our colleague the Finalizer.
   */
  private static com.acunia.wonka.Finalizer theFinalizer;

  /**
   ** Native code which is used to initialise our tables of loaded libraries.
   */
  private static native void initLibraries();

  /**
   ** The number of instances of FinaDonna created and not yet finalized.
   ** Sycnchronize on FinaDonna.class when modifying this.
   */
  private static int finalizatora  = 0;

  /**
   ** The Threads which should be started on shutdown.
   */
  private Vector shutdownHooks;

  /**
   ** Set to true iff a shutdown is in progress.
   */
  private boolean shuttingDown;

  /**
   ** Inner class used to implement runFinalization()
   */
  private class FinaDonna {

    public FinaDonna() {
      synchronized (FinaDonna.class) {
        ++finalizatora ;
      }
    }

    protected void finalize() {
      synchronized (FinaDonna.class) {
        --finalizatora ;
        FinaDonna.class.notifyAll();
      }
    }
  }
        
  /**
   ** Static initialiser: creates the singleton.
   */
  static {
    theRuntime = new Runtime();
  }

  /**
   ** Private constructor to block attempts to create a new instance.
   */
  private Runtime() {
    initLibraries();
  }

  private void permissionCheck(String permission) {
    if (com.acunia.wonka.SecurityConfiguration.USE_ACCESS_CONTROLLER) {
      java.security.AccessController.checkPermission(new RuntimePermission(permission));
    }
    else if (com.acunia.wonka.SecurityConfiguration.USE_SECURITY_MANAGER) {
      SecurityManager sm = System.getSecurityManager();
      if (sm != null) {
        sm.checkPermission(new RuntimePermission(permission));
      }
    }
  }

  /**
   ** Add a Thread to the list of threads to be started when we decide
   ** to shut down the system.
   */
  public void addShutdownHook(Thread t) {
    permissionCheck("shutdownHooks");

    synchronized(this) {
      if (shuttingDown) {
        throw new IllegalStateException();
      }

      if (shutdownHooks == null) {
        shutdownHooks = new Vector();
      }

      if (shutdownHooks.contains(t)) {
        throw new IllegalArgumentException();
      }

      shutdownHooks.add(t);
    }
  }

  /**
   ** Remove a Thread from the list of threads to be started when we decide
   ** to shut down the system.
   */
  public boolean removeShutdownHook(Thread t) {
    permissionCheck("shutdownHooks");

    synchronized(this) {
      if (shutdownHooks == null) {

        return false;

      }

      if (shuttingDown) {
        throw new IllegalStateException();
      }

      return shutdownHooks.remove(t);
    }
  }

  /*
  ** Method which should be called to run all the shutdown hooks.
  ** The LIFO behaviour is not mandated by the specification.
  */
  private void runShutdownHooks() {
    synchronized(this) {
      if (shuttingDown || shutdownHooks == null) {
        return;
      }

      shuttingDown = true;
    }

    int n = shutdownHooks.size();
    while (n > 0) {
      Thread t = (Thread)shutdownHooks.elementAt(--n);
      System.out.println("Runtime: starting shutdown hook "+t);
      t.start();
    }

    n = shutdownHooks.size();
    while (n > 0) {
      Thread t = (Thread)shutdownHooks.elementAt(n-1);
      try {
        t.join();
        --n;
      }
      catch (InterruptedException ie) {}
      System.out.println("Runtime: joined shutdown hook "+t);
      shutdownHooks.removeElementAt(n);
    }
    shutdownHooks = null;

    System.out.println("Runtime: all shutdown hooks have completed");
  }

  /**
  ** Return the solitary instance of Runtime.
  */
  public static Runtime getRuntime() {
    return theRuntime;
  }

  /**
   ** Native method called by exit() after performing all checks.
   */
  private native void exit0(int status);

  /**
   ** Exit the VM, after running all shutdown hooks.
   */
  public void exit(int status)
    throws SecurityException
  {
    permissionCheck("VMexit");

    System.err.println("Thread "+Thread.currentThread()+" called Runtime.exit("+status+").");
    runShutdownHooks();
    exit0(status);
  }
    
/**
**	Not allowed ...
**	@remark always throws a SecurityException
*/
  public Process exec(String command) throws IOException, SecurityException {
     return exec(command, null);
  }

/**
**	Not allowed ...
**	@remark always throws a SecurityException
*/
  public Process exec(String command, String envp[]) throws IOException, SecurityException {
    throw new SecurityException("execution of commands is not allowed");
  }

/**
**	Not allowed ...
**	@remark always throws a SecurityException
*/
  public Process exec(String cmdarray[]) throws IOException, SecurityException{
    return exec(cmdarray, null);
  }

/**
**	Not allowed ...
**	@remark always throws a SecurityException
*/
  public Process exec(String cmdarray[], String envp[]) throws IOException, SecurityException{
    return exec0(cmdarray, envp);
    // throw new SecurityException("execution of commands is not allowed");
  }

  /*
  ** Go native to execute the process.
  */
  
  private native Process exec0(String cmdarray[], String envp[]) throws IOException;

/**
**	Not allowed ...
**	@remark always throws a SecurityException
*/
  public void load(String filename) throws SecurityException, UnsatisfiedLinkError{
       	throw new SecurityException("load is not allowed");
  }

  /**
   ** Tell the garbage collector to do some work.
   */
  public void gc(){
    GarbageCollector theGC = GarbageCollector.getInstance();
    int n = (int)((totalMemory() - freeMemory()) / 1000000);
    for (int i = 0; i < n; ++i) {
      // Leave the loop if two successive requests fail to free up any memory
      if (theGC.request(500000) == 0 && theGC.request(500000) == 0) {
        break;
      }
    }
  }

  /**
   ** Ask the garbage collector to run finalizers.  Actually what we really
   ** do is to create an instance of FinaDonna and then make it unreachable,
   ** kick the Garbage Collector, and wait for the finalizer to be run.
   */
  public void runFinalization(){
    GarbageCollector theGC = GarbageCollector.getInstance();
    FinaDonna bella = new FinaDonna();
    synchronized(FinaDonna.class) {
      bella = null;
      try {
        while (finalizatora  > 0) {
          theGC.kick();
          FinaDonna.class.wait(1000);
        }
      }
      catch (InterruptedException ie) {}
    }
  }

/**
** 	@status not implemented
*/
  public void traceInstructions(boolean on){}
/**
** 	@status not implemented
*/
  public void traceMethodCalls(boolean on){}


  /**
  **	@depricated
  **    @remark always throws a runtime exception
  */
  public InputStream getLocalizedInputStream(InputStream in){
   	throw new RuntimeException("Depricated Method");
  }
  /**
  **	@depricated
  **    @remark always throws a runtime exception
  */
  public OutputStream getLocalizedOutputStream(OutputStream out){
   	throw new RuntimeException("Depricated Method");
  }

  public native long totalMemory();
  public native long freeMemory();

  public void loadLibrary(String libname) throws SecurityException, UnsatisfiedLinkError
  {
    SecurityManager sm = System.getSecurityManager();

    if (sm != null) {
//      sm.checkLink(new RuntimePermission(libname));
    }

    loadLibrary0(libname);
  }
    
  private native void loadLibrary0(String libname) throws UnsatisfiedLinkError;
}

