/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/**
 * $Id: HashMap.java,v 1.1.1.1 2001/11/21 10:41:35 pollock Exp $
 */

package java.util;

import java.io.*;

public class HashMap extends AbstractMap implements Map, Cloneable,
Serializable {

  private transient int capacity;
  private float loadFactor;
  private transient int occupancy;
  private int threshold;

  private transient final static Object nullKey = new Object();

  private final static float DEFAULT_LOADFACTOR = (float).75;
  private final static int DEFAULT_CAPACITY = 101;

  private transient Object[] key;
  private transient Object[] value;
  private transient int modCount=0;

  public HashMap() {
    this(DEFAULT_CAPACITY, DEFAULT_LOADFACTOR);
  }

  public HashMap(int initialCapacity) {
    this(initialCapacity, DEFAULT_LOADFACTOR);
  }

  public HashMap(int initialCapacity, float loadFactor) {
    if ( initialCapacity < 0 || loadFactor <= 0.0f )
    	throw new IllegalArgumentException("HashMap needs positive numbers");
    this.capacity = (initialCapacity < 5 ? 5 :initialCapacity);
    this.key = new Object[capacity];
    this.value = new Object[capacity];
    this.loadFactor = (loadFactor > 1.0f ? DEFAULT_LOADFACTOR : loadFactor );
  }

  public HashMap (Map t) {
    this(DEFAULT_CAPACITY, DEFAULT_LOADFACTOR);
    Iterator it = t.entrySet().iterator();
    Map.Entry me;
    while (it.hasNext()) {
  	 me = (Map.Entry) it.next();
  	 put(me.getKey(),me.getValue());
    }
  }

  public int size() {
    return this.occupancy;
  }

  public boolean isEmpty() {
    return (this.occupancy == 0);
  }

  public boolean containsValue(Object value) {
    return contains(value);
  }

  public boolean containsKey(Object key) {
    int nextSlot = firstBusySlot(0);
    if (key==null) key = nullKey;
    while(nextSlot>=0) {
      if(key.equals(this.key[nextSlot])) {
        return true;
      }
      nextSlot = firstBusySlot(nextSlot+1);
    }
    return false;
  } 

  public Object get(Object key)  {
// System.out.println("I need to get key "+key+" from "+this);
    if(key==null) {  key = nullKey; }

    if (this.key==null) {
      return null;
    }
    int i;
    int j = 0;
    int hash = key.hashCode();

    while(true) {
      i = probe(hash, j);
      if(this.key[i] == null) {
        return null;
      } else if(key.equals(this.key[i])) {
        return this.value[i];
      } else ++j;
    }
  }          

  public Object put(Object key, Object newvalue)  {
    if(key==null) { key = nullKey; }
    int i;
    int j = 0;
    int hashcode = key.hashCode();
    int newsize;
    if (this.key==null) {
      this.key = new Object[DEFAULT_CAPACITY];
      this.value = new Object[DEFAULT_CAPACITY];
      this.capacity = DEFAULT_CAPACITY;
      this.threshold = (int)(capacity*loadFactor);
    }
    while(true) {
      i = probe(hashcode, j);
      if(this.key[i] == null) {
// System.out.println("HASH for "+key+": "+hashcode);
// System.out.println("PUT "+key+" in "+this+" at position "+i+" (j="+j+")");
        this.key[i] = key;
        this.value[i] = newvalue;
        ++this.occupancy;
        newsize = this.checkSize(+1);
        if(newsize!=0) {
          resize(newsize);
        }
        modCount++;
        return null;
      } 
      else if(key.equals(this.key[i])) {
        Object oldvalue = this.value[i];
        this.value[i] = newvalue;
        modCount++;
        return oldvalue;
      }
      ++j;
    }
  }

  public Object remove(Object key)  {
    if(key==null) { key = nullKey; }
    int   i;
    int   j = 0;
    int   hashcode = key.hashCode();
    int   newsize;

    if (this.key==null) {
      return null;
    }

    while(true) {
      i = probe(hashcode,j);
      if(this.key[i] == null) {
        return null; 
      } else {
        if(key.equals(this.key[i])) {
          Object oldvalue = this.value[i];
          deleteSlot(i);
//          --this.occupancy;
          newsize = this.checkSize(-1);
          if(newsize!=0) {
            resize(newsize);
          }
          modCount++;
          return oldvalue;
        }
      }
      ++j;
    }
  }

  public void putAll(Map t) {
    super.putAll(t);
  }

  public void clear() {
     if (occupancy > 0) {
     key = new Object[capacity];
     value = new Object[capacity];
     occupancy = 0;
     modCount++;
     }
  }

  public synchronized Object clone() {
    Object that = null;
    try {
      that = super.clone();
      HashMap that_hashmap = (HashMap)that;

      if (this.key!=null) {
        that_hashmap.key = new Object[capacity];
        System.arraycopy(this.key,0,that_hashmap.key,0,capacity);
      }
      if (this.value!=null) {
        that_hashmap.value = new Object[capacity];
        System.arraycopy(this.value,0,that_hashmap.value,0,capacity);
      }
    }
    catch (CloneNotSupportedException cnse) {
    }
    return that;
  } 

  public Set keySet() {
    return new HashMapKeySet();
  }

  public Collection values() {
    return new HashMapCollection();
  }
              
  public Set entrySet() {
    return new HashMapSet();
  }

  private boolean contains(Object value) {
    int nextSlot = firstBusySlot(0);
    while(nextSlot>=0) {
      if((value==null ? this.value[nextSlot]==null :value.equals(this.value[nextSlot]))) {
        return true;
      }
      nextSlot = firstBusySlot(nextSlot+1);
    }
    return false;
  } 

  private int firstBusySlot(int i) {
    int j;
    if (i<0 || i>=capacity || this.key==null) {
      return -1;
    }
    for(j=i;j<capacity;++j)
      if (this.key[j] != null)  {
        return j;
      }
    return -1;
  } 

  private int probe(int hashcode, int sequence) {
    int rehash = (hashcode-sequence) % this.capacity;
    if (rehash<0) rehash = rehash + this.capacity;
    return rehash;
  }

 private int checkSize(int direction) {
    float targetsquared, current, currentsquared;

    targetsquared = loadFactor * loadFactor;
    current = (float)occupancy / (float)capacity;
    currentsquared = current * current;

    if (direction>=0 && currentsquared<loadFactor
     || direction<=0 && targetsquared>current
    ) {

      return 0;

    }

    int newsize = (int)((float)occupancy/loadFactor);

    if (newsize<occupancy+2) newsize = occupancy+2;

    if (direction>=0 && newsize<=capacity) newsize = 0;
    if (direction<=0 && newsize>=capacity) newsize = 0;

    return newsize;

  } 

  private void resize(int newsize) {
    int oldsize = this.capacity;
    Object[] oldkeys = this.key;
    Object[] oldvalues = this.value;
    Object[] newkeys = new Object[newsize];
    Object[] newvalues = new Object[newsize];

    this.capacity = newsize;
    this.threshold = (int)(capacity*loadFactor);
    this.occupancy = 0;
    this.key = newkeys;
    this.value = newvalues;

    int oldindex, newindex;

    for (oldindex=0;oldindex<oldsize;++oldindex) {
      Object key = oldkeys[oldindex];
      if (key != null)
        this.put(key,oldvalues[oldindex]);
    }
  } 

  private void deleteSlot(int slotIndex) {
    int vacant, current, home, distance1, distance2;
    boolean happy;

    --occupancy;
    current = slotIndex;

    while(true) {
    // R1
      key[current] = null;
      vacant = current;

    // R2
      happy = true;
      while(happy) {
        current = current==0 ? capacity-1 : current-1;

    // R3
        if (key[current] == null) {

          return; // normal termination

        }
        home = probe(key[current].hashCode(),0);
        distance1 = current<=home ? home - current : capacity + home - current;
        distance2 = current<=vacant ? vacant - current : capacity + vacant - current;

        happy = distance1<distance2;
/*      oops -- Same bug as in hashtable;
        home = probe(key[current].hashCode(),0);
        distance1 = current>=home ? current-home : current+capacity-home;
        distance2 = current>=vacant ? current-vacant : current+capacity-vacant;
        happy = distance1<distance2;
*/
      // back to R2
      }

    // R4
      key[vacant] = key[current];
      value[vacant] = value[current];

    // repeat from R1 ...
    }
  }

  private void deleteInHM(Object key) {
   	remove(key);
  }


  protected class HashMapKeyIterator implements Iterator {
  	int pos=firstBusySlot(0); //points to the next free slot
  	int oldpos=-1; // points to the current slot
  	int mc=modCount;
  	int status=0;


  public  boolean hasNext() {
	//System.out.println("DEBUG -- in HT keySetIterator -- hasNext is called ->"+(pos < buf.size()));
 		return  firstBusySlot(pos)!=-1;
  }

  public Object next() {
	//System.out.println("DEBUG -- in HT keySetIterator -- next is called");
        if (modCount != mc) throw new ConcurrentModificationException("don't change the set please");
               	status = 1;
                if (firstBusySlot(pos)==-1) throw new NoSuchElementException("no elements left");
		oldpos = pos;
                pos = firstBusySlot(pos+1);			
		return (key[oldpos]== nullKey ? null : key[oldpos]);                 	
  }

  public void remove() {
	//System.out.println("DEBUG -- in HT keySetIterator -- remove is called");
	if (modCount != mc) throw new ConcurrentModificationException("don't change the set please");
 	if (status != 1 ) throw new IllegalStateException("do a next() operation before remove()");
  	deleteInHM(key[oldpos]);
        mc++;
        status=-1;
  }
 }
  protected class HashMapValueIterator implements Iterator {
  	int pos=firstBusySlot(0); //points to the next free slot
  	int oldpos=-1; // points to the current slot
  	int mc=modCount;
  	int status=0;


  public  boolean hasNext() {
	//System.out.println("DEBUG -- in HT keySetIterator -- hasNext is called ->"+(pos < buf.size()));
 		return  firstBusySlot(pos)!=-1;
  }

  public Object next() {
	//System.out.println("DEBUG -- in HT keySetIterator -- next is called");
        if (modCount != mc) throw new ConcurrentModificationException("don't change the set please");
               	status = 1;
                if (firstBusySlot(pos)==-1) throw new NoSuchElementException("no elements left");
		oldpos = pos;
                pos = firstBusySlot(pos+1);			
		return value[oldpos];                 	
  }

  public void remove() {
	//System.out.println("DEBUG -- in HT keySetIterator -- remove is called");
	if (modCount != mc) throw new ConcurrentModificationException("don't change the set please");
 	if (status != 1 ) throw new IllegalStateException("do a next() operation before remove()");
  	deleteInHM(key[oldpos]);
        mc++;
        status=-1;
  }
 }
  protected class HashMapIterator implements Iterator {
  	int pos=firstBusySlot(0); //points to the next free slot
  	int oldpos=-1; // points to the current slot
  	int mc=modCount;
  	int status=0;


  public  boolean hasNext() {
	//System.out.println("DEBUG -- in HT keySetIterator -- hasNext is called ->"+(pos < buf.size()));
 		return  firstBusySlot(pos)!=-1;
  }

  public Object next() {
	//System.out.println("DEBUG -- in HT keySetIterator -- next is called");
        if (modCount != mc) throw new ConcurrentModificationException("don't change the set please");
               	status = 1;
                if (firstBusySlot(pos)==-1) throw new NoSuchElementException("no elements left");
		oldpos = pos;
                pos = firstBusySlot(pos+1);			
		return new Entry((key[oldpos]==nullKey ? null : key[oldpos]),value[oldpos]);                 	
  }

  public void remove() {
	//System.out.println("DEBUG -- in HT keySetIterator -- remove is called");
	if (modCount != mc) throw new ConcurrentModificationException("don't change the set please");
 	if (status != 1 ) throw new IllegalStateException("do a next() operation before remove()");
  	deleteInHM(key[oldpos]);
  	mc++;
  	pos = firstBusySlot(pos);
        status=-1;
  }
 }
  protected class HashMapCollection extends AbstractCollection {

    public Iterator iterator() {
      return new HashMapValueIterator();
    }

    public int size() {
      return occupancy;
    }
  }

  protected class HashMapKeySet extends AbstractSet {
    public Iterator iterator() {
      return new HashMapKeyIterator();
    }

    public int size() {
      return occupancy;
    }
  }

  protected class HashMapSet extends AbstractSet {
    public Iterator iterator() {
      return new HashMapIterator();
    }

    public int size() {
      return occupancy;
    }
  }


  static class Entry implements Map.Entry {

    private Object key;
    private Object value;

    Entry (Object key, Object value) {
      this.key = key;
      this.value = value;
    }

    public Object getKey() {
      return this.key;
    }

    public Object getValue() {
      return this.value;
    }

    public Object setValue(Object o) throws UnsupportedOperationException, ClassCastException, IllegalArgumentException, NullPointerException {
      throw new UnsupportedOperationException();
    }

    public boolean equals (Object o) {
      boolean answer = false;
      if (o instanceof HashMap.Entry) {
        HashMap.Entry e2 = (HashMap.Entry)o;
        answer = (key==null ? e2.getKey()==null : key.equals(e2.getKey())) &&
         (value==null ?  e2.getValue()==null : value.equals(e2.getValue()));
      }
      return answer;
    }

    public int hashCode() {
      return (key==null ? 0 : key.hashCode()) ^ (value==null ? 0 : value.hashCode());
    }

    public String toString() {
      String answer = "HashMap Entry: "+key+"="+value;
      return answer;
    }

  }

  private void readObject (ObjectInputStream s) throws IOException, ClassNotFoundException {
    int local_oc;
    s.defaultReadObject();
    this.capacity=s.readInt();
    local_oc=s.readInt();
    this.key = new Object[this.capacity];
    this.value = new Object[this.capacity];
    Object key,value;
    for (int i=0;i<local_oc;i++) {
      key=s.readObject();
      value=s.readObject();
      this.put(key,value);
    }
  }

  private void writeObject (ObjectOutputStream s) throws IOException {
    s.defaultWriteObject();
    s.writeInt (capacity);
    s.writeInt (occupancy);
    int nextSlot=firstBusySlot(0);
    for (int i=0;i<occupancy;i++) {
      s.writeObject((key[nextSlot]==nullKey ? null :key[nextSlot]));
      s.writeObject(value[nextSlot]);
      nextSlot=firstBusySlot(nextSlot+1);
    }               
  }

}
