/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/**
 * $Id: HashSet.java,v 1.1.1.1 2001/11/21 10:41:35 pollock Exp $
 */

package java.util;

import java.io.Serializable;

public class HashSet extends AbstractSet implements Set, Cloneable, Serializable {

  private static Object nullKey = new Object();
  private static int defaultSize = 101;
  private static float defaultLoadFactor= 0.75f;

  private transient int modCount=0;

  private transient float loadFactor;
  private transient int occupancy=0;
  private transient int capacity;
  private transient Object[] key;

  public HashSet() {
  	this(defaultSize,defaultLoadFactor);
  }

  public HashSet(Collection c){
  	this((int)(c.size()*2),defaultLoadFactor);
        addAll(c);
  }

  public HashSet(int initCap){
  	this(initCap,defaultLoadFactor);
  }

  public HashSet(int initCap, float loadFactor){
        if (initCap < 0 || loadFactor < 0.0f) {
         	throw new IllegalArgumentException();
        }
        capacity = (initCap < 5 ? 5 : initCap);
        this.loadFactor = (loadFactor > 1.0f ? defaultLoadFactor : loadFactor);
  }

  public boolean add(Object key) {
    if(key==null) { key = nullKey; }
    int i;
    int j = 0;
    int hashcode = key.hashCode();
    int newsize;
    if (this.key==null) {
      this.key = new Object[capacity];
    }
    while(true) {
      i = probe(hashcode, j);
      if(this.key[i] == null) {
        modCount++;
        this.key[i] = key;
        ++this.occupancy;
        newsize = this.checkSize(+1);
        if(newsize!=0) {
          resize(newsize);
        }
        return true;
      }
      else if(key.equals(this.key[i])) {
        return false;
      }
      ++j;
    }
  }

  public int size() {
    return occupancy;
  }

  public boolean isEmpty() {
   	return (occupancy == 0);
  }

  public boolean remove(Object key){
    if(key==null) { key = nullKey; }
    int   i;
    int   j = 0;
    int   hashcode = key.hashCode();
    int   newsize;

    if (this.key==null) {
      return false;
    }

    while(true) {
      i = probe(hashcode,j);
      if(this.key[i] == null) {
        return false;
      } else {
        if(key.equals(this.key[i])) {
          modCount++;
          deleteSlot(i);
          newsize = this.checkSize(-1);
          if(newsize!=0) {
            resize(newsize);
          }
          return true;
        }
      }
      ++j;
    }
  }

  public Iterator iterator() {
    return new HashSetIterator();
  }

  public void clear(){
        modCount++;
  	key = null;
        occupancy=0;
  }

  public Object clone(){
  	HashSet h = null;
  	try {
  	 	h = (HashSet) super.clone();
  	}
  	catch(CloneNotSupportedException cnse){}
  	if (key != null) {
  	   	h.key = (Object[])this.key.clone();
  	}
  	return h;
  }

  public boolean contains(Object key) {
    int nextSlot = firstBusySlot(0);
    if (key==null) key = nullKey;
    while(nextSlot>=0) {
      if(key.equals(this.key[nextSlot])) {
        return true;
      }
      nextSlot = firstBusySlot(nextSlot+1);
    }
    return false;

  }


  private int firstBusySlot(int i) {
    int j;
    if (i<0 || i>=capacity || this.key==null) {
      return -1;
    }
    for(j=i;j<capacity;++j)
      if (this.key[j] != null)  {
        return j;
      }
    return -1;
  }

  private int probe(int hashcode, int sequence) {
    int rehash = (hashcode-sequence) % this.capacity;
    if (rehash<0) rehash = rehash + this.capacity;
    return rehash;
  }

 private int checkSize(int direction) {
    float targetsquared, current, currentsquared;

    targetsquared = loadFactor * loadFactor;
    current = (float)occupancy / (float)capacity;
    currentsquared = current * current;
    if (direction>=0 && currentsquared<loadFactor
     || direction<=0 && targetsquared>current
    ) {
      return 0;
    }
    int newsize = (int)((float)occupancy/loadFactor);
    if (newsize<occupancy+2) newsize = occupancy+2;
    if (direction>=0 && newsize<=capacity) newsize = 0;
    if (direction<=0 && newsize>=capacity) newsize = 0;

    return newsize;
  }


  private void resize(int newsize) {
    int oldsize = this.capacity;
    Object[] oldkeys = this.key;
    Object[] newkeys = new Object[newsize];
    this.capacity = newsize;
    this.occupancy = 0;
    this.key = newkeys;
    int oldindex, newindex;
    for (oldindex=0;oldindex<oldsize;++oldindex) {
      Object key = oldkeys[oldindex];
      if (key != null)
        this.add(key);
    }
  }


  private void deleteSlot(int slotIndex) {
    int vacant, current, home, distance1, distance2;
    boolean happy;
    --occupancy;
    current = slotIndex;
    while(true) {
    // R1
      key[current] = null;
      vacant = current;
    // R2
      happy = true;
      while(happy) {
        current = current==0 ? capacity-1 : current-1;
    // R3
        if (key[current] == null) {
          return; // normal termination
        }
        home = probe(key[current].hashCode(),0);
        distance1 = current<=home ? home - current : capacity + home - current;
        distance2 = current<=vacant ? vacant - current : capacity + vacant - current;
        happy = distance1<distance2;
      // back to R2
      }
    // R4
      key[vacant] = key[current];
    // repeat from R1 ...
    }
  }

  private class HashSetIterator implements Iterator {

  	private int mc;
  	private int pos;
  	private int oldpos;
  	
  	public HashSetIterator(){
  		mc = modCount;
  		pos = firstBusySlot(0);
  		oldpos = pos;
  	}	
  	
  	public boolean hasNext() {
  	       	if (mc != modCount) {
  	        	throw new ConcurrentModificationException();
  	       	}
  	       	return (pos != -1);
  	}
  	
  	public Object next() {
  	       	if (pos == -1) {
  	       	 	throw new NoSuchElementException();  	       	 	
  	       	}
  	       	if (mc != modCount) {
  	        	throw new ConcurrentModificationException();
  	       	}
  	       	Object e = key[pos];
  	       	oldpos = pos;
  	       	pos = firstBusySlot(pos+1);
  	       	return (nullKey.equals(e) ? null : e);		  	
  	} 	
  	
  	public void remove() {
  	       	if (oldpos == pos) {
  	       	 	throw new IllegalStateException();
  	       	}
  	       	if (mc != modCount) {
  	        	throw new ConcurrentModificationException();
  	       	}
  	       	deleteSlot(oldpos);
  	       	oldpos = pos;
  	
  	}

  }
}
