/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: Properties.java,v 1.1.1.1 2001/11/21 10:41:35 pollock Exp $
**
** TODO/REMARKS:
**
** The current implementation of the load() mehtod reads bytes directly from the InputStream
** without any buffering. Moreover, it doesn't care about the character enconding being used.
** - Proper implementation should make use of an InputStreamReader to ensure proper byte to
**   character conversion.
** - One should also make use of buffered IO for greater performance.
** - Same remarks are applicable for the save()/store() functions.
*/
package java.util;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.IOException;
import java.io.BufferedWriter;
import java.io.OutputStreamWriter;
import java.io.BufferedReader;
import java.io.InputStreamReader;

public class Properties extends Hashtable {

  private static final String keyValueSeparators = "= :\t\r\n\f";
  private static final String commentChars = "#!";
  private static final long serialVersionUID = 4112578634029874840L;

  protected Properties defaults;

  public Properties() {
    this(null);
  }

  public Properties(Properties deflts) {
    this.defaults = deflts;
  }

  public String getProperty(String key) {
    String val = (String) get(key);
    if (val == null && defaults != null){
      val = defaults.getProperty(key);
    }
    return val;
  }

  public String getProperty(String key, String defaultValue){
    String val = getProperty(key);
    return (val == null) ? defaultValue : val;
  }

  public Object setProperty(String key, String value){
    return put(key, value);
  }

  /**
   * Copy all key/value pairs in a new hashtable merged with the
   * default values if any.
   *
   * Note: we first expand the 'defaults' then override them with the specific key/values.
   */
  private Hashtable flatten() {
    if (defaults==null){
      return (Hashtable)this.clone();
    }
    else {
      Hashtable h = defaults.flatten();
      h.putAll(this);
      return h;
    }
  }

  public Enumeration propertyNames() {
    return this.flatten().keys();
  }

  /**
   * Returns true if the line ends with a '\' character.
   * Note that the '\' might be escaped ('\\').
   */
  private boolean isBrokenLine(String line) {
    int i=line.length()-1;
    int count=0;
    while( (i>=0) && (line.charAt(i) == '\\') ) {
      i--;
      count++;
    }
    return ((count % 2) == 1);
  }

  /**
   * Left trim (remove white spaces from the left)
   */
  private String ltrim(String s) {
    if ("".equals(s)){
      return s;
    }
    if(!Character.isWhitespace(s.charAt(0))){
      return s;
    }
    int len = s.length();
    int i   = 1;
    while( (i < len) && Character.isWhitespace(s.charAt(i)) )
        i++;

    if( i == len )
        return "";
    else
        return s.substring(i);
  }

  /**
   * Gets a complete line from the reader (handles broken lines)
   */
  private StringBuffer getLine(BufferedReader br) throws IOException {
    StringBuffer resultLine = new StringBuffer();
    while(true) {
      String line = br.readLine();
      if(line == null){
        return null;
      }
      line = ltrim(line);
  	  if(line.equals("")){
  	    continue;
      }
      if(commentChars.indexOf(line.charAt(0),0) != -1 ){
  	    while (isBrokenLine(line)){
  	      line = br.readLine();
  	      if(line == null){
  	        return null;
  	      }
  	    }
  	    continue;
      }
      while (isBrokenLine(line)) {
  	    resultLine.append(line.substring(0, line.length()-1) );
  	    line = br.readLine();
  	    line = (line == null) ? "" : ltrim(line);
      }
      resultLine.append(line);
      break;
    	
    }
    return resultLine;
  }


  /**
   * Load from an input stream.
   */
  public void load(InputStream inStream) throws IOException {
    BufferedReader br = new BufferedReader(new InputStreamReader(inStream, "8859_1"));
    while(true) {
      StringBuffer line = getLine(br);
      if(line == null){
  	    return;
      }	
      int len = line.length();
      boolean gotIt = false;
      int index=1;
      for(; index < len ; index++ )
  	    if( keyValueSeparators.indexOf(line.charAt(index),0) != -1 ) {
          gotIt = true;
          break;
  	  }

      if(gotIt) {
  	    String key = line.substring(0, index).trim();
  	    String value = "";
  	    if( index < len-1 ) {
          value = ltrim( line.substring(index+1, len) );
          if (value.length() > 0 && (value.charAt(0)=='=' ||value.charAt(0)==':')){
            value = ltrim(value.substring(1,value.length()));
       	  }	
  	    }
  	    put(key, value);
      }
      else {
        put(line.toString(), "");
      }
    }
  }


  public synchronized void save(OutputStream out, String header) {
    try {
      store(out, header);
    }catch(IOException ioe){}
  }

  public synchronized void store(OutputStream out, String header) throws IOException {
    BufferedWriter bout = new BufferedWriter(new OutputStreamWriter(out));
    if(header != null) {
      bout.write("#"+header+'\n');
    }
    bout.write("#"+new Date().toString()+'\n');
    String key;
    String value;
    Enumeration enum = keys();
    while(enum.hasMoreElements()){
      key = (String) enum.nextElement();
      value = (String) get(key);
      bout.write(key + "=" + value+'\n');
    }
    bout.flush();
  }

  public void list(PrintStream out) throws NullPointerException {
    if (out == null){
      throw new NullPointerException();
    }
    Hashtable h = flatten();
    String key;
    String value;
    Enumeration enum = h.keys();
    while(enum.hasMoreElements()) {
      Object o = enum.nextElement();
      key = (String)o;
      value = (String) h.get(key);
      if( value.length() > 40 ){
  	value = value.substring(0,37)+"...";
      }
      out.println(key + "=" + value);
    }
  }

  public void list(PrintWriter out) throws NullPointerException  {
    Hashtable h = flatten();
    String key;
    String value;
    Enumeration enum = h.keys();
    while(enum.hasMoreElements()) {
      key = (String) enum.nextElement();
      value = (String) h.get(key);
      if( value.length() > 40 ){
  	  value = value.substring(0,37)+"...";
      }
      out.println(key + "=" + value);
    }
  }
}
