/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/**
 * $Id: WeakHashMap.java,v 1.3 2002/01/08 16:55:09 ruelens Exp $
 */

package java.util;

import java.lang.ref.*;

public class WeakHashMap extends AbstractMap implements Map {

  final static Object nullKey = new Object();

  private final static float DEFAULT_LOADFACTOR = (float).75;
  private final static int DEFAULT_CAPACITY = 101;

  private WeakHashReference[] entries;
  private ReferenceQueue queue;

  private int threshold;
  private int capacity;
  private float loadFactor;

  int occupancy;
  int modCount=0;

  public WeakHashMap() {
    this(DEFAULT_CAPACITY, DEFAULT_LOADFACTOR);
  }

  public WeakHashMap(int initialCapacity) {
    this(initialCapacity, DEFAULT_LOADFACTOR);
  }

  public WeakHashMap(int initialCapacity, float loadFactor) {
    if ( initialCapacity < 0 || loadFactor <= 0.0f ) {
    	throw new IllegalArgumentException("HashMap needs positive numbers");
    }
    this.capacity = (initialCapacity < 5 ? 5 :initialCapacity);
    entries = new WeakHashReference[capacity];
    this.loadFactor = (loadFactor > 1.0f ? DEFAULT_LOADFACTOR : loadFactor );
    this.queue = new ReferenceQueue();
  }

  public void clear() {
    if (occupancy > 0) {
      entries = new WeakHashReference[capacity];
      occupancy = 0;
      modCount++;
    }
  }

  public boolean containsKey(Object key) {
    checkQueue();
    if(key == null) {
      key = nullKey;
    }

    int j = 0;
    int hash = key.hashCode();

    while(true) {
      int i = probe(hash, j);
      if(entries[i] == null) {
        return false;
      } else if(key.equals(entries[i].get())) {
        return true;
      } else ++j;
    }
  }

  public Set entrySet(){
    checkQueue();
    return new WeakHashMapSet();
  }

  public Object get(Object key)  {
    checkQueue();

    if(key==null) {
      key = nullKey;
    }

    int j = 0;
    int hash = key.hashCode();

    while(true) {
      int i = probe(hash, j);
      if(entries[i] == null) {
        return null;
      } else if(key.equals(entries[i].get())) {
        return entries[i].value;
      } else ++j;
    }
  }

  public boolean isEmpty() {
    checkQueue();
    return (this.occupancy == 0);
  }

  public Object put(Object key, Object newvalue)  {
    if(key==null) {
      key = nullKey;
    }

    checkQueue();

    int j = 0;
    int hashcode = key.hashCode();

    while(true) {
      int i = probe(hashcode, j);
      if(entries[i] == null) {
        entries[i] = new WeakHashReference(key, newvalue, queue);
        ++this.occupancy;
        int newsize = this.checkSize(+1);
        if(newsize!=0) {
          resize(newsize);
        }
        modCount++;
        return null;
      }
      else if(key.equals(entries[i].get())) {
        Object oldvalue = entries[i].value;
        entries[i].value = newvalue;
        modCount++;
        return oldvalue;
      }
      ++j;
    }
  }

  public Object remove(Object key)  {
    checkQueue();

    if(key==null) {
      key = nullKey;
    }

    int j = 0;
    int hashcode = key.hashCode();

    while(true) {
      int i = probe(hashcode,j);
      if(entries[i] == null) {
        return null;
      } else {
        if(key.equals(entries[i].get())) {
          Object oldvalue = entries[i].value;
          deleteSlot(i);
          int newsize = this.checkSize(-1);
          if(newsize!=0) {
            resize(newsize);
          }
          modCount++;
          return oldvalue;
        }
      }
      ++j;
    }
  }

  public int size() {
    checkQueue();
    return this.occupancy;
  }

  void checkQueue(){
    WeakHashReference ref = (WeakHashReference) queue.poll();
    while(ref != null){
      int place = probe(ref.hashValue,0);
      while(ref != entries[place]){
        place = (place - 1 + capacity) % capacity;
      }
      deleteSlot(place);
      ref = (WeakHashReference) queue.poll();
    }
  }

  int firstBusySlot(int i) {
    int j;
    if (i<0 || i>=capacity) {
      return -1;
    }
    for(j=i;j<capacity;++j)
      if (entries[j] != null)  {
        return j;
      }
    return -1;
  } 

  void deleteSlot(int slotIndex) {
    int vacant, current, home, distance1, distance2;
    boolean happy;

    --occupancy;
    current = slotIndex;

    while(true) {
      // R1
      entries[current] = null;
      vacant = current;

      // R2
      happy = true;
      while(happy) {
        current = current==0 ? capacity-1 : current-1;

        // R3
        if (entries[current] == null) {

          return; // normal termination

        }
        home = probe(entries[current].hashValue,0);
        distance1 = current<=home ? home - current : capacity + home - current;
        distance2 = current<=vacant ? vacant - current : capacity + vacant - current;

        happy = distance1<distance2;
        // back to R2
      }

      // R4
      entries[vacant] = entries[current];

      // repeat from R1 ...
    }
  }

  private int probe(int hashcode, int sequence) {
    int rehash = (hashcode-sequence) % this.capacity;
    if (rehash<0) {
      rehash = rehash + this.capacity;
    }
    return rehash;
  }

  private int checkSize(int direction) {
    float targetsquared, current, currentsquared;

    targetsquared = loadFactor * loadFactor;
    current = (float)occupancy / (float)capacity;
    currentsquared = current * current;

    if (direction>=0 && currentsquared<loadFactor || direction<=0 && targetsquared>current) {
      return 0;
    }

    int newsize = (int)((float)occupancy/loadFactor);

    if (newsize<occupancy+2) {
      newsize = occupancy+2;
    }

    if (direction>=0 && newsize<=capacity) newsize = 0;
    if (direction<=0 && newsize>=capacity) newsize = 0;

    return newsize;
  }

  private void resize(int newsize) {
    int oldsize = this.capacity;
    WeakHashReference[] oldkeys = entries;
    entries = new WeakHashReference[newsize];

    this.capacity = newsize;
    this.threshold = (int)(capacity*loadFactor);

    for (int oldindex=0; oldindex < oldsize; ++oldindex) {
      WeakHashReference key = oldkeys[oldindex];
      if (key != null){
        int place = probe(key.hashValue,0);
        while(entries[place] != null){
          place = (place - 1 + newsize) % newsize;
        }
        entries[place] = key;
      }
    }
  }

//inner classes ...

  protected class WeakHashMapSet extends AbstractSet {
    public Iterator iterator() {
      return new WeakHashMapIterator();
    }

    public int size() {
      return WeakHashMap.this.size();
    }
  }

  protected class WeakHashMapIterator implements Iterator {
  	int pos=firstBusySlot(0); //points to the next free slot
  	int mc=modCount;
  	Object refKey;

    public  boolean hasNext() {
   		return pos !=-1;
    }

    public Object next() {
      if (modCount != mc) {
        throw new ConcurrentModificationException("don't change the set please (this.modCount == "+mc+", outer.modCount == "+modCount+")");
      }
      if (pos == -1) {
        throw new NoSuchElementException("no elements left");
      }
  		WeakHashReference entry = entries[pos];
      refKey = entry.get();
      pos = firstBusySlot(pos+1);
  		return entry;
    }

    public void remove() {
     	if (refKey == null ) {
     	  throw new IllegalStateException("do a next() operation before remove()");
     	}
    	if (modCount != mc) {
    	  throw new ConcurrentModificationException("don't change the set please");
    	}
    	WeakHashMap.this.remove(refKey);
    	mc++;
    	refKey = null;
    }
  }
}
