/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** References of the form [J+JVM ...] are to the book _Java and the Java
** Virtual Machine_ by Robert Staerk, Joachim Schmid, and Egon Boerger.
*/

#include <string.h>

#include "arrays.h"
#include "clazz.h"
#include "constant.h"
#include "descriptor.h"
#include "exception.h"
#include "fields.h"
#include "interpreter.h"
#include "loading.h"
#include "locks.h"
#include "oswald.h"
#include "hashtable.h"
#include "heap.h"
#include "ts-mem.h"
#include "methods.h"
#include "strings.h"
#include "threads.h"
#include "stack.h"
#include "ieee754.h"
#include "dispatcher.h"
#include "list.h"
#include "wonka.h"
#include "bar.h"

extern w_object object_sentinel;

/*
** Each classloader has associated with it a hashtable of loaded classes.
** For the "primordial" class loader this is a static item.
*/

w_hashtable system_class_hashtable;

/*
** The clazz we clone arrays from. See also comments in wonka.h
*/

w_clazz clazz_Array;


w_clazz clazz_boolean;
w_clazz clazz_char;
w_clazz clazz_float;
w_clazz clazz_double;
w_clazz clazz_byte;
w_clazz clazz_short;
w_clazz clazz_int;
w_clazz clazz_long;
w_clazz clazz_void;

/*
** Three often used clazzes.
*/

w_clazz clazzArrayOf_Object;
w_clazz clazzArrayOf_Class;
w_clazz clazzArrayOf_String;
w_clazz clazzArrayOf_Constructor;
w_clazz clazzArrayOf_Field;
w_clazz clazzArrayOf_Method;

static void parseMethodCodeAttributes(w_method, w_bar, w_instance *loading_problem);

/*
** Function used by x_snprintf to print out the name of a clazz (format %k).
** N.B. Truncates non-ASCII chars instead of converting to UTF*. <== FIXME
*/
char *print_clazz(char *buffer, int *remain, void *c, int w, int p, unsigned int f) {
  w_clazz clazz = c;
  w_char *chars;
  w_char *end;

  if (c == NULL) {
    strncpy(buffer, (char*)"<NULL>", *remain);
    if (*remain < 6) {
      *buffer += *remain;
      *remain = 0;
    }
    else {
      *buffer += 6;
      *remain -= 6;
    }

    return buffer;

  }

  chars = clazz->dotified->chars;
  end = chars + clazz->dotified->length;
  while (*remain > 0 && chars < end) {
    *buffer++ = *chars++;
    *remain -= 1;
  }

  return buffer;
}

/*
** Functions to get an unsigned integer (respectively 1, 2 and 4 bytes
** long) from the class file
*/

inline static u1 get_u1(w_bar s) {
  w_byte ch;
  w_int lread;

  bar_read(s, &ch, 1, &lread);

  return ch;

}

inline static u2 get_u2(w_bar s) {
  u2 value;

  value  = get_u1(s) << 8;
  value |= get_u1(s);

  return value;

}

inline static u4 get_u4(w_bar s) {

  u4 value;

  value  = get_u1(s) << 24;
  value |= get_u1(s) << 16;
  value |= get_u1(s) <<  8;
  value |= get_u1(s); 

  return value;

}

static w_instance check_length(w_bar s, int min) {
  if (bar_avail(s) < min) {
    woempa(9,"Need to read %d bytes from classfile, only %d bytes left\n", min, bar_avail(s));

    return createExceptionInstance(currentWonkaThread, clazzClassFormatError, string_error_too_short);

  }

  return NULL;
}

/*
** Get the "magic number" 0xcafebabe from the class file
*/

inline static w_instance get_cafebabe(w_bar s) {

  u4 cafebabe;
  w_instance problem = check_length(s, 4);

  if (!problem) {
    cafebabe = get_u4(s);

    if (cafebabe != 0xcafebabe) {
      woempa(9,"Uh oh, cafebabe is called `%08x' today\n", cafebabe);

      return createExceptionInstance(currentWonkaThread, clazzClassFormatError, string_error_bad_magic);

    }
  }

  return problem;
}

/*
** Get the major and minor version number from the class file
*/

inline static w_instance get_version(w_clazz clazz, w_bar s) {
  w_instance problem = check_length(s, 4);

  if (!problem) {
    clazz->cminor = get_u2(s);
    clazz->cmajor = get_u2(s);
  }

  return problem;
}

/*
** parseConstant will increment *idx by the number of slots occupied
** by the constant (1 or 2). It will read the classfile
** data and put it in the constant pool.
*/

static void bumpUtf8Refcount(w_clazz clazz, w_int utf8index) {
  ++clazz->constants[utf8index].info.utf8.refcount;
  woempa(1, "Incremented reference count of constant[%d], is now %d\n", utf8index, clazz->constants[utf8index].info.utf8.refcount);
}

static void doUtf8Refcounts(w_clazz clazz, w_int i) {
  w_cnt      c = &clazz->constants[i];
  w_ushort   utf8index;

  switch (c->cnt_tag) {
    case CONSTANT_CLASS:
      utf8index = c->info.clindex;
      woempa(1, "Constant[%d] is a class: utf8 is %d.\n", i, utf8index);
      bumpUtf8Refcount(clazz, utf8index);
      break;
      
    case CONSTANT_STRING:
      utf8index = c->info.strindex;
      woempa(1, "Constant[%d] is a string: utf8 is %d.\n", i, utf8index);
      bumpUtf8Refcount(clazz, utf8index);
      break;
      
    case CONSTANT_FIELD:
      woempa(1, "Constant[%d] is a field: class is %d, name-and-type is %d.\n", i, c->info.member.class_index, c->info.member.name_and_type_index);
      utf8index = clazz->constants[c->info.member.name_and_type_index].info.nat.name_index;
      bumpUtf8Refcount(clazz, utf8index);
      utf8index = clazz->constants[c->info.member.name_and_type_index].info.nat.descriptor_index;
      bumpUtf8Refcount(clazz, utf8index);
      break;
      
    case CONSTANT_METHOD:
      woempa(1, "Constant[%d] is a method: class is %d, name-and-type is %d.\n", i, c->info.member.class_index, c->info.member.name_and_type_index);
      utf8index = clazz->constants[c->info.member.name_and_type_index].info.nat.name_index;
      bumpUtf8Refcount(clazz, utf8index);
      utf8index = clazz->constants[c->info.member.name_and_type_index].info.nat.descriptor_index;
      bumpUtf8Refcount(clazz, utf8index);
      break;
      
    case CONSTANT_IMETHOD:
      woempa(1, "Constant[%d] is an imethod: class is %d, name-and-type is %d.\n", i, c->info.member.class_index, c->info.member.name_and_type_index);
      utf8index = clazz->constants[c->info.member.name_and_type_index].info.nat.name_index;
      bumpUtf8Refcount(clazz, utf8index);
      utf8index = clazz->constants[c->info.member.name_and_type_index].info.nat.descriptor_index;
      bumpUtf8Refcount(clazz, utf8index);
      break;
      
    default:
      
  }
}

static w_instance parseConstant(w_clazz clazz, w_bar s, w_int *idx) {
  /*
  ** We check that at least 3 bytes are available, 1 for the tag and at
  ** least 2 bytes thereafter.  CONSTANT_UTF8, CONSTANT_CLASS, and 
  ** CONSTANT_STRING just need these 2  bytes: the other types need
  ** more, so we insert an extra check when we get there.
  */
  w_cnt c = &clazz->constants[*idx];
  w_instance problem = check_length(s, 3);
  w_tag      tag;
  u2         length;
  w_int      lread;

  if (problem) {

    return problem;

  }

  tag = get_u1(s);
  c->cnt_tag = tag;

  switch (tag) {
    case CONSTANT_UTF8: 
      length = get_u2(s);

      c->get = get_unresolved_Other;
      c->info.utf8.refcount = 0;
      c->info.utf8.length = length;
      if (length > 0) {
        char *buffer = allocMem_system((w_size)length);
        bar_read(s, buffer, length, &lread);
        c->info.utf8.string = utf2String(W_Thread_system, buffer, length);
        releaseMem(buffer);
      }
      else {
        c->info.utf8.string = utf2String(W_Thread_system, "", 0);
        woempa(7, "Utf8 constant with length %d!\n", length);
      }
      woempa(1, "Constant[%d] = `%w'\n", *idx, c->info.utf8.string);
      
      *idx += 1;
      break;
      
    case CONSTANT_INTEGER:
      problem = check_length(s, 4);
      if (!problem) {
        c->get = get_resolved;
        c->info.intval = (w_int)get_u4(s);
        *idx += 1;
      }
      break;
      
    case CONSTANT_FLOAT:
      problem = check_length(s, 4);
      if (!problem) {
        c->get = get_resolved;
        c->info.floatval = (w_float)get_u4(s);
        *idx += 1;
      }
      break;
      
    case CONSTANT_LONG:
      problem = check_length(s, 8);
      if (!problem) {
        c->get = get_resolved;
        c->info.longval.words[WORD_MSW] = get_u4(s);
        c->info.longval.words[WORD_LSW] = get_u4(s);
        (c+1)->get = get_wrong_word;
        (c+1)->cnt_tag = NO_VALID_ENTRY;
        *idx += 2;
      }
      break;
      
    case CONSTANT_DOUBLE:
      problem = check_length(s, 8);
      if (!problem) {
        c->get = get_resolved;
        c->info.doubleval.words[WORD_MSW] = get_u4(s);
        c->info.doubleval.words[WORD_LSW] = get_u4(s);
        (c+1)->get = get_wrong_word;
        (c+1)->cnt_tag = NO_VALID_ENTRY;
        *idx += 2;
      }
      break;
      
    case CONSTANT_CLASS:
      c->get = get_unresolved_Class;
      c->info.clindex = get_u2(s);
      *idx += 1;
      break;
      
    case CONSTANT_STRING:
      c->get = get_unresolved_String;
      c->info.strindex = get_u2(s);
      *idx += 1;
      break;
      
    case CONSTANT_FIELD:
      problem = check_length(s, 4);
      if (!problem) {
        c->get = get_unresolved_Field;
        c->info.member.class_index = get_u2(s);
        c->info.member.name_and_type_index = get_u2(s);
        *idx += 1;
      }
      break;
      
    case CONSTANT_METHOD:
      problem = check_length(s, 4);
      if (!problem) {
        c->get = get_unresolved_Method;
        c->info.member.class_index = get_u2(s);
        c->info.member.name_and_type_index = get_u2(s);
        *idx += 1;
      }
      break;
      
    case CONSTANT_IMETHOD:
      problem = check_length(s, 4);
      if (!problem) {
        c->get = get_unresolved_IMethod;
        c->info.member.class_index = get_u2(s);
        c->info.member.name_and_type_index = get_u2(s);
        *idx += 1;
      }
      break;
      
    case CONSTANT_NAME_AND_TYPE:
      problem = check_length(s, 4);
      if (!problem) {
        c->get = get_unresolved_Other;
        c->info.nat.name_index = get_u2(s);
        c->info.nat.descriptor_index = get_u2(s);
        woempa(1, "Constant[%d] is a name-and-type: name is %d, type is %d.\n", *idx, c->info.nat.name_index, c->info.nat.descriptor_index);
        *idx += 1;
      }
      break;

    default:
      c->get = get_unresolved_Other;
      problem = createExceptionInstance(currentWonkaThread, clazzClassFormatError, string_error_bad_constant);
      break;      
  }

  return problem;

}

/*
** Scan through the constant pool and copy the data to clazz->constants[].
*/

inline static w_instance get_constantpool(w_clazz clazz, w_bar s) {

  w_thread thread = currentWonkaThread;
  w_int i;
  w_instance problem = check_length(s, 2);

  if (problem) {

    return problem;

  }

  clazz->numConstants = get_u2(s);

  woempa(1, "Number of constant pool entries is %d, using %d bytes each\n", clazz->numConstants - 1, sizeof(w_Cnt));
  clazz->constants = allocMem(thread, (clazz->numConstants + 1) * sizeof(w_Cnt));

  for (i = 1; i < (w_int)clazz->numConstants; i++) {
    clazz->constants[i].get = get_unresolved_Other;
  }

  for (i = 1; i < (w_int)clazz->numConstants && !problem; ) {
    woempa(2, "parsing constant pool entry %d of %d\n", i, clazz->numConstants - 1);
     problem = parseConstant(clazz, s, &i);
  }

  if (problem) {

    return problem;

  }

  for (i = 1; i < (w_int)clazz->numConstants; i++) {
    doUtf8Refcounts(clazz, i);
  }

  return NULL;
}

/*
** Parse a field attribute (a name/value pair).
*/
static w_instance parseFieldAttribute(w_field field, w_bar s) {

  w_thread thread = currentWonkaThread;
  u2 attribute_name_index = 0;
  u4 attribute_length = 0;
  w_string attributeName;
  w_instance problem = check_length(s, 6);
  
  if (!problem) {
    attribute_name_index = get_u2(s);
    attribute_length = get_u4(s);
    problem = check_length(s, (w_int)attribute_length);
  }

  if (problem) {

    return problem;

  }

  attributeName = resolveUtf8(field->declaring_clazz, attribute_name_index, thread);
  woempa(1, "Parsing attribute %w for field %w of %k.\n", attributeName, NM(field), field->declaring_clazz);

  /*
  ** See which attribute we have here.
  */

  if (attributeName == string_ConstantValue) {
    problem = check_length(s, 2);
    if (problem) {

      return problem;

    }
  
    field->initval = get_u2(s);
  }
/* Not supported
  else if (attributeName == string_Deprecated) {
    field->flags |= ACC_DEPRECATED;
  }
  else if (attributeName == string_Synthetic) {
    field->flags |= ACC_SYNTHETIC;
  }
*/
  else {
    u4 i;

    /*
    ** Unknown attributes are silently ignored.
    */

    woempa(8, "Unknown attribute %w\n", attributeName);
    for (i = 0; i < attribute_length; i++) {
      get_u1(s);
    }
  }

  return NULL;
}

/*
** Parse a field entry in the classfile.  The entry consists of a
** flags set (u2), name index (u2), descriptor index (u2), number of
** attributes (u2), followed by the attributes (see previous function).
**
** We store the field entries sorted as follows:
**   first, all static fields;
**   then all instance fields which store references;
**   lastly, instance fields which store primitives.
** This ordering will simplify later processing (class preparation, etc.).
*/
static w_instance parseField(w_clazz clazz, w_size idx, w_bar s) {

  w_thread thread = currentWonkaThread;
  u2 tmp_u2;
  int attributeNumber;
  w_field  field;
  w_flags  flags;
  w_string name;
  w_string desc_name;
  w_desc   desc;
  w_instance problem = check_length(s, 8);

  if (problem) {

    return problem;

  }

  flags = get_u2(s);
  tmp_u2 = get_u2(s);
  name = resolveUtf8(clazz, tmp_u2, thread);
  
  tmp_u2 = get_u2(s);
  desc_name = resolveUtf8(clazz, tmp_u2, thread);
  desc = newFieldDescriptor(thread, desc_name);
  deregisterString(desc_name);

  if (desc == NULL) {
    wabort(ABORT_WONKA,"whoops - failed to parse field descriptor %p\n", NULL);
  }

  field = &clazz->fields[idx];
  field->declaring_clazz = clazz;
  field->label = (char *) "field";
  field->flags = flags;
  if (isSet(field->flags, ACC_STATIC)) {
    ++clazz->numStaticFields;
  }
  field->name  = name;
  field->desc  = desc;
  field->initval = 0;
  if (field->desc->tag == D_Primitive) {
    switch (field->desc->body.primitive.type) {
      case P_long:
      case P_double:
        setFlag(field->flags, FIELD_IS_LONG);
      default:
    }
  }
  else {
    setFlag(field->flags, FIELD_IS_REFERENCE);
  }

  /*
  ** Parse its attributes.
  */

  tmp_u2 = get_u2(s);
  if (tmp_u2) {
    for (attributeNumber = 0; attributeNumber < tmp_u2; attributeNumber++) {
      parseFieldAttribute(field, s);
    }
  }

  return NULL;
}

static void parseMethodCodeExceptionTable(w_method method, w_bar s, w_instance *loading_problem) {

  w_thread thread = currentWonkaThread;
  w_exception exception;
  w_int i;

  method->numExceptions = get_u2(s);
  if (method->numExceptions > 0) {
    method->exceptions = allocMem(thread, method->numExceptions * sizeof(w_Exception));
    for (i = 0; i < method->numExceptions; i++) {
      exception = &method->exceptions[i];
      exception->start_pc = get_u2(s);
      exception->end_pc = get_u2(s);
      exception->handler_pc = get_u2(s);
      exception->type_index = get_u2(s);
    }
  }
}

static void parseMethodCode(w_method method, w_bar s) {

  w_int lread;

  method->stack_i = get_u2(s);
  method->local_i = get_u2(s);
  method->code_length = get_u4(s);
  if (method->code_length > 0) {
    method->code = allocMem(currentWonkaThread, (w_size)method->code_length);
    bar_read(s, method->code, (signed)method->code_length, &lread);
  }

  
}

static void parseMethodAttribute(w_method method, w_bar s, w_instance *loading_problem) {
  w_thread  thread = currentWonkaThread;
  u2 attribute_name_index;
  u4 attribute_length;
  w_int i;
  w_string attributeName;
  
  attribute_name_index = get_u2(s);
  attribute_length = get_u4(s);
  attributeName = resolveUtf8(method->clazz, attribute_name_index, thread);

  woempa(3, "Parsing %w attribute\n", attributeName);

  /*
  ** See which attribute we have here.
  */

  if (attributeName == string_Code) {
    parseMethodCode(method, s);
    parseMethodCodeExceptionTable(method, s, loading_problem);

    if (*loading_problem) {
      return;
    }

    parseMethodCodeAttributes(method, s, loading_problem);

    if (*loading_problem) {
      return;
    }
    woempa(4, "Read code attribute; %d bytes, %d exception handlers, throws %d exceptions\n", method->code_length, method->numExceptions, method->numThrows);
  }
  else if (attributeName == string_Exceptions) {
    method->numThrows = get_u2(s);
    method->throws = allocMem(thread, method->numThrows * sizeof(w_ushort));
    for (i = 0; i < method->numThrows; i++) {
      method->throws[i] = get_u2(s);
    }
  }
  else if (attributeName == string_LocalVariableTable) {
    method->numLocalVars = get_u2(s);
    if (method->numLocalVars > 0) {
      method->localVars = allocMem(thread, method->numLocalVars * sizeof(w_LocalVar));
      for (i = 0; i < method->numLocalVars; i++) {
        method->localVars[i].start_pc = get_u2(s);
        method->localVars[i].end_pc = method->localVars[i].start_pc + get_u2(s);
        method->localVars[i].name_index = get_u2(s);
        method->localVars[i].desc_index = get_u2(s);
        method->localVars[i].slot = get_u2(s);
      }
    }
  }
  else if (attributeName == string_LineNumberTable) {
    method->numLineNums = get_u2(s);
    if (method->numLineNums > 0) {
      method->lineNums = allocMem(thread, method->numLineNums * sizeof(w_LineNum));
      for (i = 0; i < method->numLineNums; i++) {
        method->lineNums[i].start_pc = get_u2(s);
        method->lineNums[i].line_number = get_u2(s);
      }
    }
  }
/* Not supported
  else if (attributeName == string_Deprecated) {
    method->flags |= ACC_DEPRECATED;
  }
  else if (attributeName == string_Synthetic) {
    method->flags |= ACC_SYNTHETIC;
  }
*/
  else {
    /* XXX should we throw something here ? */
    woempa(8, "Unknown attribute '%w'\n", attributeName);
    for (i=0;(u4)i<attribute_length;++i) {
      get_u1(s);
    }
  }

}

/*
** This function can not be made inline. It's called by parseMethodAttribute and this function calls
** this parseMethodAttribute function recursively...
*/

static void parseMethodCodeAttributes(w_method method, w_bar s, w_instance *loading_problem) {

  w_int i;
  w_int attr_count;

  attr_count = get_u2(s);
  if (attr_count > 0) {
    for (i = 0; i < attr_count; i++) {
      parseMethodAttribute(method, s, loading_problem);
      if (*loading_problem) {

        return;

      }
    }
  }

}

static w_method parseMethod(w_clazz clazz, w_bar s, w_instance *loading_problem) {

  w_thread thread = currentWonkaThread;
  w_method method = allocClearedMem(thread, sizeof(w_Method));
  u2 numAttributes;
  w_size i;
  
  method->label = (char *)"method";
  method->clazz = clazz;
  method->flags = get_u2(s);
  method->slot = SLOT_NOT_ALLOCATED;
  method->code = NULL;
  method->function.void_fun = NULL;
  method->numThrows = 0;
  method->numExceptions = 0;
  method->numLocalVars = 0;
  method->numLineNums = 0;
  
  /*
  ** Get the method name and the descriptor.
  */

  method->name = resolveUtf8(clazz,get_u2(s), thread);
  method->desc_index = get_u2(s);

  /*
  ** Parse the attributes of this method.
  */

  numAttributes = get_u2(s);
  if (numAttributes > 0) {
    for (i = 0; i < numAttributes; i++) {
      parseMethodAttribute(method, s, loading_problem);
      if (*loading_problem) {
        return NULL;
      }
    }
  }
  
  return method;
  
}


/*
** Register/deregister a w_Clazz structure in a class hashtable.  
*/

void registerClazz(w_clazz clazz, w_hashtable hashtable) {
  w_clazz  existing;

  woempa(1, "Registering clazz %k in hashtable %p.\n", clazz, hashtable);

  existing = (w_clazz)ht_write(hashtable, (w_word)clazz->dotified, (w_word)clazz);

  if (existing && existing != clazz) {
    wabort(ABORT_WONKA, "Fascinating. The class %k (w_clazz %p) was already registered in %s as w_clazz %p.\n", clazz, clazz, hashtable->label, existing);
  }

}

void deregisterClazz(w_clazz clazz, w_hashtable hashtable) {
  w_clazz  existing;

  woempa(8, "Deregistering clazz %k from hashtable %p.\n", clazz, hashtable);

  existing = (w_clazz)ht_erase(hashtable, (w_word)clazz->dotified);

  if (!existing) {
    woempa(9, "Fascinating. When we wanted to deregister class %k, we couldn't find it in %s\n", clazz, hashtable->label);
  }

}

// return exception or NULL
inline static w_instance get_superclass(w_clazz clazz, w_bar s) {
  w_thread thread = clazz->loading_thread;
  w_instance loader = clazz2loader(clazz);
  w_int super_index;
  w_string super_slashed;
  w_string super_dotified;
  w_clazz super_clazz = NULL;
  w_instance problem = NULL;

  super_index = get_u2(s);
  if (super_index) {
    super_slashed = resolveUtf8(clazz, clazz->constants[super_index].info.clindex, thread);
    super_dotified = slashes2dots(thread, super_slashed);
    deregisterString(super_slashed);
    woempa(1, "  superclass: %w\n", super_dotified);

    super_clazz = namedClassMustBeLoaded(loader, super_dotified, &problem);

    /*
    ** [J+JVM Constraint 4.1.3]
    **
    ** We suppress this check while loading class java.lang.Object,
    ** otherwise we have bootstrapping problems (e.g. when we try to
    ** load java.lang.Throwable.
    */
    if (!problem && clazzObject && super_clazz->loading_thread == thread) {
      problem = createExceptionInstance(thread, clazzClassCircularityError, super_clazz->dotified);
    }

    if (!problem) {
      clazz->superClazz = super_clazz;
      clazz->numSuperClasses = clazz->superClazz->numSuperClasses + 1;
    
      woempa(1, "Direct superclass of %k is %w (1 of %d).\n", clazz, super_dotified, clazz->numSuperClasses);

      clazz->flags |= super_clazz->flags & (CLAZZ_HERITABLE_FLAGS);
    }
    else {
      woempa(1, "Class %k has no direct superclass.\n", clazz);
      clazz->superClazz = NULL;
      clazz->numSuperClasses = 0;
    }
  }

  return problem;
}

// return exception or NULL
inline static w_instance get_interfaces(w_clazz clazz, w_bar s) {
  w_thread thread = clazz->loading_thread;
  w_instance loader = clazz2loader(clazz);
  w_size i;
  w_int interface_index;
  w_string interface_slashed;
  w_string interface_dotified;
  w_clazz interfaze;
  w_instance problem = NULL;

  clazz->numInterfaces = get_u2(s);
  if (clazz->numInterfaces > 0) {
    clazz->interfaces = allocMem(thread, clazz->numInterfaces * sizeof(w_clazz));

    for (i = 0; i < clazz->numInterfaces; i++) {
      interface_index = get_u2(s);
      interface_slashed = resolveUtf8(clazz,clazz->constants[interface_index].info.clindex, thread);
      interface_dotified = slashes2dots(thread, interface_slashed);
      deregisterString(interface_slashed);
      woempa(1,"  interface[%d]: '%w'\n",i,interface_dotified);

      interfaze = namedClassMustBeLoaded(loader, interface_dotified, &problem);
      /*
      ** [J+JVM Constraint 4.1.3]
      */
      if (!problem && interfaze->loading_thread == thread) {
        problem = createExceptionInstance(thread, clazzClassCircularityError, interfaze->dotified);
      }

      if (problem) {

        return problem;

      }

      clazz->interfaces[i] = interfaze;
      woempa(1, "Direct superinterface %d of %k is %w.\n",i, clazz, interface_dotified);
    }
  }

  return NULL;
}

static void swapFields(w_clazz clazz, w_int i, w_int j) {
  w_Field f;

  woempa(1, "Swapping fields [%d] (`%w') and [%d] (`%w') of %k\n", i, clazz->fields[i].name, j, clazz->fields[j].name, clazz);
  memcpy(&f, &clazz->fields[i], sizeof(w_Field));
  memcpy(&clazz->fields[i], &clazz->fields[j], sizeof(w_Field));
  memcpy(&clazz->fields[j], &f, sizeof(w_Field));
}

inline static void sortFields(w_clazz clazz) {
  w_int i;
  w_int j;

  woempa(1, "Sorting fields of %k\n", clazz);
  i = 0;
  j = clazz->numFields - 1;
  while (i < j) {
    if (isSet(clazz->fields[i].flags, ACC_STATIC)) {
      woempa(1, "Field [%d] (`%w') is static, leaving it be\n",  i, clazz->fields[i].name);
      ++i;
    }
    else if (isNotSet(clazz->fields[j].flags, ACC_STATIC)) {
      woempa(1, "Field [%d] (`%w') is non-static, leaving it be\n",  j, clazz->fields[j].name);
      --j;
    }
    else {
      swapFields(clazz, i, j);
      ++i;
      --j;
    }
  }

  woempa(1, "Sorting static fields of %k\n", clazz);
  i = 0;
  j = clazz->numStaticFields - 1;
  while (i < j) {
    if (isSet(clazz->fields[i].flags, FIELD_IS_REFERENCE)) {
      woempa(1, "Static field [%d] (`%w') is  reference, leaving it be\n",  i, clazz->fields[i].name);
      ++i;
    }
    else if (isNotSet(clazz->fields[j].flags, FIELD_IS_REFERENCE)) {
      woempa(1, "Static field [%d] (`%w') is primitive, leaving it be\n",  j, clazz->fields[j].name);
      --j;
    }
    else {
      swapFields(clazz, i, j);
    }
  }

  woempa(1, "Sorting non-static fields of %k\n", clazz);
  i = clazz->numStaticFields;
  j = clazz->numFields - 1;
  while (i < j) {
    if (isSet(clazz->fields[i].flags, FIELD_IS_REFERENCE)) {
      woempa(1, "Static field [%d] (`%w') is  reference, leaving it be\n",  i, clazz->fields[i].name);
      ++i;
    }
    else if (isNotSet(clazz->fields[j].flags, FIELD_IS_REFERENCE)) {
      woempa(1, "Static field [%d] (`%w') is primitive, leaving it be\n",  j, clazz->fields[j].name);
      --j;
    }
    else {
      swapFields(clazz, i, j);
    }
  }

}

inline static w_instance get_fields(w_clazz clazz, w_bar s) {
  w_thread thread = currentWonkaThread;
  w_instance problem;
  w_size i;
  w_size n;

  clazz->staticFields = NULL;
  clazz->numStaticFields = 0;
  clazz->numStaticWords = 0;
  n = get_u2(s);
  clazz->numFields = n;
  if (n > 0) {
    woempa(1, "Allocating %d w_Field's of %d bytes each\n", clazz->numFields, sizeof(w_Field));
    clazz->fields = allocMem(thread, clazz->numFields * sizeof(w_Field));
    for (i = 0; i < n; i++) {
      problem = check_length(s, 8);

      if (!problem) {
        problem = parseField(clazz, i, s);
      }

      if (problem) {

        return problem;

      }
    }
  }

  if (n > 1) {
    sortFields(clazz);
  }

  return NULL;
}

inline static void get_methods(w_clazz clazz, w_bar s, w_instance *loading_problem) {

  w_thread thread = currentWonkaThread;
  w_size i;

  clazz->numMethods = get_u2(s);
  clazz->numInheritableMethods = 0;
  if (clazz->numMethods > 0) {
    clazz->methods = allocMem(thread, clazz->numMethods * sizeof(w_method));
    for (i = 0; i < clazz->numMethods; i++) {
      clazz->methods[i] = parseMethod(clazz, s, loading_problem); 
      if (*loading_problem) {

        return;

      }
      woempa(1, "parsed method %d of %d (%w)\n", i + 1, clazz->numMethods, NM(clazz->methods[i]));
      woempa(1, "  Takes %d argument words; returns %d words; needs %d stack words\n", clazz->methods[i]->arg_i,  clazz->methods[i]->return_i, clazz->methods[i]->stack_i);
    }
  }
}

static void parseClassAttributes(w_clazz clazz, w_bar s) {

  w_thread thread = currentWonkaThread;
  u2 attribute_name_index;
  u4 attribute_length;
  w_string attributeName;
  w_int  i;
  
  attribute_name_index = get_u2(s);
  attribute_length = get_u4(s);
  attributeName = resolveUtf8(clazz, attribute_name_index, thread);
  
  if (attributeName == string_SourceFile) {
    clazz->source = resolveUtf8(clazz, get_u2(s), currentWonkaThread);
  }
  else {
    woempa(1, "Unknown attribute '%w'\n", attributeName);
    for (i = 0; (u4)i < attribute_length; i++) {
      get_u1(s);
    }
  }

}

inline static void get_attributes(w_clazz clazz, w_bar s) {

  u2 numberOfAttributes;

  numberOfAttributes = get_u2(s);
  if (numberOfAttributes > 0) {
    parseClassAttributes(clazz,s);
  }

}

/*
** Create the w_Clazz structure corresponding to a class,
** reading the class file from stream s.  The class will
** be registered in the hashtable specified.
*/

// TODO, rewrite createClazz to use native exceptions in stead of passing the 'loading_problem' pointer
// down the chain...

w_clazz createClazz(w_thread thread, w_string name, w_bar dev, w_instance loader, w_instance *loading_problem) {

  w_int thisClassIndex;
  w_hashtable class_hashtable = loader2hashtable(loader);
  w_clazz clazz;
  w_string slashed;

  clazz = allocClazz(thread);
  // We initialise instanceSize to an impossible value, so any attempt
  // to allocate an instance of an unresolved clazz will fail.
  clazz->instanceSize = 0x80000000;  // mugtrap
  clazz->aclazz = NULL;
  clazz->loader = loader;
  clazz->resolution_monitor = allocMem(thread, sizeof(x_Monitor));
  x_monitor_create(clazz->resolution_monitor);
  clazz->loading_thread = thread;

  *loading_problem = get_cafebabe(dev);
  if (*loading_problem == NULL) {
    *loading_problem = get_version(clazz, dev);
  }

  if (*loading_problem) {
    releaseMem(clazz);

    return NULL;

  }

  *loading_problem = get_constantpool(clazz, dev);
  clazz->flags = get_u2(dev) & (ACC_PUBLIC | ACC_FINAL | ACC_SUPER | ACC_INTERFACE | ACC_ABSTRACT);

  /**
  ** this lines were added to cope with bad compilers
  */
  if(isSet(clazz->flags, ACC_INTERFACE) && isNotSet(clazz->flags, ACC_ABSTRACT)){
    woempa(9,"How rude, clazz %w is has ACC_INTERFACE set but not ACC_ABSTRACT\n",name);
    woempa(9,"Fixing this ...\n");
    setFlag(clazz->flags, ACC_ABSTRACT);
  } 

  thisClassIndex = get_u2(dev);
  slashed = resolveUtf8(clazz, clazz->constants[thisClassIndex].info.clindex, thread);
  clazz->dotified = slashes2dots(W_Thread_system, slashed);
  deregisterString(slashed);
  if (name && clazz->dotified != name) {
    woempa(9,"Impostor! The class file `%w' contains class `%k'.\n",name,clazz);
    *loading_problem = createExceptionInstance(thread,clazzClassFormatError, string_error_wrong_name);
    destroyClazz(clazz);

    return NULL;

  }

  if (clazzClass && clazzClass->prepare == alreadyPrepared) {
    woempa(1, "attaching instance of Class to class %k.\n", clazz);
    attachClassInstance(clazz);
  }

  createClazzDescriptor(clazz);
  registerClazz(clazz,class_hashtable);  

  clazz->verify = verifyClass;
  clazz->prepare = prepareClass;
  clazz->resolve = resolveClass;
  clazz->initialize = initializeClass;
 
  *loading_problem = get_superclass(clazz, dev);
  if (*loading_problem == NULL) {
    *loading_problem = get_interfaces(clazz, dev);
    if (*loading_problem == NULL) {
      get_fields(clazz, dev);
      get_methods(clazz, dev, loading_problem);
      if (*loading_problem == NULL) {
        get_attributes(clazz, dev);
      }
    }
  }

/* Why does this fail ???
  if (*loading_problem == NULL && bar_read(dev, &ch, 1, &lread) == 1) {
    woempa(9,"ClassFormatError when loading %k - junk at end of file\n",clazz);
    *loading_problem = createExceptionInstance(thread,clazzClassFormatError, string_error_too_long);
  }
*/

  if (*loading_problem) {
    destroyClazz(clazz);

    return NULL;

  }

  setClazzState(clazz, CLAZZ_STATE_LOADED);
  clazz->loading_thread = NULL;

#ifdef CACHE_INSTANCES
  /*
  ** Set up the cache fields for this clazz. Note that the bytes_needed field is set up during the
  ** loading of the clazz, when we create the field table.
  */
  
  clazz->cache_hits = 0;
  clazz->cache_miss = 0;
  clazz->cache_lock = allocMem(thread, sizeof(x_Mutex));
  x_mutex_create(clazz->cache_lock);
  clazz->cache = NULL;
#endif

  return clazz;

}

/*
** Attach an instance of java.lang.Class to this clazz.
*/

w_instance attachClassInstance(w_clazz clazz) {
  w_thread   thread = currentWonkaThread;
  w_instance Class;

  if (clazzClass == NULL) {
    wabort(ABORT_WONKA, "Cannot create instance of Class for %k, as clazzClass not yet defined\n", clazz);
    return NULL;
  }

  Class = allocInstanceWithFlags(thread, clazzClass, O_SYSTEM);

  if (Class) {
    clazz->Class = Class;
    Class[F_Class_loader] = (w_word)clazz->loader;
    Class[F_Class_wotsit] = (w_word)clazz;
    woempa(1, "Attached instance of Class (%p) to %k.\n", Class, clazz);
  }

  return Class;
}

/*
** Create the descriptor for this class. 
*/

void createClazzDescriptor(w_clazz clazz) {
  w_thread thread = currentWonkaThread;
  w_char * desc_buffer;
  w_string desc_string;
  w_size   i;
  w_char   ch;

  woempa(1, "Creating descriptor for %k.\n", clazz);

  desc_buffer = allocMem_system((clazz->dotified->length + 2) * sizeof(w_char));
  desc_buffer[0] = (w_char)'L';
  for (i = 0; i < clazz->dotified->length; ++i) {
    ch = clazz->dotified->chars[i];
    desc_buffer[i + 1] = ch == '.' ? '/' : ch;
  }
  desc_buffer[clazz->dotified->length + 1] = ';';

  desc_string = unicode2String(thread, desc_buffer, clazz->dotified->length + 2);
  clazz->desc = newFieldDescriptor(thread, desc_string);
  if(clazz->desc == NULL) {
    wabort(ABORT_WONKA,"whoops - failed to parse clazz descriptor %w\n", desc_string);
  }

  deregisterString(desc_string);
  releaseMem(desc_buffer);

  woempa(1, "Registered descriptor %w for %k.\n", NM(clazz->desc), clazz);
}

w_clazz _allocClazz(const char *file, int line, w_thread thread) {

  w_clazz clazz = allocClearedMem(thread, sizeof(w_Clazz));
  clazz->label = (char *) "clazz";
  clazz->file = (char *) file;
  clazz->line = line;
  
  return clazz; 

}

/*
** Destroy the w_Clazz structure corresponding to a class.
*/

void destroyClazz(w_clazz clazz) {
  w_int  i;
  
  woempa(7,"Destroying class %k\n",clazz);
// undo this ...  list_init(clazz->cache_list);
// and this ...  clazz->cache_list = (w_object) &clazz->cache_next;
// TODO: move this to dissolveClazz, when cache creation is also moved
// return all cached instances ... and then
//  woempa(7,"Destroying the class cache lock\n");
//  x_mutex_delete(clazz->cache_lock);
//  releaseMem(clazz->cache_lock);
//  release the cache wordset too ...
/* TODO: (or rather undo ...)
        get_attributes(clazz, s);
      get_methods(clazz, s, loading_problem);
      get_fields(clazz, s);
*/

  for (i = 1; i < (w_int)clazz->numConstants; i++) {
    dissolveConstant(clazz, i);
  }

  if (clazz->interfaces) {
    releaseMem(clazz->interfaces);
  }
  woempa(7,"Deregistering the class (isn't that bass-ackwards?\n");
  deregisterClazz(clazz, loader2hashtable(clazz->loader));  
//  destroyClazzDescriptors(clazz->thread,clazz);
  woempa(7,"Deregistering string `%w'\n",clazz->dotified);
  deregisterString(clazz->dotified);

  // TODO: if (clazz->aclazz), clean up clazz->aclazz
  // TODO: release all strings, methods, fields
  releaseMem(clazz->constants);
  x_monitor_delete(clazz->resolution_monitor);
  releaseMem(clazz->resolution_monitor);
  if (clazz->failure_message) {
    deregisterString(clazz->failure_message);
  }
  woempa(7,"Finally releasing memory of clazz itself\n");
  releaseMem(clazz);
}

/*
** To find a field, we have to search recursively over the superclass
** and all superinterfaces.
*/
w_field getField(w_clazz clazz, w_string name) {
  w_clazz current_clazz = clazz;
  w_field field;
  w_size  i;
  w_size  j;

  woempa(1, "Looking for field `%w' in %k\n", name, clazz);
  while (current_clazz) {
    field = current_clazz->fields;
    for (i = 0; i < current_clazz->numFields; ++i) {
      woempa(1, "Trying `%w'\n", field->name);
      if (field->name == name) {
        woempa(1, "Success!  Found field `%w' in %k\n", name, current_clazz);

        return field;

      }
      ++field;
    }
    for (j = 0; j < current_clazz->numInterfaces; ++j) {
      woempa(1, "Looking in superinterface %w of %w\n", current_clazz->interfaces[j], current_clazz);
      field = getField(current_clazz->interfaces[j], name);
      if (field) {

        return field;

      }
    }
    woempa(1, "Looking in superclass %w of %w\n", current_clazz->superClazz, current_clazz);
    current_clazz = current_clazz->superClazz;
  }

  return NULL;
}

w_method getMethod_name_sig(w_clazz clazz, w_string name, w_string sig) {
  w_method *methodptr = clazz->methods;
  w_size    i;

  woempa(1, "Looking for method `%w%w' in %k\n", name, sig, clazz);
  for (i = 0; i < clazz->numMethods; ++i) {
    woempa(1, "Trying `%m'\n", *methodptr);
    if ((*methodptr)->name == name && (*methodptr)->param_list == sig) {
      woempa(1, "Found `%m'\n", *methodptr);

      return *methodptr;

    }
    ++methodptr;
  }

  return NULL;
}

w_clazz identifyClazz(w_string name, w_instance loader) {
  w_clazz result = NULL;

  if (loader == NULL) {
    result = (w_clazz)ht_read_no_lock(system_class_hashtable, (w_word)name);
  }
  else {
    if (loader == systemClassLoader) {
      result = (w_clazz)ht_read_no_lock(system_class_hashtable, (w_word)name);
    }
    if (!result) {
      result = (w_clazz)ht_read_no_lock(loader2hashtable(loader), (w_word)name);
    }
  }

  if (!result) {
    result = allocMem_system(sizeof(w_UnloadedClazz));

    result->dotified = registerString(name);
    result->label    = (char *)"clazz(unloaded)";
    result->flags    = 0; 
    result->loader   = loader;
  }

  return result;
}

