/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: descriptor.c,v 1.8 2002/03/26 10:44:14 gray Exp $
*/

#include <stdio.h>
#include <string.h>

#include "arrays.h"
#include "clazz.h"
#include "descriptor.h"
#include "hashtable.h"
#include "heap.h"
#include "ieee754.h"
#include "loading.h"
#include "ts-mem.h"
#include "oswald.h"
#include "strings.h"
#include "core-classes.h"

w_desc atype2desc[12];
w_desc primitive2desc[12];

w_desc desc_byte;
w_desc desc_char;
w_desc desc_double;
w_desc desc_float;
w_desc desc_int;
w_desc desc_long;
w_desc desc_short;
w_desc desc_boolean;
w_desc desc_void;
w_desc desc_String;
w_desc desc_Object;
w_desc desc_Class;
w_desc desc_ArrayOf_Object;
w_desc desc_ArrayOf_Class;
w_desc desc_ArrayOf_String;

/*
** An array that maps primitive descriptors to a character. It corresponds with
** the P_Type enum declaration in wonka.h
*/

const char primitive2char[] = {  
  '*', 
  '*', 
  '*', 
  '*', 
  'Z', 
  'C', 
  'F', 
  'D', 
  'B', 
  'S', 
  'I', 
  'J', 
  'V'
};

/*
** An array that maps primitive descriptors to a C string. It corresponds with
** the P_Type enum declaration in wonka.h
*/

const char *primitive2name[] = {  
  "<invalid primitive>", 
  "<invalid primitive>", 
  "<invalid primitive>", 
  "<invalid primitive>", 
  "boolean", 
  "char", 
  "float", 
  "double", 
  "byte", 
  "short", 
  "int", 
  "long", 
  "void"
};

/*
** An array that maps descriptors to their size in bits. It corresponds with
** the P_Type enum declaration in wonka.h
*/

const unsigned char desc2bits[] = {
   0,
   0,
   0,
   0,
   1,  /*  4: P_Type: boolean */
  16,  /*  5: P_Type: char    */
  32,  /*  6: P_Type: float   */
  64,  /*  7: P_Type: double  */
   8,  /*  8: P_Type: byte    */
  16,  /*  9: P_Type: short   */
  32,  /* 10: P_Type: int     */
  64,  /* 11: P_Type: long    */
   0,  /* 12: P_Type: void    */
  32,  /* 13: P_Type: Object  */
  32,  /* 14: D_Type: Class   */
  32,  /* 15: D_Type: Array   */
   0
};

/*
** An array that maps descriptors to their size in words (32 bits in a word). It corresponds with
** the P_Type enum declaration in wonka.h
*/

const unsigned char desc2words[] = {
   0,
   0,
   0,
   0,
   1,  /*  4: P_Type: boolean */
   1,  /*  5: P_Type: char    */
   1,  /*  6: P_Type: float   */
   2,  /*  7: P_Type: double  */
   1,  /*  8: P_Type: byte    */
   1,  /*  9: P_Type: short   */
   1,  /* 10: P_Type: int     */
   2,  /* 11: P_Type: long    */
   0,  /* 12: P_Type: void    */
   1,  /* 13: P_Type: Object  */
   1,  /* 14: D_Type: Class   */
   1,  /* 15: D_Type: Array   */
   0
};

w_instance primitive2classInstance[13];
w_clazz primitive2clazz[13];
w_clazz primitive2wrapper[13];
w_size primitive2wrapperSlot[13];

/*
** To keep track of his descriptors, Willie Wonka puts them in a hashtable.
** The "key" of a hashtable entry is the address of the UTF8 bytestring of
** the registered w_string holding the descriptor in its textual form:
** the "value" is the address of the w_desc structure used internally by
** Wonka. 
*/

#define DESCRIPTOR_HASHTABLE_SIZE   397

static w_hashtable descriptor_hashtable;
static x_sem descriptor_semaphore;

static w_desc parseAndCreateField(w_thread thread, w_char *charbuf, w_int length);

/*
** Calculate the length of the FieldDescriptor held in the
** 'unicode' character array. The length may not exceed 
** 'length' characters.  If length <= 0 or if no valid 
** descriptor is found, 0 is returned.
**
** There are three different cases:
**
** 1) The unicode string begins with one or more '['
**    characters, then count and strip the '[' characters,
**    then process the rest of the descriptor.
**
** OR
**
** 2) The unicode string begins with 'L', then everything up 
**    to the next ';' is part of the descriptor.
**
** OR
**
** 3) Assume it's a primitive and return 1 as the length.
*/

w_int getFieldDescriptorLength(w_char * unicode, w_int length) {

  w_char * current = unicode;
  w_char * end  = unicode + length;

  if (length <= 0) {
    return 0;
  }

  while (length > 0 && *current == '[') {
    current++;
    length--;
  }

  if(*current++ == 'L') {
    while(*current++ != ';') {
      if(current >= end) {
        return 0;
      }
    }
  }

  return (current - unicode);

}

/*
** Parse a field, this function is only called for non primitive types since
** the primitive types allready live prepopulated in the hashtable, so this
** function is never called for them. When we get a primitive type, we return
** an error.
*/

static w_boolean parseField(w_thread thread, w_char *charbuf, w_int length, w_desc desc) {

  w_desc component;
  
  switch (*charbuf) {

    /*
    ** Starts with '[' -> create array descriptor and recurse
    */

    case '[': {
      desc->tag = D_Array;
      component = parseAndCreateField(thread, charbuf + 1, length - 1);
      if (! component) {
        break;
      }
      desc->body.array.component = component;
      return WONKA_TRUE;
    }

    /*
    ** Starts with 'L' -> instance of a class.  (We'll worry about
    ** which one later, when we use this field/parameter).
    */

    case 'L': {
      desc->tag = D_Class;
      return WONKA_TRUE;
    }

    /*
    ** else primitive
    */

    default: {
      woempa(1, "Field descriptor was not well-formed\n");
    }

  }

  return WONKA_FALSE;
  
}

/*
** Parse the text of a Field Descriptor and create a w_desc from it.
** The descriptor may or may not be identical to the one we already parsed.
*/

static w_desc parseAndCreateField(w_thread thread, w_char *charbuf, w_int length) {

  w_desc desc = NULL;
  w_size descLength;
  w_string name;

  /*
  ** Check that 'charbuf' contains a well-formed descriptor, neither
  ** more nor less
  */

  descLength = getFieldDescriptorLength(charbuf, length);

  if (descLength != (w_size)length) {
    return NULL;
  }

  /*
  ** Convert 'charbuf' to a registered w_string.  Do we already have
  ** a descriptor by this name?
  ** N.B. primitive descriptors are pre-loaded, so they will always
  ** be existing.
  */

  name = unicode2String(thread, charbuf, descLength);
  desc = (w_desc)ht_read(descriptor_hashtable, (w_word)name);
  
  if (! desc) {
    desc = allocMem(thread, sizeof(w_Desc));
    desc->name = name;

    if (! parseField(thread, charbuf, length, desc)) {
      woempa(9, "Field descriptor %w is not well-formed!\n", name);
#ifdef DEBUG      
      dumpDescriptors();
#endif
      releaseMem(desc);
      deregisterString(name);
      return NULL;
    }
  }

  return registerDescriptor(thread, desc);
  
}

/*
** Analyse a w_string holding a FieldDescriptor, to either create a new
** w_desc or identify an existing one
*/

w_desc _newFieldDescriptor(const char *f, int l, w_thread thread, w_string string) {

  w_desc desc;

  desc = (w_desc)ht_read(descriptor_hashtable, (w_word)string);

  if (! desc) {
    desc = parseAndCreateField(thread, string->chars, (w_int)string->length);
  }
  else {
    desc = _registerDescriptor(f, l, thread, desc);
  }

  return desc;
      
}

/*
** Free the memory associated with a descriptor, and maybe of its ancestors.
** In the worst case, this function could recurse 255 times.
*/

static void releaseDescriptor(w_desc desc) {

  w_int i;

  if (desc->registered > 0) {
    wabort(ABORT_WONKA, "Freeing descriptor %w R %d !!\n", desc->name, desc->registered);
  }
  
  woempa(9, "Freeing descriptor %p %w.\n", desc, NM(desc));

  switch (desc->tag) {
    case D_Primitive: 
    case D_Class: 
      break;
    
    case D_Method: {
      deregisterDescriptor(desc->body.method.retDesc);
      for (i = 0; i < desc->body.method.numArgs; i++) {
        deregisterDescriptor(desc->body.method.args[i]);
      }
      break;
    }
    
    case D_Array: {
      deregisterDescriptor(desc->body.array.component);
      break;
    }
    
  }

  deregisterString(desc->name);
  releaseMem(desc);

}

w_desc _registerDescriptor(const char *f, int l, w_thread thread, w_desc desc) {

  w_desc result;
  w_desc garbage = NULL;

  if (! desc) {
    wabort(ABORT_WONKA, "Descriptor is NULL!\n");
  }

  /*
  ** Note: the descriptor->name is registered allready by parseAndCreateField, so we don't have to register the
  ** name again here, just the descriptor itself.
  */

  woempa(1, "(%s:%d) Registering %w %3d ('%w') in hashtable.\n", f, l, desc->name, desc->registered, desc->name);
  x_sem_get(descriptor_semaphore, x_eternal);
  result = (w_desc)ht_read(descriptor_hashtable, (w_word)desc->name);
  if (result) {
    result->registered += 1;
    woempa(1, "(%s:%d) Found %w %3d ('%w') in hashtable.\n", f, l, result->name, result->registered, result->name);
    if (result != desc) {
      garbage = desc;
    }
  }
  else {
    desc->registered = 1;
    ht_write(descriptor_hashtable, (w_word)desc->name, (w_word)desc);
    woempa(1, "(%s:%d) Wrote new %w %3d ('%w') in hashtable.\n", f, l, desc->name, desc->registered, desc->name);
    result = desc;
  }
  x_sem_put(descriptor_semaphore);

  if (garbage) {
    releaseDescriptor(garbage);
  }

  return result;

}

void _deregisterDescriptor(const char *f, int l, w_desc desc) {

  w_desc found;
  w_desc garbage = NULL;
  
  x_sem_get(descriptor_semaphore, x_eternal);
  woempa(1, "(%s:%d) Deregistering %w %3d ('%w') in hashtable.\n", f, l, desc->name, desc->registered, desc->name);
  found = (w_desc)ht_read(descriptor_hashtable, (w_word)desc->name);
  if (! found) {
    wabort(ABORT_WONKA, "(%s:%d) Did NOT find descriptor %V in hashtable!\n", f, l, desc);
  }
  else {
    found->registered -= 1;
    woempa(1, "(%s:%d) Deregistered %w %3d ('%w') in hashtable.\n", f, l, found->name, found->registered, found->name);
    if (found->registered == 0) {
      woempa(9, "(%s:%d) descriptor %w R %d has become garbage...\n", f, l, found->name, found->registered);
      garbage = found;
    }
  }
  x_sem_put(descriptor_semaphore);
  
  if (garbage) {
    ht_erase(descriptor_hashtable, (w_word)garbage->name);
    releaseDescriptor(garbage);
  }

}

void registerStaticDescriptor(w_desc desc) {

  woempa(7, "register static descriptor %V.\n", desc);
  ht_write(descriptor_hashtable, (w_word)desc->name, (w_word)desc);     

}

void startDescriptors() {

  descriptor_semaphore = allocMem(W_Thread_system, sizeof(x_Sem));

  desc_byte = allocMem(W_Thread_system, sizeof(w_Desc));
  desc_byte->registered = 1234567; 
  desc_byte->tag = D_Primitive; 
  desc_byte->name = cstring2String(W_Thread_system,"B",1);
  desc_byte->body.primitive.type = P_byte; 
 
  desc_char = allocMem(W_Thread_system, sizeof(w_Desc));
  desc_char->registered = 1234567;
  desc_char->tag = D_Primitive; 
  desc_char->name = cstring2String(W_Thread_system,"C",1);
  desc_char->body.primitive.type = P_char; 
 
  desc_double = allocMem(W_Thread_system, sizeof(w_Desc));
  desc_double->registered = 1234567; 
  desc_double->tag = D_Primitive;
  desc_double->name = cstring2String(W_Thread_system,"D",1);
  desc_double->body.primitive.type = P_double; 
 
  desc_float = allocMem(W_Thread_system, sizeof(w_Desc));
  desc_float->registered = 1234567; 
  desc_float->tag = D_Primitive; 
  desc_float->name = cstring2String(W_Thread_system,"F",1);
  desc_float->body.primitive.type = P_float; 
 
  desc_int = allocMem(W_Thread_system, sizeof(w_Desc));
  desc_int->registered = 1234567; 
  desc_int->tag = D_Primitive;
  desc_int->name = cstring2String(W_Thread_system,"I",1);
  desc_int->body.primitive.type = P_int; 
 
  desc_long = allocMem(W_Thread_system, sizeof(w_Desc));
  desc_long->registered = 1234567;
  desc_long->tag = D_Primitive;
  desc_long->name = cstring2String(W_Thread_system,"J",1);
  desc_long->body.primitive.type = P_long; 
 
  desc_short = allocMem(W_Thread_system, sizeof(w_Desc));
  desc_short->registered = 1234567;
  desc_short->tag = D_Primitive;
  desc_short->name = cstring2String(W_Thread_system,"S",1);
  desc_short->body.primitive.type = P_short; 
 
  desc_boolean = allocMem(W_Thread_system, sizeof(w_Desc));
  desc_boolean->registered = 1234567;
  desc_boolean->tag = D_Primitive; 
  desc_boolean->name = cstring2String(W_Thread_system,"Z",1);
  desc_boolean->body.primitive.type = P_boolean; 
 
  desc_void = allocMem(W_Thread_system, sizeof(w_Desc));
  desc_void->registered = 1234567;
  desc_void->tag = D_Primitive;
  desc_void->name = cstring2String(W_Thread_system,"V",1);
  desc_void->body.primitive.type = P_void;
 
  /*
  ** Make a well know descriptor for Ljava/lang/String;
  */

  desc_String = allocMem(W_Thread_system, sizeof(w_Desc));
  desc_String->tag = D_Class;
  desc_String->name = cstring2String(W_Thread_system,"Ljava/lang/String;",18);
  desc_String->registered = 1234567;

  primitive2desc[0] = NULL;
  primitive2desc[1] = NULL;
  primitive2desc[2] = NULL;
  primitive2desc[3] = NULL;
  primitive2desc[P_boolean] = desc_boolean;
  primitive2desc[P_char]    = desc_char;
  primitive2desc[P_float]   = desc_float;
  primitive2desc[P_double]  = desc_double;
  primitive2desc[P_byte]    = desc_byte;
  primitive2desc[P_short]   = desc_short;
  primitive2desc[P_int]     = desc_int;
  primitive2desc[P_long]    = desc_long;

  /*
  ** Allocate the descriptor_hashtable, and seed it with some entries.
  */

  descriptor_hashtable = ht_create(W_Realm_system, (char*)"hashtable:descriptors", DESCRIPTOR_HASHTABLE_SIZE, NULL , NULL, 0, 0);
  woempa(1, "created descriptor_hashtable at %p\n", descriptor_hashtable);
  x_sem_create(descriptor_semaphore, 1);

  registerDescriptor(W_Thread_system, desc_byte);
  registerDescriptor(W_Thread_system, desc_char);
  registerDescriptor(W_Thread_system, desc_double);
  registerDescriptor(W_Thread_system, desc_float);
  registerDescriptor(W_Thread_system, desc_int);
  registerDescriptor(W_Thread_system, desc_long);
  registerDescriptor(W_Thread_system, desc_short);
  registerDescriptor(W_Thread_system, desc_boolean);
  registerDescriptor(W_Thread_system, desc_void);
  registerDescriptor(W_Thread_system, desc_String);
  
}

#ifdef DEBUG

const char *desc2char(w_desc desc) {
  switch (desc->tag) {
    case 0:
      return "Primitive";
    case 14:
      return "Class    ";
    case 15:
      return "Array    ";
    default:
      return "(desc2char) Huh...";
  }
}

static void dumpOneDescriptor(w_word string_word, w_word desc_word) {

  w_desc desc = (w_desc)desc_word;
  w_string string = (w_string)string_word;

  woempa(9, "  %s R %3d %w ('%w')\n", desc2char(desc), desc->registered, desc->name, string);

}

void dumpDescriptors() {

  woempa(9,"**************** Descriptor dump ********************************************************\n");
  ht_every(descriptor_hashtable,dumpOneDescriptor);
  woempa(9,"*****************************************************************************************\n");

}

#endif

/*
** See if a clazz is a primitive clazz, e.g. 'int.class' and return
** the appropriate index if that is the case, i.e. one that can be used
** to index the primitive2... arrays. Returns 0 if the clazz is not a primitive 
** clazz.
*/

w_int isPrimitiveClazz(w_clazz clazz) {

  w_int i;

  for (i = 4; i <= P_long; i++) {
    if (primitive2clazz[i] == clazz) {
      return i;
    }
  }
  
  return 0;

}

/*
** Create a descriptor for an array of 'clazz' with 'dim' dimensions.
*/

w_desc createArrayDescriptor(w_thread thread, w_clazz clazz, w_int dim) {

  w_char *chars;
  w_char *slider;
  w_string string;
  w_desc desc;
  w_int i;
  
  chars = allocMem(thread, (clazz->desc->name->length + dim) * sizeof(w_char));
  slider = chars;
  for (i = 0; i < dim; i++) {
    *slider++ = '[';
  }
  memcpy(slider, clazz->desc->name->chars, clazz->desc->name->length * sizeof(w_char));
  string = unicode2String(thread, chars, clazz->desc->name->length + dim);
  releaseMem(chars);
  woempa(1, "New descriptor for '%w' with %d new dimensions is '%w'.\n", clazz->desc->name, dim, string);
  desc = newFieldDescriptor(thread, string);
// TODO: error if desc==NULL ...
 
  /*
  ** The newFieldDescriptor bumped up the registration count of string, so we have to deregister it
  ** locally here...
  */
  
  deregisterString(string); 
  
  return desc;
  
}

w_clazz parseDescriptor(w_string descriptor, w_size *start, w_size end, w_instance loader) {
  w_clazz result = NULL;
  w_string desc_string;
  w_string temp_string;
  w_string dotified;
  w_size   dims = 0;
  w_size   k;
    /*
    ** Fallback values, used only if parsing fails.
    */
    result = NULL;

    dims = 0;
    while (descriptor->chars[*start] == '[') {
      ++dims;
      ++*start;
      if (*start >= end) {

        return NULL;

      }
    }
    if (dims) {
      switch (descriptor->chars[(*start)++]) {
      case 'Z':
      case 'C':
      case 'F':
      case 'D':
      case 'B':
      case 'S':
      case 'I':
      case 'J':
        temp_string = unicode2String(currentWonkaThread, &descriptor->chars[*start - dims - 1], dims + 1);
        result = identifyClazz(temp_string, loader);
        break;
  
      case 'L':
        k = *start;
        while (k < end) {
          if (descriptor->chars[k] == ';') {
            break;
          }
          ++k;
        }
        if (k == end) {

          return NULL;

        }
        desc_string = unicode2String(currentWonkaThread, &descriptor->chars[*start - dims - 1], k - *start + dims + 2);
        *start = k + 1;
        temp_string = undescriptifyClassName(desc_string);
        result = identifyClazz(temp_string, loader);
        deregisterString(desc_string);
// TODO: find out why this next line causes a crash
//        deregisterString(temp_string);
        break;
  
      default:
        temp_string = unicode2String(currentWonkaThread, &descriptor->chars[*start - dims - 1], dims + 1);

        return NULL;

      }
      deregisterString(temp_string);
    }
    else {
      switch (descriptor->chars[(*start)++]) {
      case 'Z':
        result = clazz_boolean;
        break;
  
      case 'C':
        result = clazz_char;
        break;
  
      case 'F':
        result = clazz_float;
        break;

      case 'D':
        result = clazz_double;
        break;

      case 'B':
        result = clazz_byte;
        break;

      case 'S':
        result = clazz_short;
        break;
  
      case 'I':
        result = clazz_int;
        break;
  
      case 'J':
        result = clazz_long;
        break;

      case 'V':
        result = clazz_void;
        break;

      case 'L':
        k = *start;
        while (k < end) {
          if (descriptor->chars[k] == ';') {
            break;
           }
          ++k;
        }
        if (k == end) {

          return NULL;

        }

        temp_string = unicode2String(currentWonkaThread, &descriptor->chars[*start], k - *start);
        dotified = slashes2dots(currentWonkaThread, temp_string);
        result = identifyClazz(dotified, loader);
        deregisterString(dotified);
        deregisterString(temp_string);
        *start = k + 1;
        break;

      default:
        woempa(9, "Illegal char %c\n", descriptor->chars[*start-1]);

        return NULL;

      }
    }

  woempa(1, "Returning %k (loaded by %k@%p)\n", result, instance2clazz(result->loader), result->loader);

  return result;
}

