/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

#include <string.h>

#include "loopback.h"

static void loopback_initDevice(w_device device) {
  w_loopback loop;

  // do not initialise the mother device
  if(device->familyMember != 0) {
    device->control = allocMem_system(sizeof(w_Loopback));
    loop = (w_loopback)device->control;
  
    loop->mutx = allocMem_system(sizeof(x_mutex));
    x_mutex_create(loop->mutx);

    loop->qmem = allocMem_system(4 * 100);
    loop->q = allocMem_system(sizeof(x_Queue));
    x_queue_create(loop->q, loop->qmem, 100);
  }
}

static w_void loop_freeQueue(void *ee) {
  w_loopback_queueelem e = (w_loopback_queueelem)ee;
  
  releaseMem(e->data);
  releaseMem(e);
}

static void loopback_termDevice(w_device device) {
  w_loopback loop = (w_loopback)device->control;

  // do not destuct the mother device
  if(device->familyMember != 0) {
    if (loop->par != NULL) {
      releaseMem(loop->par->data);
      releaseMem(loop->par);
    }
    x_queue_flush(loop->q, loop_freeQueue);

    x_queue_delete(loop->q);
    releaseMem(loop->qmem);
    releaseMem(loop->q);

    x_mutex_delete(loop->mutx);
    releaseMem(loop->mutx);

    releaseMem(loop);
  }
}

static w_ubyte _devnr;

static w_device loopback_open(w_device device, w_driver_perm mode) {
  char name[] = "loop__";
  w_device dev;
  
  // only mother thread will generate new devices on opening
  if(device->familyMember == 0){
    // maximum 32 unzip devices present
    _devnr = _devnr & 31;
  
    name[4] = (_devnr / 10) + '0';
    name[5] = (_devnr % 10) + '0';
  
    // it is possible that this is NULL because device is already present, fix it !!!!
    dev = registerDevice(name, "loop", 10 + _devnr, wdt_byte_serial);

    _devnr += 1;

    device->driver->usage += 1;
    dev->usage = 1;

    return dev;
  }
  else {
    return device;
  }
}

static w_void loopback_close(w_device device) {
  device->usage -= 1;

  // we cleanup device completely, so do this here
  device->driver->usage -= 1;

  deregisterDevice(device->name);
}

static w_driver_status loopback_write(w_device device, w_ubyte *bytes, w_int length, w_int *lwritn, x_sleep timeout) {
  w_loopback l = (w_loopback)device->control;
  x_status s;
  w_loopback_queueelem qe = allocMem_system(sizeof(w_Loopback_QueueElem));

  if (!qe) {
    return wds_internal_error;
  }

  qe->data = allocMem_system((unsigned)length);
  if (!qe->data) {
    releaseMem(qe);
    return wds_internal_error;
  }

  memcpy(qe->data, bytes, (unsigned)length);
  qe->size = length;

  s = x_queue_send(l->q, qe, timeout);
  switch (s) {
    case xs_success:
      *lwritn = length;
      return wds_success;
    case xs_no_instance:
      *lwritn = 0;
      releaseMem(qe);
      return wds_space_exhausted;    
    default:
      *lwritn = 0;
      releaseMem(qe);
      return wds_internal_error;    
  }
}

static w_driver_status loopback_read(w_device device, w_ubyte *bytes, w_int length, w_int *lread, x_sleep timeout) {
  w_loopback l = (w_loopback)device->control;
  w_loopback_queueelem qe;
  x_status s;
  w_int offset;

  // Get lock
  switch (x_mutex_lock(l->mutx, timeout)) {
    case xs_success:
      break;
    case xs_no_instance:
      return wds_no_instance;
    default:
      return wds_internal_error;
  }

  // TODO : decrease timeout if we waited in the x_mutex_lock

  // Here we try to return as much as possible of the requested size, if necesarry reading
  // multiple blocks from the queue
  *lread = 0;
  offset = 0;
  while (1) {
    if (l->par != NULL) {
      // we already have what we need, just return
      if (offset == length){
        x_mutex_unlock(l->mutx);
        return wds_success;
        exit(0);
      }
      
      // we have partial data, copy it
      else if ((l->par->size - l->offset) <= (length - offset)) {
        // more data requested than we have, give everything
        memcpy(bytes + offset, l->par->data + l->offset, (unsigned)(l->par->size - l->offset));

        *lread += l->par->size - l->offset;
        offset += l->par->size - l->offset;

        if (l->par->data != NULL) releaseMem(l->par->data);
        releaseMem(l->par);
        l->par = NULL;
        l->offset = 0;
      }
      else {
        // give what they ask 
        memcpy(bytes + offset, l->par->data + l->offset, (unsigned)(length - offset));
        l->offset += length - offset;
        *lread += length - offset;
        offset += length - offset;

        // buffer is now full, so move out
        x_mutex_unlock(l->mutx);
        return wds_success;
      }
    }
    else {
      // no data, get some
      s = x_queue_receive(l->q, (w_void **)&qe, timeout);

      // TODO : decrease timeout if we had to wait

      switch (s) {
        case xs_success:
          if (qe == NULL) {
            *lread = 0;

            woempa(7, "AIAIAI, not good, allocation failure somewhere.\n");

            x_mutex_unlock(l->mutx);
            return wds_internal_error;
          } 
          else {
            l->par = qe;
            l->offset = 0;
          }
          break;
        case xs_no_instance:
          // no data received during given timeout
          // if we had data, return it, else return lread 0 and wds_no_instance

          if (offset > 0) {
            x_mutex_unlock(l->mutx);
            return wds_success;
          }
          else {
            x_mutex_unlock(l->mutx);
            return wds_no_instance;    
          }
        default:
          // AIAI, an error occured, throw away all data and return error
          *lread = 0;

          // Release lock (no error checking necessary)
          x_mutex_unlock(l->mutx);
          return wds_internal_error;    
      }  
    }
  }
  // Release lock (no error checking necessary) -- Redundant code, only for safety
  x_mutex_unlock(l->mutx);
  return wds_internal_error;    
}
  
static w_driver_status loopback_seek(w_device device, int posn, x_sleep timeout) {
  return wds_not_implemented;
}

static w_driver_status loopback_set(w_device device, w_word command, w_word param, x_sleep timeout) {
  return wds_not_implemented;
}
 
static w_driver_status loopback_query(w_device device, w_word query, w_word *reply, x_sleep timeout) {
  return wds_not_implemented;
}

w_Driver_ByteSerial loopback_driver = {
  "loopback-driver-v0.1",
  "loop",
  0,
  NULL,
  loopback_initDevice,
  loopback_termDevice,
  loopback_open,
  loopback_close,
  loopback_read,
  loopback_write,
  loopback_seek,
  loopback_set,
  loopback_query,
};
