/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: comms.c,v 1.1 2002/03/21 09:11:22 dfrederickx Exp $
*/

#include <stdio.h>
#include <unistd.h>
#include "comms.h"
#include "sio.h"
#include "wt.h"
#include "ts-mem.h"
#include "realms.h"
#include "oswald.h"
#include "wonka.h"
#include "driver_byteserial.h"

/*
** Currently we have 4 hard-coded comm ports, 3 hard-coded
** file descriptors, and one wonkaterm.
** Should be made smarter.
** For each comm port we have a w_Sio structure, a w_Device,
** and two termios buffers; one to store the state when
** Wonka was started up (so we can restore it on exit),
** the other to store (and manipulate) the current state.
** For each file descriptor we have a w_Wfd structure and
** a w_Device, for each wonkaterm we have a w_Wt structure 
** and a w_Device.
*/

static struct termios sio0_oldtermios;
static struct termios sio0_newtermios;
static struct termios sio1_oldtermios;
static struct termios sio1_newtermios;
static struct termios sio2_oldtermios;
static struct termios sio2_newtermios;
static struct termios sio3_oldtermios;
static struct termios sio3_newtermios;

/*
** The pathnames for the four serial ports
*/
char * pathname[4];

typedef struct w_Control_Sio {
  struct pollfd pollfd;
/*
** clock speed (bits/sec)
*/
  w_int      bitrate;
/*
** Receive buffer: size, start, point where next char will be read/written,
** number of bytes currently used.
*/
  w_int      rxBufSize;
  w_ubyte    *rxBuffer;
  w_ubyte    *rxReadPtr;
  w_ubyte    *rxWritePtr;
  w_int      rxBytesInBuf;
/*
** Value of rxBytesInBuf at which DATA_AVAILABLE will be signalled
*/
  w_int      rxThreshold;
/*
** Counting semaphores used to synchronize producers and consumers of
** received bytes, sent bytes, and i/o events.
*/
  x_Sem     rxSem;    // count = bytes in buffer
  x_Sem     txSem;    // count = bytes buffered in producer
  x_Sem     evtSem;   // count = outstanding events
/*
** Which flags (DTR, RTS, DSR, CTS, CD, RI) are supported by this UART
*/
  w_short   validflags;
/*
** Values of the IOFLAGS for this sio
**   oldflags = already reported via ioevt
**   newflags = latest available values
**   thisflag: only one bit is set, indicating which bit of newflags
**             caused ioevt to return control to its caller last time.
**   thisvalue: either zero or equal to thisflag, depending on whether
**              the value of the flag identified by thisflag was 0 or 1.
*/
  w_short   oldflags;
  w_short   newflags;
  w_short   thisflag;
  w_short   thisvalue;
/*
** Other flags used internally for congestion control etc.
*/
  w_word     miscflags;
/*
** Field to store the result of a 'wait_for_event'
*/
  w_word     event_data;
/*
** Pointers to the termios structure as it was before we grabbed the sio,
** and as it is now
*/
  w_termios  oldtermios;
  w_termios  newtermios;
/*
** The name of the actual device e.g. '/dev/ttyS0'
*/
  char       *devicename;
  
} w_Control_Sio;

typedef w_Control_Sio *w_control_sio;

/*
** Somewhat ad-hoc function to set the path of an sio device
*/
void sio_set_path(int n, char *s) {
  if (n >= 0 && n < 4) {
    char *buf = allocMem_system(strlen(s+1));
    strcpy(s, buf);
    pathname[n] = buf;
  }
  else {
    woempa(9, "sio devnum %d out of range\n", n);
  }
}

/*
** Prototypes 
*/

w_void sio_initDevice(w_device device); 
w_void sio_termDevice(w_device device); 
w_device sio_open(w_device device, w_driver_perm mode);
w_void sio_close(w_device device);
w_driver_status sio_read(w_device device, w_ubyte *bytes, w_int length, w_int *lread, x_sleep timeout);
w_driver_status sio_write(w_device device, w_ubyte *bytes, w_int length, w_int *lwritn, x_sleep timeout);
w_driver_status sio_seek(w_device device, w_int offset, x_sleep timeout);
w_driver_status sio_set(w_device device, w_driver_ioctl command, w_word param, x_sleep timeout);
w_driver_status sio_query(w_device device, w_driver_ioctl query, w_word *reply, x_sleep timeout);

/*
** sio_init sets the port state to default values, and saves
** the current state so that it can be restored later.
*/

w_void sio_initDevice(w_device device) {
  w_control_sio c;
  device->control = allocClearedMem_system(sizeof(w_Control_Sio));
  c = device->control;

  c->validflags = IOFLAG_DTR|IOFLAG_RTS|IOFLAG_DSR|IOFLAG_CTS|IOFLAG_CD|IOFLAG_RI;
  switch(device->familyMember) {
    case 0: c->bitrate = SIO0_DEFAULT_BITRATE;
            c->devicename = pathname[0];
            c->oldtermios = &sio0_oldtermios;
            c->newtermios = &sio0_newtermios;
            break;
    case 1: c->bitrate = SIO1_DEFAULT_BITRATE;
            c->devicename = pathname[1];
            c->oldtermios = &sio1_oldtermios;
            c->newtermios = &sio1_newtermios;
            break;
    case 2: c->bitrate = SIO2_DEFAULT_BITRATE;
            c->devicename = pathname[2];
            c->oldtermios = &sio2_oldtermios;
            c->newtermios = &sio2_newtermios;
            break;
    case 3: c->bitrate = SIO3_DEFAULT_BITRATE;
            c->devicename = pathname[3];
            c->oldtermios = &sio3_oldtermios;
            c->newtermios = &sio3_newtermios;
            break;
    default:
  }
            
  // atexit(sio_term);
}

w_void sio_termDevice(w_device device) {
  w_control_sio sio = device->control;
  
  if (sio->pollfd.fd>=0) {
    tcsetattr(sio->pollfd.fd, TCSAFLUSH, sio->oldtermios);
  }
  sio_close(device);

  releaseMem(device->control);
}


w_device sio_open(w_device device, w_driver_perm mode) {
  w_control_sio sio = device->control;
  
  if(device->usage == 0) {
    
    sio->miscflags = mode = wdp_read|wdp_write;
  
//  if (isNotSet(sio->miscflags, wdp_read|wdp_write)) {
// Somehow this if always returns false... Not so good...
  
      sio->pollfd.fd = open(sio->devicename,O_RDWR);
      sio->pollfd.events = POLLIN|POLLOUT;
      
      /*
      ** If a tty, save current (pre-wonka) settings in oldtermios.  (If not a tty,
      ** this call returns nonzero and we skip the rest of the tty-related stuff).
      */
      if (tcgetattr(sio->pollfd.fd,sio->oldtermios)==0) {
        /*
        ** Copy oldtermios to newtermios and make some changes:
        ** echo off, canonical mode off, extended i/p proc off, signal chars off
        ** no SIGINT on BREAK, CR>NL off, i/p parity off, strip8 off, CTS/RTS off
        ** size 8 bits, no parity check, no output processing
        ** read is blocking with no timeout
        */
        memcpy(sio->newtermios,sio->oldtermios,sizeof(struct termios));
        sio->newtermios->c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
        sio->newtermios->c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
        sio->newtermios->c_cflag &= ~(CSIZE | PARENB);
        sio->newtermios->c_cflag |= CREAD | CS8 | CLOCAL;
        sio->newtermios->c_oflag &= ~(OPOST);
        sio->newtermios->c_cc[VMIN] = 1;
        sio->newtermios->c_cc[VTIME] = 0;
        cfsetispeed(sio->newtermios,(w_size)sio->bitrate);
        cfsetospeed(sio->newtermios,(w_size)sio->bitrate);
        tcsetattr(sio->pollfd.fd,TCSAFLUSH,sio->newtermios);
      }
  
      sio->oldflags = 0;
      sio->newflags = 0;
      x_sem_create(&sio->rxSem,0);
      x_sem_create(&sio->txSem,1);
      x_sem_create(&sio->evtSem,0);
//  }

//  if (mode & wdp_read) {
 
      if (sio->rxBuffer) releaseMem(sio->rxBuffer);
      if (sio->rxBufSize==0) {
        sio->rxBufSize = sio->bitrate/RX_BUF_SIZE_DIVISOR;
      }
      sio->rxBuffer = allocMem(W_Thread_system, (w_size)sio->rxBufSize);
      sio->rxReadPtr = sio->rxBuffer;
      sio->rxWritePtr = sio->rxBuffer;
      sio->rxBytesInBuf = 0;
//  }
    
  }

  device->usage += 1;
  device->driver->usage += 1;
  
  return device;
}


w_void sio_close(w_device device) {
  w_control_sio sio = device->control;

  if(device->usage == 1) {
  
    if (sio->miscflags & (wdp_read | wdp_write)) {
      sio->miscflags &= ~(wdp_read | wdp_write);
      if (sio->pollfd.fd>=0) close(sio->pollfd.fd);

      sio->rxBufSize = 0;
      if (sio->rxBuffer) {
        releaseMem(sio->rxBuffer);
        sio->rxBuffer = NULL;
      }
    }
  
    x_sem_delete(&sio->rxSem);
    x_sem_delete(&sio->txSem);
    x_sem_delete(&sio->evtSem);
    
  }

  device->usage -= 1;
  device->driver->usage -= 1;
}


w_driver_status sio_read(w_device device, w_ubyte *bytes, w_int length, w_int *lread, x_sleep timeout) {
  w_control_sio sio = device->control;

  *lread = read(sio->pollfd.fd, bytes, (w_word)length);

  return wds_success;
  
}


w_driver_status sio_write(w_device device, w_ubyte *bytes, w_int length, w_int *lwritn, x_sleep timeout) {
  w_control_sio sio = device->control;

  *lwritn = write(sio->pollfd.fd, bytes, (w_word)length);

  return wds_success;
}


w_driver_status sio_seek(w_device device, w_int offset, x_sleep timeout) {
  return wds_not_implemented;
}


w_driver_status sio_set(w_device device, w_driver_ioctl command, w_word param, x_sleep timeout) {
  return wds_no_such_command;
}


w_driver_status sio_query(w_device device, w_driver_ioctl query, w_word *reply, x_sleep timeout) {
  return wds_no_such_command;
}


w_Driver_ByteSerial sio_driver = {
  "serial-I/O-driver-linux_v0.1",
  "sio",
  0,
  NULL,
  sio_initDevice,
  sio_termDevice,  
  sio_open,
  sio_close,  
  sio_read,
  sio_write,
  sio_seek,
  sio_set,
  sio_query
};


/********************************************************************/
/**            T h e   F i l e D e s c   d r i v e r               **/
/** This is a thin wrapper for stdin, stdout, or stderr.           **/
/********************************************************************/
 
typedef struct w_Control_Wfd {
  int fd;
} w_Control_Wfd;

typedef w_Control_Wfd *w_control_wfd;


w_void wfd_initDevice(w_device device) {
  w_control_wfd wfd;
  device->control = allocMem_system(sizeof(w_Control_Wfd));
  wfd = device->control;
  wfd->fd = device->familyMember;
}


w_void wfd_termDevice(w_device device) {
}


w_device wfd_open(w_device device, w_driver_perm mode) {
  // nowt to do: fd's 0, 1 & 2 are pre-opened by the host OS
  device->usage += 1;
  device->driver->usage += 1;
  
  return device;
}


w_void wfd_close(w_device device) {
  // ignore 
  device->usage -= 1;
  device->driver->usage -= 1;
}


w_driver_status wfd_read(w_device device, w_ubyte *bytes, w_int length, w_int *lread, x_sleep timeout) {
  w_control_wfd wfd = device->control;

  *lread = read(wfd->fd, bytes, (w_word)length);

  if (*lread<0) {
    return wds_data_exhausted;
  }
  else {
    return wds_success;
  }
  
}


w_driver_status wfd_write(w_device device, w_ubyte *bytes, w_int length, w_int *lwritn, x_sleep timeout) {
  w_control_wfd wfd = device->control;

  *lwritn = write(wfd->fd, bytes, (w_word)length);

  if (*lwritn<0) {
    return wds_space_exhausted;
  }
  else {
    return wds_success;
  }
}


w_driver_status wfd_seek(w_device device, w_int offset, x_sleep timeout) {
  return wds_not_implemented;
}


w_driver_status wfd_set(w_device device, w_driver_ioctl command, w_word param, x_sleep timeout) {
  return wds_not_implemented;
}


w_driver_status wfd_query(w_device device, w_driver_ioctl query, w_word *reply, x_sleep timeout) {
  w_control_wfd wfd = device->control;
  
  switch (query) {
    case(wdi_get_available):
      if (ioctl(wfd->fd,FIONREAD,reply)!=0) {
        *reply = (w_word)-1;
      }
      return wds_success;
      
    case(wdi_get_bitrate):
    case(wdi_get_databits):
    case(wdi_get_stopbits):
    case(wdi_get_flowcontrol):
    case(wdi_get_parity):
    case(wdi_get_signals):
      return wds_success;

    default:
      return wds_no_such_command;
  }

  return wds_no_such_command;
}


w_Driver_ByteSerial wfd_driver = {
  "filedesc-driver-linux_v0.1",
  "wfd",
  0,
  NULL,
  wfd_initDevice,
  wfd_termDevice,  
  wfd_open,
  wfd_close,  
  wfd_read,
  wfd_write,
  wfd_seek,
  wfd_set,
  wfd_query
};


/********************************************************************/
/**          T h e   W o n k a T e r m   d r i v e r               **/
/** This simulates a port by listening to a socket.                **/
/********************************************************************/
/* 
static void wt_init(void);
static void wt_term(void);

static w_int wt_open(void *cb, int mode);
static void wt_close(void *cb);
static w_driver_status wt_read(void *cb,w_byte *bytes, w_size length, w_size *lwritn);
static w_driver_status wt_write(void *cb,w_byte *bytes, w_size length, w_size *lwritn);
static w_driver_status wt_seek (void *cb, int posn);
static w_driver_status wt_set(void* cb,w_word command, w_word param);
static w_driver_status wt_query(void* cb,w_word command, w_word *reply);
static w_driver_status wt_event(void* cb,w_word *event, w_word *param);

*/

/*
** This is the way to listen to a socket, according to Richie Stevens:
**    int sockfd, newsockfd, clilen;
**    struct sockaddr_in cli_addr, serv_addr;
**    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0) < 0) {
**      -- error: can't open stream socket --
**    }
**    bzero((char*)&serv_addr, sizeof(serv_addr);
**    serv_addr.sin_family = AF_INET;
**    serv_addr.sin_addr,s_addr = htonl(INADDR_ANY);
**    serv_addr.sin_port = htons(1234);
**    if (bind(sockfd, (struct sockaddr *) % serv_addr, sizeof(serv_addr)) < 0) {
**      -- error: can't bind to local socket --
**    }
**    listen(sockfd, 5);
**    for (;;) {
**      clilen = sizeof(cli_addr);
**      newsockfd = accept(sockfd, (struct sockaddr *)&cli_addr, &clilen);
**      if (newsockfd < 0) {
**        -- error in accept --
**      }
**      -- process request (newsockfd) --
**      close(newsockfd);
**    }
*/

/*
static w_StreamDriver wt_driver = {
  WDRV_BYTE_SERIAL,
  "driver:linux-socket",
  wt_init,
  wt_term,
  wt_open,
  wt_close,
  wt_read,
  wt_write,
  wt_seek,
  wt_set,
  wt_query,
  wt_event
};
*/

/*
** wt_init sets the port state to default values, and saves
** the current state so that it can be restored later.
*/

// void wt_init (void) {
// Following silliness is because I can't get htons to work with -Wconversion.
// The idea is that wt0 connects to localhost port 8192.  |8-[
//  wt0.myaddr.sin_port = ((8192 & 0x00ff) << 8) | ((8192 & 0xff00) >> 8);
//  wt0.myaddr.sin_addr.s_addr = htonl(0x7f000001);

//  if ( (wt0.fd = socket(PF_INET,SOCK_STREAM,6)) < 0) {
//    perror("init wt0: open socket");
//  }
//  else {
//    woempa(9,"Created DEBUG socket, connecting fd %d to port %d (%x)\n",wt0.fd,8192,wt0.myaddr.sin_port);
//    if ( connect(wt0.fd,(struct sockaddr*)&(wt0.myaddr),sizeof(struct sockaddr_in)) < 0) {
//      perror("init wt0: connect");
//    }
//  }


//  atexit(wt_term);

//}

//void wt_term (void) {
//  int rc;

//  if (wt0.fd >= 0) {
//    rc = close(wt0.fd);
//    wt0.fd = -1;
//  }

//}
  
/*
w_driver_status wt_open (void *cb,int mode) {
  w_wt wt = (w_wt)cb;

  wt->opened = 1;

  return WDRVERR_NO_ERROR;
}

void wt_close (void *cb) {
  w_wt wt = (w_wt)cb;

  wt->opened = 0;
}

w_int wt_read (void *cb, w_byte *dataloc, w_size length, w_size *lread) {
  w_wt wt = (w_wt)cb;
  int  read_result;

  read_result = read(wt->fd,dataloc,length);

  if (read_result > 0) {
    *lread = read_result;
    return WDRVERR_NO_ERROR;
  }
  else {
    return WDRVERR_DATA_EXHAUSTED; // TODO: Derive something from perror?
  }
}

w_driver_status wt_seek (void *cb, int posn) {
// not implemented
  return WDRVERR_NOT_IMPLEMENTED;
}

w_int wt_write(void* cb,w_byte *bytes, w_size length, w_size *lwritn) {
  w_wt wt;

  wt = (w_wt)cb;

  *lwritn = write(wt->fd,bytes,length);

  if (*lwritn<0) {
    return *lwritn; // TODO: Derive something from perror?
  }
  else {
    return WDRVERR_NO_ERROR;
  }
}

w_driver_status wt_set (void* cb,w_word command, w_word param) {
// not implemented
  return WDRVERR_NO_ERROR;
}

w_driver_status wt_query (void* cb,w_word query, w_word *reply) {
  w_wt wt;

  wt = (w_wt)cb;

  switch (query) {
  case(IOCON_GET_AVAILABLE):
    if (ioctl(wt->fd,FIONREAD,reply)!=0) {
      *reply = (w_word)-1;
    }

    return WDRVERR_NO_ERROR;

  default:
// not implemented
  }
  return WDRVERR_NO_ERROR;
}

w_driver_status wt_event (void* cb,w_word *eventloc, w_word *paramloc) {
// not implemented
  return WDRVERR_NO_ERROR;
}
*/

void startCommports() {
}

