/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/



#include <string.h>
#include <setjmp.h>
#ifdef LINUX
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#endif

#include "ficl.h"
// Wonka stuff
#include "clazz.h"
#include "comms.h"
#include "core-classes.h"
#include "descriptor.h"
#include "hashtable.h"
#include "heap.h"
#include "interpreter.h"
#include "loading.h"
#include "locks.h"
#include "ts-mem.h"
#include "methods.h"
#include "ps-mem.h"
#include "repository.h"
#include "oswald.h"
#include "stack.h"
#include "strings.h"
#include "threads.h"
#include "vfs.h"
#include "wonka.h"
#include "driver_byteserial.h"

/**************************************************************************
            T h e   W o n k a   V o c a b u l a r y
**************************************************************************/

static w_device debug_device;

static JNIEnv *claudia_env;

#define WONKA_PAD_SIZE 256

char wonkaPad[WONKA_PAD_SIZE];

#define TEMP_W_STRING_SIZE 256

// N.B. We assume Claudia is compiled last, so we can use this symbol
// as an upper bound for sanePtr ...
// int ficl_end_marker;

/**************************************************************************
                           b a n n e r T e x t

** Pushes the address and length of the banner onto the stack

**************************************************************************/

static void bannerText(FICL_VM *pVM) {
  stackPushPtr(pVM->pStack, (char*)banner);
  stackPushUNS(pVM->pStack, strlen(banner));
}

/**************************************************************************
                    h e a p I n s t a n c e s

** Pushes the number of bytes of heap in use onto the stack

**************************************************************************/

static void heapInstances(FICL_VM *pVM) {
  stackPushUNS(pVM->pStack, x_mem_total() - x_mem_avail());
}

/**************************************************************************
                  a c t i v e I n s t a n c e s

** Pushes the number of active instances onto the stack

**************************************************************************/

static void activeInstances(FICL_VM *pVM) {
  stackPushUNS(pVM->pStack, instance_use);
}

/**************************************************************************
                   f r e e d I n s t a n c e s

** Pushes the number of freed instances onto the stack

**************************************************************************/

static void freedInstances(FICL_VM *pVM) {
  stackPushUNS(pVM->pStack, instance_returned);
}

/**************************************************************************
                 c u r r e n t T i m e r T i c k s

** Pushes the number timer ticks since system startup onto the stack

**************************************************************************/

static void currentTimerTicks(FICL_VM *pVM) {
  stackPushUNS(pVM->pStack, x_time_get());
}

/**************************************************************************
                        m e m o r y R e p o r t           

** Pops an integer from the stack and calls reportMemStat.
** The integer should be 1 for a sort on hits, or 2 for a sort by file.

**************************************************************************/

static void memoryReport(FICL_VM *pVM) {
#ifdef DEBUG
  reportMemStat(stackPopINT(pVM->pStack));
#else
  vmThrowErr(pVM, (char*)"Sorry, statistics are not available");
#endif
}

/**************************************************************************
                            i n s t 2 o b j

** Given the address of an instance on the stack, replace this by a pointer
** to the object header.

**************************************************************************/

static void inst2obj(FICL_VM *pVM) {
  w_instance inst = stackPopPtr(pVM->pStack);
  if (sanePtr(inst)) {
    stackPushPtr(pVM->pStack, instance2object(inst));
  }
  else vmThrowErr(pVM, (char*)"Error: bad instance pointer");
}

/**************************************************************************
                            o b j 2 c h n k

** Given the address of an object on the stack, replace this by a pointer
** to the chunk header.

**************************************************************************/

static void obj2chnk(FICL_VM *pVM) {
  w_object obj = stackPopPtr(pVM->pStack);
  if (sanePtr(obj)) {
    stackPushPtr(pVM->pStack, block2chunk(obj));
  }
  else vmThrowErr(pVM, (char*)"Error: bad object pointer");
}

/**************************************************************************
                            c h n k 2 o b j

** Given the address of a chunk on the stack, replace this by a pointer
** to the chunk data.

**************************************************************************/

static void chnk2obj(FICL_VM *pVM) {
  w_chunk chnk = stackPopPtr(pVM->pStack);
  if (sanePtr(chnk)) {
    stackPushPtr(pVM->pStack, (chnk->data));
  }
  else vmThrowErr(pVM, (char*)"Error: bad chunk pointer");
}

/**************************************************************************
                            o b j 2 i n s t

** Given the address of an object on the stack, replace this by a pointer
** to the instance it contains.

**************************************************************************/

static void obj2inst(FICL_VM *pVM) {
  w_object obj = stackPopPtr(pVM->pStack);
  if (sanePtr(obj)) {
    stackPushPtr(pVM->pStack, (obj->fields));
  }
  else vmThrowErr(pVM, (char*)"Error: bad object pointer");
}

/**************************************************************************
                            i n s t 2 p o b j

** Given the address of a persistent instance on the stack, 
** replace this by a pointer to the object header.

**************************************************************************/

static void inst2pobj(FICL_VM *pVM) {
  w_instance inst = stackPopPtr(pVM->pStack);
  if (sanePtr(inst)) {
    stackPushPtr(pVM->pStack, instance2pobject(inst));
  }
  else vmThrowErr(pVM, (char*)"Error: bad instance pointer");
}

/**************************************************************************
                          p o b j 2 c h n k

** Given the address of a persistent object on the stack, replace this by 
** a pointer to the chunk header.

**************************************************************************/

static void pobj2chnk(FICL_VM *pVM) {
  stackPushPtr(pVM->pStack, PSblock2chunk(stackPopPtr(pVM->pStack)));
}

/**************************************************************************
                            c h n k 2 p o b j

** Given the address of a PSchunk on the stack, replace this by a pointer
** to the chunk data.

**************************************************************************/

static void chnk2pobj(FICL_VM *pVM) {
  stackPushPtr(pVM->pStack, ((w_PSchunk*)stackPopPtr(pVM->pStack))->contents);
}

/**************************************************************************
                          p o b j 2 i n s t

** Given the address of an object on the stack, replace this by a pointer
** to the persistent instance it contains.

**************************************************************************/

static void pobj2inst(FICL_VM *pVM) {
  stackPushPtr(pVM->pStack, ((w_pobject)stackPopPtr(pVM->pStack))->fields);
}

#ifdef ENABLE_PS

/**************************************************************************
                              p u s h S           

** Push onto the stack the address of the first persistent object, i.e. S

**************************************************************************/
static void pushS(FICL_VM *pVM) {
  stackPushPtr(pVM->pStack, firstPSblock());
}

/**************************************************************************
                              n e x t S           

** Pop from the stack the address of a persistent object and push the
** address of the next object in the persistent storage area.

**************************************************************************/

static void nextS(FICL_VM *pVM) {
  stackPushPtr(pVM->pStack, nextPSblock(stackPopPtr(pVM->pStack)));
}
#endif

/**************************************************************************
                      f i r s t H a s h t a b l e

** Push onto the stack the address of the first hashtable in our linked
** list of hashtables.

**************************************************************************/

static void firstHashtable(FICL_VM *pVM) {
  stackPushPtr(pVM->pStack, wonkaPointers->first_hashtable);
}

/**************************************************************************
                        n e x t H a s h t a b l e

** Pop the address of a hashtable and push the address of the next one
** in our linked list of hashtables.

**************************************************************************/

static void nextHashtable(FICL_VM *pVM) {
  w_hashtable this_hashtable = (w_hashtable)stackPopPtr(pVM->pStack);

  if (sanePtr(this_hashtable) && sanePtr(this_hashtable->next)) {
    stackPushPtr(pVM->pStack, this_hashtable->next);
  }
  else vmThrowErr(pVM, (char*)"Error: bad hashtable pointer");
}

/**************************************************************************
                        p r e v H a s h t a b l e

** Pop the address of a hashtable and push the address of the previous one
** in our linked list of hashtables.

**************************************************************************/

static void prevHashtable(FICL_VM *pVM) {
  w_hashtable this_hashtable = (w_hashtable)stackPopPtr(pVM->pStack);

  if (sanePtr(this_hashtable) && sanePtr(this_hashtable->previous)) {
    stackPushPtr(pVM->pStack, this_hashtable->previous);
  }
  else vmThrowErr(pVM, (char*)"Error: bad hashtable pointer");
}

/**************************************************************************
                        f i n d H a s h t a b l e

** ( c-addr count -- a-addr 0|true )
** Given the label of a hashtable (as c-addr count), find the       
** hashtable which matches this label.  If TOS is zero then
** no match was found and a-address is meaningless, else it is the
** address of the hashtable.

**************************************************************************/

static void findHashtable(FICL_VM *pVM) {
  w_int desc_length = stackPopINT(pVM->pStack);
  char* desc_chars  = stackPopPtr(pVM->pStack);
  w_hashtable h     = wonkaPointers->first_hashtable;
  w_int found = -1;

  if (desc_length>=WONKA_PAD_SIZE-1) desc_length = WONKA_PAD_SIZE - 2;
  wonkaPad[0] = desc_length;
  memcpy(wonkaPad+1,desc_chars,(w_size)desc_length);
  wonkaPad[desc_length+1] = 0;

  while (sanePtr(h) && strcmp(wonkaPad+1,h->label)) {
    h = h->next;
    if (h==wonkaPointers->first_hashtable)  {
      found = 0;
      break;
    }
  }

  if (sanePtr(h)) {
    stackPushPtr(pVM->pStack, h);
    stackPushINT(pVM->pStack, found);
  }
  else vmThrowErr(pVM, (char*)"Error: bad hashtable pointer");
}

/**************************************************************************
                         s h o w H a s h t a b l e

** Pop the address of a Wonka hashtable and prepare a string describing its
** current state.  Push the address and length of the string onto the stack.

**************************************************************************/

static void showHashtable(FICL_VM *pVM) {
  w_hashtable h = (w_hashtable)stackPopPtr(pVM->pStack);

  if (sanePtr(h)) {
    x_snprintf(wonkaPad,WONKA_PAD_SIZE,"%-64s size %5d occupancy %5d",h->label,h->currentsize,h->occupancy);

    stackPushPtr(pVM->pStack, wonkaPad);
    stackPushUNS(pVM->pStack, strlen(wonkaPad));
  }
  else vmThrowErr(pVM, (char*)"Error: bad hashtable pointer");
}

/**************************************************************************
                           l o c k H a s h t a b l e

** Lock the given hashtable.  Remember to unlock it!  (And don't lock it 2ce)
**************************************************************************/

static void lockHashtable(FICL_VM *pVM) {
  w_hashtable h = (w_hashtable)stackPopPtr(pVM->pStack);

  if (sanePtr(h)) {
    ht_lock(h);
  }
  else vmThrowErr(pVM, (char*)"Error: bad hashtable pointer");
}

/**************************************************************************
                        u n l o c k H a s h t a b l e

** Unlock the given hashtable.
**************************************************************************/

static void unlockHashtable(FICL_VM *pVM) {
  w_hashtable h = (w_hashtable)stackPopPtr(pVM->pStack);

  if (sanePtr(h)) {
    ht_unlock(h);
  }
  else vmThrowErr(pVM, (char*)"Error: bad hashtable pointer");
}


/**************************************************************************
                       i t e r a t e H a s h t a b l e

** Given a starting point (minimum:0) and a hashtable address, find the
** first occupied slot in the hashtable and overwite the starting point
** with the index of this slot.  If no occupied slot remains, return -1
** as the index.  Typical use:
**    0 h                        ( 0 hashtable )
**    begin iterate-hashtable    ( -1|n hashtable )
**    over -1 <>                 ( -1|n hashtable flase|true )
**    if 2dup                    ( n hashtable n hashtable )
**      do-something then        ( n hashtable )
**    swap 1+ dup rot            ( 0|n+1 hashtable 0|n+1 )
**    until                      ( 0 hashtable )
**    2drop                      ( )
**
** N.B. Does not lock hashtable!  Use lockHashtable and unlockHashtable 
**      if needed!
**************************************************************************/

static void iterateHashtable(FICL_VM *pVM) {
  w_hashtable h = (w_hashtable)stackPopPtr(pVM->pStack);
  w_int   idx = stackPopINT(pVM->pStack);

  if (sanePtr(h)) {
    while (idx<h->currentsize && !ht_occupied(h,idx)) ++idx;

    if (idx==h->currentsize) {
      stackPushINT(pVM->pStack, -1);
    }
    else stackPushINT(pVM->pStack, idx);
    stackPushPtr(pVM->pStack, h);
  }
  else vmThrowErr(pVM, (char*)"Error: bad hashtable pointer");
}

/**************************************************************************
                       s l o t O f H a s h t a b l e

** Given the index of a slot in a hashtable and the address of the hashtable,
** replace these on the stack by the key and the value found in that slot.
** This could well be the first word of "do-something" in the above example.
**************************************************************************/

static void slotOfHashtable(FICL_VM *pVM) {
  w_hashtable h = (w_hashtable)stackPopPtr(pVM->pStack);
  w_int   idx = stackPopINT(pVM->pStack);

  if (sanePtr(h)) {
    stackPushUNS(pVM->pStack, h->keys[idx]);
    stackPushUNS(pVM->pStack, h->values ? h->values[idx] : h->keys[idx]);
  }
  else vmThrowErr(pVM, (char*)"Error: bad hashtable pointer");
}

/**************************************************************************
                         r e a d H a s h t a b l e

** Given a key and the address of a hashtable, pops these from the stack
** and returns the value which is associated with that key in the hashtable.
**************************************************************************/

static void readHashtable(FICL_VM *pVM) {
  w_hashtable h = (w_hashtable)stackPopPtr(pVM->pStack);
  w_word  key   = stackPopUNS(pVM->pStack);

  if (sanePtr(h)) {
    stackPushUNS(pVM->pStack, ht_read(h,key));
  }
  else vmThrowErr(pVM, (char*)"Error: bad hashtable pointer");
}

/**************************************************************************
                           s h o w I n s t a n c e      

** Pop the address of a Wonka instance and prepare a string describing its
** current state.  Push the address and length of the string onto the stack.

**************************************************************************/

static void showInstance(FICL_VM *pVM) {
  w_instance i = (w_instance)stackPopPtr(pVM->pStack);

  if (isNullInstance(i)) {
    x_snprintf(wonkaPad,WONKA_PAD_SIZE,"the null instance %p",i);
  }
  else x_snprintf(wonkaPad,WONKA_PAD_SIZE,"instance %p of %k",i,instance2clazz(i));

  stackPushPtr(pVM->pStack, wonkaPad);
  stackPushUNS(pVM->pStack, strlen(wonkaPad));
}

/**************************************************************************
                           s h o w T h r e a d

** Pop the address of a Wonka thread and prepare a string describing its
** current state.  Push the address and length of the string onto the stack.

**************************************************************************/

static void showThread(FICL_VM *pVM) {
  w_thread t = (w_thread)stackPopPtr(pVM->pStack);
  char *threadPad = threadDescription(t);

  stackPushPtr(pVM->pStack, threadPad);
  stackPushUNS(pVM->pStack, strlen(threadPad));
}


/**************************************************************************
                       w a k e u p T h r e a d

** Pop a flag and the address of a Wonka thread and wake up the thread.

**************************************************************************/

static void wakeupThread(FICL_VM *pVM) {
  w_word   f = stackPopUNS(pVM->pStack);
  w_thread t = (w_thread)stackPopPtr(pVM->pStack);
  setFlag(t->flags, f);
  x_thread_stop_waiting(t->kthread);
}


/**************************************************************************
                           s h o w G r o u p 

** Pop the address of a Wonka threadgroup and prepare a string describing
** it.  Push the address and length of the string onto the stack.

**************************************************************************/

static void showGroup(FICL_VM *pVM) {
  w_group g = (w_group)stackPopPtr(pVM->pStack);
  char *groupPad = groupDescription(g);

  stackPushPtr(pVM->pStack, groupPad);
  stackPushUNS(pVM->pStack, strlen(groupPad));
}


/**************************************************************************
                       w o n k a f y S t r i n g

** Convert a Forth string (passed as c-addr u ) into a Wonka string (as 
** an a-addr).  Note that cstring2String has the side-effect of registering
** the string: if the string is new (refcount==1) then wonkafyString
** should return zero (meaning "not found").  Otherwise it will push
** the address of the canonical w_string.

**************************************************************************/

static void wonkafyString(FICL_VM *pVM) {
  w_size  length = stackPopUNS(pVM->pStack);
  char  * bytes = (char*)stackPopPtr(pVM->pStack);
  w_string result = cstring2String(W_Thread_system,bytes,length);
  w_size   refcount = ht_read(string_hashtable,(w_word)result);

  if (refcount>1) {
    stackPushUNS(pVM->pStack,(w_word)result);
  }
  else {
    stackPushUNS(pVM->pStack,0);
  }

  deregisterString(result);
}

/**************************************************************************
                     d e w o n k a f y S t r i n g

** Convert a Wonka string (passed as a a-addr) into a Forth string as 
** ( c-addr u ).

**************************************************************************/

static void dewonkafyString(FICL_VM *pVM) {
  w_string s = (w_string)stackPopPtr(pVM->pStack);
  w_size   l = s->length;
  w_size   i;

  for (i=0;i<WONKA_PAD_SIZE-1 && i<l;++i) {
    wonkaPad[i] = (char)(s->chars[i] & 0xff);
  }
  wonkaPad[i] = 0;

  stackPushPtr(pVM->pStack, wonkaPad);
  stackPushUNS(pVM->pStack, l);
}

/**************************************************************************
                           s h o w C l a s s

** Tell all about a Wonka clazz, in c-addr u format. 

**************************************************************************/

static void showClass(FICL_VM *pVM) {
  w_clazz c = (w_clazz)stackPopPtr(pVM->pStack);

  (void)x_snprintf(wonkaPad,WONKA_PAD_SIZE,"Class %k extends %k implements %d interfaces has %d methods %d static fields %d instance fields, instance size is %d",c,c->superClazz,c->numInterfaces,c->numMethods,c->numStaticFields,c->numPrimitiveFields+c->numReferenceFields,c->instanceSize);

  stackPushPtr(pVM->pStack, wonkaPad);
  stackPushUNS(pVM->pStack, strlen(wonkaPad));
}

/**************************************************************************
                             s e t T r i g 

** Set the trigger level for a specific source file or for all Wonka (if
** name begins with '*'). Stack diagram: ( level c-addr count -- )

**************************************************************************/

static void setTrig(FICL_VM *pVM) {
  w_size  filenameLength = stackPopUNS(pVM->pStack);
  char  * filenameChars = (char*)stackPopPtr(pVM->pStack);
  w_size  newLevel = stackPopUNS(pVM->pStack);

  if (*filenameChars=='*') {
    setAllTriggerLevel((w_int)newLevel);
  }
  else {
    if (filenameLength>=WONKA_PAD_SIZE) filenameLength = WONKA_PAD_SIZE-1;
    memcpy(wonkaPad,filenameChars,filenameLength);
    wonkaPad[filenameLength]=0;
    _setTriggerLevel(filenameChars,(w_int)newLevel);
  }
}

/**************************************************************************
                                   k e y           

** Wait for a key to be pressed, and push its character code.

**************************************************************************/

static void key(FICL_VM *pVM) {
  unsigned char   c;
  w_size lread;

  if(!debug_device) debug_device = deviceBSOpen("debug0", wdp_read_write);

  if (debug_device) {

    woempa(9,"Device is %s\n",debug_device->name);

    deviceBSRead(debug_device, &c, 1, &lread, x_eternal);
    stackPushUNS(pVM->pStack,c);
  }
  else {
    woempa(9,"Couldn't find debug device!\n");
    stackPushUNS(pVM->pStack,0);
  }
}

/**************************************************************************
                           n e w I n s t a n c e

** Pop a w_clazz pointer from the stack, and replace it by a pointer to
** an object of that class (or abort if failure)

**************************************************************************/

static void newInstance(FICL_VM *pVM) {
  w_clazz    clazz = (w_clazz)stackPopPtr(pVM->pStack);
  w_instance new;

  new = Class_newInstance0(claudia_env,clazz2Class(clazz));
  if (new) {
    stackPushPtr(pVM->pStack,instance2object(new));
  }
  else {
    vmTextOut(pVM, (char*)"Could not create a new instance", 1);
    vmThrow(pVM, VM_ABORT);
  }
}


/**************************************************************************
                           s t a r t T h r e a d

** Pop a w_object pointer from the stack, and if it corresponds to an
** instance of Thread then start the thread.

**************************************************************************/

static void startThread(FICL_VM *pVM) {
  w_instance instance = (w_instance)stackPopPtr(pVM->pStack);

  if (isSet(instance2clazz(instance)->flags,CLAZZ_IS_THREAD)) {
    Thread_start(claudia_env,instance);
  }
  else {
    vmTextOut(pVM, (char*)"Not an instance of Thread", 1);
    vmThrow(pVM, VM_ABORT);
  }
}

/**************************************************************************
                                c a l l M a i n

** Pop a w_clazz pointer from the stack, and if it has a main(String[])
** method then call it.

**************************************************************************/

static void callMain(FICL_VM *pVM) {
  w_clazz  clazz;
  w_thread thread;
  w_method method;
  w_instance exception;
  w_frame  bottom;
  int cmdline_len;
  char *cmdline_chars;

  clazz = stackPopPtr(pVM->pStack);
  thread = JNIEnv2w_thread(claudia_env);
// TODO: args!
  cmdline_len = stackPopINT(pVM->pStack);
  cmdline_chars = stackPopPtr(pVM->pStack);

  exception = classMustBeInitialized(clazz,thread);
  if (exception) {
    vmTextOut(pVM, (char*)"Unable to initialize class", 1);
    vmThrow(pVM, VM_ABORT);
  }
  else {
    w_string string_main = cstring2String(thread, "main", 4);
    w_string string_array_of_String = cstring2String(thread, "([Ljava/lang/String;)",21);
    method = getMethod_name_sig(clazz, string_main, string_array_of_String);
    deregisterString(string_main);
    deregisterString(string_array_of_String);
    if (method) {
      bottom = allocBottomFrame(thread, method);
      if (bottom) {
        bottom->locals[0] = 0; // will be: String[]
        thread->state = W_THREAD_READY; // ???
        exception = runThread(thread, bottom);
        releaseBottomFrame(bottom);
      }
      else {
        vmTextOut(pVM, (char*)"FICL internall error - couldn't alllocBottomFrame", 1);
        vmThrow(pVM, VM_ABORT);
      }
    }
    else {
      vmTextOut(pVM, (char*)"Class has no main(String[])", 1);
      vmThrow(pVM, VM_ABORT);
    }
  }
  // TODO: simething with exception ...
}

/**************************************************************************
                            l o a d C l a s s

** Pops a counted string and tries to load the class it names, using the
** system class loader.  Class name must be in dotified form.

***************************************************************************/

static void loadClass(FICL_VM *pVM) {
  w_size classname_len = stackPopINT(pVM->pStack);
  char *classname_chars = stackPopPtr(pVM->pStack);
  w_thread thread;
  w_instance exception = NULL;

  if ((thread = JNIEnv2w_thread(claudia_env))) { 
    namedClassMustBeLoaded(NULL, cstring2String(thread,classname_chars,classname_len),&exception);
//    mustBeLoaded(identifyClazz(cstring2String(thread,classname_chars,classname_len), NULL), &exception);

    if (exception) {
      vmTextOut(pVM, (char*)"Unable to load class", 1);
      vmThrow(pVM, VM_ABORT);
    }
  }
  else {
    vmTextOut(pVM, (char*)"Unable to load classes just yet, try later",1);
    vmThrow(pVM, VM_ABORT);
  }
}

/**************************************************************************
                      g e t W o n k a P o i n t e r s

** Pushes the address of wonkaPointers. 

**************************************************************************/

static void getWonkaPointers(FICL_VM *pVM) {
  stackPushPtr(pVM->pStack, wonkaPointers); 
}
  

/**************************************************************************
                      t e r m i n a t e W o n k a

** Terminate Wonka using exit(n), where n is popped from the stack.

**************************************************************************/

static void terminateWonka(FICL_VM *pVM) {
  exit(stackPopINT(pVM->pStack));
}

/**************************************************************************
                         d u m p I n s t a n c e s
`
** Dump a listing of the contents of the instance heap to the tty

**************************************************************************/

extern w_hashtable instance_hashtable;

static FICL_VM *this_is_not_thread_safe;

static void dumpInstances_callback(w_word keyword, w_word valword) {
  w_instance instance = (w_instance)keyword;
  w_object   object  = instance2object(instance);

  x_snprintf(wonkaPad, WONKA_PAD_SIZE, "%k at %p %s", object->clazz, instance, isSet(object->flags, O_GREEN)?"GREEN":"");
  vmTextOut(this_is_not_thread_safe, wonkaPad, 1);

  if (object->clazz == clazzString) {
    x_snprintf(wonkaPad, WONKA_PAD_SIZE, "  `%w'", object->fields[F_String_wotsit]);
    vmTextOut(this_is_not_thread_safe, wonkaPad, 1);
  }
}

static void dumpInstances(FICL_VM *pVM) {
  x_mem_lock(x_eternal);
  this_is_not_thread_safe = pVM;
  vmTextOut(pVM, (char*)"--- Start of instance listing ---", 1);
  ht_every(instance_hashtable, dumpInstances_callback);
  vmTextOut(pVM, (char*)"---- End of instance listing ----", 1);
  x_mem_unlock();
}

/**************************************************************************
                       a d d W o n k a W o r d s

** Add to the vocabulary the special words needed to probe/control Wonka.

**************************************************************************/

void addWonkaWords(void) {
  ficlBuild((char*)"banner",          bannerText,         FW_DEFAULT);
  ficlBuild((char*)"heap-instances",  heapInstances,      FW_DEFAULT);
  ficlBuild((char*)"active-instances", activeInstances,   FW_DEFAULT);
  ficlBuild((char*)"freed-instances", freedInstances,     FW_DEFAULT);
  ficlBuild((char*)"memory-report",   memoryReport,       FW_DEFAULT);
  ficlBuild((char*)"current-timer-ticks", currentTimerTicks, FW_DEFAULT);
  ficlBuild((char*)"instance2object", inst2obj,           FW_DEFAULT);
  ficlBuild((char*)"object2chunk",    obj2chnk,           FW_DEFAULT);
  ficlBuild((char*)"chunk2object",    chnk2obj,           FW_DEFAULT);
  ficlBuild((char*)"object2instance", obj2inst,           FW_DEFAULT);
  ficlBuild((char*)"instance2pobject", inst2pobj,         FW_DEFAULT);
  ficlBuild((char*)"pobject2chunk",   pobj2chnk,          FW_DEFAULT);
  ficlBuild((char*)"chunk2pobject",   chnk2pobj,          FW_DEFAULT);
  ficlBuild((char*)"pobject2instance", pobj2inst,         FW_DEFAULT);
#ifdef ENABLE_PS
  ficlBuild((char*)"first-pobject",   pushS,              FW_DEFAULT);
  ficlBuild((char*)"next-pobject",    nextS,              FW_DEFAULT);
#endif
  ficlBuild((char*)"first-hashtable", firstHashtable,     FW_DEFAULT);
  ficlBuild((char*)"next-hashtable",  nextHashtable,      FW_DEFAULT);
  ficlBuild((char*)"prev-hashtable",  prevHashtable,      FW_DEFAULT);
  ficlBuild((char*)"show-hashtable",  showHashtable,      FW_DEFAULT);
  ficlBuild((char*)"find-hashtable",  findHashtable,      FW_DEFAULT);
  ficlBuild((char*)"lock-hashtable",  lockHashtable,      FW_DEFAULT);
  ficlBuild((char*)"unlock-hashtable", unlockHashtable,   FW_DEFAULT);
  ficlBuild((char*)"iterate-hashtable", iterateHashtable, FW_DEFAULT);
  ficlBuild((char*)"slot-of-hashtable", slotOfHashtable,  FW_DEFAULT);
  ficlBuild((char*)"read-hashtable",  readHashtable,      FW_DEFAULT);
  ficlBuild((char*)"show-instance",   showInstance,       FW_DEFAULT);
  ficlBuild((char*)"show-thread",     showThread,         FW_DEFAULT);
  ficlBuild((char*)"wakeup-thread",   wakeupThread,       FW_DEFAULT);
  ficlBuild((char*)"show-group",      showGroup,          FW_DEFAULT);
  ficlBuild((char*)"wonkafy",         wonkafyString,      FW_DEFAULT);
  ficlBuild((char*)"dewonkafy",       dewonkafyString,    FW_DEFAULT);
  ficlBuild((char*)"show-class",      showClass,          FW_DEFAULT);
  ficlBuild((char*)"key",             key,                FW_DEFAULT);
  ficlBuild((char*)"(set-trigger-level)", setTrig,        FW_DEFAULT);
  ficlBuild((char*)"(new)",           newInstance,        FW_DEFAULT);
  ficlBuild((char*)"(run)",           startThread,        FW_DEFAULT);
  ficlBuild((char*)"(main)",          callMain,           FW_DEFAULT);
  ficlBuild((char*)"(load)",          loadClass,          FW_DEFAULT);
  ficlBuild((char*)"get-wonka-pointers", getWonkaPointers, FW_DEFAULT);
  ficlBuild((char*)"instances",        dumpInstances,     FW_DEFAULT);
  ficlBuild((char*)"terminate",        terminateWonka,    FW_DEFAULT);
}


static w_Pointers WonkaPointers;

FICL_VM FiclVM;
char PStack[sizeof (FICL_STACK) + FICL_DEFAULT_STACK * sizeof (CELL)];
char RStack[sizeof (FICL_STACK) + FICL_DEFAULT_STACK * sizeof (CELL)];

/**************************************************************************
                            i n i t F i c l

** Special startup code for starting initial ficlVM in Wonka:
**   - uses static memory allocation
**   - knows it is the first VM, so initialises vmList and loads softcore.

**************************************************************************/

#ifndef CLAUDIADIR
#define CLAUDIADIR "system"
#endif

#ifndef CLAUDIAFILE
#define CLAUDIAFILE "claudia.fr"
#endif

static char             *claudia_data;
static int              fd;
static struct vfs_STAT  statbuf;
static w_word           claudia_size;
static const char      *claudia_filename = CLAUDIADIR "/" CLAUDIAFILE;

void initFicl() {
  FICL_STACK *pStack;
  FICL_STACK *rStack;

  wonkaPointers = &WonkaPointers;
  ficlInitSystem(NULL,-1);
  ficlVM = &FiclVM;
  pStack = (FICL_STACK*)&PStack;
  pStack->nCells = FICL_DEFAULT_STACK;
  pStack->sp     = pStack->base;
  pStack->pFrame = NULL;
  ficlVM->pStack = pStack;
  rStack = (FICL_STACK*)&RStack;
  rStack->nCells = FICL_DEFAULT_STACK;
  rStack->sp     = rStack->base;
  rStack->pFrame = NULL;
  ficlVM->rStack = rStack;
  ficlVM->textOut = ficlTextOut;
  vmReset(ficlVM);
  wonkaPointers->ficlVM = ficlVM;
  ficlVM->link = wonkaPointers->vmList;
  wonkaPointers->vmList = ficlVM;
  ficlCompileSoftCore((FICL_VM*)(wonkaPointers->ficlVM));
  addWonkaWords();

  // smartBoot();
}


#define listIsSane(f) (sanePtr(f) && sanePtr((f)->next) && sanePtr((f)->previous) \
      && (f)->next->previous==(f) && (f)->previous->next==(f))

#define WONKA_TIB_SIZE 1024

char wonkaTib[WONKA_TIB_SIZE];

char *getLine(w_device device) {
  w_size lread;

  deviceBSRead(device, wonkaTib, WONKA_TIB_SIZE-1, &lread, x_eternal);
  wonkaTib[lread] = 0;

  return wonkaTib;
}

/*
w_int smartBoot(void) {

  if (wonkaPointers->bootmode==SMARTBOOT_WARM_BOOT
   && listIsSane(wonkaPointers->first_hashtable)
   && listIsSane(wonkaPointers->systemRepository)
     ) {

    ficlExec(ficlVM,(char*)
      ".ver cr "
      ".( Seems like Wonka is already loaded.  Cool... ) cr "
      ".( Type 'warm boot' to restart the existing image, 'cold boot' to reload.) cr "
    );

    if(!debug_device) debug_device = deviceBSOpen("debug0", wdp_read_write);

    if (debug_device) {
      for (;;) {
        int rc = ficlExec((FICL_VM*)(wonkaPointers->ficlVM), getLine(debug_device));
        if (rc==VM_USEREXIT) {
            break;
        }
      }
    }

    return SMARTBOOT_WARM_BOOT;
  }
  else {
    wonkaPointers->first_hashtable = NULL;
    wonkaPointers->systemRepository = NULL;
    wonkaPointers->bootmode = SMARTBOOT_COLD_BOOT;

    return SMARTBOOT_COLD_BOOT;
  }

}
*/

#define WATCHDOG_SLEEP_TIME 10000

void claudia(void* data) {
  w_word avail;
  char   *command;
  JavaVM *allCreatedJavaVMs[1];
  JavaVM *mrWonka = NULL;
  jint nVMs;
  jint rc;
  Wonka_AttachArgs waa;

  if(!debug_device) {
    debug_device = deviceBSOpen("debug0", wdp_read_write);
  }

  if (vfs_stat(claudia_filename, &statbuf) == 0) {
    claudia_size = statbuf.st_size;
    claudia_data = allocMem(W_Thread_system, claudia_size+1);
    fd = vfs_open(claudia_filename, VFS_O_RDONLY, 0);
    vfs_read(fd, claudia_data, claudia_size);
    vfs_close(fd);
    claudia_data[claudia_size] = 0;
    ficlExec((FICL_VM*)wonkaPointers->ficlVM, claudia_data);
    releaseMem(claudia_data);
    PutString((char*)"Found claudia definitions file '");
  }
  else {
    PutString((char*)"Could not find claudia definitions file '");
  }
  PutString((char *)claudia_filename);
  PutString((char*)"'\n");

  if (debug_device) {
    PutString((char*)"Debug device is '");
    PutString(debug_device->name);
    PutString((char*)"'\n");

    for (;;) {
      x_thread_sleep(x_millis2ticks(WATCHDOG_SLEEP_TIME));

      deviceBSQuery(debug_device, wdi_get_available, &avail, x_eternal);

      if (avail) {
        waa.name = (char*)"Claudia";
        rc = JNI_GetCreatedJavaVMs(allCreatedJavaVMs,1,&nVMs);
        woempa(9,"JNI_GetCreatedJavaVMs returned %d\n",rc);
        if (rc == 0 && nVMs > 0) {
          mrWonka = allCreatedJavaVMs[0];
          rc = ((*mrWonka)->AttachCurrentThread)(mrWonka, &claudia_env, &waa);
          woempa(9,"AttachCurrentThread returned %d\n",rc);
        }
        ficlExec((FICL_VM*)(wonkaPointers->ficlVM),(char*)
          "find-hashtables "
          "welcome "
          "quit "
         );
        for (;;) {
          command = getLine(debug_device);
          if (command[0]) { // non-empty line
            rc = ficlExec((FICL_VM*)(wonkaPointers->ficlVM), command);
            if (rc==VM_USEREXIT) {
              break;
            }
          }
          else {
            x_thread_sleep(125);
          }
        }
        if (rc==0 && *mrWonka) {
          ((*mrWonka)->DetachCurrentThread)(mrWonka);
        }
      }
    }
  }
  else {
    vmTextOut((FICL_VM*)(wonkaPointers->ficlVM), (char*)"claudia: Could not open debug device", 1);
    vmThrow((FICL_VM*)(wonkaPointers->ficlVM), VM_ABORT);
  }
}

x_Thread claudia_thread;
#define claudia_stacksize (1024 * 16)
unsigned char claudia_stack[claudia_stacksize];

void startClaudia (void)
{
	memset(claudia_stack, 0x00, claudia_stacksize);
	memset(&claudia_thread, 0x00, sizeof(x_Thread));

	x_thread_create (
		&claudia_thread,
		claudia,
		NULL,
		claudia_stack,
		claudia_stacksize,
		1,
		TF_START);
}

void smartAbort() {

  ficlExec((FICL_VM*)(wonkaPointers->ficlVM),(char*)"find-hashtables abwelcome quit ");

  if(!debug_device) debug_device = deviceBSOpen("debug0", wdp_read_write);

  if (debug_device) {
    for (;;) {
      int rc = ficlExec((FICL_VM*)(wonkaPointers->ficlVM), getLine(debug_device));
      if (rc==VM_USEREXIT) {
        break;
      }
    }
  }
  ficlExec((FICL_VM*)(wonkaPointers->ficlVM),(char*)"cold boot ");
}

