/*******************************************************************
** m a t h 6 4 . c
** Forth Inspired Command Language - 64 bit math support routines
** Author: John Sadler (john_sadler@alum.mit.edu)
** Created: 25 January 1998
** Rev 2.03: Support for 128 bit DP math. This file really ouught to
** be renamed!
*******************************************************************/

#include "ficl.h"
#include "math64.h"


/**************************************************************************
                        m 6 4 A b s
** Returns the absolute value of an DPINT
**************************************************************************/
void m64Abs(DPINT x, DPINT *result)
{
    if (m64IsNegative(*result))
        m64Negate(result);
}


/**************************************************************************
                        m 6 4 F l o o r e d D i v I
** 
** FROM THE FORTH ANS...
** Floored division is integer division in which the remainder carries
** the sign of the divisor or is zero, and the quotient is rounded to
** its arithmetic floor. Symmetric division is integer division in which
** the remainder carries the sign of the dividend or is zero and the
** quotient is the mathematical quotient rounded towards zero or
** truncated. Examples of each are shown in tables 3.3 and 3.4. 
** 
** Table 3.3 - Floored Division Example
** Dividend        Divisor Remainder       Quotient
** --------        ------- ---------       --------
**  10                7       3                1
** -10                7       4               -2
**  10               -7      -4               -2
** -10               -7      -3                1
** 
** 
** Table 3.4 - Symmetric Division Example
** Dividend        Divisor Remainder       Quotient
** --------        ------- ---------       --------
**  10                7       3                1
** -10                7      -3               -1
**  10               -7       3               -1
** -10               -7      -3                1
**************************************************************************/
void m64FlooredDivI(DPINT num, FICL_INT den, INTQR *result)
{
    UNSQR uqr;
    int signRem = 1;
    int signQuot = 1;

    if (m64IsNegative(num))
    {
        m64Negate(&num);
        signQuot = -signQuot;
    }

    if (den < 0)
    {
        den      = -den;
        signRem  = -signRem;
        signQuot = -signQuot;
    }

    ficlLongDiv(m64CastIU(num), (FICL_UNS)den, &uqr);
//    m64CastQRUI(uqr, result);
    result->quot = uqr.quot;
    result->rem = uqr.rem;
    if (signQuot < 0)
    {
        result->quot = -result->quot;
        if (result->rem != 0)
        {
            result->quot--;
            result->rem = den - result->rem;
        }
    }

    if (signRem < 0)
        result->rem = -result->rem;

}


/**************************************************************************
                        m 6 4 I s N e g a t i v e
** Returns TRUE if the specified DPINT has its sign bit set.
**************************************************************************/
int m64IsNegative(DPINT x)
{
    return (x.hi < 0);
}


/**************************************************************************
                        m 6 4 M a c
** Mixed precision multiply and accumulate primitive for number building.
** Multiplies DPUNS u by FICL_UNS mul and adds FICL_UNS add. Mul is typically
** the numeric base, and add represents a digit to be appended to the 
** growing number. 
** Returns the result of the operation
**************************************************************************/
void m64Mac(DPUNS u, FICL_UNS mul, FICL_UNS add, DPUNS *result)
{
    DPUNS resultLo;
    DPUNS resultHi;
    ficlLongMul(u.lo, mul, &resultLo);
    ficlLongMul(u.hi, mul, &resultHi);
    resultLo.hi += resultHi.lo;
    resultHi.lo = resultLo.lo + add;

    if (resultHi.lo < resultLo.lo)
        resultLo.hi++;

    resultLo.lo = resultHi.lo;

    *result = resultLo;
}


/**************************************************************************
                        m 6 4 M u l I
** Multiplies a pair of FICL_INTs and returns an DPINT result.
**************************************************************************/
void m64MulI(FICL_INT x, FICL_INT y, DPINT *result)
{
    DPUNS prod;
    int sign = 1;

    if (x < 0)
    {
        sign = -sign;
        x = -x;
    }

    if (y < 0)
    {
        sign = -sign;
        y = -y;
    }

    ficlLongMul((w_word)x, (w_word)y, &prod);
    *result = m64CastUI(prod);
    if (sign < 0)
      m64Negate(result);
}


/**************************************************************************
                        m 6 4 N e g a t e
** Negates an DPINT by complementing and incrementing.
**************************************************************************/
void m64Negate(DPINT *x)
{
    x->hi = ~x->hi;
    x->lo = ~x->lo;
    x->lo ++;
    if (x->lo == 0)
        x->hi++;
}


/**************************************************************************
                        m 6 4 P u s h
** Push an DPINT onto the specified stack in the order required
** by ANS Forth (most significant cell on top)
** These should probably be macros...
**************************************************************************/
void  i64Push(FICL_STACK *pStack, DPINT i64)
{
    stackPushINT(pStack, i64.lo);
    stackPushINT(pStack, i64.hi);
    return;
}

void  u64Push(FICL_STACK *pStack, DPUNS u64)
{
    stackPushINT(pStack, (w_int)u64.lo);
    stackPushINT(pStack, (w_int)u64.hi);
    return;
}


/**************************************************************************
                        m 6 4 P o p
** Pops an DPINT off the stack in the order required by ANS Forth
** (most significant cell on top)
** These should probably be macros...
**************************************************************************/
void i64Pop(FICL_STACK *pStack, DPINT *result)
{
    result->hi = stackPopINT(pStack);
    result->lo = stackPopINT(pStack);
}

void u64Pop(FICL_STACK *pStack, DPUNS *result)
{
    result->hi = (w_word)stackPopINT(pStack);
    result->lo = (w_word)stackPopINT(pStack);
}


/**************************************************************************
                        m 6 4 S y m m e t r i c D i v
** Divide an DPINT by a FICL_INT and return a FICL_INT quotient and a
** FICL_INT remainder. The absolute values of quotient and remainder are not
** affected by the signs of the numerator and denominator (the operation
** is symmetric on the number line)
**************************************************************************/
void m64SymmetricDivI(DPINT num, FICL_INT den, INTQR *result)
{
    UNSQR uqr;
    int signRem = 1;
    int signQuot = 1;

    if (m64IsNegative(num))
    {
        m64Negate(&num);
        signRem  = -signRem;
        signQuot = -signQuot;
    }

    if (den < 0)
    {
        den      = -den;
        signQuot = -signQuot;
    }

    ficlLongDiv(m64CastIU(num), (FICL_UNS)den, &uqr);

    result->quot = uqr.quot;
    result->rem = uqr.rem;

    if (signRem < 0)
        result->rem = -result->rem;

    if (signQuot < 0)
        result->quot = -result->quot;

}


/**************************************************************************
                        m 6 4 U M o d
** Divides a DPUNS by base (an UNS16) and returns an UNS16 remainder.
** Writes the quotient back to the original DPUNS as a side effect.
** This operation is typically used to convert an DPUNS to a text string
** in any base. See words.c:numberSignS, for example.
** Mechanics: performs 4 ficlLongDivs, each of which produces 16 bits
** of the quotient. C does not provide a way to divide an FICL_UNS by an
** UNS16 and get an FICL_UNS quotient (ldiv is closest, but it's signed,
** unfortunately), so I've used ficlLongDiv.
**************************************************************************/
#if (BITS_PER_CELL == 32)

#define UMOD_SHIFT 16
#define UMOD_MASK 0x0000ffff

#elif (BITS_PER_CELL == 64)

#define UMOD_SHIFT 32
#define UMOD_MASK 0x00000000ffffffff

#endif

w_word m64UMod(DPUNS *pUD, w_word base)
{
    DPUNS ud;
    UNSQR qr;
    DPUNS result;

    result.hi = result.lo = 0;

    ud.hi = 0;
    ud.lo = pUD->hi >> UMOD_SHIFT;
    ficlLongDiv(ud, (FICL_UNS)base, &qr);
    result.hi = qr.quot << UMOD_SHIFT;

    ud.lo = (qr.rem << UMOD_SHIFT) | (pUD->hi & UMOD_MASK);
    ficlLongDiv(ud, (FICL_UNS)base, &qr);
    result.hi |= qr.quot & UMOD_MASK;

    ud.lo = (qr.rem << UMOD_SHIFT) | (pUD->lo >> UMOD_SHIFT);
    ficlLongDiv(ud, (FICL_UNS)base, &qr);
    result.lo = qr.quot << UMOD_SHIFT;

    ud.lo = (qr.rem << UMOD_SHIFT) | (pUD->lo & UMOD_MASK);
    ficlLongDiv(ud, (FICL_UNS)base, &qr);
    result.lo |= qr.quot & UMOD_MASK;

    *pUD = result;

    return (UNS16)(qr.rem);
}


/**************************************************************************
** Contributed by
** Michael A. Gauland	gaulandm@mdhost.cse.tek.com  
**************************************************************************/
#if PORTABLE_LONGMULDIV != 0
/**************************************************************************
                        m 6 4 A d d
** 
**************************************************************************/
void m64Add(DPUNS x, DPUNS y, DPUNS *result)
{
    int carry;
    
    result->hi = x.hi + y.hi;
    result->lo = x.lo + y.lo;


    carry  = ((x.lo | y.lo) & CELL_HI_BIT) && !(result->lo & CELL_HI_BIT);
    carry |= ((x.lo & y.lo) & CELL_HI_BIT);

    if (carry)
    {
        result->hi++;
    }

}


/**************************************************************************
                        m 6 4 S u b
** 
**************************************************************************/
void m64Sub(DPUNS x, DPUNS y, DPUNS *result)
{
    result->hi = x.hi - y.hi;
    result->lo = x.lo - y.lo;

    if (x.lo < y.lo) 
    {
        result->hi--;
    }
}


/**************************************************************************
                        m 6 4 A S L
** 64 bit left shift
**************************************************************************/
void m64ASL( DPUNS x, DPUNS *result )
{
    result->hi = x.hi << 1;
    if (x.lo & CELL_HI_BIT) 
    {
        result->hi++;
    }

    result->lo = x.lo << 1;
}


/**************************************************************************
                        m 6 4 A S R
** 64 bit right shift (unsigned - no sign extend)
**************************************************************************/
void m64ASR( DPUNS x, DPUNS *result )
{
    result->lo = x.lo >> 1;
    if (x.hi & 1) 
    {
        result->lo |= CELL_HI_BIT;
    }

    result->hi = x.hi >> 1;
}


/**************************************************************************
                        m 6 4 O r
** 64 bit bitwise OR
**************************************************************************/
void m64Or( DPUNS x, DPUNS y, DPUNS *result)
{
    result->hi = x.hi | y.hi;
    result->lo = x.lo | y.lo;
}


/**************************************************************************
                        m 6 4 C o m p a r e
** Return -1 if x < y; 0 if x==y, and 1 if x > y.
**************************************************************************/
int m64Compare(DPUNS x, DPUNS y)
{
    int result;
    
    if (x.hi > y.hi) 
    {
        result = +1;
    } 
    else if (x.hi < y.hi) 
    {
        result = -1;
    } 
    else 
    {
        /* High parts are equal */
        if (x.lo > y.lo) 
        {
            result = +1;
        } 
        else if (x.lo < y.lo) 
        {
            result = -1;
        } 
        else 
        {
            result = 0;
        }
    }
    
    return result;
}


/**************************************************************************
                        f i c l L o n g M u l
** Portable versions of ficlLongMul and ficlLongDiv in C
** Contributed by:
** Michael A. Gauland	gaulandm@mdhost.cse.tek.com  
** [CG] moved result to a &parameter
**************************************************************************/
void ficlLongMul(FICL_UNS x, FICL_UNS y, DPUNS *result)
{
    DPUNS addend;
    
    result->lo = 0;
    result->hi = 0;

    addend.lo = y;
    addend.hi = 0; /* No sign extension--arguments are unsigned */
    
    while (x != 0) 
    {
        if ( x & 1) 
        {
            m64Add(*result, addend, result);
        }
        x >>= 1;
        m64ASL(addend,&addend);
    }
}


/**************************************************************************
                        f i c l L o n g D i v
** Portable versions of ficlLongMul and ficlLongDiv in C
** Contributed by:
** Michael A. Gauland	gaulandm@mdhost.cse.tek.com  
**************************************************************************/
void ficlLongDiv(DPUNS q, FICL_UNS y, UNSQR *result)
{
    DPUNS quotient;
    DPUNS subtrahend;
    DPUNS mask;

    quotient.lo = 0;
    quotient.hi = 0;
    
    subtrahend.lo = y;
    subtrahend.hi = 0;
    
    mask.lo = 1;
    mask.hi = 0;
    
    while ((m64Compare(subtrahend, q) < 0) &&
           (subtrahend.hi & CELL_HI_BIT) == 0)
    {
        m64ASL(mask,&mask);
        m64ASL(subtrahend,&subtrahend);
    }
    
    while (mask.lo != 0 || mask.hi != 0) 
    {
        if (m64Compare(subtrahend, q) <= 0) 
        {
            m64Sub( q, subtrahend, &q);
            m64Or(quotient, mask, &quotient);
        }
        m64ASR(mask,&mask);
        m64ASR(subtrahend,&subtrahend);
    }
    
    result->quot = quotient.lo;
    result->rem = q.lo;
}

#endif

