/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: instance.c,v 1.16 2002/03/20 13:04:20 gray Exp $
*/

#include "arrays.h"
#include "checks.h"
#include "clazz.h"
#include "core-classes.h"
#include "descriptor.h"
#include "exception.h"
#include "hashtable.h"
#include "loading.h"
#include "list.h"
#include "ts-mem.h"
#include "stack.h"
#include "strings.h"
#include "threads.h"
#include "wordset.h"

extern w_hashtable lock_hashtable;

#define ANNOY_THE_COLLECTOR
#undef ANNOY_THE_COLLECTOR
#define ANNOY_TRIGGER 100 // 1280
#define ANNOY_TIME 40

/*
** The instance_hashtable holds all active instances (with key==value).
*/
w_hashtable instance_hashtable;

inline static void i_addLocalReference(w_thread thread, w_instance instance) {
  w_frame frame = thread->top;

  if (frame->firstRef) {
    addToWordset(frame->thread, &frame->localRefs, (w_word)instance);
  }
  else {
    frame->firstRef = instance;
  }

  setFlag(instance2flags(instance),O_GREEN);
  newGlobalReference(instance);
}

/*
** We create a non inlined version of addLocalReference so that others can 
** also call this function.  This one has markTransientRef() built in, for
** good measure.
*/

void addLocalReference(w_thread thread, w_instance instance) {
  i_addLocalReference(thread, instance);
  markTransientRef(instance);
}

w_clazz
instance2clazz(w_instance ref) {
  if (ref == NULL) {

    return clazzObject;

  }

  return instance2object(makeLocal(ref))->clazz;
}

w_size instance_use = 0;
w_size instance_allocated = 0;
w_size instance_returned = 0;
w_object instance_first = NULL;

// cached clazz...
w_clazz cache_clazz = NULL;
w_int cache_hits = 0;

inline static void registerObject(w_object object) {
  ht_write(instance_hashtable,(w_word)object->fields,(w_word)object->fields);
  instance_use += 1;
  instance_allocated += 1;
}

inline static void checkClazz(w_clazz clazz) {

  if (! clazz) {
    wabort(ABORT_WONKA, "Clazz is NULL\n");
  }

  /*
  ** Check if bytes needed is negative or larger than 16 Mbytes.
  */
    
  if (clazz->bytes_needed & 0xff000000) {
    wabort(ABORT_WONKA, "Clazz %k impossible size words = 0x%x bytes = 0x%x\n", clazz, clazz->instanceSize, clazz->bytes_needed);
  }

}

inline static w_object tryCache(w_clazz clazz) {

#ifdef CACHE_INSTANCES
  w_object object = NULL;
  x_status status;
  
  status = x_mutex_lock(clazz->cache_lock, x_eternal);
  if (status == xs_success) {
    if (clazz->cache && sizeOfWordset(&clazz->cache) > 0) {
      object = (w_object)takeFirstFromWordset(&clazz->cache);
      clazz->cache_hits += 1;
    }
    else {
      clazz->cache_miss += 1;
    }
    x_mutex_unlock(clazz->cache_lock);
  }

  return object;
#else
  return NULL;
#endif  
}

/* 
** Zero out all the field words of an object. Note that we only have fields with word sizes, 
** even with arrays, so 'size' indicates the number of fields that should be cleared to 0x00000000.
*/

inline static void clearObject(w_object object, w_size size) {
  woempa(1, "Clearing %d words of object %p\n", size, object->fields);

  while (size--) {
    object->fields[size] = 0x00000000;
  }
  
}

void newReferenceTo(w_instance child) {

  w_word *child_flags = instance2flagsptr(child);

  woempa(1,"Child flags: 0x%08x\n",*child_flags);

  if (isNotSet(*child_flags, O_BLACK | O_NEAR_BLACK | O_GRAY)) {
    markTransientRef(child);
  }
  unsetFlag(*child_flags, O_GREEN);
}

w_boolean heap_request(w_realm realm, w_int bytes) {
  if (gc_instance == NULL) {
    woempa(1, "Called before gc exists: gc_instance = %p.\n", gc_instance);

    return WONKA_TRUE;

  }

  if (currentWonkaThread == gc_thread) {
    woempa(1, "Called by gc thread.\n");

    return WONKA_TRUE;

  }


  if (x_mem_avail() - bytes < min_heap_free) {
    woempa(9, "Sorry: only %d bytes free, minimum is %d\n", x_mem_avail() - bytes, min_heap_free);

    return WONKA_FALSE;

  }

  gc_reclaim(bytes, NULL);

  if (checkRealmMemoryQuota(realm, bytes)) {

    return WONKA_TRUE;

  }

  gc_reclaim(bytes, NULL);

  if (checkRealmMemoryQuota(realm, bytes)) {

    return WONKA_TRUE;

  }

  return WONKA_FALSE;
}

/*
** We have now
**
** 1) w_instance _norm_allocInstance(w_thread thread, w_clazz clazz) 
**    An allocator function for NORMAL operation of Wonka.
**
** 2) w_instance _boot_allocInstance(w_thread thread, w_clazz clazz) 
**    An allocator function for use during BOOTING of Wonka.
**
** Either 1) or 2) will be put in the _xxxx_allocInstance function pointer, based on wether we are
** booting or not.
**
** 3) w_instance _norm_allocInstanceWithFlags(w_thread thread, w_clazz clazz, w_flags flags)
**    An allocator function with flags for NORMAL operation of Wonka.
**
** 4) w_instance _boot_allocInstanceWithFlags(w_thread thread, w_clazz clazz, w_flags flags)
**    An allocator function for use during BOOTING of Wonka.
**
** Either 3) or 4) will be put in the _xxxx_allocInstanceWithFlags function pointer, based on wether we are
** booting or not.
**
** 5) w_instance _norm_AllocArrayInstance(const char *f, int l, w_thread thread, w_clazz clazz, w_flags flags)
**    An array instance allocator function with flags for NORMAL operation of Wonka.
**
** 6) w_instance _boot_AllocArrayInstance(const char *f, int l, w_thread thread, w_clazz clazz, w_flags flags)
**    An array instance allocator function for use during BOOTING of Wonka.
**
** Either 5) or 6) will be put in the _xxxx_allocArrayInstance function pointer, based on wether we are
** booting or not.
*/

inline static void handleLocalRef(w_thread thread, w_object object) {

  w_frame frame = thread->top;
  
  if (frame && frame->native) {
    woempa(1, "addLocalReference %k@%p to thread %p.\n", object->clazz, object->fields, thread);
    i_addLocalReference(thread, object->fields);
    setFlag(object->flags,O_GREEN);
  }
  else {
    woempa(1, "marking %k@%p GREEN.\n", object->clazz, object->fields);
    setFlag(object->flags,O_GREEN);
  }
  markTransientRef(object->fields);

}

w_instance _xxxx_norm_allocInstance(w_thread thread, w_clazz clazz) {

  w_object object = NULL;
  w_realm realm = NULL;

  if (thread->group) {
    realm = thread2realm(thread);
  }

  /*
  ** The class must be initialized.
  */
  
  classMustBeInitialized(clazz, thread);

  checkClazz(clazz);  

  woempa(1, "clazz is %k at %p, requested size is %d words, instance needs %d bytes.\n", clazz, clazz, clazz->instanceSize, clazz->bytes_needed);

  /*
  ** The test for CLAZZ_IS_THROWABLE is really a proxy for `ignore quotas'.
  ** Maybe we should have a separate allocator for priority cases?
  **
  ** First we try to allocate from cache, and if that fails we do an allocMem.
  */

  if (heap_request(realm, (w_int)clazz->bytes_needed) || isSet(clazz->flags, CLAZZ_IS_THROWABLE)) {

    object = tryCache(clazz);
  
    if (object) {
      resetMemOwnerRealm(object, realm);
    }
    else {
      object = allocMem_realm(realm, clazz->bytes_needed);
      woempa(1, "Allocated a %k at %p, need %d bytes\n", clazz, object->fields, clazz->bytes_needed);
      if (object) {
        object->clazz = clazz;
      }
    }
  }

  /*
  ** If object is still NULL, we are out of memory.
  */
  if (!object) {
    if (thread->Thread) {
      woempa(9, "Could not allocate a %k for thread %w, throwing OutOfMemoryError\n", clazz, thread->name);
      throwOutOfMemoryError(thread);
    }
    else {
      woempa(9, "Drat.  Can't allocate instance, and no Thread to throw an exception in.\n");
    }

    return NULL;

  }

  /*
  ** These object fields need to be intialized each time again.
  */
  
  object->flags = (O_CACHEABLE);
  if (isSet(clazz->flags, CLAZZ_HAS_FINALIZER)) {
    woempa(1, "Clazz %k has a finalizer, setting instance at %p FINALIZABLE\n", clazz, object->fields);
    setFlag(object->flags, O_FINALIZABLE);
  }
  if (isSet(clazz->flags, CLAZZ_IS_REFERENCE)) {
    woempa(7, "Clazz %k is reference, setting instance at %p ENQUEUEABLE\n", clazz, object->fields);
    setFlag(object->flags, O_ENQUEUEABLE);
  }

  clearObject(object, clazz->instanceSize);

  /*
  ** If inside a native method, add new instance to local reference table.
  */
  handleLocalRef(thread, object);

  registerObject(object);
  
  //woempa(1, "Allocated instance %p (object %p) of %k, %d words, realm %p.\n", object->fields, object, clazz, clazz->instanceSize, realm);

  //woempa(9, "CACHE class %k has %d avail, %d hits, %d misses, %d instance size...\n", clazz, sizeOfWordset(&clazz->cache), clazz->cache_hits, clazz->cache_miss, clazz->bytes_needed);

  return object->fields;
  
}

w_instance _xxxx_boot_allocInstance(w_thread thread, w_clazz clazz) {

  w_object object;

  /*
  ** The class must be initialized.
  */
  
  if (clazzObject && clazz != clazzObject && clazzClass && clazz != clazzClass) {
    classMustBeInitialized(clazz, thread);
  }

  checkClazz(clazz);  

  woempa(1, "Allocating a %k, instance size is %d words, instance needs %d bytes.\n", clazz, clazz->instanceSize, clazz->bytes_needed);

  /*
  ** We don't use the cache in the booting version.
  */
  
  object = allocMem_system(clazz->bytes_needed);

  if (object == NULL) {
    wabort(ABORT_WONKA, "Unable to allocate memory for instance!\n");
  }

  object->clazz = clazz;
  object->flags = O_CACHEABLE;
  if (isSet(clazz->flags, CLAZZ_HAS_FINALIZER)) {
    woempa(1, "Clazz %k has a finalizer, setting instance at %p FINALIZABLE\n", clazz, object->fields);
    setFlag(object->flags, O_FINALIZABLE);
  }
  if (isSet(clazz->flags, CLAZZ_IS_REFERENCE)) {
    woempa(7, "Clazz %k is reference, setting instance at %p ENQUEUEABLE\n", clazz, object->fields);
    setFlag(object->flags, O_ENQUEUEABLE);
  }

  clearObject(object, clazz->instanceSize);

  registerObject(object);
  
  //woempa(1, "Allocated instance %p (object %p) of %k, %d words\n", object->fields, object, clazz, clazz->instanceSize);

  return object->fields;
  
}

w_instance _xxxx_norm_allocInstanceWithFlags(w_thread thread, w_clazz clazz, w_flags flags) {

  w_object object = NULL;
  w_realm realm = NULL;

  if (thread->group) {
    realm = thread2realm(thread);
  }

  /*
  ** The class must be initialized.
  */
  
  classMustBeInitialized(clazz, thread);

  checkClazz(clazz);  

  woempa(1, "clazz is %k at %p, requested size is %d words, instance needs %d bytes.\n", clazz, clazz, clazz->instanceSize, clazz->bytes_needed);

  /*
  ** The test for CLAZZ_IS_THROWABLE is really a proxy for `ignore quotas'.
  ** Maybe we should have a separate allocator for priority cases?
  **
  ** First we try to allocate from cache, and if that fails we do an allocMem.
  */
  if (isSet(clazz->flags, CLAZZ_IS_THROWABLE) || heap_request(realm, (w_int)clazz->bytes_needed)) {

    object = tryCache(clazz);

    /*
    ** See if we found a cached object and if not, still allocate one.
    */
  
    if (! object) {
      object = allocMem_realm(realm, clazz->bytes_needed);
      woempa(1, "Allocated a %k at %p, need %d bytes\n", clazz, object->fields, clazz->bytes_needed);
      if (object) {
        object->clazz = clazz;
      }
    }
  }

  /*
  ** If object is still NULL, we are out of memory.
  */
  if (!object) {
    throwOutOfMemoryError(thread);

    return NULL;

  }

  /*
  ** These object fields need to be intialized each time again.
  */
  
  object->flags = (O_CACHEABLE | flags);
  if (isSet(clazz->flags, CLAZZ_HAS_FINALIZER)) {
    woempa(1, "Clazz %k has a finalizer, setting instance at %p FINALIZABLE\n", clazz, object->fields);
    setFlag(object->flags, O_FINALIZABLE);
  }
  if (isSet(clazz->flags, CLAZZ_IS_REFERENCE)) {
    woempa(7, "Clazz %k is reference, setting instance at %p ENQUEUEABLE\n", clazz, object->fields);
    setFlag(object->flags, O_ENQUEUEABLE);
  }

  clearObject(object, clazz->instanceSize);

  /*
  ** If inside a native method, add new instance to local reference table.
  */
  handleLocalRef(thread, object);

  registerObject(object);
  
  //woempa(1, "Allocated instance %p (object %p) of %k, %d words, realm %p.\n", object->fields, object, clazz, clazz->instanceSize, realm);

  //woempa(9, "CACHE class %k has %d avail, %d hits, %d misses, %d instance size...\n", clazz, sizeOfWordset(&clazz->cache), clazz->cache_hits, clazz->cache_miss, clazz->bytes_needed);

  return object->fields;
  
}

w_instance _xxxx_boot_allocInstanceWithFlags(w_thread thread, w_clazz clazz, w_flags flags) {

  w_object object;

  /*
  ** The class must be initialized.
  */
  
  if (clazzObject && clazz != clazzObject && clazzClass && clazz != clazzClass) {
    classMustBeInitialized(clazz, thread);
  }

  checkClazz(clazz);  

  woempa(9, "clazz is %k at %p, requested size is %d words, instance needs %d bytes.\n", clazz, clazz, clazz->instanceSize, clazz->bytes_needed);

  /*
  ** We don't use the cache in the booting version.
  */
  
  object = allocMem_system(clazz->bytes_needed);
  woempa(1, "Allocated a %k at %p, need %d bytes\n", clazz, object->fields, clazz->bytes_needed);
  
  if (object == NULL) {
    wabort(ABORT_WONKA, "Unable to allocate memory for instance!\n");
  }

  object->clazz = clazz;
  object->flags = (O_CACHEABLE | flags);
  if (isSet(clazz->flags, CLAZZ_HAS_FINALIZER)) {
    woempa(1, "Clazz %k has a finalizer, setting instance at %p FINALIZABLE\n", clazz, object->fields);
    setFlag(object->flags, O_FINALIZABLE);
  }
  if (isSet(clazz->flags, CLAZZ_IS_REFERENCE)) {
    woempa(7, "Clazz %k is reference, setting instance at %p ENQUEUEABLE\n", clazz, object->fields);
    setFlag(object->flags, O_ENQUEUEABLE);
  }

  clearObject(object, clazz->instanceSize);

  registerObject(object);
  markTransientRef(object->fields);
  
  woempa(1, "Allocated instance %p (object %p) of %k, %d words\n", object->fields, object, clazz, clazz->instanceSize);

  return object->fields;
}

w_instance _xxxx_norm_allocArrayInstance(w_thread thread, w_clazz clazz, w_size size, w_flags flags) {

  w_object object;
  w_size   bytes_needed;
  w_realm realm = NULL;

  if (thread->group) {
    realm = thread2realm(thread);
  }

  checkClazz(clazz);  
  
  classMustBeInitialized(clazz, thread);

  /*
  ** Calculate the number of bytes we need. Note that checking the requirements with the quota has been
  ** done allready by the calling function.
  */
  
  bytes_needed = sizeof(w_Object) + (size * sizeof(w_word));
  heap_request(realm, (w_int)bytes_needed);

  object = allocMem_realm(realm, bytes_needed);
  woempa(1, "Allocated a %k at %p, need %d bytes\n", clazz, object->fields, bytes_needed);
  
  if (object == NULL) {
    woempa(9, "Unable to allocate memory for instance!\n");

    return NULL;
  }

  if (clazz == atype2clazz[P_byte]) {
    woempa(1, "Allocated %d bytes at %p.\n", bytes_needed, object);
  }

  object->clazz = clazz;
  object->flags = flags;

  clearObject(object, size);

  // heapCheck;

  handleLocalRef(thread, object);

  registerObject(object);

  return object->fields;
  
}

w_instance _xxxx_boot_allocArrayInstance(w_thread thread, w_clazz clazz, w_size size, w_flags flags) {

  w_object object;
  w_size   bytes_needed;

  checkClazz(clazz);  
  
  classMustBeInitialized(clazz, thread);

  if (clazz == atype2clazz[P_byte]) {
    woempa(1, "clazz is %k, requested size is %d words.\n", clazz, size);
  }

  /*
  ** Calculate the number of bytes we need. Note that checking the requirements with the quota has been
  ** done allready by the calling function.
  */
  
  bytes_needed = sizeof(w_Object) + (size * sizeof(w_word));

  object = allocMem_system(bytes_needed);
  woempa(1, "Allocated a %k at %p, need %d bytes\n", clazz, object->fields, bytes_needed);
  
  if (object == NULL) {
    wabort(ABORT_WONKA, "Unable to allocate memory for instance!\n");
  }

  if (clazz == atype2clazz[P_byte]) {
    woempa(1, "Allocated %d bytes at %p.\n", bytes_needed, object);
  }

  object->clazz = clazz;
  object->flags = flags;

  clearObject(object, size);

  // heapCheck;

  handleLocalRef(thread, object);

  registerObject(object);

//  woempa(9, "ARRAY Allocated instance %p (object %p) of %k, %d words.\n", object->fields, object, clazz, size);

  return object->fields;
  
}

static inline w_size round_bytes_to_words(w_int bytes) {
  return (bytes + sizeof(w_word) - 1) / sizeof(w_word);
}

w_instance _allocateArrayInstance(w_thread thread, w_clazz clazz, w_int fixedDimensions, w_word lengths[], w_flags flags) {

  w_realm realm = thread2realm(thread);
  w_int idx;
  w_int i;
  w_int j;
  w_int peel;
  w_int length;
  w_int words;
  w_instance *currentInstances = NULL;
  w_int numCurrent;
  w_aclazz currentClazz;
  w_instance *previousInstances = NULL;
  w_instance instance;
  w_instance subarray;
  w_int numPrevious;
  w_int bytes_needed = 0;
  w_size words_needed;
  w_boolean failed = WONKA_FALSE;

  woempa(1, "Allocating array of %k with %d fixed dimensions.\n", clazz, fixedDimensions);
  length = lengths[fixedDimensions - 1];
  woempa(1, "--> dimension %d has length %d.\n", fixedDimensions, length);
  bytes_needed = ((clazz->aclazz->bits * length + 31) & 0xffffffe0) >> 3;
  woempa(1, "    last dimension will need ((%d*%d+31) & 0xffffffe0) >> 3  = %d bytes each for data\n",clazz->aclazz->bits, length, bytes_needed);
  bytes_needed += sizeof(w_Object) + (F_Array_data * sizeof(w_word));
  woempa(1, "    + %d + %d bytes overhead = %d bytes each\n", F_Array_data * sizeof(w_word), sizeof(w_Object), bytes_needed);
  words_needed = round_bytes_to_words(bytes_needed);
  woempa(1, "    -> %d words each\n", words_needed);
  for (i = 1; i < fixedDimensions; i++) {
    length = lengths[fixedDimensions-i-1];
    woempa(1, "--> dimension %d has length %d.\n", fixedDimensions-i, length);
    words_needed = round_bytes_to_words((w_int)(sizeof(w_Object) + F_Array_data * sizeof(w_word)));
    woempa(1, "    dimension %d will need %d words each\n",fixedDimensions-i, words_needed);
    woempa(1, "    plus %d*%d = %d bytes from dimension %d\n",bytes_needed,length, bytes_needed*length, fixedDimensions-i+1);
    bytes_needed = bytes_needed*length + words_needed*sizeof(w_word);
  }
  woempa(1, "    Total requirements = %d bytes\n",bytes_needed);

  if (! heap_request(realm, bytes_needed)) {
    woempa(9, "Request for %d bytes not allowed for clazz %k.\n", bytes_needed, clazz);
    if (thread->Thread) {
      throwOutOfMemoryError(thread);
    }
    else {
      woempa(9, "Drat.  Can't allocate instance, and no Thread to throw an exception in.\n");
    }

    return NULL;
  }

  peel = 1;
  while (peel <= fixedDimensions && !failed) {
    numCurrent = 1;
    currentClazz = clazz->aclazz;
    for (i = 0; i < fixedDimensions - peel; i++) {
      numCurrent *= lengths[i];
      currentClazz = currentClazz->componentClazz->aclazz;
      wassert(currentClazz);
    }
    length = lengths[i];
    /*
    ** Find the number of bits required and round of to multiple of 32. Then
    ** use this for the number of words required.
    */
    words = (((currentClazz->bits * length) + 31) & 0xffffffe0) >> 5;
    woempa(1, "%d of these instances will be created, with %d words to hold %d items of %d bit(s) each.\n", numCurrent, words, length, currentClazz->bits);
    woempa(1, "One single cell occupies %d bits.\n", currentClazz->bits);
    woempa(1, "Back array descriptor is %k\n", currentClazz->clazz);
    currentInstances = allocMem_realm(realm, numCurrent * sizeof(w_instance *));

    if (!currentInstances) {
      woempa(9, "Well waddaya know.  Couldn't allocate memory for that darned currentInstances thang\n");
      failed = WONKA_TRUE;
    }

    for (i = 0; i < numCurrent && !failed; i++) {
      /*
      ** Allocate the instances; add 1 for the extra F_Array_length word cell...
      */
      instance = _xxxx_allocArrayInstance(thread, currentClazz->clazz, (w_size)(F_Array_data + words), flags);
      if (instance) {
        instance[F_Array_length] = (w_word)length;
        woempa(1, "instance %d at %p, array data begins at %p ends before %p\n",i+1,instance,instance+F_Array_data,instance+F_Array_data+words); 
        currentInstances[i] = instance;
      }
      else {
        woempa(9,"Hey, no room all of a sudden!\n");
// Note that this shouldn't happen often, as we checked our total memory 
// requirements on the way in ...
        failed = WONKA_TRUE;
        break;
      }
    }
    if (previousInstances) {
      idx = 0;
      for (i = 0; i < numCurrent; i++) {
        instance = currentInstances[i];
        for (j = 0; j < (w_int)instance[F_Array_length]; j++) {
          subarray = previousInstances[idx++];
          woempa(1, "Filling front array %d, idx %d with %p\n", j, idx, subarray);
          setArrayReferenceField(instance, subarray, j);
          unsetFlag(instance2object(subarray)->flags, O_GREEN);
        }
      }
      releaseMem(previousInstances);
    }
    peel += 1;
    previousInstances = currentInstances;
    numPrevious = numCurrent;
  }

  if (failed) {
    if (thread->Thread) {
      throwOutOfMemoryError(thread);
    }
    else {
      woempa(9, "Drat.  Can't allocate instance, and no Thread to throw an exception in.\n");
    }

    return NULL;

  }
  else {
    instance = currentInstances[0];
    releaseMem(currentInstances);
  
    woempa(1, "Allocated array of %k (%p) with %d fixed dimensions.\n", clazz, instance, fixedDimensions);


    return instance;

  }  
}

/*
** Define the special function pointers for allocating instances. Preset them for the _boot_ functions.
*/

w_instance (*_xxxx_allocInstance)(w_thread thread, w_clazz clazz);
w_instance (*_xxxx_allocInstanceWithFlags)(w_thread thread, w_clazz clazz, w_flags flags);
w_instance (*_xxxx_allocArrayInstance)(w_thread thread, w_clazz clazz, w_size size, w_flags flags);

w_boolean isValidInstance(w_instance instance) {
  if (((char*)instance) >= PS_start && ((char*)instance) < PS_end) {
    woempa(7, "TODO: put a sensible test in here\n");

    return WONKA_TRUE;

  }
  return !!ht_read(instance_hashtable,(w_word)instance);
}

void newGlobalReference(w_instance instance) {

  w_instance found;

  found = (w_instance) ht_register(globals_hashtable, (w_word)instance);
  newReferenceTo(instance);
  woempa(1, "%s global reference to instance %p.\n", found ? "further" : "new", instance);

}

void startHeap() {
  _xxxx_allocInstance = _xxxx_boot_allocInstance;
  _xxxx_allocInstanceWithFlags = _xxxx_boot_allocInstanceWithFlags;
  _xxxx_allocArrayInstance = _xxxx_boot_allocArrayInstance;
  instance_hashtable = ht_create(W_Realm_system, (char*)"hashtable:instances", 4321, NULL, NULL, 0, 0);
  lock_hashtable = ht_create(W_Realm_system, (char*)"hashtable:locks", 131, NULL, NULL, 0, 0);
  min_heap_free = x_mem_total() /16;
  if (min_heap_free > (4 * 1024 * 1024)) {
    min_heap_free = 4 * 1024 * 1024;
  }
}

