/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

#include "clazz.h"
#include "constant.h"
#include "descriptor.h"
#include "exception.h"
#include "interpreter.h"
#include "jdwp.h"
#include "jdwp_events.h"
#include "loading.h"
#include "locks.h"
#include "methods.h"
#include "threads.h"
#include "wonka.h"
#include <string.h> /* for memcpy */

static void createInterfaceMethodTable(w_thread, w_clazz);
static void createMethodTable(w_thread, w_clazz);
static w_clazz prepareMethodTable(w_clazz);

#define MEMBER_HASHTABLE_SIZE 17

/*
** A class is said to be prepared when:
**  - it is verified;
**  - the superClass (if any) is prepared;
**  - the tables of methods and fields have been prepared;
**  - the CIFs have been created;
**  - the static fields have been initialized to their default values.
**    (Note: the static initializers are not run yet!)
**
** A class must be prepared before it can be resolved.
**
*/

w_instance prepareClass(w_clazz clazz) {
  w_thread   thread = currentWonkaThread;
  w_size     i;
  w_clazz    exception_clazz = NULL;
  w_instance exception = NULL;
  w_instance classLoader = clazz2loader(clazz);

  exception = classMustBeVerified(clazz, thread);
  if (exception) {

    return exception;

  }

  x_monitor_enter(clazz->resolution_monitor, x_eternal);
  if (clazz->prepare != prepareClass) {
    x_monitor_exit(clazz->resolution_monitor);

    return clazz->prepare(clazz);

  }

  if (isSet(clazz->flags, ACC_INTERFACE)) {
    for (i = 0; i < clazz->numInterfaces && exception == NULL; ++i) {
      exception = classMustBePrepared(clazz->interfaces[i], thread);
    }
  }
  else if (clazz->superClazz) {
    exception = classMustBePrepared(clazz->superClazz, thread);
  }

  if (exception) {
    setClazzState(clazz, CLAZZ_STATE_BROKEN);
    clazz->prepare = alreadyBroken;
    clazz->resolve = alreadyBroken;
    clazz->initialize = alreadyBroken;
    x_monitor_exit(clazz->resolution_monitor);

    return exception;

  }

  woempa(1, "Preparing class %k ...\n", clazz);
  setClazzState(clazz, CLAZZ_STATE_PREPARING);

  exception_clazz = createClazzTables(classLoader, clazz);

  woempa(1, "Finished preparing class %k, spreading the news.\n", clazz);
  if (exception_clazz) {
    woempa(7, "Exception in createClazzTables, throwing LinkageError(%k).\n", exception_clazz);
    setClazzState(clazz, CLAZZ_STATE_BROKEN);
    clazz->prepare = alreadyBroken;
    clazz->resolve = alreadyBroken;
    clazz->initialize = alreadyBroken;
    exception = createExceptionInstance(thread, clazzLinkageError, exception_clazz->dotified);
  }
  else {
    setClazzState(clazz, CLAZZ_STATE_PREPARED);
    clazz->prepare = alreadyPrepared;
  }
  x_monitor_exit(clazz->resolution_monitor);

  if(jpda_hooks) {
    jdwp_event_class_prepare(clazz);
  }

  return exception;

}

w_instance alreadyPrepared(w_clazz clazz) {
  woempa(1, "Class %k is already prepared.\n", clazz);

  return NULL;

}

static inline void
prepareDeclaredMethods(w_clazz clazz) {
  w_size   i;
  w_method method;

  woempa(1,"Class %k declares %d methods.\n", clazz,clazz->numMethods);
  if (isNotSet(clazz->flags, ACC_FINAL)) {
    for (i = 0; i < clazz->numMethods; i++) {
      method = clazz->methods[i];
      if (isNotSet(method->flags, ACC_STATIC | ACC_PRIVATE)
          && method->name != string_angle_brackets_init
         ) {
        clazz->numInheritableMethods += 1;
      }
    }
  }
  for (i = 0; i < clazz->numMethods; i++) {
    method = clazz->methods[i];


  }
}

static w_clazz parseMethodDescriptor(w_method method) {
  w_thread thread = currentWonkaThread;
  w_clazz  clazz = method->clazz;
  w_instance exception = NULL;
  w_string desc_string;
  w_string temp_string;
  w_clazz  item_clazz;
  w_size   j;
  w_size   rparen;

  method->arg_types = allocMem(currentWonkaThread, 255 * sizeof(w_clazz));
  if (!method->arg_types) {

    return clazzOutOfMemoryError;

  }

  desc_string = resolveUtf8(clazz, method->desc_index, thread);
  if (desc_string->chars[0] != '(') {
    woempa(9, "Missing '(' in %w\n", desc_string);

      return clazzClassFormatError;

  }

  /*
  ** Skip the leading `(' and start parsing the parameter list.
  */
  j = 1;
  rparen = desc_string->length;

  while(--rparen && desc_string->chars[rparen] != ')');

  if (!rparen) {

    return clazzClassFormatError;

  }

  while (j < rparen && !exception) {
    item_clazz = parseDescriptor(desc_string, &j, rparen, clazz->loader);
    if (item_clazz) {
      if (item_clazz != clazz_void) {
        woempa(1, "Method %k/%w%w arg[%d] has type %k\n", clazz, method->name, desc_string, method->nargs, item_clazz);
        method->arg_types[method->nargs++] = item_clazz;
      }
      else {
        woempa(9, "Method %k/%w%w arg[%d] has type %k\n", clazz, method->name, desc_string, method->nargs, clazz_void);

        return clazzClassFormatError;

      }
    }
    else {

      return exception ? instance2clazz(exception) : clazzNoClassDefFoundError;

    }
  }
  
  
  if (method->nargs) {
     method->arg_types = reallocMem(method->arg_types, method->nargs * sizeof(w_clazz));
  }
  else {
    releaseMem(method->arg_types);
    method->arg_types = NULL;
  }
  
    method->param_list = unicode2String(currentWonkaThread, desc_string->chars, rparen + 1);
    woempa(1, "Method %k/%w%w param_list is %w\n", clazz, method->name, desc_string, method->param_list);
    j = rparen + 1;
    item_clazz = parseDescriptor(desc_string, &j, desc_string->length, clazz->loader);
    temp_string = unicode2String(currentWonkaThread, &desc_string->chars[rparen+1], desc_string->length - rparen - 1);

    if (item_clazz) {
      woempa(1, "Method %k/%w%w return type is %k\n", clazz, method->name, desc_string, item_clazz);
      method->return_type = item_clazz;
    }
    else {

      return clazzClassFormatError;

    }
  

  return NULL;
}

w_clazz // Exception class or NULL
createClazzTables(w_instance classLoader, w_clazz clazz) {
  w_thread thread = currentWonkaThread;
  w_size   i;
  w_method method;
  w_clazz  exception_clazz = NULL;

  for (i = 0; i < clazz->numMethods && !exception_clazz; i++) {
    method = clazz->methods[i];
    exception_clazz = parseMethodDescriptor(method);

    if (exception_clazz) {

      return exception_clazz;

    }

    /*
    ** If the method is native we use the stack_i variable
    ** to store the number of words the native stack should grow.
    ** For non static functions, it is 2 (env and status pointers)
    ** for static functions, it needs third extra spot for the Class
    ** instance that is passed.
    */

    if (isSet(method->flags, ACC_NATIVE)) {
      if (isSet(method->flags, ACC_STATIC)) {
        method->stack_i = 3;
      }
      else {
        method->stack_i = 2;
      }
    }

    if (isSet(clazz->flags, ACC_NATIVE)) {
      woempa(1, "Method %M is native.\n", method);
      method->flags |= METHOD_NEEDS_LINKING;
    }
    else if (method->name == string_angle_brackets_clinit) {
      woempa(1, "Method %M is a clinit.\n", method);
      method->flags |= METHOD_IS_CLINIT;
// TODO: only one clinit ...
      clazz->clinit = method;
    }
    else if (method->name == string_angle_brackets_init) {
      woempa(1, "Method %M is a constructor.\n", method);
      method->flags |= METHOD_IS_CONSTRUCTOR;
    }
  }

  prepareDeclaredMethods(clazz);

  /*
  ** If this a regular class, create its method table; if it is
  ** an interface, create the interface method table instead.
  ** The method table contains all methods declared by this class
  ** or any of its superclasses.  The interface method table contains
  ** all interface methods declared by this interface or by any of
  ** its superinterfaces.
  */

  if (clazz->superClazz == NULL) {
    woempa(1, "Creating method table for top class %k: %d slots, %d are inheritable.\n",  clazz, clazz->numMethods, clazz->numInheritableMethods);
  }
  else if (isSet(clazz->flags, ACC_INTERFACE)) {
    createInterfaceMethodTable(thread, clazz);
  }
  else {
    createMethodTable(thread, clazz);
  }


  exception_clazz = prepareMethodTable(clazz);
  // methodTableDump(clazz);

  /*
  ** Create the correct field offsets and the
  ** tablespace for the static data.
  */

  if (exception_clazz == NULL) {
    createFieldTable(thread,clazz);
    // fieldTableDump(clazz);
  }

  return exception_clazz;
}

/*
** For an interface, we coalesce all the methods of the superinterfaces
** into our own method table.  This means that when we come to build
** the interface hashtable of a class we will only need to look at the
** method tables of its immediate superinterface, instead of fanning out
** recursively ...
*/

static void createInterfaceMethodTable(w_thread thread, w_clazz clazz) {

  w_size i;
  w_size j;
  w_int extraSlots;
  w_int offset;
  w_method method;
  w_clazz interfaze;
  w_method *newMethodTable;

  woempa(1,"Creating interface table for %k\n", clazz);
  extraSlots = 0;
  woempa(1, "Interface %k has %d own methods and %d super interfaces.\n",  clazz, clazz->numMethods, clazz->numInterfaces);
  for (i = 0; i < clazz->numInterfaces; i++) {
    interfaze = clazz->interfaces[i];
    woempa(1, "%d:  Superinterface %k has %d methods (including those of its superinterfaces).\n",i+1, interfaze,interfaze->numMethods);
    extraSlots += interfaze->numMethods;
  }


  offset = 0;
  if (clazz->numMethods > 0 && extraSlots > 0) {
    woempa(1, "Allocating %d extra slots for interface %k.\n",extraSlots, clazz);
    i = clazz->numMethods + extraSlots;
    newMethodTable = allocMem(thread, i * sizeof(w_method));
    /*
    ** Copy over all method arrays of super interfaces. Note that we
    ** have to make a copy of the methods of each superinterface since
    ** they live in different slot offsets in each interface; for each
    ** copy, we then adjust the clazz pointer and the slot to point
    ** to out current class and slot.
    **
    ** Better still, ignore the preceding note and note that the slot
    ** number of an interface method relates to the interface in which it
    ** was defined, which may be a superinterface of the current
    ** interface.  So we don't make a copy of the w_Method structure,
    ** and we leave the clazz pointer as is ...
    **   (Actually I don't believe the slot of an interface method is
    **    ever used for anything anyway...)
    **
    ** TODO: in theory an interface method can override a superinterface
    ** method.  Maybe this is the one thing for which interface method slot
    ** numbers are useful ...
    */
    for (i = 0; i < clazz->numInterfaces; i++) {
      interfaze = clazz->interfaces[i];
      for (j = 0; j < interfaze->numMethods; j++) {
        newMethodTable[j + offset] = interfaze->methods[j];
      }
      offset += interfaze->numMethods;
    }
    /*
    ** Copy over our own method table but don't adjust the offset
    ** since we will use it to loop over our own methods and adjust the
    ** method slots since they are not yet assigned.
    */
    memcpy(newMethodTable + offset, clazz->methods, (w_size)(clazz->numMethods * sizeof(w_method)));
    for (i = 0; i < clazz->numMethods; i++) {
      method = clazz->methods[i];
      method->slot = i + offset;
    }
    clazz->numMethods += extraSlots;
    clazz->numInheritableMethods = clazz->numMethods;
    releaseMem(clazz->methods);
    clazz->methods = newMethodTable;
    woempa(1,"Interface %k inherited %d methods from superinterfaces, total is %d\n", clazz,extraSlots,clazz->numMethods);
  }
  else {
    /*
    ** No need for growth. Just loop over the methods
    ** and record their slots.
    */
    for (i = 0; i < clazz->numMethods; i++) {
      method = clazz->methods[i];
      method->slot = i;
    }
  }
}

/*
** For a "normal" class (not java/lang/Object, not an interface) we have
** to consider
*/

static void createMethodTable(w_thread thread, w_clazz clazz) {

  w_size i;
  w_size j;
  w_int extraSlots = 0;
  w_method method1;
  w_method method2;
  w_method *newMethodTable;

  /*
  ** Check what number of slots we need to grow our method table with.
  ** Loop over the super class methods and see how many we override and
  ** how many we don't override. If we override a super class method,
  ** we record the slot offset allready.
  */
  woempa(1,"class %k: superclass %k has %d inheritable methods\n", clazz, clazz->superClazz,clazz->superClazz->numInheritableMethods);
  for (i = 0; i < clazz->superClazz->numInheritableMethods; i++) {
    method1 = clazz->superClazz->methods[i];
    for (j = 0; j < clazz->numMethods; j++) {
      method2 = clazz->methods[j];
      if (method2->name == method1->name) {
        /*
        ** Names match allready, check the descriptors.
        */
        if (method2->param_list == method1->param_list) {
          /*
          ** Method is overriden, record the offset allready in the method.
          */
          woempa(1, "Method %w of %k \n", NM(method2),  method2->clazz);
          woempa(1, "  overrides method of %k, slot is %d.\n",  method1->clazz, i);
          method2->slot = i;
          break;
        }
      }
    }
    if (j == clazz->numMethods) {
      extraSlots += 1;
    }
  }

  if (extraSlots != 0) {
    /*
    ** We need to modify our existing table...
    **
    ** 1) Allocate a new table with the correct size .
    ** 2) Copy over the parent class inheritable methods to our new table.
    ** 3) Loop over our old method table, inserting the overriding method
    **    slots in there
    **    OR allocate a new slot in the new table and record it in the method.
    ** 4) Adjust our method counters.
    ** 5) Free our old method table and replace it with the new one.
    */
    i = clazz->numMethods + extraSlots;
    woempa(1, "Creating space for %d method slots. Need %d extra slots.\n", i, extraSlots);
    newMethodTable = allocMem(thread, i * sizeof(w_method));
    i = clazz->superClazz->numInheritableMethods * sizeof(w_method);
    memcpy(newMethodTable, clazz->superClazz->methods, (w_size)i);
    j = clazz->superClazz->numInheritableMethods;
    for (i = 0; i < clazz->numMethods; i++) {
      method1 = clazz->methods[i];
      if (method1->slot != SLOT_NOT_ALLOCATED) {
        newMethodTable[method1->slot] = method1;
      }
      else {
        newMethodTable[j] = method1;
        method1->slot = j;
        j += 1;
      }
    }
    clazz->numMethods += extraSlots;
    clazz->numInheritableMethods += extraSlots;
    releaseMem(clazz->methods);
    clazz->methods = newMethodTable;
    woempa(1,"Class %k inherited %d methods from superclasses, total is %d\n", clazz,extraSlots,clazz->numMethods);
  }
}

void
createFieldTable(w_thread thread, w_clazz clazz) {
  w_int fieldsOffset = 0;
  w_int staticOffset = 0;
  w_size i;
  w_field field;
  w_fixup fixup;

  woempa(1,"creating field table for class %k\n", clazz);
  /*
  ** Calculate the size in words of an instance of this class
  ** and fill in the correct offsets for each field of this
  ** class.
  */

  if (clazz->superClazz != NULL) {
    fieldsOffset = clazz->superClazz->instanceSize;
    woempa(1,"superclass %k has instance size %d words and %d words of static fields\n", clazz->superClazz,fieldsOffset,staticOffset);
  }
  else {
    fieldsOffset = 0;
  }
  staticOffset = 0;

  for (i = 0; i < clazz->numStaticFields; i++) {
    field = &clazz->fields[i];
    woempa(1,"static field '%w' at slot[%d]\n",NM(field),staticOffset);
    field->slot = staticOffset;
    staticOffset +=  fieldSize(field);
  }
  for (;i < clazz->numFields; i++) {
    field = &clazz->fields[i];
    woempa(1,"instance field '%w' at slot[%d]\n",NM(field),fieldsOffset);
    field->slot = fieldsOffset;
    fieldsOffset +=  fieldSize(field);
  }

  clazz->instanceSize = fieldsOffset;
  clazz->numStaticWords = staticOffset;

  /*
  ** Fixup the clazz for wotsit fields
  ** No need to lock fixup_hashtable, as all writes preceded all reads
  */

  fixup = (w_fixup)ht_read_no_lock(fixup_hashtable,(w_word) clazz->dotified);
  woempa(1,"fixup for %k = %p\n", clazz,fixup);
  if (fixup) {
    fixup(clazz);
  }

  /*
  ** Only set the number of bytes needed after fixing up the clazz since instanceSize
  ** can grow during fixup.
  */

  clazz->bytes_needed = (clazz->instanceSize * sizeof(w_word)) + sizeof(w_Object);

  woempa(1,"class %k has instance size %d words and %d words of static fields\n", clazz,clazz->instanceSize,clazz->numStaticWords);

}

/** Swap the i'th and j'th methods of clazz, updating method->slot accordingly.
*/
void
swapMethods(w_clazz clazz, w_size i, w_size j) {
  w_method method1 = clazz->methods[i];
  w_method method2 = clazz->methods[j];

  clazz->methods[i] = method2;
  method2->slot = i;
  clazz->methods[j] = method1;
  method1->slot = j;
}

static int clazz2words(w_clazz clazz) {
  return (clazz == clazz_void) ? 0 : (clazz == clazz_long || clazz == clazz_double) ? 2 : 1;
}

/** Convert the method table from its unprepared to its prepared form.
**
**   - sort the method table so that all inherited methods remain at
**     the top, followed by all inheritable methods of this class,
**     followed by non-inheritable instance methods, with static
**     methods bringing up the rear.
*/

static w_clazz // Exception clazz or NULL 
prepareMethodTable(w_clazz clazz) {
  w_size i;
  w_size j;
  w_size numInheritedMethods;
  w_size numStaticMethods;
  w_size numPrivateMethods;
  w_method method1;
  w_method method2;
  w_int disp_row;  /* The row of the dispatcher matrix for this method. */
  w_int disp_col;  /* The column of the dispatcher matrix for this method. */

  numInheritedMethods = 0;
  if (isSet(clazz->flags, ACC_INTERFACE)) {
    for (i = 0; i < clazz->numInterfaces; ++i) {
      numInheritedMethods += clazz->interfaces[i]->numInheritableMethods;
    }
  }
  else if (clazz->superClazz) {
    numInheritedMethods = clazz->superClazz->numInheritableMethods;
  }

  numStaticMethods = 0;
  i = numInheritedMethods;
  while (i < clazz->numMethods - numStaticMethods) {
    method1 = clazz->methods[i];
    if (isSet(method1->flags,ACC_STATIC)) {
      ++numStaticMethods;
      woempa(1,"Method[%d] %M is static\n",i,method1);
      for (j = clazz->numMethods - numStaticMethods; j > i; --j) {
        method2 = clazz->methods[j];
        if (isNotSet(method2->flags,ACC_STATIC)) {
          woempa(1,"  Method[%d] %m is non-static\n",j,method2);
          swapMethods(clazz, i, j);
          woempa(1,"  --> swapped, %w is now slot[%d], %w is slot[%d], numStaticMethods is %d\n",NM(method1),method1->slot,NM(method2),method2->slot,numStaticMethods);
          break;
        }
        else {
          ++numStaticMethods;
          woempa(1,"  Method[%d] %w is static\n",j,method2->name, method2->param_list);
        }
      }
    }
    ++i;
  }

  numPrivateMethods = 0;
  i = numInheritedMethods;
  while (i < clazz->numMethods - numStaticMethods - numPrivateMethods) {
    method1 = clazz->methods[i];
    if (isSet(method1->flags,ACC_PRIVATE)
       || (method1->name == string_angle_brackets_init)
    ) {
      woempa(1,"Method[%d] %M is private/<init>\n",i,method1);
      ++numPrivateMethods;
      for (j = clazz->numMethods - numStaticMethods  - numPrivateMethods; j > i; --j) {
        method2 = clazz->methods[j];
        if (isNotSet(method2->flags,ACC_PRIVATE)
          && (method2->name != string_angle_brackets_init)
        ) {
          woempa(1,"  Method[%d] %m is non-private\n",j,method2);
          swapMethods(clazz, i, j);
          woempa(1,"  --> swapped, %m is now slot[%d], %m is slot[%d], numPrivateMethods is %d\n",method1,method1->slot,method2,method2->slot,numPrivateMethods);
          break;
        }
        else {
          ++numPrivateMethods;
          woempa(1,"  Method[%d] %m is private/<init>\n",j,method2);
        }
      }
    }
    else {
      woempa(1,"Method[%d] %M is non-private\n",i,method1);
      method1->slot = i;
    }
    ++i;
  }

  woempa(7,"clazz(%k)->numMethods is %d, of which %d inherited; numStaticMethods is %d, numPrivateMethods is %d, clazz->numInheritableMethods is %d\n",clazz, clazz->numMethods,numInheritedMethods,numStaticMethods,numPrivateMethods,clazz->numInheritableMethods);

//  reorderMethodTable(clazz, clazz->superClazz ? clazz->superClazz->numInheritableMethods : 0);

  /*
  ** Determine which dispatcher tuple for call/return we are going to give
  ** each method.  These are determined by a small calculation. See the
  ** table in dispatcher.c for more information on how it is calculated.
  ** Since we are loading, only the  non-profiling versions should be used,
  ** and only native or byte code  can be selected. So the choice is between
  ** column 1 (native, unprofiled) or column 3 (byte code, unprofiled).
  ** When this is a real byte code method (code != NULL) see if it does
  ** something more than just 'void return' (opcode 0xb1). If that is all
  ** it does, adjust the dispatcher to our special handler. We DON'T do
  ** this when the method is SYNCHRONIZED in case somebody wants this
  ** method to be used as a fencepost...
  ** [CG 20010411] Why not? 8-> They get what they deserve ...
  */
  i = 0;
  while (i < clazz->numMethods - numStaticMethods) {
    method1 = clazz->methods[i];

    if (method1->dispatcher) {
      woempa(1,"Instance method %m already has a dispatcher\n",method1);
    }
    else if (method1->code && method1->code[0] == 0xb1) {
      woempa(1,"Instance method %m does nothing, use dummy dispatcher\n",method1);
      method1->arg_i = 1;
      for (j = 0; j < (w_size)method1->nargs; j++) {
        method1->arg_i +=  clazz2words(method1->arg_types[j]);
      }
      method1->return_i = 0;
      method1->dispatcher = &dispatch_vreturn;
      method1->function = (w_function)(w_void_fun)NULL;
    }
    else {
      woempa(1,"Setting up stack and local variable information for %M of %k\n",method1);
      method1->arg_i = 1; 
      for (j = 0; j < (w_size)method1->nargs; j++) {
        method1->arg_i +=  clazz2words(method1->arg_types[j]);
      }
      method1->return_i = clazz2words(method1->return_type);
      woempa(1,"Instance method %m: arg_i = %d, return_i = %d\n",method1,method1->arg_i,method1->return_i);

      disp_row =
            (isSet(method1->flags, ACC_SYNCHRONIZED) ? 0 : 3)
          + method1->return_i;

      if (isSet(method1->flags, ACC_NATIVE)) {
        disp_col = 1;
      }
      else {
        #ifdef JSPOT
          disp_col = 2;
        #else
          disp_col = 3;
        #endif
      }
      woempa(1,"Instance method %m: dispatcher[%d,%d]\n",method1,disp_row,disp_col);
      method1->dispatcher = &dispatchers[disp_row][disp_col];
      method1->function = (w_function)(w_void_fun)NULL;

    }
    ++i;
  }

  while (i < clazz->numMethods) {
    method1 = clazz->methods[i];

    if (method1->dispatcher) {
      woempa(1,"Static method %m already has a dispatcher\n",method1);
    }
    else if (method1->code && method1->code[0] == 0xb1) {
      woempa(1,"Static method %m does nothing, use dummy dispatcher\n",method1);
      method1->arg_i = 0;
      for (j = 0; j < (w_size)method1->nargs; j++) {
        method1->arg_i +=  clazz2words(method1->arg_types[j]);
      }
      method1->return_i = 0;
      method1->dispatcher = &dispatch_vreturn;
    }
    else {
      woempa(1,"Setting up stack and local variable information for %M\n",method1);
      method1->arg_i = 0;
      for (j = 0; j < (w_size)method1->nargs; j++) {
        method1->arg_i +=  clazz2words(method1->arg_types[j]);
      }
      method1->return_i =  clazz2words(method1->return_type);
      woempa(1,"Static method %m: arg_i = %d, return_i = %d\n",method1,method1->arg_i,method1->return_i);
      disp_row = 6
        + (isSet(method1->flags, ACC_SYNCHRONIZED) ? 0 : 3)
        + method1->return_i;

      if (isSet(method1->flags, ACC_NATIVE)) {
        disp_col = 1;
      }
      else {
        #ifdef JSPOT
          disp_col = 2;
        #else
          disp_col = 3;
        #endif
      }

      woempa(1,"Static method %m: dispatcher[%d,%d]\n",method1,disp_row,disp_col);
      method1->dispatcher = &dispatchers[disp_row][disp_col];
      method1->function = (w_function)(w_void_fun)NULL;
    }

    ++i;
  }

  return NULL;
}

