/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

#include "clazz.h"
#include "checks.h"
#include "exception.h"
#include "loading.h"
#include "threads.h"
#include "ts-mem.h"
#include "wonka.h"
#include <string.h>  /* for strlen */

/*
** References of the form [J+JVM ...] are to the book _Java and the Java
** Virtual Machine_ by Robert Staerk, Joachim Schmid, and Egon Boerger.
*/

/*
** Generate an instance of VerifyError using the name of the failing class
** and two strings, one C (ISO 8859-1) and one Java.
*/
w_instance createVerifyError(w_clazz failing_clazz, char *msg, w_string str) {
  char    *cptr;
  w_char  *buffer;
  w_size   length;
  w_string message;
  w_instance result;
  w_thread thread = currentWonkaThread;

  length = failing_clazz->dotified->length;
  if (msg) {
    length += 2 + strlen(msg);
  }
  if (str) {
    length += 2 + str->length;
  }

  buffer = allocMem_system(length * sizeof(w_char));
  woempa(7, "buffer: %p, length %d\n", buffer, length);

  length = w_string2chars(failing_clazz->dotified, buffer);
  woempa(7, "Wrote `%w', length now %d chars\n", failing_clazz->dotified, length);
  if (msg) {
    buffer[length++] = ':';
    buffer[length++] = ' ';
    for (cptr = msg; *cptr;) {
      buffer[length++] = *cptr++;
    }
    woempa(7, "Wrote `%s', length now %d chars\n", msg, length);
  }
  if (str) {
    buffer[length++] = ':';
    buffer[length++] = ' ';
    length += w_string2chars(str, buffer + length);
    woempa(7, "Wrote `%w', length now %d chars\n", str, length);
  }

  message = unicode2String(thread, buffer, length);
  releaseMem(buffer);
  woempa(9,"%w: VerifyError: %w\n", failing_clazz->dotified, message);
  result = createExceptionInstance(thread, clazzVerifyError, message);
  failing_clazz->failure_message = message;

  return result;
}

/*
** Verify that the class heirarchy (superclass, superinterface) of
** a class is in order.  Note that we assume that the superclass
** and superinterface are already loaded.
** Returns an instance of VerifyError or NULL.
** TODO: extend this for exceptions and inner classes?
*/
w_instance checkHierarchy(w_clazz clazz) {
  w_clazz super_clazz = clazz->superClazz;
  w_clazz super_interface;
//  w_fifo  interface_fifo;
  w_size i;
  w_size j;

  /*
  ** [J+JVM Constraint 4.1.1, item 1 / 4.1.2, item 1]
  */
  if(super_clazz && isSet(super_clazz->flags, ACC_INTERFACE)){

    return createVerifyError(clazz, "superclass is interface",  super_clazz->dotified);

  }

  for (i = 0; i < clazz->numInterfaces; ++i) {
    super_interface = clazz->interfaces[i];
    if (isNotSet(super_interface->flags, ACC_INTERFACE)){

      return createVerifyError(clazz, "superinterface is not interface",  super_interface->dotified);

    }
    for (j = i + 1; j < clazz->numInterfaces; ++j) {
      if (super_interface == clazz->interfaces[j]) {

        return createVerifyError(clazz, "duplicate superinterface",  super_interface->dotified);

      }
    }
  }

  /*
  ** [J+JVM Constraint 4.1.1, item 2]
  */
  if(super_clazz && isSet(super_clazz->flags, ACC_FINAL)){

    return createVerifyError(clazz, "superclass is FINAL",  super_clazz->dotified);

  }

  /*
  ** [J+JVM Constraint 4.1.1, item 3]
  */
  if (isSet(clazz->flags, ACC_FINAL) && isSet(clazz->flags, ACC_ABSTRACT)) {

    return createVerifyError(clazz, "class is both FINAL and ABSTRACT",  clazz->dotified);

  }

  /*
  ** [J+JVM Constraint 4.1.4]
  */
  super_clazz = clazz->superClazz;
  if(super_clazz && isNotSet(super_clazz->flags, ACC_PUBLIC) && !sameRuntimePackage(clazz, super_clazz)) {

    return createVerifyError(clazz, "superclass is neither PUBLIC nor in same package",  super_clazz->dotified);

  }
  for (i = 0; i < clazz->numInterfaces; ++i) {
    super_interface = clazz->interfaces[i];
    if(isNotSet(super_interface->flags, ACC_PUBLIC) && !sameRuntimePackage(clazz, super_interface)) {

      return createVerifyError(clazz, "superinterface is neither PUBLIC nor in same package",  super_interface->dotified);

    }
  }

  return NULL;
}

/*
** So far we do not perform any verification ...
** TODO: go yak-trekking in Mongolia
*/
w_instance verifyClass(w_clazz clazz) {
  w_instance verify_error;

  x_monitor_enter(clazz->resolution_monitor, x_eternal);
  if (clazz->verify != verifyClass) {
    x_monitor_exit(clazz->resolution_monitor);

    return NULL;

  }

  woempa(1, "Verifying class %k\n", clazz);
  verify_error = checkHierarchy(clazz);
  if (verify_error) {
    setClazzState(clazz, CLAZZ_STATE_BROKEN);
    clazz->verify = verifyFailed;
    clazz->prepare = alreadyBroken;
    clazz->resolve = alreadyBroken;
    clazz->initialize = alreadyBroken;

    x_monitor_exit(clazz->resolution_monitor);

    return verify_error;

  }


  woempa(1, "Verified class %k\n", clazz);
  setClazzState(clazz, CLAZZ_STATE_VERIFIED);
  clazz->verify = alreadyVerified;
  x_monitor_exit(clazz->resolution_monitor);

  return NULL;
}

w_instance alreadyVerified(w_clazz clazz) {
  woempa(1,"(class %k is already verified...)\n", clazz);

  return NULL;
}

w_instance verifyFailed(w_clazz clazz) {

  return createExceptionInstance(currentWonkaThread, clazzVerifyError, clazz->failure_message);

}

