/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: DeviceInputStream.c,v 1.3 2002/02/20 08:55:25 dfrederickx Exp $
*/

#include "arrays.h"
#include "core-classes.h"
#include "comms.h"
#include "exception.h"
#include "hashtable.h"
#include "heap.h"
#include "ts-mem.h"
#include "strings.h"
#include "threads.h"
#include "driver_byteserial.h"

/**@name File DeviceInputStream.c
*/
//@{

///
/**@name Helper function for object constructor
*/
//@{
/// Create an InputStream by name
/**@type native method
** @name createFromString()I
** Initialise an instance of DeviceInputStream using the named Device.
** Called from the constructor \textsf{DeviceInputStream(String)}.
** Returns the unique id given by registerDevice().
*/
w_int
DeviceInputStream_createFromString
(  JNIEnv *env, w_instance thisInputStream,
  w_instance nameString
) {
  w_thread   thread = JNIEnv2w_thread(env);
  w_string   nameStr;
  w_device commport;
  
  w_ubyte    *name;
  w_ubyte    *dummy = NULL;
  w_int      i;

  if (isNullInstance(nameString)) {
    throwException(thread,clazzNullPointerException,NULL);
    return 0;
  }

  nameStr = String2string(nameString);
  name = allocMem(thread, nameStr->length + 1);
 
  dummy = (w_ubyte *)name;
  for (i = 0; i < (w_int)nameStr->length; ++i) {
     (*dummy) = nameStr->chars[i];
     dummy++;
  }
  (*dummy) = '\0';

  commport = getDeviceByName(name);
  if (commport->type != wdt_byte_serial) {
    woempa(9,"Device %s is of wrong type! (%d s/b %d)\n", name, commport->type, wdt_byte_serial);
    commport = NULL;
  }
  releaseMem(name);
  if (commport) {
    if (deviceBSOpen(name, wdp_read) == NULL) {
      throwException(thread,clazzIOException,NULL);
    }
    return (w_int)commport;
  }
  else {
    throwException(thread,clazzIllegalArgumentException,NULL);
  }
  return 0;
}
//@}


/// Read bytes into an array                
/**@type native method
** @name readIntoBuffer([BII)I                    
** @name read(I)I                                 
** @param devid - the device id of the input stream
** @param theByteArray - where the bytes read will be placed
** @param offset - from the start of theByteArray
** @param length - the maximum length to read
** @returns the number of bytes actually read
** Read bytes from the input stream into an array.
*/
w_int
DeviceInputStream_readIntoBuffer
(  JNIEnv *env, w_instance thisInputStream,
  w_int    devid,
  w_instance theByteArray, w_int offset, w_int length
) {
  w_thread   thread = JNIEnv2w_thread(env);
  w_device device = (w_device)devid;
  w_sbyte    *byte_array  = instance2Array_byte(theByteArray);
  w_sbyte    *dest  = byte_array + offset;
  w_size     lread;

  if (length<0 || offset<0 || length+offset > (w_int)theByteArray[F_Array_length]) {
    throwException(thread,clazzArrayIndexOutOfBoundsException,NULL);
    return -1;
  }

  if (deviceBSRead(device, dest, length, &lread, x_eternal) != wds_success) {
    throwException(thread,clazzIOException,NULL);
  }

  return lread;
}


/// Skip bytes                 
/**@type native method
** @name skip0(IL)L                    
** @param devid - the device id of the input stream
** @param n - the number of bytes to skip
** @returns the number of bytes skipped (can be less than n).
** Skip bytes from the input stream.
** We do this by asking the underlying driver to ``read into nowhere''.
*/
w_long
DeviceInputStream_skip0
(  JNIEnv *env, w_instance thisInputStream,
  w_int    devid,
  w_long n
) {
  w_thread   thread = JNIEnv2w_thread(env);
  w_device device = (w_device)devid;
  w_long     l = 0;
  w_size     lread;
  w_int      rc = wds_success;
  w_ubyte    *buffer = allocMem(thread, (w_size)(l > 32768 ? 32768 : l));

  for (l=0;l<n-32767 && rc==0;l+=32767) {
    rc = deviceBSRead(device, buffer, 32767, &lread, x_eternal);
  }
  rc = deviceBSRead(device, buffer, (w_int)l, &lread, x_eternal);

  if (rc != wds_success) {
    throwException(thread,clazzIOException,NULL);
  }

  releaseMem(buffer);
  
  return l;
}
//@}

/**@name Miscellaneous support functions
*/
//@{
///             
/**@type native method
** @name available(I)I                    
** @param devid - the device id of the input stream
** @returns the number of bytes waiting to be read
** Ask the underlying driver how many bytes of input are waiting to be read.
*/
w_int
DeviceInputStream_available
(  JNIEnv *env, w_instance thisInputStream,
  w_int    devid
) {
  w_thread   thread = JNIEnv2w_thread(env);
  w_device device = (w_device)devid;
  w_word     temp;

  if (deviceBSQuery(device, wdi_get_available, &temp, x_eternal) != wds_success) {
    throwException(thread,clazzIOException,NULL);
  }

  return temp;
}

//@}
//@}
