/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: exception.c,v 1.17 2002/03/26 10:44:15 gray Exp $
*/

#include <string.h>
#include <stdarg.h>
#include <stdio.h>

#include "clazz.h"
#include "constant.h"
#include "exception.h"
#include "heap.h"
#include "interpreter.h"
#include "loading.h"
#include "realms.h"
#include "oswald.h"
#include "Throwable.h"
#include "stack.h"
#include "strings.h"
#include "threads.h"
#include "fields.h"
#include "ts-mem.h"
#include "methods.h"
#include "descriptor.h"
#include "core-classes.h"
#include "checks.h"

#ifdef DEBUG
void dumpStackTrace (w_instance thisThrowable);
#else
#define dumpStackTrace(thrown)
#endif

#define EXCEPTION_STRING_BUFFER_SIZE 1024

void setLocalException(w_thread thread, w_instance Exception) {

  if (!thread->Thread) {
    wabort(ABORT_WONKA,"No instance of Thread found, so couldn't throw %k in %w\n",instance2clazz(Exception),NM(thread));
  }

  woempa(1,"Throwing %k in %w\n",instance2clazz(Exception),NM(thread));
  if (isNotSet(thread->flags, (WT_THREAD_THROWN | WT_THREAD_KILLED))) {
    setFlag(thread->flags, WT_THREAD_THROWN);
    setReferenceField(thread->Thread, Exception, F_Thread_thrown);
  }
}

void throwOutOfMemoryError(w_thread thread) {
  if (thread->Thread && thread->Thread[F_Thread_thrown] == 0) {
    setFlag(thread->flags, WT_THREAD_THROWN);

    /*
    ** thread->Thread[F_Thread_theOutOfMemoryError] can sometimes be NULL. this causes an abort later on (f.i. in interpreter)
    ** because WT_THREAD_THROWN is set but no exception was found to be thrown.
    */
    if(!((w_instance)thread->Thread[F_Thread_theOutOfMemoryError])){
      woempa(9,"FREE MEMORY %i TOTAL MEMORY %i\n",x_mem_avail(),  x_mem_total());
      thread->Thread[F_Thread_theOutOfMemoryError] = (w_int)allocInstance(thread, clazzOutOfMemoryError);
    }
    setReferenceField(thread->Thread,  (w_instance)thread->Thread[F_Thread_theOutOfMemoryError], F_Thread_thrown);
  }
}

void throwException(w_thread thread, w_clazz exception, w_string message) {
  w_instance throwable;
  w_instance stringInstance;
    
  if (thread->Thread && thread->Thread[F_Thread_thrown] == 0) {
    throwable = classMustBeInitialized(exception, thread);

    if (throwable == NULL) {
      throwable = allocInstance(thread, exception);
    }

    if (throwable) {
      if (message) {
        stringInstance = getStringInstance(thread, message);
        setReferenceField(throwable, stringInstance, F_Throwable_detailMessage);
      }
      setLocalException(thread,throwable);
      dumpStackTrace(throwable);
    }
    else {
      throwOutOfMemoryError(thread);
    }
  }
}

void throwException_C(w_thread thread, w_clazz exception, char *message) {
  w_string string = NULL;

  if (message) {
    string = cstring2String(thread, message, strlen(message));
  }
  throwException(thread, exception, string);
  if (string) {
    deregisterString(string);
  }
}

void throwLocalException(w_thread thread, w_clazz exception, const char *file, const char *function, w_int line, const char *fmt, ...) {

  va_list ap;
  w_instance throwable;
  w_string string;
  w_instance stringInstance;
  char *buffer;
  w_size bufsize = EXCEPTION_STRING_BUFFER_SIZE;
  
  throwable = classMustBeInitialized(exception, thread);
  if (throwable == NULL) {
    throwable = allocInstance(thread, exception);
  }

  if (throwable) {
    buffer = allocMem(thread, EXCEPTION_STRING_BUFFER_SIZE * sizeof(w_byte));
    x_snprintf(buffer, EXCEPTION_STRING_BUFFER_SIZE, "%s %s %d: ", file, function, line);
    va_start(ap, fmt);
    x_vsnprintf(buffer + strlen(buffer), bufsize, fmt, ap);
    va_end(ap);

    string = cstring2String(thread, buffer, strlen(buffer));
    releaseMem(buffer);

    stringInstance = getStringInstance(thread, string);
    // we just registered string twice ...
    deregisterString(string);
    setReferenceField(throwable, stringInstance, F_Throwable_detailMessage);

    setLocalException(thread,throwable);
  }
  else {
    throwOutOfMemoryError(thread);
  }

}

void fillInStackTrace(w_instance thisThrowable) {

  w_frame top = (w_frame)thisThrowable[F_Throwable_frame];

  destroyStackTrace((w_trace)thisThrowable[F_Throwable_trace]);

  if (top == NULL) {
    woempa(1, "Saved top is NULL, assuming JNIEnv2w_thread(env)->top as topframe.\n");
    top = currentWonkaThread->top;
    thisThrowable[F_Throwable_frame] = (w_word)top;
  }
  else {
    woempa(1, "Top %p was saved in frame.\n", top);
  }
  
  thisThrowable[F_Throwable_trace] = (w_word)createStackTrace(top);

}

w_frame handleThrowable(w_instance thrown, w_frame throw_frame, w_code throw_code, w_code *new_code) {
  w_thread thread = throw_frame->thread;
  w_frame  catch_frame;
  w_code   catch_code;
  w_clazz  catch_clazz;
  w_int i;
  w_exception exception;
  //w_Item item;
  w_word item;
  w_method method = throw_frame->method; // current method when seeking handler
  w_pc     pc;

  woempa(1, "Handling a %k at [%05d] in method %M\n", instance2clazz(thrown), throw_code - method->code, method);

  if (thrown[F_Throwable_detailMessage] != (w_word)nullInstance) {
    wassert(instance2clazz((w_instance)thrown[F_Throwable_detailMessage])==clazzString);
    woempa(1, "Exception message '%w'\n", String2string((w_instance)thrown[F_Throwable_detailMessage]));
  }

  /*
  ** We save the frame pointer where this exception occured; we want to handle 
  ** exceptions fast, but the programmer in most cases doesn't want to print 
  ** a stacktrace unless it's for debugging; We start tracing from the saved 
  ** frame since we still have our frame's parent pointer intact, the stack 
  ** information is (still now) mostly lost since we push the throwable on 
  ** the stack...
  ** If this seems strange, note that some programmers use exceptions in the 
  ** 'normal' flow of programming so if we would generate a stack trace for 
  ** each occured exception, we would loose in time...
  */

  if (thrown[F_Throwable_frame] == 0) {
    woempa(1,"%k@%p[F_Throwable_frame] == 0, setting it to %p (%w)\n", instance2clazz(thrown),thrown,throw_frame,NM(throw_frame->method));
    thrown[F_Throwable_frame] = (w_word)throw_frame;
    thrown[F_Throwable_current] = (w_word)(throw_code - 1); // was incremented in interpreter.c
    fillInStackTrace(thrown);
  // dumpStackTrace(thrown);
  }

  catch_code = throw_code;
  pc = throw_code - method->code; 
  catch_frame = throw_frame;
  woempa(1,"Throwing frame is %p, pc %d\n",throw_frame,pc);
  while (catch_frame && catch_frame->method) {
    woempa(1,"method %M has %d exception handlers\n", catch_frame->method, catch_frame->method->numExceptions);
    for (i = 0; i < catch_frame->method->numExceptions; i++) {
      exception = &catch_frame->method->exceptions[i];
      woempa(1, "--> current pc [%05d], method %M, thrown %k: handler %d has start %d, end %d.\n", pc, catch_frame->method, instance2clazz(thrown), i, exception->start_pc, exception->end_pc);
      if (exception->type_index) {
        catch_clazz = getClassConstant(method->clazz, exception->type_index);
        woempa(1, "    it catches the type %k.\n", catch_clazz);
      }
      else {
        catch_clazz = NULL;
        woempa(1, "    it catches ANY EXCEPTION.\n");
      }

      /*
      ** pc always points to the instruction just after the point where the exception was raised.
      ** We don't know how long that instruction was (could be anything from a 1-byte athrow to
      ** a five-byte invokeinterface), but subtracting 1 should be enough to make the comparison
      ** with exception->start/end_pc work ... (remember start_pc is inclusive and end_pc exclusive).
      */
      if ((pc - 1) >= exception->start_pc && (pc - 1) < exception->end_pc) {
        if (catch_clazz == NULL || isSuperClass(catch_clazz, instance2clazz(thrown))) {
          woempa(1, "--> yippediedoe, found matching handler.\n");
          woempa(1, "Clearing stack frame, pushing %k@%p (frame=%p,...)\n", instance2clazz(thrown),thrown,(w_frame)thrown[F_Throwable_frame]);
          clearJavaOperands(catch_frame);
          item = (w_word)thrown;
          catch_frame->o_top[0] = item;
          catch_frame->o_top += 1;
          // TODO unset some flags for collector ???
          *new_code = &catch_frame->method->code[exception->handler_pc];
          
          woempa(1,"--> new pc is %d of method %M\n", *new_code-catch_frame->method->code, catch_frame->method);
          /*
          ** Now clear the exception from the Thread instance...
          */
          woempa(1,"Clearing exception from thread %w Thread %p\n",NM(thread),thread->Thread);
          unsetFlag(thread->flags, WT_THREAD_KILLED | WT_THREAD_THROWN);
          setReferenceField(thread->Thread, nullInstance, F_Thread_thrown);
          woempa(1,"returning frame %p\n",catch_frame);
          
          return catch_frame;
        }
      }
    }
    
    /*
    ** See if this frame is a bottom frame. If that is the case, we need to return this frame and let the caller decide
    ** what should happen... TODO when we have appropriate 'signal' handling !!
    */
    
    if (isSet(catch_frame->flags, FRAME_BOTTOM)) {
      woempa(1, "We arrived at a bottom frame when throwing a %k in %M !\n", instance2clazz(thrown), method);
      dumpStackTrace(thrown);
      // TODO, It should allready be in the thread as the thrown field, but probably in the 'throw'
      // bytecode, it's not yet set in the thread. Check this out so that this statement can be removed.
      // DONE catch_frame->thread->Thread[F_Thread_thrown] = (w_word)thrown;
      return NULL;
    }

    catch_frame = clearJavaFrame(catch_frame, &catch_code);
    woempa(1,"Cleared a frame, now looking at %p\n",catch_frame);

    /*
    ** If frame is NULL, we have arrived at the root frame and still didn't find an
    ** exception handler. Launch the default one...
    */

    if (catch_frame == NULL) {
      break;
    }
    method = catch_frame->method;
    pc = catch_code - method->code;
    woempa(1, "--> frame was executing instruction before [%05d] in %M\n", pc, method);
    --pc; // 
  }

  // If the thread is started using Thread,start(), then the invocation
  // of _run() is wrapped in a try ... catch (Throwable t), so we can
  // never have an uncaught exception.  The only thread which is not
  // (yet) started this way is com.acunia.wonka.Init, so if we include
  // a universal catch clause in that too then the code below can never
  // be reached.
  // Conclusion: if you find yourself reading this, then add the following
  // to your favourite version of Init:
  //    try {
  //      ... whatever you have now ...
  //    } catch (Throwable t) {
  //      Thread.getThreadGroup().uncaughtException(this,t);
  //    }

  wabort(ABORT_WONKA,"Uncaught exception in %w\n",NM(throw_frame->thread));

  return NULL;

}

/*
** Create an instance of an Exception. 'clazz' is the class of the exception,
** 'message' is the message that goes with it. 
*/

w_instance createExceptionInstance(w_thread thread, w_clazz clazz, w_string message) {

  w_instance exception_inst;
  w_instance string_inst;

  classMustBeInitialized(clazz, thread);
  exception_inst = allocInstance(thread, clazz);

  if (exception_inst) {
    if (message) {
      string_inst = getStringInstance(thread, message);
      setReferenceField(exception_inst, string_inst, F_Throwable_detailMessage);
    }
    else {
      setReferenceField(exception_inst, NULL, F_Throwable_detailMessage);
    }
  }
  else {
    exception_inst = (w_instance)thread->Thread[F_Thread_theOutOfMemoryError];
  }

  return exception_inst;

}

#ifdef DEBUG
void dumpStackTrace (w_instance thisThrowable) {

  w_trace trace = (w_trace)thisThrowable[F_Throwable_trace];
  w_trace callee = NULL;
  w_size pc;
  w_size line;
  w_code current = (w_code)thisThrowable[F_Throwable_current];

  if (! trace) {
    fillInStackTrace(thisThrowable);
    trace = (w_trace)thisThrowable[F_Throwable_trace];
  }
    
  if (!isNullInstance((w_instance)thisThrowable[F_Throwable_detailMessage])) {
    woempa(2, "%w\n", String2string((w_instance)thisThrowable[F_Throwable_detailMessage]));
  }
  while (trace) {
    line = code2line(trace->method, current, &pc);
    woempa(2, "%k -> [pc = %05d, line = %04d] %m\n", trace->method->clazz, pc, line, trace->method);
    current = trace->caller_current;
    callee = trace;
    trace = trace->previous;
  }
  
  return;

}
#endif

