/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: interpreter.c,v 1.29 2002/03/27 15:34:59 dbuytaert Exp $
*/

#include <stdio.h>
#include <string.h>

#include "arrays.h"
#include "clazz.h"
#include "constant.h"
#include "core-classes.h"
#include "descriptor.h"
#include "exception.h"
#include "fields.h"
#include "hashtable.h"
#include "ieee754.h"
#include "Math.h"
#include "heap.h"
#include "interpreter.h"
#include "loading.h"
#include "locks.h"
#include "ts-mem.h"
#include "methods.h"
#include "opcodes.h"
#include "stack.h"
#include "strings.h"
#include "threads.h"
#include "checks.h"
#include "comms.h"
#include "dispatcher.h"
#include "jdwp.h"
#include "jdwp_events.h"

#ifndef INLINE
#define INLINE inline
#endif 

INLINE static w_byte truncate_i2b(w_int i) {

  int j = i & 0xff;

  return (w_char) j;

}

INLINE static w_short truncate_i2c(w_int i) {

  int j = i & 0xffff;

  return (w_char) j;

}

INLINE static w_short truncate_i2s(w_int i) {

  int j = i & 0xffff;

  return (w_short) j;

}

INLINE static w_int truncate_l2i(w_long l) {

  int j = l & 0x00000000ffffffff;

  return j;

}

/*
** Extensions: b2i and s2i are sign-extension, c2i is zero-extension.  
** i2l is also a sign-extension.
*/

INLINE static w_int extend_b2i(w_word b) {

  int i = (b&0x80) ? (b|0xffffff00) : (b&0x0000007f);

  return i;

}

INLINE static w_int extend_s2i(w_int s) {
  int i = (s&0x8000) ? (int)(s|0xffff0000) : (int)(s&0x00007fff);

  return i;

}

INLINE static w_int extend_c2i(w_int c) {

  return c & 0x0000ffff;

}

INLINE static w_long extend_i2l(w_int i) {

  return i;

}


INLINE static w_boolean checkArrayrefAndIndex(w_frame frame, w_instance Array, w_int slot) {

  woempa(1, "%s %p index %d\n", Array == NULL ? "null reference" : "reference to ", Array, slot);

  if (isNullInstance(Array)) {
    throwException(frame->thread, clazzNullPointerException, NULL);

    return WONKA_FALSE;

  }

  if (slot < 0 || slot >= instance2Array_length(Array)) {
    woempa(5, "ArrayIndexOutOfBoundsException in %M\n", frame->method);
    woempa(5, "Array at %p has length %d, index used is %d\n", Array, instance2Array_length(Array), slot);
    throwException(frame->thread, clazzArrayIndexOutOfBoundsException, NULL);

    return WONKA_FALSE;

  }

  return WONKA_TRUE;

}

INLINE static void pushInt (w_frame frame, w_int i) {

  stack_pops_pushes(frame, 0, 1);

  *frame->o_top++ = i;
}

INLINE static void pushLong (w_frame frame, w_u64 j) {

  stack_pops_pushes(frame, 0, 2);

  *frame->o_top++ = j.words[WORD_LSW];
  *frame->o_top++ = j.words[WORD_MSW];
}

INLINE static w_int popInt(w_frame frame) {
  stack_pops_pushes(frame, 1, 0);
  
  return *--frame->o_top;
}

INLINE static w_char popChar(w_frame frame) {

  w_word c;

  stack_pops_pushes(frame, 1, 0);

  c = *--frame->o_top;
  
  woempa(1, "Popped char %d (0x%4x) '%c'\n", (w_int)c, (w_int)c, (w_byte)c);

  return (w_char)c;

}

INLINE static w_short popShort(w_frame frame) {

  w_word s;

  stack_pops_pushes(frame, 1, 0);

  s = *--frame->o_top;

  woempa(1, "Popped short %d (0x%4x)\n", (w_short)s, (w_short)s);

  return (w_short)s;

}

INLINE static w_byte popByte(w_frame frame) {

  w_word b;

  stack_pops_pushes(frame, 1, 0);

  b = *--frame->o_top;

  woempa(1, "Popped byte %d (0x%2x)\n", (w_byte)b, (w_byte)b);

  return (w_byte)b;

}

INLINE static w_float popFloat(w_frame frame) {
  stack_pops_pushes(frame, 1, 0);

  return (w_float)(*--frame->o_top);
}

INLINE static w_double popDouble(w_frame frame) {

  w_u64 d;

  stack_pops_pushes(frame, 2, 0);
  
  d.words[WORD_MSW] = *--frame->o_top;
  d.words[WORD_LSW] = *--frame->o_top;

  return (w_double) d.u64;  

}

INLINE static w_long popLong(w_frame frame) {

  w_u64 j;

  stack_pops_pushes(frame, 2, 0);
  
  j.words[WORD_MSW] = *--frame->o_top;
  j.words[WORD_LSW] = *--frame->o_top;

  return j.s64;

}

INLINE static void pushFloat(w_frame frame, w_float f) {

  stack_pops_pushes(frame, 0, 1);

  *frame->o_top++ = f;
}

INLINE static void pushDouble(w_frame frame, w_u64 d) {

  stack_pops_pushes(frame, 0, 2);

  *frame->o_top++ = d.words[WORD_LSW];
  *frame->o_top++ = d.words[WORD_MSW];

}

INLINE static void pushReference(w_frame frame, w_instance ref) {

  stack_pops_pushes(frame, 0, 1);

  *frame->o_top++ = (w_word)makeLocal(ref);
}

INLINE static void pushAddress(w_frame frame, w_int addr) {

  stack_pops_pushes(frame, 0, 1);

  *frame->o_top++ = addr;
}

INLINE static w_instance popReference(w_frame frame) {
  stack_pops_pushes(frame, 1, 0);
  
  return (w_instance)(*--frame->o_top);
}

INLINE static w_instance peekReference(w_frame frame, w_method method) {
  return (w_instance)(*(frame->o_top - method->arg_i));
}

#define popArrayRef(f) popReference(f)

#define popThrowable(f) popReference(f)

INLINE static w_boolean checkReference(w_frame frame, w_instance instance) {

  if (instance == NULL) {
    throwException(frame->thread, clazzNullPointerException, NULL);

    return WONKA_FALSE;
  }

  return WONKA_TRUE;

}

INLINE static void loadConstant(w_frame frame, int idx) {

  w_clazz clazz = frame->method->clazz;
  w_cnt cnt = &clazz->constants[idx];

  if (cnt->cnt_tag == CONSTANT_STRING) {
    w_string string = getStringConstant(clazz, idx);
    w_instance instance = getStringInstance(frame->thread, string);;

    pushReference(frame, instance);
    if (instance) {
      woempa(1,"UnGREENING String at %p (pushed by ldc)\n", instance);
      unsetFlag(instance2flags(instance),O_GREEN);
    }
  }
  else if (cnt->cnt_tag == CONSTANT_INTEGER) {
    pushInt(frame, getIntegerConstant(clazz, idx));
  }
  else if (cnt->cnt_tag == CONSTANT_FLOAT) {
    pushFloat(frame, getFloatConstant(clazz, idx));
  }
  else if (cnt->cnt_tag == CONSTANT_LONG) {
    pushLong(frame, getLongConstant(clazz, idx));
  } 
  else if (cnt->cnt_tag == CONSTANT_DOUBLE) {
    pushDouble(frame, getDoubleConstant(clazz, idx));
  }
  else {
    wabort(ABORT_WONKA, "Bad constant type %d\n", cnt->cnt_tag);
  }

}

INLINE static void loadLocalInt(w_frame frame, w_int idx) {

  stack_pops_pushes(frame, 0, 1);
  
  pushInt(frame, (w_int)frame->locals[idx]);
}

INLINE static void loadLocalLong(w_frame frame, w_int idx) {

  stack_pops_pushes(frame, 0, 2);

  woempa(1, "loading long from locals[%d,%d] at %p.\n", idx, idx + 1, frame->locals + idx);

  frame->o_top[0] = frame->locals[idx + 0];
  frame->o_top[1] = frame->locals[idx + 1];
  
  frame->o_top += 2;

}

INLINE static void loadLocalFloat(w_frame frame, w_int idx) {

  stack_pops_pushes(frame, 0, 1);

  woempa(1, "loading float from local[%d] at %p.\n", idx, frame->locals + idx);

  frame->o_top[0] = frame->locals[idx];

  frame->o_top += 1;

}

INLINE static void loadLocalDouble(w_frame frame, w_int idx) {

  stack_pops_pushes(frame, 0, 2);

  woempa(1, "loading double from locals[%d,%d] at %p.\n", idx, idx + 1, frame->locals + idx);

  frame->o_top[0] = frame->locals[idx + 0];
  frame->o_top[1] = frame->locals[idx + 1];
  
  frame->o_top += 2;

}

INLINE static void loadLocalReference(w_frame frame, w_int idx) {

  stack_pops_pushes(frame, 0, 1);

  woempa(1, "loading reference from local[%d] at %p.\n", idx, frame->locals + idx);

  frame->o_top[0] = frame->locals[idx];

  frame->o_top += 1;

}

INLINE static void storeLocalInt(w_frame frame, w_int idx) {

  stack_pops_pushes(frame, 1, 0);

  frame->locals[idx] = frame->o_top[-1];

  frame->o_top -= 1;

  woempa(1, "stored int %d into local[%d] at %p.\n", (w_int) (frame->locals[idx]), idx, frame->locals + idx);

}

INLINE static void storeLocalLong(w_frame frame, w_int idx) {

  stack_pops_pushes(frame, 2, 0);

  woempa(1, "storing long into locals[%d,%d] at %p.\n", idx, idx + 1, frame->locals + idx);

  frame->o_top -= 2;

  frame->locals[idx + 1] = frame->o_top[1];
  frame->locals[idx + 0] = frame->o_top[0];
  
}

INLINE static void storeLocalFloat(w_frame frame, w_int idx) {

  stack_pops_pushes(frame, 1, 0);

  woempa(1, "storing float into local[%d] at %p.\n", idx, frame->locals + idx);

  frame->locals[idx] = frame->o_top[-1];

  frame->o_top -= 1;

}

INLINE static void storeLocalDouble(w_frame frame, w_int idx) {

  stack_pops_pushes(frame, 2, 0);

  woempa(1, "storing double into locals[%d,%d] at %p.\n", idx, idx + 1, frame->locals + idx);

  frame->o_top -= 2;

  frame->locals[idx + 1] = frame->o_top[1];
  frame->locals[idx + 0] = frame->o_top[0];

}

INLINE static void storeLocalReferenceOrAddress(w_frame frame, w_int idx) {

  stack_pops_pushes(frame, 1, 0);

  frame->locals[idx] = frame->o_top[-1];
  
  frame->o_top -= 1;

}

INLINE static unsigned char get_byte_operand(w_code * codeptr) {

  return (unsigned char) * (*codeptr)++;

}

INLINE static w_short get_short_operand(w_code *codeptr) {

  w_short temp = get_byte_operand(codeptr);

  temp = (temp << 8) | get_byte_operand(codeptr);

  return temp;
  
}

INLINE static w_int get_wide_operand(w_code *codeptr) {

  w_int temp = get_byte_operand(codeptr);

  temp = (temp << 8) | get_byte_operand(codeptr);
  temp = (temp << 8) | get_byte_operand(codeptr);
  temp = (temp << 8) | get_byte_operand(codeptr);

  return temp;

}

/*
** TODO:
** All bar one of the 2-byte immediate operands in the JVM definition
** are unsigned indices (the exception is the second operand of a wide
** iinc).  So short_operand should really be a w_char, but when I try
** that nasty things happen ... to be investigated.     -- CG 19991216
**
** (The 4-byte operands of lookupswitch and tableswitch are signed, so
** w_int is correct).
*/

void interpreterEntry(void* athread) {

  w_thread thread = (w_thread)athread;
  w_instance exception;
  
  woempa(1,"preparing to run thread %w using frame %p method %w\n",NM(thread),thread->top,NM(thread->top->method));
  handleSynchronization(thread->top, thread->top->method);
  woempa(1,"got required lock (if any)\n");
  exception = runThread(thread, thread->top);

  if(jpda_hooks) {
    jdwp_event_thread_end(thread);
  }

  if (exception) {
    woempa(1,"thread %w has terminated abnormally, exception = %k\n",NM(thread),instance2clazz(exception));
  }
  else {
    woempa(1,"thread %w has terminated normally\n",NM(thread));
  }

  requestThreadTermination(thread);
  
}

/*
** handleException creates the object Thread.thrown if it does not
** already exist, saves the pc, and then calls handleThrowable.
*/

void handleException(w_frame* frameptr, w_code *codeptr) {

  w_frame  old_frame = *frameptr;
  w_code   old_code  = *codeptr;
  w_thread thread = old_frame->thread;
  w_instance thrown;
  w_instance Thread = thread->Thread;

  if (Thread == NULL) {
    wabort(ABORT_WONKA, "Panic! Exception thrown in non-Java thread '%w', don't know what to do.\n", NM(thread));
  }

  thrown = (w_instance)Thread[F_Thread_thrown];

  if (thrown == NULL) {
    wabort(ABORT_WONKA, "Thread %w flags are 0x%08x but %k at %p has thrown==null\n", thread->name, thread->flags, instance2clazz(Thread), Thread);
  }
  (*frameptr)->pc = *codeptr - 1 - (*frameptr)->method->code;
  *frameptr = handleThrowable(thrown,old_frame,old_code,codeptr);
  woempa(1,"New frame is %p, code pointer is %p\n",*frameptr,*codeptr);
}

/*
** checkForAsynchException returns WONKA_TRUE if no asynchronous
** exception is pending on this thread.  If an asynchronous
** exception was pending then checkForAsynchException clears
** the flag and calls handleException before returning WONKA_FALSE.
*/

INLINE static w_boolean checkForAsynchException(w_frame *frameptr, w_code *codeptr) {

  w_thread thread = (*frameptr)->thread;

  if (isSet(thread->flags, WT_THREAD_KILLED)) {
    unsetFlag(thread->flags, WT_THREAD_KILLED);
    handleException(frameptr, codeptr);
    return WONKA_FALSE;
  }

  return WONKA_TRUE;

}

#define UPDATE_PC (*frameptr)->pc = *codeptr - (*frameptr)->method->code;

INLINE static void opcode_nop(w_frame *frameptr, w_code *codeptr) {
}

INLINE static void opcode_aconst_null (w_frame *frameptr, w_code *codeptr) {
  pushReference(*frameptr, NULL);
}

INLINE static void opcode_iconst_m1 (w_frame *frameptr, w_code *codeptr) {
  pushInt(*frameptr, -1);
}

INLINE static void opcode_iconst_0 (w_frame *frameptr, w_code *codeptr) {
  pushInt(*frameptr, 0);
}

INLINE static void opcode_iconst_1 (w_frame *frameptr, w_code *codeptr) {
  pushInt(*frameptr, 1);
}

INLINE static void opcode_iconst_2 (w_frame *frameptr, w_code *codeptr) {
  pushInt(*frameptr, 2);
}

INLINE static void opcode_iconst_3 (w_frame *frameptr, w_code *codeptr) {
  pushInt(*frameptr, 3);
}

INLINE static void opcode_iconst_4 (w_frame *frameptr, w_code *codeptr) {
  pushInt(*frameptr, 4);
}

INLINE static void opcode_iconst_5 (w_frame *frameptr, w_code *codeptr) {
  pushInt(*frameptr, 5);
}

INLINE static void opcode_lconst_0 (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 0, 2);
  
  frame->o_top[0] = 0;
  frame->o_top[1] = 0;

  frame->o_top += 2;

}

INLINE static void opcode_lconst_1 (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 0, 2);

  frame->o_top[0] = 1;
  frame->o_top[1] = 0;

  frame->o_top += 2;

}

INLINE static void opcode_fconst_0 (w_frame *frameptr, w_code *codeptr) {

  pushFloat(*frameptr, F_ZERO);

}

INLINE static void opcode_fconst_1 (w_frame *frameptr, w_code *codeptr) {

  pushFloat(*frameptr, F_ONE);

}

INLINE static void opcode_fconst_2 (w_frame *frameptr, w_code *codeptr) {

  pushFloat(*frameptr, F_TWO);

}

INLINE static void opcode_dconst_0 (w_frame *frameptr, w_code *codeptr) {

  w_u64 d;
  
  d.u64 = D_ZERO;

  pushDouble(*frameptr, d);
  
}

INLINE static void opcode_dconst_1 (w_frame *frameptr, w_code *codeptr) {

  w_u64 d;
  
  d.u64 = D_ONE;

  pushDouble(*frameptr, d);
  
}

INLINE static void opcode_bipush (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 0, 1);

  frame->o_top[0] = (w_int) (signed char) get_byte_operand(codeptr);

  frame->o_top += 1;

}

INLINE static void opcode_sipush (w_frame *frameptr, w_code *codeptr) {

  pushInt(*frameptr, extend_s2i(get_short_operand(codeptr)));

}

INLINE static void opcode_ldc (w_frame *frameptr, w_code *codeptr) {
  loadConstant(*frameptr, get_byte_operand(codeptr));
}

INLINE static void opcode_ldc_w (w_frame *frameptr, w_code *codeptr) {
  loadConstant(*frameptr, get_short_operand(codeptr));
}

INLINE static void opcode_ldc2_w (w_frame *frameptr, w_code *codeptr) {
  loadConstant(*frameptr, get_short_operand(codeptr));
}

INLINE static void opcode_iload (w_frame *frameptr, w_code *codeptr) {

  loadLocalInt(*frameptr, get_byte_operand(codeptr));

}

INLINE static void opcode_wide_iload (w_frame *frameptr, w_code *codeptr) {

  loadLocalInt(*frameptr, get_short_operand(codeptr));

}

INLINE static void opcode_lload (w_frame *frameptr, w_code *codeptr) {

  loadLocalLong(*frameptr, get_byte_operand(codeptr));

}

INLINE static void opcode_wide_lload (w_frame *frameptr, w_code *codeptr) {

  loadLocalLong(*frameptr, get_short_operand(codeptr));

}

INLINE static void opcode_fload (w_frame *frameptr, w_code *codeptr) {

  loadLocalFloat(*frameptr, get_byte_operand(codeptr));

}

INLINE static void opcode_wide_fload (w_frame *frameptr, w_code *codeptr) {

  loadLocalFloat(*frameptr, get_short_operand(codeptr));

}

INLINE static void opcode_dload (w_frame *frameptr, w_code *codeptr) {

  loadLocalDouble(*frameptr, get_byte_operand(codeptr));

}

INLINE static void opcode_wide_dload (w_frame *frameptr, w_code *codeptr) {

  loadLocalDouble(*frameptr, get_short_operand(codeptr));

}

INLINE static void opcode_aload (w_frame *frameptr, w_code *codeptr) {

  loadLocalReference(*frameptr, get_byte_operand(codeptr));

}

INLINE static void opcode_wide_aload (w_frame *frameptr, w_code *codeptr) {

  loadLocalReference(*frameptr, get_short_operand(codeptr));

}

INLINE static void opcode_iload_0 (w_frame *frameptr, w_code *codeptr) {

  loadLocalInt(*frameptr, 0);

}

INLINE static void opcode_iload_1 (w_frame *frameptr, w_code *codeptr) {

  loadLocalInt(*frameptr, 1);

}

INLINE static void opcode_iload_2 (w_frame *frameptr, w_code *codeptr) {

  loadLocalInt(*frameptr, 2);

}

INLINE static void opcode_iload_3 (w_frame *frameptr, w_code *codeptr) {

  loadLocalInt(*frameptr, 3);

}

INLINE static void opcode_lload_0 (w_frame *frameptr, w_code *codeptr) {

  loadLocalLong(*frameptr, 0);

}

INLINE static void opcode_lload_1 (w_frame *frameptr, w_code *codeptr) {

  loadLocalLong(*frameptr, 1);

}

INLINE static void opcode_lload_2 (w_frame *frameptr, w_code *codeptr) {

  loadLocalLong(*frameptr, 2);

}

INLINE static void opcode_lload_3 (w_frame *frameptr, w_code *codeptr) {

  loadLocalLong(*frameptr, 3);

}

INLINE static void opcode_fload_0 (w_frame *frameptr, w_code *codeptr) {

  loadLocalFloat(*frameptr, 0);

}

INLINE static void opcode_fload_1 (w_frame *frameptr, w_code *codeptr) {

  loadLocalFloat(*frameptr, 1);

}

INLINE static void opcode_fload_2 (w_frame *frameptr, w_code *codeptr) {

  loadLocalFloat(*frameptr, 2);

}

INLINE static void opcode_fload_3 (w_frame *frameptr, w_code *codeptr) {

  loadLocalFloat(*frameptr, 3);

}

INLINE static void opcode_dload_0 (w_frame *frameptr, w_code *codeptr) {

  loadLocalDouble(*frameptr, 0);

}

INLINE static void opcode_dload_1 (w_frame *frameptr, w_code *codeptr) {

  loadLocalDouble(*frameptr, 1);

}

INLINE static void opcode_dload_2 (w_frame *frameptr, w_code *codeptr) {

  loadLocalDouble(*frameptr, 2);

}

INLINE static void opcode_dload_3 (w_frame *frameptr, w_code *codeptr) {

  loadLocalDouble(*frameptr, 3);

}

INLINE static void opcode_aload_0 (w_frame *frameptr, w_code *codeptr) {

  loadLocalReference(*frameptr, 0);

}

INLINE static void opcode_aload_1 (w_frame *frameptr, w_code *codeptr) {

  loadLocalReference(*frameptr, 1);

}

INLINE static void opcode_aload_2 (w_frame *frameptr, w_code *codeptr) {

  loadLocalReference(*frameptr, 2);

}

INLINE static void opcode_aload_3 (w_frame *frameptr, w_code *codeptr) {

  loadLocalReference(*frameptr, 3);

}

INLINE static void opcode_iaload (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_int      int_value;
  w_frame    frame = *frameptr;

  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    int_value = instance2Array_int(arrayref)[arrayindex];
    pushInt(frame, int_value);
  }
}

INLINE static void opcode_laload (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_u64      element;
  w_frame    frame = *frameptr;

  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    element.s64 = instance2Array_long(arrayref)[arrayindex];
    pushLong(frame, element);
  }
}

INLINE static void opcode_faload (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_float    floatvalue;
  w_frame    frame = *frameptr;

  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    floatvalue = instance2Array_float(arrayref)[arrayindex];
    pushFloat(frame,floatvalue);
  }
}

INLINE static void opcode_daload (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_frame    frame = *frameptr;

  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    pushDouble(frame, (w_u64)instance2Array_double(arrayref)[arrayindex]);
  }
}

INLINE static void opcode_aaload (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_instance objectref;
  w_frame    frame = *frameptr;

  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));
  woempa(1,"Loading element %d of %p\n",arrayindex,arrayref);

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    objectref = instance2Array_instance(arrayref)[arrayindex];
    woempa(1,"  = %p\n",objectref);
    pushReference(frame, objectref);
  }
}

INLINE static void opcode_baload (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_byte     bytevalue;
  w_frame    frame = *frameptr;

  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    if(instance2clazz(arrayref)->aclazz->element->body.primitive.type==P_boolean) {
      int byteindex = arrayindex / 8;
      int bitindex = arrayindex % 8;

      bytevalue = instance2Array_byte(arrayref)[byteindex];
      bytevalue = (bytevalue >> bitindex) & 1;
      pushInt(frame, bytevalue);
    }
    else {
      bytevalue = instance2Array_byte(arrayref)[arrayindex];
      pushInt(frame, extend_b2i(bytevalue));
    }
  }
}

INLINE static void opcode_caload (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_char     charvalue;
  w_frame    frame = *frameptr;

  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    charvalue = instance2Array_char(arrayref)[arrayindex];
    pushInt(frame, extend_c2i(charvalue));
  }
}

INLINE static void opcode_saload (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_short    shortvalue;
  w_frame    frame = *frameptr;

  arrayindex = popInt(frame);
  arrayref =  makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    shortvalue = instance2Array_short(arrayref)[arrayindex];
    pushInt(frame, extend_s2i(shortvalue));
  }
}

INLINE static void opcode_istore (w_frame *frameptr, w_code *codeptr) {

  woempa(1,"TODO: sort out this signed/unsigned byte/short business\n");

  storeLocalInt(*frameptr, get_byte_operand(codeptr));

}

INLINE static void opcode_wide_istore (w_frame *frameptr, w_code *codeptr) {

  woempa(1,"TODO: sort out this signed/unsigned byte/short business\n");

  storeLocalInt(*frameptr, get_short_operand(codeptr));

}

INLINE static void opcode_lstore (w_frame *frameptr, w_code *codeptr) {

  woempa(1,"TODO: sort out this signed/unsigned byte/short business\n");

  storeLocalLong(*frameptr, get_byte_operand(codeptr));

}

INLINE static void opcode_wide_lstore (w_frame *frameptr, w_code *codeptr) {

  woempa(1,"TODO: sort out this signed/unsigned byte/short business\n");

  storeLocalLong(*frameptr, get_short_operand(codeptr));

}

INLINE static void opcode_fstore (w_frame *frameptr, w_code *codeptr) {

  woempa(1,"TODO: sort out this signed/unsigned byte/short business\n");

  storeLocalFloat(*frameptr, get_byte_operand(codeptr));

}

INLINE static void opcode_wide_fstore (w_frame *frameptr, w_code *codeptr) {

  woempa(1,"TODO: sort out this signed/unsigned byte/short business\n");

  storeLocalFloat(*frameptr, get_short_operand(codeptr));

}

INLINE static void opcode_dstore (w_frame *frameptr, w_code *codeptr) {

  woempa(1,"TODO: sort out this signed/unsigned byte/short business\n");

  storeLocalDouble(*frameptr, get_byte_operand(codeptr));

}

INLINE static void opcode_wide_dstore (w_frame *frameptr, w_code *codeptr) {

  woempa(1,"TODO: sort out this signed/unsigned byte/short business\n");

  storeLocalDouble(*frameptr, get_short_operand(codeptr));

}

INLINE static void opcode_astore (w_frame *frameptr, w_code *codeptr) {

  woempa(1,"TODO: sort out this signed/unsigned byte/short business\n");

  storeLocalReferenceOrAddress(*frameptr, get_byte_operand(codeptr));

}

INLINE static void opcode_wide_astore (w_frame *frameptr, w_code *codeptr) {

  woempa(1,"TODO: sort out this signed/unsigned byte/short business\n");

  storeLocalReferenceOrAddress(*frameptr, get_short_operand(codeptr));

}

INLINE static void opcode_istore_0 (w_frame *frameptr, w_code *codeptr) {

  storeLocalInt(*frameptr, 0);

}

INLINE static void opcode_istore_1 (w_frame *frameptr, w_code *codeptr) {

  storeLocalInt(*frameptr, 1);

}

INLINE static void opcode_istore_2 (w_frame *frameptr, w_code *codeptr) {

  storeLocalInt(*frameptr, 2);

}

INLINE static void opcode_istore_3 (w_frame *frameptr, w_code *codeptr) {

  storeLocalInt(*frameptr, 3);

}

INLINE static void opcode_lstore_0 (w_frame *frameptr, w_code *codeptr) {

  storeLocalLong(*frameptr, 0);

}

INLINE static void opcode_lstore_1 (w_frame *frameptr, w_code *codeptr) {

  storeLocalLong(*frameptr, 1);

}

INLINE static void opcode_lstore_2 (w_frame *frameptr, w_code *codeptr) {

  storeLocalLong(*frameptr, 2);

}

INLINE static void opcode_lstore_3 (w_frame *frameptr, w_code *codeptr) {

  storeLocalLong(*frameptr, 3);

}

INLINE static void opcode_fstore_0 (w_frame *frameptr, w_code *codeptr) {

  storeLocalFloat(*frameptr, 0);

}

INLINE static void opcode_fstore_1 (w_frame *frameptr, w_code *codeptr) {

  storeLocalFloat(*frameptr, 1);

}

INLINE static void opcode_fstore_2 (w_frame *frameptr, w_code *codeptr) {

  storeLocalFloat(*frameptr, 2);

}

INLINE static void opcode_fstore_3 (w_frame *frameptr, w_code *codeptr) {

  storeLocalFloat(*frameptr, 3);

}

INLINE static void opcode_dstore_0 (w_frame *frameptr, w_code *codeptr) {

  storeLocalDouble(*frameptr, 0);

}

INLINE static void opcode_dstore_1 (w_frame *frameptr, w_code *codeptr) {

  storeLocalDouble(*frameptr, 1);

}

INLINE static void opcode_dstore_2 (w_frame *frameptr, w_code *codeptr) {

  storeLocalDouble(*frameptr, 2);

}

INLINE static void opcode_dstore_3 (w_frame *frameptr, w_code *codeptr) {

  storeLocalDouble(*frameptr, 3);

}

INLINE static void opcode_astore_0 (w_frame *frameptr, w_code *codeptr) {

  storeLocalReferenceOrAddress(*frameptr, 0);

}

INLINE static void opcode_astore_1 (w_frame *frameptr, w_code *codeptr) {

  storeLocalReferenceOrAddress(*frameptr, 1);

}

INLINE static void opcode_astore_2 (w_frame *frameptr, w_code *codeptr) {

  storeLocalReferenceOrAddress(*frameptr, 2);

}

INLINE static void opcode_astore_3 (w_frame *frameptr, w_code *codeptr) {

  storeLocalReferenceOrAddress(*frameptr, 3);

}

INLINE static void opcode_iastore (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_int      int_value;
  w_frame    frame = *frameptr;

  int_value = popInt(frame);
  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    instance2Array_int(arrayref)[arrayindex] = int_value;
#ifdef ENABLE_PS
    if (isLocalCopy(arrayref)) {
      besmirchLocalCopy(arrayref);
    }
#endif
  }
}

INLINE static void opcode_lastore (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_long     long_value;
  w_frame    frame = *frameptr;
  
  long_value = popLong(frame);
  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    instance2Array_long(arrayref)[arrayindex] = long_value;
#ifdef ENABLE_PS
    if (isLocalCopy(arrayref)) {
      besmirchLocalCopy(arrayref);
    }
#endif
  }
}

INLINE static void opcode_fastore (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_float    float_value;
  w_frame    frame = *frameptr;

  float_value = popFloat(frame);
  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    instance2Array_float(arrayref)[arrayindex] = float_value;
#ifdef ENABLE_PS
    if (isLocalCopy(arrayref)) {
      besmirchLocalCopy(arrayref);
    }
#endif
  }
}
 
INLINE static void opcode_dastore (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_double   double_value;
  w_frame    frame = *frameptr;
  
  double_value = popDouble(frame);
  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    instance2Array_double(arrayref)[arrayindex] = double_value;
#ifdef ENABLE_PS
    if (isLocalCopy(arrayref)) {
      besmirchLocalCopy(arrayref);
    }
#endif
  }
}

INLINE static void opcode_aastore (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_instance ref_value;
  w_frame    frame = *frameptr;

  ref_value = popReference(frame);
  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));
  woempa(1,"Storing %p in element %d of %p\n",ref_value,arrayindex,arrayref);

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    if (ref_value && !isAssignmentCompatible(instance2clazz(ref_value), instance2clazz(arrayref)->aclazz->componentClazz)) {
      throwException(frame->thread, clazzArrayStoreException, NULL);
    }
    else {
      setArrayReferenceField(arrayref, ref_value, arrayindex);
    }
  }
}

INLINE static void opcode_bastore (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_byte     bytevalue;
  w_frame    frame = *frameptr;

  bytevalue = popByte(frame);
  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    if (instance2clazz(arrayref)->aclazz->element->body.primitive.type==P_boolean) {
      int byteindex = arrayindex / 8;
      int bitindex = arrayindex % 8;
      w_byte work;

      woempa(1,"boolean array: use bit %d of byte %d\n", bitindex, byteindex);
      work = instance2Array_byte(arrayref)[byteindex];
      woempa(1,"byte currently contains %02x, mask with %02x\n", work, 0xff ^ (1 << bitindex));
      work &= 0xff ^ (1 << bitindex);
      work |= (bytevalue & 1) << (bitindex);
      woempa(1,"byte now holds %02x\n",work);
      instance2Array_byte(arrayref)[byteindex] = work;
    }
    else {
     instance2Array_byte(arrayref)[arrayindex] = truncate_i2b(bytevalue);
    }

#ifdef ENABLE_PS
    if (isLocalCopy(arrayref)) {
      besmirchLocalCopy(arrayref);
    }
#endif
  }
}

INLINE static void opcode_castore (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_char     charvalue;
  w_frame    frame = *frameptr;

  charvalue = popChar(frame);
  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    instance2Array_char(arrayref)[arrayindex] = truncate_i2c(charvalue);
#ifdef ENABLE_PS
    if (isLocalCopy(arrayref)) {
      besmirchLocalCopy(arrayref);
    }
#endif
  }
}

INLINE static void opcode_sastore (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      arrayindex;
  w_short    shortvalue;
  w_frame    frame = *frameptr;
  
  shortvalue = popShort(frame);
  arrayindex = popInt(frame);
  arrayref = makeLocal(popReference(frame));

  if (checkArrayrefAndIndex(frame, arrayref, arrayindex)) {
    instance2Array_short(arrayref)[arrayindex] = truncate_i2s(shortvalue);
#ifdef ENABLE_PS
    if (isLocalCopy(arrayref)) {
      besmirchLocalCopy(arrayref);
    }
#endif
  }
}

INLINE static void opcode_pop (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 1, 0);
  
  frame->o_top -= 1;

}

INLINE static void opcode_pop2 (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 2, 0);

  frame->o_top -= 2;

}

INLINE static void opcode_dup (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 0, 1);

  frame->o_top[0] = frame->o_top[-1];

  frame->o_top += 1;

}

INLINE static void opcode_dup_x1 (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 0, 1);

  frame->o_top += 1;
  frame->o_top[-1] = frame->o_top[-2];
  frame->o_top[-2] = frame->o_top[-3];
  frame->o_top[-3] = frame->o_top[-1];

}

INLINE static void opcode_dup_x2 (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 0, 1);

  frame->o_top += 1;
  frame->o_top[-1] = frame->o_top[-2];
  frame->o_top[-2] = frame->o_top[-3];
  frame->o_top[-3] = frame->o_top[-4];
  frame->o_top[-4] = frame->o_top[-1];

}

INLINE static void opcode_dup2 (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 0, 2);

  frame->o_top += 2;
  frame->o_top[-1] = frame->o_top[-3];
  frame->o_top[-2] = frame->o_top[-4];

}

INLINE static void opcode_dup2_x1 (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 0, 2);

  frame->o_top += 2;
  frame->o_top[-1] = frame->o_top[-3];
  frame->o_top[-2] = frame->o_top[-4];
  frame->o_top[-3] = frame->o_top[-5];
  frame->o_top[-4] = frame->o_top[-1];
  frame->o_top[-5] = frame->o_top[-2];

}

INLINE static void opcode_dup2_x2 (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 0, 2);

  frame->o_top += 2;
  frame->o_top[-1] = frame->o_top[-3];
  frame->o_top[-2] = frame->o_top[-4];
  frame->o_top[-3] = frame->o_top[-5];
  frame->o_top[-4] = frame->o_top[-6];
  frame->o_top[-5] = frame->o_top[-1];
  frame->o_top[-6] = frame->o_top[-2];

}

INLINE static void opcode_swap (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  w_word tmp;
  
  tmp = frame->o_top[-1];
  frame->o_top[-1] = frame->o_top[-2];
  frame->o_top[-2] = tmp;

}

INLINE static void opcode_iadd (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 1);

  (w_int) frame->o_top[-2] += (w_int) frame->o_top[-1];

  frame->o_top -= 1;

}

INLINE static void opcode_ladd (w_frame *frameptr, w_code *codeptr) {

  w_u64 j1;
  w_u64 j2;
  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 4, 2);
  
  j2.words[WORD_MSW] = frame->o_top[-1];
  j2.words[WORD_LSW] = frame->o_top[-2];
  j1.words[WORD_MSW] = frame->o_top[-3];
  j1.words[WORD_LSW] = frame->o_top[-4];

  j1.s64 += j2.s64;

  /*
  ** Note. Gcc for ARM generates shorter code if the next index you use is one used before; e.g., the
  ** last index we used was -4, so let's start with -4 below...
  */

  frame->o_top[-4] = j1.words[WORD_LSW];
  frame->o_top[-3] = j1.words[WORD_MSW];

  frame->o_top -= 2;

}

INLINE static void opcode_fadd (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 1);

  frame->o_top[-2] = float32_add(frame->o_top[-2], frame->o_top[-1]);

  frame->o_top -= 1;

}

INLINE static void opcode_dadd (w_frame *frameptr, w_code *codeptr) {

  w_u64 d1;
  w_u64 d2;
  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 4, 2);

  d2.words[WORD_MSW] = frame->o_top[-1];
  d2.words[WORD_LSW] = frame->o_top[-2];
  d1.words[WORD_MSW] = frame->o_top[-3];
  d1.words[WORD_LSW] = frame->o_top[-4];

  d1.u64 = float64_add(d1.u64, d2.u64);

  frame->o_top[-4] = d1.words[WORD_LSW];
  frame->o_top[-3] = d1.words[WORD_MSW];

  frame->o_top -= 2;

}

INLINE static void opcode_isub (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 1);

  (w_int) frame->o_top[-2] -= (w_int) frame->o_top[-1];

  frame->o_top -= 1;
  
}

INLINE static void opcode_lsub (w_frame *frameptr, w_code *codeptr) {

  w_u64 j2;
  w_u64 j1;
  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 4, 2);

  j2.words[WORD_MSW] = frame->o_top[-1];
  j2.words[WORD_LSW] = frame->o_top[-2];
  j1.words[WORD_MSW] = frame->o_top[-3];
  j1.words[WORD_LSW] = frame->o_top[-4];
  
  j1.s64 -= j2.s64;
  
  frame->o_top[-4] = j1.words[WORD_LSW];
  frame->o_top[-3] = j1.words[WORD_MSW];
  
  frame->o_top -= 2;

}

INLINE static void opcode_fsub (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 1);

  frame->o_top[-2] = float32_sub(frame->o_top[-2], frame->o_top[-1]);
  
  frame->o_top -= 1;
  
}

INLINE static void opcode_dsub (w_frame *frameptr, w_code *codeptr) {

  w_u64 d1;
  w_u64 d2;
  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 4, 2);

  d2.words[WORD_MSW] = frame->o_top[-1];
  d2.words[WORD_LSW] = frame->o_top[-2];
  d1.words[WORD_MSW] = frame->o_top[-3];
  d1.words[WORD_LSW] = frame->o_top[-4];

  d1.u64 = float64_sub(d1.u64, d2.u64);

  frame->o_top[-4] = d1.words[WORD_LSW];
  frame->o_top[-3] = d1.words[WORD_MSW];
  
  frame->o_top -= 2;

}

INLINE static void opcode_imul (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 1);

  (w_int) frame->o_top[-2] *= (w_int) frame->o_top[-1];
  
  frame->o_top -= 1;
  
}

INLINE static void opcode_lmul (w_frame *frameptr, w_code *codeptr) {

  w_u64 j2;
  w_u64 j1;
  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 4, 2);

  j2.words[WORD_MSW] = frame->o_top[-1];
  j2.words[WORD_LSW] = frame->o_top[-2];
  j1.words[WORD_MSW] = frame->o_top[-3];
  j1.words[WORD_LSW] = frame->o_top[-4];
  
  j1.s64 *= j2.s64;
  
  frame->o_top[-4] = j1.words[WORD_LSW];
  frame->o_top[-3] = j1.words[WORD_MSW];
  
  frame->o_top -= 2;
  
}

INLINE static void opcode_fmul (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 1);

  frame->o_top[-2] = float32_mul(frame->o_top[-2], frame->o_top[-1]);

  frame->o_top -= 1;
  
}

INLINE static void opcode_dmul (w_frame *frameptr, w_code *codeptr) {

  w_u64 d1;
  w_u64 d2;
  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 4, 2);

  d2.words[WORD_MSW] = frame->o_top[-1];
  d2.words[WORD_LSW] = frame->o_top[-2];
  d1.words[WORD_MSW] = frame->o_top[-3];
  d1.words[WORD_LSW] = frame->o_top[-4];

  d1.u64 = float64_mul(d1.u64, d2.u64);

  frame->o_top[-4] = d1.words[WORD_LSW];
  frame->o_top[-3] = d1.words[WORD_MSW];

  frame->o_top -= 2;

}

INLINE static void opcode_idiv (w_frame *frameptr, w_code *codeptr) {

  w_int int_value1;
  w_int int_value2;
  w_int int_result;
  w_frame    frame = *frameptr;
  
  int_value2 = popInt(frame);
  int_value1 = popInt(frame);

  if (int_value2 == 0) {
    throwException(frame->thread, clazzArithmeticException, NULL);
  }
  else {
    if ( int_value1 >= 0 && int_value2 > 0) {
      int_result = int_value1 / int_value2;
    }
    else if (int_value1 == (w_int)0x80000000) {
      if (int_value2 == -1) {
        int_result = 0x80000000;
      }
      else {
        int_result = int_value1 / int_value2;      
      }
    }
    else if (int_value1 < 0 && int_value2 < 0) {
      int_value1 = -int_value1;
      int_value2 = -int_value2;
      int_result = int_value1 / int_value2;
    }
    else if (int_value1 < 0) { 
    /* int_value2>0 */
      int_value1 = -int_value1;
      int_result = -(int_value1 / int_value2);
    }
    else {
    /*  int_value1>0, int_value2<0 */
      int_value2 = -int_value2;
      int_result = -(int_value1 / int_value2);
    }
    pushInt(frame, int_result);
  }

}

INLINE static void opcode_ldiv (w_frame *frameptr, w_code *codeptr) {

  w_long long_value1;
  w_long long_value2;
  w_u64  long_result;
  w_frame    frame = *frameptr;

  long_value2 = popLong(frame);
  long_value1 = popLong(frame);

  if (long_value2 == 0) {
    throwException(frame->thread, clazzArithmeticException, NULL);
  }
  else {
    if (long_value1 >= 0 && long_value2 > 0) {
      long_result.s64 = long_value1/long_value2;
    }
    else if (long_value1 == (w_long)0x8000000000000000){
      if (long_value2 == -1) {
        long_result.s64 = 0x8000000000000000;
      }
      else {
        long_result.s64 = long_value1/long_value2;     
      }
    }
    else if (long_value1 < 0 && long_value2 < 0) {
      long_value1 = -long_value1;
      long_value2 = -long_value2;
      long_result.s64 = long_value1/long_value2;
    }
    else if (long_value1 < 0) {
    /*  long_value2>0 */
      long_value1 = -long_value1;
      long_result.s64 = -(long_value1/long_value2);
    }
    else {
    /* long_value1 > 0, long_value2 < 0 */
      long_value2 = -long_value2;
      long_result.s64 = -(long_value1/long_value2);
    }

    pushLong(frame, long_result);
  }

}

INLINE static void opcode_fdiv (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 1);
  
  frame->o_top[-2] = float32_div(frame->o_top[-2], frame->o_top[-1]);

  frame->o_top -= 1;

}

INLINE static void opcode_ddiv (w_frame *frameptr, w_code *codeptr) {

  w_u64 d1;
  w_u64 d2;
  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 4, 2);

  d2.words[WORD_MSW] = frame->o_top[-1];
  d2.words[WORD_LSW] = frame->o_top[-2];
  d1.words[WORD_MSW] = frame->o_top[-3];
  d1.words[WORD_LSW] = frame->o_top[-4];

  d1.u64 = float64_div(d1.u64, d2.u64);

  frame->o_top[-4] = d1.words[WORD_LSW];
  frame->o_top[-3] = d1.words[WORD_MSW];

  frame->o_top -= 2;

}

INLINE static void opcode_irem (w_frame *frameptr, w_code *codeptr) {

  w_int int_value1;
  w_int int_value2;
  w_int int_result=0;
  w_frame    frame = *frameptr;

  int_value2 = popInt(frame);
  int_value1 = popInt(frame);

  if (int_value2 == 0) {
    throwException(frame->thread, clazzArithmeticException, NULL);
  }
  else {
    if (int_value1 != (w_int)0x80000000 || int_value2 != -1) {
      int_result = int_value1-int_value2 * (int)(int_value1/int_value2);
    }
  
    pushInt(frame,int_result );
  }

}

INLINE static void opcode_lrem (w_frame *frameptr, w_code *codeptr) {

  w_long long_value1;
  w_long long_value2;
  w_u64  long_result;
  w_frame    frame = *frameptr;
  
  long_result.s64 = 0x0000000000000000LL;

  long_value2 = popLong(frame);
  long_value1 = popLong(frame);

  if (long_value2 == 0) {
    throwException(frame->thread, clazzArithmeticException, NULL);
  }
  else { 
    if (long_value1 != (w_long)0x8000000000000000 || long_value2 != -1) {
      long_result.s64 = long_value1 - long_value2 * (w_long)(long_value1 / long_value2);
    }

    pushLong(frame, long_result);
  }
  
}

INLINE static void opcode_frem (w_frame *frameptr, w_code *codeptr) {
  w_float float_value1;
  w_float float_value2;
  w_float float_result;
  w_frame    frame = *frameptr;

  float_value2 = popFloat(frame);
  float_value1 = popFloat(frame);

  if (float32_is_NaN(float_value1) || float32_is_NaN(float_value2) || float32_eq(float_value2 , F_ZERO) || float32_is_Infinite(float_value1)) {
    float_result = (w_float) F_NAN;
  }
  else {
    if (float32_eq(float_value1 , F_ZERO) || float32_is_Infinite(float_value2)) {
      float_result = float_value1;
    }
    else   {
      float_result = float32_mul(float32_abs(float_value2), F_FLOAT_MAX_VALUE);
      if (float32_lt(float_result, float32_abs(float_value1))) {
        woempa(7,"fmod: division would overflow, returning zero as remainder\n");
        if (float32_is_negative(float_value1)) {
          float_result = F_MINUS_ZERO;
        }
        else {
          float_result = F_ZERO;
        }
      }
      else {
        float_result = float32_div(float_value1, float_value2);
        float_result = int32_to_float32(float32_to_int32_round_to_zero(float_result));
        float_result = float32_mul(float_value2,float_result);
        float_result = float32_sub(float_value1,float_result);
      }
    }
  }
  pushFloat(frame, float_result);

}

INLINE static void opcode_drem (w_frame *frameptr, w_code *codeptr) {
  w_double double_value1;
  w_double double_value2;
  w_double double_result;
  w_frame    frame = *frameptr;

  double_value2 = popDouble(frame);
  double_value1 = popDouble(frame);
  double_result = float64_div(double_value1, double_value2);
  if (float64_is_NaN(double_value1) || float64_is_NaN(double_value2) || float64_eq(double_value2 , D_ZERO) || float64_is_Infinite(double_value1)) {
    double_result = D_NAN; 
  }       	
  else {	
    if (float64_eq(double_value1 , D_ZERO) || float64_is_Infinite(double_value2)) {
      double_result = double_value1;
    }
    else {
      double_result = float64_mul(float64_abs(double_value2), D_DOUBLE_MAX_VALUE);
      if (float64_lt(double_result, float64_abs(double_value1))) {
        woempa(7,"fmod: division would overflow, returning zero as remainder\n");
        if (float64_is_negative(double_value1)) {
          double_result = D_MINUS_ZERO;
        }
        else {
          double_result = D_ZERO;
        }
      }
      else {
        double_result = float64_div(double_value1, double_value2);

        /**
        ** rounding is not what we would like since it doesn't round to zero ...
        */
        double_result = float64_round_to_int(double_result);

        double_result = float64_mul(double_value2,double_result);
        double_result = float64_sub(double_value1,double_result);

        /**
        ** the sign of double_result must be equal to sign of double_value1 this is not guaranteed
        ** anymore because of the rounding
        */

        if( float64_signBit(double_value1) && (! float64_signBit(double_result) && double_result) ){
          if(float64_signBit(double_value2)){
            double_result = float64_add(double_value2,double_result);
          }
          else{
            double_result = float64_sub(double_result,double_value2);
          }
        }
        else if( (! float64_signBit(double_value1)) && float64_signBit(double_result) && double_result){
          if(float64_signBit(double_value2)){
            double_result = float64_sub(double_result,double_value2);
          }
          else{
            double_result = float64_add(double_value2,double_result);
          }
        }
      }
    }
  }
  pushDouble(frame, (w_u64)double_result);

}

INLINE static void opcode_ineg (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 1, 1);

  frame->o_top[-1] = - (w_int) frame->o_top[-1];

}

INLINE static void opcode_lneg (w_frame *frameptr, w_code *codeptr) {

  w_u64 l;
  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 2);
  
  l.words[WORD_MSW] = frame->o_top[-1];
  l.words[WORD_LSW] = frame->o_top[-2];

  l.s64 = -l.s64;
  
  frame->o_top[-2] = l.words[WORD_LSW];
  frame->o_top[-1] = l.words[WORD_MSW];
  
}

INLINE static void opcode_fneg (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 1, 1);

  /*
  ** TODO The following can be made shorter. See how this is done in ieee754.c
  */
  
  frame->o_top[-1] = float32_negate(frame->o_top[-1]);

}

INLINE static void opcode_dneg (w_frame *frameptr, w_code *codeptr) {

  w_u64 d;
  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 2);
  
  /*
  ** TODO The following can be made shorter. Take a look at how float64_negate is implemented
  ** in ieee754.c
  */
  
  d.words[WORD_MSW] = frame->o_top[-1];
  d.words[WORD_LSW] = frame->o_top[-2];
  
  d.u64 = float64_negate(d.u64);
  
  frame->o_top[-2] = d.words[WORD_LSW];
  frame->o_top[-1] = d.words[WORD_MSW];

}

INLINE static void opcode_ishl (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 1);

  (w_int) frame->o_top[-2] <<= (frame->o_top[-1] & 0x0000001f);
  
  frame->o_top -= 1;
  
}

INLINE static void opcode_lshl (w_frame *frameptr, w_code *codeptr) {

  w_long long_value1;
  w_int  int_value2;
  w_u64 result;
  w_frame    frame = *frameptr;

  int_value2 = popInt(frame);
  long_value1 = popLong(frame);
  result.s64 = long_value1 << (int_value2 & 0x0000003f);
  pushLong(frame, result);

}

INLINE static void opcode_ishr (w_frame *frameptr, w_code *codeptr) {

  w_int i;
  w_int s;
  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 2, 1);
    
  i = (w_int) frame->o_top[-2];
  s = frame->o_top[-1] & 0x0000001f;
  
  if (i > 0) {
    frame->o_top[-2] = i >> s;
  }
  else {
  /* Chris holds his breath... */
    frame->o_top[-2] = -1 - (( -1 - i) >> s);
  }
  
  frame->o_top -= 1;
  
}

INLINE static void opcode_lshr (w_frame *frameptr, w_code *codeptr) {

  w_long long_value1;
  w_int int_value2;
  w_u64 result;
  w_frame    frame = *frameptr;

  int_value2 = popInt(frame);
  long_value1 = popLong(frame);

  if (long_value1 > 0) {
    long_value1 >>= int_value2 & 0x0000003f;
  } 
  else { /* take a deep breath ... */
    long_value1 = -1-((-1-long_value1) >> (int_value2 & 0x0000003f));
  }

  result.s64 = long_value1;
  pushLong(frame, result);

}

INLINE static void opcode_iushr (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 1);
  
  frame->o_top[-2] >>= (frame->o_top[-1] & 0x0000001f);
  
  frame->o_top -= 1;
  
}

INLINE static void opcode_lushr (w_frame *frameptr, w_code *codeptr) {

  w_int s;
  w_u64 j;
  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 3, 2);
  
  /*
  ** The variable 'j' must be used unsigned since we must zero extend this value !
  */
  
  s = frame->o_top[-1] & 0x0000003f;
  j.words[WORD_MSW] = frame->o_top[-2];
  j.words[WORD_LSW] = frame->o_top[-3];

  j.u64 >>= s;

  frame->o_top[-3] = j.words[WORD_LSW];
  frame->o_top[-2] = j.words[WORD_MSW];

  frame->o_top -= 1;

}

INLINE static void opcode_iand (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 1);

  frame->o_top[-2] &= frame->o_top[-1];
  
  frame->o_top -= 1;
  
}

INLINE static void opcode_land (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 4, 2);

  /*
  ** We AND together MSW and LSW separately.
  */
  
  frame->o_top[-3] &= frame->o_top[-1];
  frame->o_top[-4] &= frame->o_top[-2];
  
  frame->o_top -= 2;

}

INLINE static void opcode_ior (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 1);

  frame->o_top[-2] |= frame->o_top[-1];
  
  frame->o_top -= 1;
  
}

INLINE static void opcode_lor (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 4, 2);

  /*
  ** We OR together MSW and LSW separately.
  */

  frame->o_top[-4] |= frame->o_top[-2];
  frame->o_top[-3] |= frame->o_top[-1];
  
  frame->o_top -= 2;
  
}

INLINE static void opcode_ixor (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 2, 1);

  frame->o_top[-2] ^= frame->o_top[-1];
  
  frame->o_top -= 1;
  
}

INLINE static void opcode_lxor (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 4, 2);

  /*
  ** We XOR together MSW and LSW separately.
  */

  frame->o_top[-3] ^= frame->o_top[-1];
  frame->o_top[-4] ^= frame->o_top[-2];
  
  frame->o_top -= 2;

}

INLINE static void opcode_iinc (w_frame *frameptr, w_code *codeptr) {

  int local_var_index;
  int increment_amount;

  woempa(1,"TODO: sort out this signed/unsigned byte/short business\n");

  local_var_index = get_byte_operand(codeptr);
  increment_amount = extend_b2i((w_word)get_byte_operand(codeptr));
  woempa(1, "incrementing local[%d] by %d\n", local_var_index, increment_amount);
  (w_int) (*frameptr)->locals[local_var_index] += increment_amount;

}

INLINE static void opcode_wide_iinc (w_frame *frameptr, w_code *codeptr) {

  int local_var_index;
  int increment_amount;

  woempa(1,"TODO: sort out this signed/unsigned byte/short business\n");

  local_var_index = get_short_operand(codeptr);
  increment_amount = extend_s2i(get_short_operand(codeptr));
  woempa(1, "incrementing local[%d] by %d\n", local_var_index, increment_amount);
  (w_int) (*frameptr)->locals[local_var_index] += increment_amount;

}

INLINE static void opcode_i2l (w_frame *frameptr, w_code *codeptr) {

  w_int int_value;
  w_u64 long_value;
  w_frame frame = *frameptr;
  
  int_value = popInt(frame);
  long_value.s64 = extend_i2l(int_value);
  pushLong(frame, long_value);

}

INLINE static void opcode_i2f (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 1, 1);
  
  frame->o_top[-1] = int32_to_float32((w_int) frame->o_top[-1]);

}

INLINE static void opcode_i2d (w_frame *frameptr, w_code *codeptr) {

  w_int    int_value;
  w_u64    double_value;
  w_frame frame = *frameptr;
  
  int_value = popInt(frame);
  double_value.u64 = int32_to_float64(int_value);
  pushDouble(frame, double_value);

}

INLINE static void opcode_l2i (w_frame *frameptr, w_code *codeptr) {

  w_long long_value;
  w_frame frame = *frameptr;

  long_value = popLong(frame);
  pushInt(frame, truncate_l2i(long_value));

}

INLINE static void opcode_l2f (w_frame *frameptr, w_code *codeptr) {

  w_long long_value;
  w_frame frame = *frameptr;

  long_value = popLong(frame);
  pushFloat(frame, int64_to_float32(long_value));

}

INLINE static void opcode_l2d (w_frame *frameptr, w_code *codeptr) {

  w_u64 d;
  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 2, 2);
    
  d.words[WORD_MSW] = frame->o_top[-1];
  d.words[WORD_LSW] = frame->o_top[-2];
  
  d.u64 = int64_to_float64(d.s64);

  frame->o_top[-2] = d.words[WORD_LSW];
  frame->o_top[-1] = d.words[WORD_MSW];

}

INLINE static void opcode_f2i (w_frame *frameptr, w_code *codeptr) {

  w_float float_value;
  w_int    int_value;
  w_frame frame = *frameptr;

  float_value = popFloat(frame);

  if (float32_is_NaN(float_value)) {
    int_value = 0;
  }
  else {
    int_value = (w_int) float32_to_int32_round_to_zero(float_value);
  }
  pushInt(frame, int_value);

}

INLINE static void opcode_f2l (w_frame *frameptr, w_code *codeptr) {

  w_float float_value;
  w_u64  long_result;
  w_frame frame = *frameptr;

  float_value = popFloat(frame);
  if (float32_is_NaN(float_value)) {
    long_result.s64 = 0x0000000000000000LL;
  }
  else {
    long_result.s64 = float32_to_int64_round_to_zero(float_value);
  }
  pushLong(frame, long_result);

}

INLINE static void opcode_f2d (w_frame *frameptr, w_code *codeptr) {

  w_float  float_value;
  w_u64    double_value;
  w_frame frame = *frameptr;

  float_value = popFloat(frame);
  double_value.u64 = float32_to_float64(float_value);
  pushDouble(frame, double_value);

}

INLINE static void opcode_d2i (w_frame *frameptr, w_code *codeptr) {

  w_double double_value;
  w_int    int_value;
  w_frame frame = *frameptr;

  double_value = popDouble(frame);

  if (float64_is_NaN(double_value)) {
    int_value = 0;
  }
  else {
    int_value = (w_int) float64_to_int32_round_to_zero(double_value);
  }

  pushInt(frame, int_value);

}

INLINE static void opcode_d2l (w_frame *frameptr, w_code *codeptr) {

  w_u64 d;
  w_u64 l;
  w_frame frame = *frameptr;

  stack_pops_pushes(frame, 2, 2);
  
  l.words[WORD_MSW] = 0x00000000;
  l.words[WORD_LSW] = 0x00000000;
  d.words[WORD_MSW] = frame->o_top[-1];
  d.words[WORD_LSW] = frame->o_top[-2];

  if (! float64_is_NaN(d.u64)) {
    l.s64 = float64_to_int64_round_to_zero(d.u64);
  }

  frame->o_top[-2] = l.words[WORD_LSW];
  frame->o_top[-1] = l.words[WORD_MSW];

}
 
INLINE static void opcode_d2f (w_frame *frameptr, w_code *codeptr) {

  w_double double_value;
  w_float float_value;
  w_frame frame = *frameptr;

  double_value = popDouble(frame);
  float_value = float64_to_float32(double_value);
  pushFloat(frame, float_value);

}
 
INLINE static void opcode_i2b (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 1, 1);

  frame->o_top[-1] = extend_b2i(truncate_i2b((w_int) frame->o_top[-1]));

}
 
INLINE static void opcode_i2c (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 1, 1);

  frame->o_top[-1] = extend_c2i(truncate_i2c((w_int) frame->o_top[-1]));

}
 
INLINE static void opcode_i2s (w_frame *frameptr, w_code *codeptr) {

  w_frame frame = *frameptr;
  
  stack_pops_pushes(frame, 1, 1);

  frame->o_top[-1] = extend_s2i(truncate_i2s((w_int) frame->o_top[-1])); 

}
 
INLINE static void opcode_lcmp (w_frame *frameptr, w_code *codeptr) {

  w_long long_value1;
  w_long long_value2;
  w_int  int_result;
  w_frame    frame = *frameptr;

  long_value2 = popLong(frame);
  long_value1 = popLong(frame);

  if (long_value1 > long_value2) {
    int_result = 1;
  }
  else if (long_value1 == long_value2) {
    int_result = 0;
  }
  else {
    int_result = -1;
  }

  pushInt(frame, int_result);

}
 
INLINE static void opcode_fcmpg (w_frame *frameptr, w_code *codeptr) {

  w_float float_value1;
  w_float float_value2;
  w_int   int_result;
  w_frame    frame = *frameptr;

  float_value2 = popFloat(frame);
  float_value1 = popFloat(frame);

  if (float32_is_signaling_nan(float_value1) || float32_is_signaling_nan(float_value2)) {
    int_result = 1;
  }
  else if (float32_eq(float_value1, float_value2)) {
    int_result = 0;
  }
  else if (float32_lt(float_value1, float_value2)) {
    int_result = -1;
  }
  else {
    int_result = 1;
  }

  pushInt(frame, int_result);

}

INLINE static void opcode_fcmpl (w_frame *frameptr, w_code *codeptr) {

  w_float float_value1;
  w_float float_value2;
  w_int   int_result;
  w_frame    frame = *frameptr;

  float_value2 = popFloat(frame);
  float_value1 = popFloat(frame);

  if (float32_is_NaN(float_value1) || float32_is_NaN(float_value2) ) {
    int_result = -1;
  }
  else if (float32_eq(float_value1, float_value2)) {
    int_result = 0;
  }
  else if (float32_lt(float_value1, float_value2)) {
    int_result = -1;
  }
  else {
    int_result = 1;
  }

  pushInt(frame, int_result);

}

INLINE static void opcode_dcmpg (w_frame *frameptr, w_code *codeptr) {

  w_double double_value1;
  w_double double_value2;
  w_int    int_result;
  w_frame    frame = *frameptr;

  double_value2 = popDouble(frame);
  double_value1 = popDouble(frame);

  woempa ( 1, "double_value1 %0x %0x double_value2 %0x %0x\n", double_value1, double_value2);

  if (float64_is_NaN(double_value1) || float64_is_NaN(double_value2) ) {
    int_result = 1;
  }
  else if (float64_eq(double_value1, double_value2)) {
    int_result = 0;
  }
  else if (float64_lt(double_value1, double_value2)) {
    int_result = -1;
  }
  else {
    int_result = 1;
  }

  pushInt(frame, int_result);

}

INLINE static void opcode_dcmpl (w_frame *frameptr, w_code *codeptr) {

  w_double double_value1;
  w_double double_value2;
  w_int    int_result;
  w_frame    frame = *frameptr;

  double_value2 = popDouble(frame);
  double_value1 = popDouble(frame);

  woempa ( 1, "double_value1 %0x %0x double_value2 %0x %0x\n", double_value1, double_value2);

  if (float64_is_NaN(double_value1) || float64_is_NaN(double_value2)) {
    int_result = -1;
  }
  else if (float64_eq(double_value1, double_value2)) {
    int_result = 0;
  }
  else if (float64_lt(double_value1, double_value2)) {
    int_result = -1;
  }
  else {
    int_result = 1;
  }

  pushInt(frame, int_result);

}

INLINE static void opcode_ifeq (w_frame *frameptr, w_code *codeptr) {

  w_short  short_operand; 

  short_operand = get_short_operand(codeptr);

  if (popInt(*frameptr) == 0) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_ifne (w_frame *frameptr, w_code *codeptr) {

  w_short  short_operand; 

  short_operand = get_short_operand(codeptr);

  if (popInt(*frameptr) != 0) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_iflt (w_frame *frameptr, w_code *codeptr) {

  w_short  short_operand; 

  short_operand = get_short_operand(codeptr);

  if (popInt(*frameptr) < 0) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_ifge (w_frame *frameptr, w_code *codeptr) {

  w_short  short_operand; 

  short_operand = get_short_operand(codeptr);

  if (popInt(*frameptr) >= 0) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_ifgt (w_frame *frameptr, w_code *codeptr) {

  w_short  short_operand; 

  short_operand = get_short_operand(codeptr);

  if (popInt(*frameptr) > 0) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_ifle (w_frame *frameptr, w_code *codeptr) {

  w_short  short_operand; 

  short_operand = get_short_operand(codeptr);

  if (popInt(*frameptr) <= 0) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_if_icmpeq (w_frame *frameptr, w_code *codeptr) {

  w_int  int_value1;
  w_int  int_value2;
  w_short  short_operand; 
  w_frame    frame = *frameptr;

  short_operand = get_short_operand(codeptr);

  int_value2 = popInt(frame);
  int_value1 = popInt(frame);
  if (int_value1 == int_value2) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_if_icmpne (w_frame *frameptr, w_code *codeptr) {

  w_int  int_value1;
  w_int  int_value2;
  w_short  short_operand; 
  w_frame    frame = *frameptr;

  short_operand = get_short_operand(codeptr);

  int_value2 = popInt(frame);
  int_value1 = popInt(frame);
  if (int_value1 != int_value2) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_if_icmplt (w_frame *frameptr, w_code *codeptr) {

  w_int  int_value1;
  w_int  int_value2;
  w_short  short_operand; 
  w_frame    frame = *frameptr;

  short_operand = get_short_operand(codeptr);

  int_value2 = popInt(frame);
  int_value1 = popInt(frame);
  if (int_value1 < int_value2) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_if_icmpge (w_frame *frameptr, w_code *codeptr) {

  w_int  int_value1;
  w_int  int_value2;
  w_short  short_operand; 
  w_frame    frame = *frameptr;

  short_operand = get_short_operand(codeptr);

  int_value2 = popInt(frame);
  int_value1 = popInt(frame);
  if (int_value1 >= int_value2) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_if_icmpgt (w_frame *frameptr, w_code *codeptr) {

  w_int  int_value1;
  w_int  int_value2;
  w_short  short_operand; 
  w_frame    frame = *frameptr;

  short_operand = get_short_operand(codeptr);

  int_value2 = popInt(frame);
  int_value1 = popInt(frame);
  if (int_value1 > int_value2) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_if_icmple (w_frame *frameptr, w_code *codeptr) {

  w_int  int_value1;
  w_int  int_value2;
  w_short  short_operand; 
  w_frame    frame = *frameptr;

  short_operand = get_short_operand(codeptr);

  int_value2 = popInt(frame);
  int_value1 = popInt(frame);
  if (int_value1 <= int_value2) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_if_acmpeq (w_frame *frameptr, w_code *codeptr) {

  w_instance ref_value1;
  w_instance ref_value2;
  w_short  short_operand; 
  w_frame    frame = *frameptr;

  short_operand = get_short_operand(codeptr);

  ref_value2 = makeLocal(popReference(frame));
  ref_value1 = makeLocal(popReference(frame));

  if (ref_value1 == ref_value2) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_if_acmpne (w_frame *frameptr, w_code *codeptr) {

  w_instance ref_value1;
  w_instance ref_value2;
  w_short  short_operand; 
  w_frame    frame = *frameptr;

  short_operand = get_short_operand(codeptr);

  ref_value2 = makeLocal(popReference(frame));
  ref_value1 = makeLocal(popReference(frame));

  if (ref_value1!=ref_value2) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_goto (w_frame *frameptr, w_code *codeptr) {

  w_short  short_operand; 

  short_operand = get_short_operand(codeptr);
  *codeptr += short_operand - 3;
  UPDATE_PC

}

INLINE static void opcode_goto_w (w_frame *frameptr, w_code *codeptr) {

  w_int    wide_operand; 

  wide_operand = get_wide_operand(codeptr);
  *codeptr +=  wide_operand - 5;
  UPDATE_PC

}

INLINE static void opcode_jsr (w_frame *frameptr, w_code *codeptr) {

  w_short  short_operand; 
  w_frame frame = *frameptr;

  short_operand = get_short_operand(codeptr);
  pushAddress(frame, (*codeptr) - frame->method->code);
  *codeptr += short_operand - 3;
  UPDATE_PC

}

INLINE static void opcode_jsr_w (w_frame *frameptr, w_code *codeptr) {

  w_int    wide_operand; 
  w_frame frame = *frameptr;
  
  wide_operand = get_wide_operand(codeptr);
  pushAddress(frame, (*codeptr) - frame->method->code);
  *codeptr += wide_operand - 5;
  UPDATE_PC

}

INLINE static void opcode_ret (w_frame *frameptr, w_code *codeptr) {

  w_short  short_operand; 
  w_frame frame = *frameptr;

  short_operand = get_byte_operand(codeptr);
  *codeptr = frame->method->code + frame->locals[short_operand];
  UPDATE_PC

}

INLINE static void opcode_wide_ret (w_frame *frameptr, w_code *codeptr) {

  w_short  short_operand; 
  w_frame frame = *frameptr;

  short_operand = get_short_operand(codeptr);
  *codeptr = frame->method->code + frame->locals[short_operand];
  UPDATE_PC

}

INLINE static void opcode_tableswitch (w_frame *frameptr, w_code *codeptr) {

  w_code this_instruction = *codeptr - 1;
  w_code scanptr;
  w_int  indx;
  w_int  defaut;
  w_int  low;
  w_int  high;
  w_int  wide_jump;
  w_frame    frame = *frameptr;

  scanptr = frame->method->code + (((*codeptr) - frame->method->code + 3) & 0xfffffffc);
  defaut = get_wide_operand(&scanptr);
  low = get_wide_operand(&scanptr);
  high = get_wide_operand(&scanptr);

  woempa(1,"tableswitch with default jump by %d, low key %d (0x%x), high key %d (0x%x) \n",defaut,low,low,high,high);
          
  indx = popInt(frame);

  woempa(1,"key = %d (0x%x)\n",indx,indx);

  if ((indx < low) || (indx > high)) {
    wide_jump = defaut;
  }
  else {
    scanptr += (indx - low) * 4;
    wide_jump = get_wide_operand(&scanptr);
  }

  woempa(1, "jumping %d bytes, [%04d] here we come ...\n", wide_jump, (w_int)(this_instruction - (*frameptr)->method->code + wide_jump));
  *codeptr = this_instruction + wide_jump;
  UPDATE_PC

}

INLINE static void opcode_lookupswitch (w_frame *frameptr, w_code *codeptr) {

  w_code this_instruction = *codeptr - 1;
  w_int  key;
  w_int  match;
  w_code scanptr;
  w_int  npairs;
  w_int  pairnum;
  w_int  wide_operand; 
  w_int  wide_jump;
  w_frame    frame = *frameptr;

  scanptr = frame->method->code + (((*codeptr) - frame->method->code + 3) & 0xfffffffc);
  wide_jump = get_wide_operand(&scanptr);
  npairs = get_wide_operand(&scanptr);
  woempa(1,"lookupswitch with default jump by %d and %d match/offset pairs\n",wide_jump,npairs);

  key = popInt(frame);
  woempa(1,"key = %d (0x%x)\n",key,key);

  for(pairnum = 0; pairnum < npairs; pairnum++) {
    match = get_wide_operand(&scanptr);
    wide_operand = get_wide_operand(&scanptr);
    woempa(1,"entry %d: matches %d, jump by %d\n",pairnum,match,wide_operand);

    if (match == key) {
      wide_jump = wide_operand;
      woempa(1,"just the ticket, jumping %d bytes\n",wide_jump);
      break;
    }
    else if (match > key) {
      woempa(1,"key can't be present, getting out of here\n");
      break;
    }
  }

  woempa(1, "[%04d] here we come ...\n", (w_int)(this_instruction - frame->method->code + wide_jump));
  *codeptr = this_instruction + wide_jump;
  UPDATE_PC

}

/*
** The generic return code. It will dispatch to the correct return function handler.
*/

INLINE static void opcode_return (w_frame *frameptr, w_code *codeptr) {

  (*frameptr)->method->dispatcher->do_return(frameptr, codeptr);
  UPDATE_PC

}

INLINE static void opcode_ifnull (w_frame *frameptr, w_code *codeptr) {

  w_short  short_operand; 
  w_instance ref_value;

  short_operand = get_short_operand(codeptr);
  ref_value = popReference(*frameptr); 

  if (ref_value == NULL) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_ifnonnull (w_frame *frameptr, w_code *codeptr) {

  w_short  short_operand; 
  w_instance ref_value;

  short_operand = get_short_operand(codeptr);
  ref_value = popReference(*frameptr); 

  if (ref_value) {
    *codeptr += short_operand - 3;
    UPDATE_PC
  }

}

INLINE static void opcode_multianewarray (w_frame *frameptr, w_code *codeptr) {
  int        idx;
  w_clazz    array_clazz;
  w_instance arrayref;
  w_byte     dimensions;
  int  i,j,q;
  w_frame    frame = *frameptr;

  idx = get_short_operand(codeptr);
  array_clazz = getClassConstant(frame->method->clazz, idx);
  if ((*frameptr)->thread->Thread && (*frameptr)->thread->Thread[F_Thread_thrown]) {
    return;
  }
  
  dimensions = get_byte_operand(codeptr);

  /*
  ** Naughtiness warning:
  ** We pop all the dimensions off the stack and check that they are
  ** all nonnegative integers: but "of course" the dimensions are
  ** really still in memory, at locations (*frameptr)->o_top ..
  ** (*frameptr)->o_top + dimensions-1.
  */

  q = 0;
  for (i = 0; i < dimensions; i++) {
    j = popInt(frame);
    q |= j < 0;
  }

  if (q) {
    throwException(frame->thread, clazzNegativeArraySizeException, NULL);
    return;
  }

  arrayref = allocArrayInstance(frame->thread, array_clazz, dimensions, frame->o_top);

  if (arrayref) {
    pushReference(frame, arrayref);
    woempa(1,"UnGREENING %k at %p (pushed by multianewarray)\n", instance2clazz(arrayref),arrayref);
    unsetFlag(instance2flags(arrayref),O_GREEN);
  }
  else {
    throwException(frame->thread, clazzOutOfMemoryError, NULL);
  }

}

INLINE static void opcode_monitorenter (w_frame *frameptr, w_code *codeptr) {

  w_instance objectref;
  w_frame    frame = *frameptr;

  objectref = popReference(frame);

  if (checkReference(frame, objectref)) {
    objectref = makeLocal(objectref);
    frame->thread->state = W_THREAD_LOCKING;
    enterMonitor(objectref);
    frame->thread->state = W_THREAD_READY;
  }
  else {
    woempa(9, "instance %p is not valid!\n", objectref);
  }

}

INLINE static void opcode_monitorexit (w_frame *frameptr, w_code *codeptr) {

  w_instance objectref;
  w_frame    frame = *frameptr;

  objectref = popReference(frame);

  if ( ! checkReference(frame, objectref)) {
    woempa(9, "instance %p is not valid!\n", objectref);
    return;
  }

  objectref = makeLocal(objectref);

  /*
  ** Note that if thread is not owner of lock, exitMonitor will throw
  ** IllegalMonitorStateException ...
  */
  
  woempa(1, "Unlocking instance %p\n", objectref);

  exitMonitor(objectref);

}

INLINE static void opcode_athrow (w_frame *frameptr, w_code *codeptr) {

  w_instance throwable;
  w_frame    frame = *frameptr;
          
  throwable = popThrowable(frame);

  if (checkReference(frame, throwable)) {
    woempa(7, "throwing %p (%k) at [%5d] of %w\n", throwable, instance2clazz(throwable), (int)(*codeptr - frame->method->code), NM(frame->method));
    throwable = makeLocal(throwable);
    setLocalException(frame->thread, throwable);
  }

  handleException(frameptr, codeptr);

}

INLINE static void opcode_getstatic (w_frame *frameptr, w_code *codeptr) {
  int        idx;
  w_instance initException;
  w_field    field;
  w_frame    frame = *frameptr;

  idx = get_short_operand(codeptr);
  field = getFieldConstant(frame->method->clazz, idx);
  if (!field) {
    throwException_C((*frameptr)->thread, clazzLinkageError, "getstatic");
    return;
  }

  initException = classMustBeInitialized(field->declaring_clazz, (*frameptr)->thread);

  if (initException) {
    woempa(9, "Unable to intialize class %k for getstatic.\n", field->declaring_clazz);
    setLocalException(frame->thread, initException);
    return;
  }

  if (isSet(field->flags, FIELD_IS_LONG)) {
    woempa(1, "Getting data from static field[%d..%d] '%w' of %k\n", field->slot, field->slot+1, NM(field), field->declaring_clazz);
    frame->o_top[0] = field->declaring_clazz->staticFields[field->slot + 0];
    frame->o_top[1] = field->declaring_clazz->staticFields[field->slot + 1];
    frame->o_top += 2;
  }
  else {
    woempa(1, "Getting data from static field[%d] '%w' of %k (%p)\n", field->slot, NM(field), field->declaring_clazz,&field->declaring_clazz->staticFields[field->slot]);
    frame->o_top[0] = field->declaring_clazz->staticFields[field->slot];
    frame->o_top += 1;
  }

}

INLINE static void opcode_putstatic (w_frame *frameptr, w_code *codeptr) {
  int        idx;
  w_instance initException;
  w_field    field;
  w_frame    frame = *frameptr;

  idx = get_short_operand(codeptr);
  field = getFieldConstant(frame->method->clazz, idx);
  if (!field) {
    throwException_C((*frameptr)->thread, clazzLinkageError, "putstatic");
    return;
  }

  initException = classMustBeInitialized(field->declaring_clazz, (*frameptr)->thread);

  if (initException) {
    woempa(9, "Unable to intialize class %k for putstatic.\n", field->declaring_clazz);
    setLocalException(frame->thread, initException);
    return;
  }

  woempa(1, "GC Putting data in static field '%w' of %k (%p)\n", NM(field), field->declaring_clazz,&field->declaring_clazz->staticFields[field->slot]);

  if (isSet(field->flags, FIELD_IS_LONG)) {
    field->declaring_clazz->staticFields[field->slot + 1] = frame->o_top[-1];
    field->declaring_clazz->staticFields[field->slot + 0] = frame->o_top[-2];
    frame->o_top -= 2;
#ifdef ENABLE_PS
    {
      w_instance Class = clazz2Class(field->clazz);
      if (isLocalCopy(Class)) {
        besmirchLocalCopy(Class);
      }
    }
#endif
  }
  else {
    if (field->desc->tag != D_Primitive) {
      setStaticReferenceField(field->declaring_clazz, field->slot, (w_instance) frame->o_top[-1]);
    }
    else {
      field->declaring_clazz->staticFields[field->slot] = frame->o_top[-1];
#ifdef ENABLE_PS
      {
        w_instance Class = clazz2Class(field->clazz);
        if (isLocalCopy(Class)) {
          besmirchLocalCopy(Class);
        }
      }
#endif
    }
    frame->o_top -= 1;
  }

}

INLINE static void opcode_getfield (w_frame *frameptr, w_code *codeptr) {
  int        idx;
  w_instance objectref;
  w_field    field;
  w_frame    frame = *frameptr;

  idx = get_short_operand(codeptr);
  field = getFieldConstant(frame->method->clazz, idx);
  if (!field) {
    throwException_C((*frameptr)->thread, clazzLinkageError, "getfield");
    return;
  }
  objectref = popReference(frame);

  if ( ! checkReference(frame, objectref)) {
    return;
  }

  objectref = makeLocal(objectref);
  woempa(1, "Get field[%d] '%w' (%k) of instance %p (%k), words = %d\n", field->slot, NM(field), field->declaring_clazz, objectref, instance2clazz(objectref), fieldSize(field));

  if (isSet(field->flags, FIELD_IS_LONG)) {
    frame->o_top[0] = objectref[field->slot + 0];
    frame->o_top[1] = objectref[field->slot + 1];
    frame->o_top += 2;
  }
  else {
    frame->o_top[0] = objectref[field->slot];
    frame->o_top += 1;
  }

}

INLINE static void opcode_putfield (w_frame *frameptr, w_code *codeptr) {
  int        idx;
  w_instance objectref;
  w_field    field;
  w_frame    frame = *frameptr;

  idx = get_short_operand(codeptr);
  field = getFieldConstant(frame->method->clazz, idx);
  if (!field) {
    throwException_C((*frameptr)->thread, clazzLinkageError, "putfield");
    return;
  }

  /*
  ** To avoid checking for the size of the field (32 or 64 bits) twice. We peek for the object
  ** reference on the stack, by means of the following statement. We don't pop items from the
  ** stack yet. We should not pop here since if we were interrupted here, a possible popped
  ** reference is exposed to the garbage collector, before it was written in the reference field.
  ** We will do the pop when an error occured or after writing into the reference field.
  */

  objectref = (w_instance) frame->o_top[ - fieldSize(field) - 1];

  if ( ! checkReference(frame, objectref)) {

    /*
    ** An error occured but we still have to pop the items from the stack since we only peeked
    ** at the stack at this time.
    */

    frame->o_top -= (fieldSize(field) + 1);
    return;
  }

  /*
  ** Everything is OK here, we proceed...
  */

  objectref = makeLocal(objectref);
  if (isSet(field->flags, FIELD_IS_LONG)) {
    objectref[field->slot + 1] = frame->o_top[-1];
    objectref[field->slot + 0] = frame->o_top[-2];
#ifdef ENABLE_PS
    if (isLocalCopy(objectref)) {
      besmirchLocalCopy(objectref);
    }
#endif

    woempa(1, "Put long 0x%08x%08x in field[%d..%d] '%w' of instance %p of class %k (%k) .\n", frame->o_top[-1], frame->o_top[-2], field->slot, field->slot + 1, NM(field), objectref, instance2clazz(objectref), field->declaring_clazz);
    frame->o_top -= 3;
  }
  else {
    if (field->desc->tag != D_Primitive) {
      setReferenceField(objectref, (w_instance) frame->o_top[-1], field->slot);
    }
    else {
      objectref[field->slot] = frame->o_top[-1];
#ifdef ENABLE_PS
      if (isLocalCopy(objectref)) {
        besmirchLocalCopy(objectref);
      }
#endif
    }

    woempa(1, "Put word 0x%08x in field[%d] '%w' (from %k) of instance %p of class %k.\n", frame->o_top[-1], field->slot, NM(field), field->declaring_clazz, objectref, instance2clazz(objectref));
    frame->o_top -= 2;
  }

}

INLINE static void opcode_invokevirtual (w_frame *frameptr, w_code *codeptr) {
  int        idx;
  w_method   method;
  w_instance objectref;
  w_clazz    object_clazz;

  if ((*frameptr)->depth >= MAX_STACK_DEPTH) {
    woempa(9, "Thread %w: stack overflow at [%d] in %M\n", (*frameptr)->thread->name, *codeptr - (*frameptr)->method->code, (*frameptr)->method);
    throwException((*frameptr)->thread, clazzStackOverflowError, NULL);

    return;

  }

  idx = get_short_operand(codeptr);
  method = getMethodConstant((*frameptr)->method->clazz, idx);
  if (!method) {
    throwException_C((*frameptr)->thread, clazzLinkageError, "invokevirtual");
    return;
  }

  if (isSet(method->flags, ACC_STATIC)) {
    throwException((*frameptr)->thread, clazzIncompatibleClassChangeError, method->name);
    return;
  }

  objectref = peekReference(*frameptr, method);

  if (!objectref) {
    throwException((*frameptr)->thread, clazzNullPointerException, NULL);
    return;
  }

  if (isSet(method->flags, METHOD_NEEDS_LINKING)) {
    /* Don't worry, dispatcher will fix this ... */
  }

  objectref = makeLocal(objectref);
  object_clazz = instance2clazz(objectref);

  if ( isSet(method->clazz->flags, ACC_FINAL)
    || isSet(method->flags, ACC_FINAL | ACC_PRIVATE)
    ) {
    if (isSet(method->flags,ACC_ABSTRACT)) {
      throwException((*frameptr)->thread, clazzAbstractMethodError, method->name);
      return;
    }

    stack_pops_pushes(*frameptr, (w_int)method->arg_i, (w_int)method->return_i);
    woempa(1,"invoke special method %M directly\n", method);
    method->dispatcher->do_call(frameptr, codeptr, method);
  }
  else {
    woempa(1, "Doing a virtual lookup of %M using object ref %k@%p\n", method, object_clazz, objectref);
    method = object_clazz->methods[method->slot];

    if (isSet(method->flags,ACC_ABSTRACT)) {
      throwException((*frameptr)->thread, clazzAbstractMethodError, method->name);
      return;
    }

    stack_pops_pushes(*frameptr, (w_int)method->arg_i, (w_int)method->return_i);
    woempa(1,"invoke special method %M\n", method);
    method->dispatcher->do_call(frameptr, codeptr, method);
  }
 
  UPDATE_PC
}

INLINE static void opcode_invokespecial (w_frame *frameptr, w_code *codeptr) {
  int        idx; 
  w_method   method;
  w_instance objectref;

  if ((*frameptr)->depth >= MAX_STACK_DEPTH) {
    woempa(9, "Thread %w: stack overflow at [%d] in %M\n", (*frameptr)->thread->name, *codeptr - (*frameptr)->method->code, (*frameptr)->method);
    throwException((*frameptr)->thread, clazzStackOverflowError, NULL);

    return;

  }

  idx = get_short_operand(codeptr);
  method = getMethodConstant((*frameptr)->method->clazz, idx);
  if (!method) {
    throwException((*frameptr)->thread, clazzLinkageError, NULL);
    return;
  }

  if (isSet(method->flags, ACC_STATIC)) {
    throwException((*frameptr)->thread, clazzIncompatibleClassChangeError, method->name);
    return;
  }

  objectref = peekReference(*frameptr, method);
  objectref = makeLocal(objectref);

  if (!objectref) {
    throwException((*frameptr)->thread, clazzNullPointerException, NULL);
    return;
  }

  if (isSet(method->flags, METHOD_NEEDS_LINKING)) {
    /* Don't worry, dispatcher will fix this ... */
  }

  if ( (method->clazz->flags & (ACC_FINAL | ACC_SUPER)) != ACC_SUPER
//    || method->name == string_angle_brackets_init
    || isSet(method->flags, ACC_PRIVATE | METHOD_IS_CONSTRUCTOR)
    ) {
    if (isSet(method->flags,ACC_ABSTRACT)) {
      throwException((*frameptr)->thread, clazzAbstractMethodError, method->name);
      return;
    }

    stack_pops_pushes(*frameptr, (w_int)method->arg_i, (w_int)method->return_i);
    woempa(1,"invoke special method %M directly\n", method);
    method->dispatcher->do_call(frameptr, codeptr, method);
  }
  else {
    w_clazz  super_clazz = (*frameptr)->method->clazz->superClazz;

    if (isSuperClass(method->clazz, super_clazz)) {
      woempa(1, "Declaring class of %M is a superclass of current method %M, doing a virtual lookup from %k\n", method, (*frameptr)->method, super_clazz);
      method = super_clazz->methods[method->slot];
    }
    if (isSet(method->flags,ACC_ABSTRACT)) {
      throwException((*frameptr)->thread, clazzAbstractMethodError, method->name);
      return;
    }

    stack_pops_pushes(*frameptr, (w_int)method->arg_i, (w_int)method->return_i);
    woempa(1,"invoke special method %M\n", method);
    method->dispatcher->do_call(frameptr, codeptr, method);
  }


  UPDATE_PC
}

INLINE static void opcode_invokestatic (w_frame *frameptr, w_code *codeptr) {
  int        idx;
  w_method  method;
  w_frame frame = *frameptr;

  if ((*frameptr)->depth >= MAX_STACK_DEPTH) {
    woempa(9, "Thread %w: stack overflow at [%d] in %M\n", (*frameptr)->thread->name, *codeptr - (*frameptr)->method->code, (*frameptr)->method);
    throwException((*frameptr)->thread, clazzStackOverflowError, NULL);

    return;

  }

  idx = get_short_operand(codeptr);
  method = getMethodConstant(frame->method->clazz, idx);
  if (!method) {
    throwException_C((*frameptr)->thread, clazzLinkageError, "invokestatic -- could not resolve method constant");
    return;
  }

  if ((method->flags & (ACC_STATIC | METHOD_IS_CLINIT)) != ACC_STATIC) {
    if (isSet(method->flags, METHOD_IS_CLINIT)) {
      throwException_C((*frameptr)->thread, clazzIncompatibleClassChangeError, "invokestatic -- cannot invoke <clinit>");
    }
    else {
      throwException_C((*frameptr)->thread, clazzIncompatibleClassChangeError, "invokestatic -- not a static method");
    }
    return;
  }

  if (isSet(method->flags, METHOD_NEEDS_LINKING)) {
    /* Don't worry, dispatcher will fix this ... */
  }

  method->dispatcher->do_call(frameptr, codeptr, method);

  UPDATE_PC
}


INLINE static void opcode_invokeinterface (w_frame *frameptr, w_code *codeptr) {
  int        idx;
  w_instance initException;
  w_method   interf_method;
  w_method   actual_method;
  w_instance objectref;
  w_clazz    actual_clazz;

  if ((*frameptr)->depth >= MAX_STACK_DEPTH) {
    woempa(9, "Thread %w: stack overflow at [%d] in %M\n", (*frameptr)->thread->name, *codeptr - (*frameptr)->method->code, (*frameptr)->method);
    throwException((*frameptr)->thread, clazzStackOverflowError, NULL);

    return;

  }

  idx = get_short_operand(codeptr);
  interf_method = getIMethodConstant((*frameptr)->method->clazz, idx);
  if (!interf_method) {
    throwException_C((*frameptr)->thread, clazzLinkageError, "invokeinterface");
    return;
  }

  /*
  ** Skip over the useless <nargs> and zero byte
  */
  (void)get_short_operand(codeptr);
/*
** TODO : exception = ...
*/
    classMustBeInitialized(interf_method->clazz, (*frameptr)->thread);
    objectref = peekReference(*frameptr,interf_method);

    if ( ! checkReference(*frameptr, objectref)) {

      return;

    }

    objectref = makeLocal(objectref);
    actual_clazz = instance2clazz(objectref);
    woempa(1,"invoke interface method %M on %p (%k)\n", interf_method, objectref,actual_clazz);
    actual_method = (w_method)ht2k_read(interface_hashtable, (w_word)actual_clazz, (w_word)interf_method);
    if(actual_method==NULL) {
      throwException((*frameptr)->thread, clazzIncompatibleClassChangeError, NULL);

      return;

    }

    if(isSet(actual_method->flags, ACC_STATIC)) {
      throwException((*frameptr)->thread, clazzIncompatibleClassChangeError, NULL);

      return;

    }

    woempa(1,"will use method %M\n", actual_method);
    if (!isSet(actual_method->flags, ACC_PUBLIC)) {
      throwException((*frameptr)->thread, clazzIllegalAccessError, NULL);

      return;

    }
    if (isSet(actual_method->flags, ACC_ABSTRACT)) {
      throwException((*frameptr)->thread, clazzAbstractMethodError, NULL);

      return;

    }

    initException = classMustBeInitialized(actual_method->clazz, (*frameptr)->thread);
    if (initException) {
      woempa(9,"Unable to intialize class %k for invokeinterface\n",actual_method->clazz);
      setLocalException((*frameptr)->thread,initException);

      return;
    }


  /*
  ** Now invoke the interface method.
  */

    actual_method->dispatcher->do_call(frameptr, codeptr, actual_method);
    UPDATE_PC

}

INLINE static void opcode_new (w_frame *frameptr, w_code *codeptr) {
  int        idx;
  w_clazz    new_clazz;
  w_instance objectref;
  w_instance initException;
  w_frame frame = *frameptr;

  idx = get_short_operand(codeptr);
  new_clazz = getClassConstant(frame->method->clazz, idx);
  if ((*frameptr)->thread->Thread && (*frameptr)->thread->Thread[F_Thread_thrown]) {
    return;
  }
  
  woempa(1,"Class %k must be intialized ...\n",new_clazz);
/*
** TODO : exception = ...
*/
  initException = classMustBeInitialized(new_clazz, (*frameptr)->thread);

  if (initException) {
    woempa(9, "Unable to intialize class %k for new.\n", new_clazz);
    setLocalException(frame->thread, initException);
    return;
  }

  woempa(1,"Allocating an instance of %k\n",new_clazz);
  objectref = allocInstance(frame->thread, new_clazz);
  if(objectref) {
    pushReference(frame, objectref);
    woempa(1,"UnGREENING %k at %p (pushed by new)\n",instance2clazz(objectref),objectref);
    unsetFlag(instance2flags(objectref),O_GREEN);
  }

}

INLINE static void opcode_newarray (w_frame *frameptr, w_code *codeptr) {

  w_byte     atype;
  w_clazz    array_clazz;
  w_instance arrayref;
  w_int      size;
  w_frame frame = *frameptr;

  size = popInt(frame);
  atype = get_byte_operand(codeptr);
  array_clazz = atype2clazz[atype];

  if (size < 0) {
    throwException((*frameptr)->thread, clazzNegativeArraySizeException, NULL);
    return;
  }

  woempa(1, "Allocating single dimensional array of size %d.\n", size); 
  arrayref = allocArrayInstance(frame->thread, array_clazz, 1, &size);

  if (arrayref) {
    pushReference(frame, arrayref);
    woempa(1,"UnGREENING %k at %p (pushed by newarray)\n",instance2clazz(arrayref),arrayref);
    unsetFlag(instance2flags(arrayref),O_GREEN);
  }
  else {
    throwException((*frameptr)->thread, clazzOutOfMemoryError, NULL);
  }

}

INLINE static void opcode_anewarray (w_frame *frameptr, w_code *codeptr) {
  int        idx;
  w_clazz    component_clazz;
  w_clazz    array_clazz;
  w_int      size;
  w_instance thenewarray;
  w_frame frame = *frameptr;
  w_instance  exception = NULL;

  idx = get_short_operand(codeptr);
  component_clazz = getClassConstant(frame->method->clazz, idx);
  if ((*frameptr)->thread->Thread && (*frameptr)->thread->Thread[F_Thread_thrown]) {
    return;
  }
 
  size = popInt(frame);

  if (size < 0) {
    throwException_C((*frameptr)->thread, clazzNegativeArraySizeException, "anewarray");
    return;
  }

  if (!component_clazz->adesc) {
    w_char * desc_buffer;
    w_string desc_string;
    w_size   i;
    w_char   ch;

    desc_buffer = allocMem(frame->thread, (component_clazz->dotified->length + 3) * sizeof(w_char));
    desc_buffer[0] = (w_char)'[';
    desc_buffer[1] = (w_char)'L';
    for (i = 0; i < component_clazz->dotified->length; ++i) {
      ch = component_clazz->dotified->chars[i];
      if (ch == '.') {
        desc_buffer[i + 2] = '/';
      }
      else {
        desc_buffer[i + 2] = ch;
      }
    }
    desc_buffer[component_clazz->dotified->length + 2] = ';';
    desc_string = unicode2String(frame->thread, desc_buffer, component_clazz->dotified->length + 3);
    component_clazz->adesc = newFieldDescriptor(frame->thread, desc_string);
    if (component_clazz->adesc == NULL) {
      wabort(ABORT_WONKA,"whoops - failed to parse descriptor `%w'\n", desc_string);
    }

    deregisterString(desc_string);

    woempa(1, "Registered descriptor %w for array of %k.\n", NM(component_clazz->adesc), component_clazz);

  }
  array_clazz = seekClazzByDesc(component_clazz->adesc, loader2hashtable(clazz2loader(component_clazz)));

  if (array_clazz == NULL) {
    array_clazz = createArrayClazzFromDesc(clazz2loader(component_clazz), component_clazz->adesc, &exception);
  } 

  if (exception) {
    woempa(9, "Unable to resolve class %k for anewarray\n", component_clazz);
    setLocalException(frame->thread, exception);
    return;
  }

  if (array_clazz != NULL) {  
    thenewarray = allocArrayInstance(frame->thread, array_clazz, 1, &size);
  }
  else {
    thenewarray = NULL;
  }

  if (array_clazz && thenewarray) {
    pushReference(frame, thenewarray);
    woempa(1,"UnGREENING %k at %p (pushed by anewarray)\n",instance2clazz(thenewarray),thenewarray);
    unsetFlag(instance2flags(thenewarray),O_GREEN);
  }
  else {
    woempa(1, "Could not allocate %d reference cells...\n", size);
    throwException((*frameptr)->thread, clazzOutOfMemoryError, NULL);
  }

}

INLINE static void opcode_arraylength (w_frame *frameptr, w_code *codeptr) {

  w_instance arrayref;
  w_int      length;
  w_frame frame = *frameptr;

  arrayref = popArrayRef(frame);
  if (checkReference(frame, arrayref)) {
   arrayref = makeLocal(arrayref);
   length = (w_int)arrayref[F_Array_length];
   woempa(1, "Length of %p (%k) is %d.\n", arrayref, instance2clazz(arrayref), length);
   pushInt(frame, length);
 }

}

INLINE static void opcode_checkcast (w_frame *frameptr, w_code *codeptr) {
  int        idx;
  w_clazz    targetclazz;
  w_instance sourceinstance;
  w_frame frame = *frameptr;

  idx = get_short_operand(codeptr);
  targetclazz = getClassConstant(frame->method->clazz, idx);
  if ((*frameptr)->thread->Thread && (*frameptr)->thread->Thread[F_Thread_thrown]) {
    return;
  }
  
  sourceinstance = (w_instance) frame->o_top[-1];
  if (sourceinstance && !isAssignmentCompatible(instance2clazz(sourceinstance), targetclazz)) {
    throwException((*frameptr)->thread, clazzClassCastException, NULL);
  }

}

INLINE static void opcode_instanceof (w_frame *frameptr, w_code *codeptr) {
  int        idx;
  w_clazz    targetclazz;
  w_instance sourceinstance;
  w_int      int_result;
  w_frame frame = *frameptr;

  idx = get_short_operand(codeptr);
  targetclazz = getClassConstant(frame->method->clazz, idx);
  if ((*frameptr)->thread->Thread && (*frameptr)->thread->Thread[F_Thread_thrown]) {
    return;
  }
  
  sourceinstance = popReference(frame);

  if (sourceinstance == NULL) {
    pushInt(frame, 0);
  }
  else {
    sourceinstance = makeLocal(sourceinstance);
    int_result = isAssignmentCompatible(instance2clazz(sourceinstance), targetclazz);
    pushInt(frame, int_result);
  }

}

#define CHECK_FOR_ASYNCH_EXCEPTION \
  if (!checkForAsynchException(frameptr, codeptr)) {break;}

#define CHECK_FOR_SYNCH_EXCEPTION  \
  if ((*frameptr) && isSet((*frameptr)->thread->flags, (WT_THREAD_THROWN | WT_THREAD_KILLED))) \
    {handleException(frameptr, codeptr);}        \
  if (*frameptr == NULL) { proceed = WONKA_FALSE; }

/*
** Perform a wide opcode. Since only a few opcodes exists in the 'wide' variant, we have made a special
** handler for these that is executed right after the 'wide' opcode.
*/

static w_boolean perform_wide_opcode (w_frame *frameptr, w_code *codeptr) {

  w_opcode  opcode;
 
  opcode = *(*codeptr)++;

    woempa(1, "Thread %w:  %M\n", NM((*frameptr)->thread), (*frameptr)->method);
    woempa(1, "Thread %w:  CODE[%04d] %-16s(wide)\n", NM((*frameptr)->thread), (int)(*codeptr - 1 - (*frameptr)->method->code), opc2name(opcode));

    switch (opcode) {
      case (iload):
        opcode_wide_iload(frameptr, codeptr);
        break;
 
      case (lload):
        opcode_wide_lload(frameptr, codeptr);
        break;
 
      case (fload):
        opcode_wide_fload(frameptr, codeptr);
        break;

      case (dload):
        opcode_wide_dload(frameptr, codeptr);
        break;

      case (aload):
        opcode_wide_aload(frameptr, codeptr);
        break;

      case (istore):
        opcode_wide_istore(frameptr, codeptr);
        break;

      case (lstore):
        opcode_wide_lstore(frameptr, codeptr);
        break;
    
      case (fstore):
        opcode_wide_fstore(frameptr, codeptr);
        break;

      case (dstore):
        opcode_wide_dstore(frameptr, codeptr);
        break;

      case (astore):
        opcode_wide_astore(frameptr, codeptr);
        break;

      case (iinc):
        opcode_wide_iinc(frameptr, codeptr);
        break;
  
      case (ret):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_wide_ret(frameptr, codeptr);
        break;
  
      default:
        wabort(ABORT_WONKA, "Duizend bommen en granaten! Wide opcode '%s' is not defined!.\n", opc2name(opcode));
        break;
        
    }

  return (*frameptr != NULL);
}

#undef S_CHECK_FOR_ASYNCH_EXCEPTION
#define S_CHECK_FOR_ASYNCH_EXCEPTION {                  \
  if (! checkForAsynchException(frameptr, codeptr)) {   \
    break;                                              \
  }                                                     \
}

#ifdef DEBUG
extern int woempa_bytecodecount;
extern int woempa_bytecodetrigger;
#endif

/*
** Returns an instance of Throwable if an exception was thrown, else NULL.
*/
w_instance 
runThread(w_thread thread, w_frame startFrame) {

  w_frame   c_frame = startFrame;
  w_code    c_codeptr = c_frame->method->code;
  w_boolean proceed = WONKA_TRUE;
  w_frame * frameptr;
  w_code * codeptr;
  w_opcode  opcode;
  w_code    break_point = NULL;

  frameptr = &c_frame;
  codeptr = &c_codeptr;

  classMustBeInitialized(c_frame->method->clazz, thread);


  if (isSet(c_frame->method->flags, ACC_NATIVE)) {
    w_method m = c_frame->method;
    m->dispatcher->do_call(&c_frame, &c_codeptr, m);

    woempa(1, "Returning from native method '%w', flags are 0x%08x.\n", NM(startFrame->method), thread->flags);
    if (isNotSet(thread->flags, (WT_THREAD_THROWN | WT_THREAD_KILLED))) {
      return NULL;
    }
    else { 
      w_instance thrown = (w_instance)thread->Thread[F_Thread_thrown];
      setReferenceField(thread->Thread, NULL, F_Thread_thrown);
      unsetFlag(thread->flags, WT_THREAD_KILLED | WT_THREAD_THROWN);
      return thrown;
    }
  }
    
  while (proceed) {

#ifdef DEBUG
    woempa_bytecodecount++;
#endif
    opcode = *(*codeptr)++;

    woempa(1, "Thread %w:  %M\n", NM((*frameptr)->thread), (*frameptr)->method);
    woempa(1, "Thread %w:  CODE[%04d] %-16s\n", NM((*frameptr)->thread), (int)(*codeptr - 1 - (*frameptr)->method->code), opc2name(opcode));

    switch (opcode) {
      case (nop):
        opcode_nop(frameptr, codeptr);
        break;

      case (aconst_null):
        opcode_aconst_null(frameptr, codeptr);
        break;
        
      case (iconst_m1):
        opcode_iconst_m1(frameptr, codeptr);
        break;

      case (iconst_0):
        opcode_iconst_0(frameptr, codeptr);
        break;

      case (iconst_1):
        opcode_iconst_1(frameptr, codeptr);
        break;

      case (iconst_2):
        opcode_iconst_2(frameptr, codeptr);
        break;

      case (iconst_3):
        opcode_iconst_3(frameptr, codeptr);
        break;

      case (iconst_4):
        opcode_iconst_4(frameptr, codeptr);
        break;

      case (iconst_5):
        opcode_iconst_5(frameptr, codeptr);
        break;

      case (lconst_0):
        opcode_lconst_0(frameptr, codeptr);
        break;

      case (lconst_1):
        opcode_lconst_1(frameptr, codeptr);
        break;

      case (fconst_0):
        opcode_fconst_0(frameptr, codeptr);
        break;
        
      case (fconst_1):
        opcode_fconst_1(frameptr, codeptr);
        break;

      case (fconst_2):
        opcode_fconst_2(frameptr, codeptr);
        break;

      case (dconst_0):
        opcode_dconst_0(frameptr, codeptr);
        break;

      case (dconst_1):
        opcode_dconst_1(frameptr, codeptr);
        break;
        
      case (bipush):
        opcode_bipush(frameptr, codeptr);
        break;

      case (sipush):
        opcode_sipush(frameptr, codeptr);
        break;

      case (ldc):
        opcode_ldc(frameptr, codeptr);
        break;

      case (ldc_w):
        opcode_ldc_w(frameptr, codeptr);
        break;

      case (ldc2_w):
        opcode_ldc2_w(frameptr, codeptr);
        break;

      case (iload):
        opcode_iload(frameptr, codeptr);
        break;
 
      case (lload):
        opcode_lload(frameptr, codeptr);
        break;
 
      case (fload):
        opcode_fload(frameptr, codeptr);
        break;

      case (dload):
        opcode_dload(frameptr, codeptr);
        break;

      case (aload):
        opcode_aload(frameptr, codeptr);
        break;

      case (iload_0):
        opcode_iload_0(frameptr, codeptr);
        break;
 
      case (iload_1):
        opcode_iload_1(frameptr, codeptr);
        break;
 
      case (iload_2):
        opcode_iload_2(frameptr, codeptr);
        break;
 
      case (iload_3):
        opcode_iload_3(frameptr, codeptr);
        break;
 
      case (lload_0):
        opcode_lload_0(frameptr, codeptr);
        break;
 
      case (lload_1):
        opcode_lload_1(frameptr, codeptr);
        break;
 
      case (lload_2):
        opcode_lload_2(frameptr, codeptr);
        break;
 
      case (lload_3):
        opcode_lload_3(frameptr, codeptr);
        break;
 
      case (fload_0):
        opcode_fload_0(frameptr, codeptr);
        break;
 
      case (fload_1):
        opcode_fload_1(frameptr, codeptr);
        break;
 
      case (fload_2):
        opcode_fload_2(frameptr, codeptr);
        break;
 
      case (fload_3):
        opcode_fload_3(frameptr, codeptr);
        break;
 
      case (dload_0):
        opcode_dload_0(frameptr, codeptr);
        break;
 
      case (dload_1):
        opcode_dload_1(frameptr, codeptr);
        break;
 
      case (dload_2):
        opcode_dload_2(frameptr, codeptr);
        break;
 
      case (dload_3):
        opcode_dload_3(frameptr, codeptr);
        break;
 
      case (aload_0):
        opcode_aload_0(frameptr, codeptr);
        break;
 
      case (aload_1):
        opcode_aload_1(frameptr, codeptr);
        break;
 
      case (aload_2):
        opcode_aload_2(frameptr, codeptr);
        break;
 
      case (aload_3):
        opcode_aload_3(frameptr, codeptr);
        break;
 
      case (iaload):
        opcode_iaload(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (laload):
        opcode_laload(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (faload):
        opcode_faload(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (daload):
        opcode_daload(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (aaload):
        opcode_aaload(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (baload):
        opcode_baload(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (caload):
        opcode_caload(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (saload):
        opcode_saload(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (istore):
        opcode_istore(frameptr, codeptr);
        break;

      case (lstore):
        opcode_lstore(frameptr, codeptr);
        break;
    
      case (fstore):
        opcode_fstore(frameptr, codeptr);
        break;

      case (dstore):
        opcode_dstore(frameptr, codeptr);
        break;

      case (astore):
        opcode_astore(frameptr, codeptr);
        break;

      case (istore_0):
        opcode_istore_0(frameptr, codeptr);
        break;

      case (istore_1):
        opcode_istore_1(frameptr, codeptr);
        break;

      case (istore_2):
        opcode_istore_2(frameptr, codeptr);
        break;

      case (istore_3):
        opcode_istore_3(frameptr, codeptr);
        break;

      case (lstore_0):
        opcode_lstore_0(frameptr, codeptr);
        break;

      case (lstore_1):
        opcode_lstore_1(frameptr, codeptr);
        break;

      case (lstore_2):
        opcode_lstore_2(frameptr, codeptr);
        break;

      case (lstore_3):
        opcode_lstore_3(frameptr, codeptr);
        break;

      case (fstore_0):
        opcode_fstore_0(frameptr, codeptr);
        break;

      case (fstore_1):
        opcode_fstore_1(frameptr, codeptr);
        break;

      case (fstore_2):
        opcode_fstore_2(frameptr, codeptr);
        break;

      case (fstore_3):
        opcode_fstore_3(frameptr, codeptr);
        break;

      case (dstore_0):
        opcode_dstore_0(frameptr, codeptr);
        break;

      case (dstore_1):
        opcode_dstore_1(frameptr, codeptr);
        break;

      case (dstore_2):
        opcode_dstore_2(frameptr, codeptr);
        break;

      case (dstore_3):
        opcode_dstore_3(frameptr, codeptr);
        break;

      case (astore_0):
        opcode_astore_0(frameptr, codeptr);
        break;

      case (astore_1):
        opcode_astore_1(frameptr, codeptr);
        break;

      case (astore_2):
        opcode_astore_2(frameptr, codeptr);
        break;

      case (astore_3):
        opcode_astore_3(frameptr, codeptr);
        break;

      case (iastore):
        opcode_iastore(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (lastore):
        opcode_lastore(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (fastore):
        opcode_fastore(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (dastore):
        opcode_dastore(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (aastore):
        opcode_aastore(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (bastore):
        opcode_bastore(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (castore):
        opcode_castore(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (sastore):
        opcode_sastore(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (pop):
        opcode_pop(frameptr, codeptr);
        break;

      case (pop2):
        opcode_pop2(frameptr, codeptr);
        break;

      case (dup):
        opcode_dup(frameptr, codeptr);
        break;

      case (dup_x1):
        opcode_dup_x1(frameptr, codeptr);
        break;

      case (dup_x2):
        opcode_dup_x2(frameptr, codeptr);
        break;

      case (dup2):
        opcode_dup2(frameptr, codeptr);
        break;

      case (dup2_x1):        
        opcode_dup2_x1(frameptr, codeptr);
        break;

      case (dup2_x2):
        opcode_dup2_x2(frameptr, codeptr);
        break;

      case (swap):
        opcode_swap(frameptr, codeptr);
        break;

      case (iadd):
        opcode_iadd(frameptr, codeptr);
        break;

      case (ladd):
        opcode_ladd(frameptr, codeptr);
        break;

      case (fadd):
        opcode_fadd(frameptr, codeptr);
        break;
   
      case (dadd):
        opcode_dadd(frameptr, codeptr);
        break;
   
      case (isub):
        opcode_isub(frameptr, codeptr);
        break;

      case (lsub):
        opcode_lsub(frameptr, codeptr);
        break;

      case (fsub):
        opcode_fsub(frameptr, codeptr);
        break;
   
      case (dsub):
        opcode_dsub(frameptr, codeptr);
        break;
   
      case (imul):
        opcode_imul(frameptr, codeptr);
        break;

      case (lmul):
        opcode_lmul(frameptr, codeptr);
        break;

      case (fmul):
        opcode_fmul(frameptr, codeptr);
        break;
   
      case (dmul):
        opcode_dmul(frameptr, codeptr);
        break;
   
      case (idiv):
        opcode_idiv(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (ldiv):
        opcode_ldiv(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (fdiv):
        opcode_fdiv(frameptr, codeptr);
        break;
   
      case (ddiv):
        opcode_ddiv(frameptr, codeptr);
        break;
   
      case (irem):
        opcode_irem(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;
   
      case (lrem):
        opcode_lrem(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;
   
      case (frem):
        opcode_frem(frameptr, codeptr);
        break;
   
      case (drem):
        opcode_drem(frameptr, codeptr);
        break;
   
      case (ineg):
        opcode_ineg(frameptr, codeptr);
        break;

      case (lneg):
        opcode_lneg(frameptr, codeptr);
        break;

      case (fneg):
        opcode_fneg(frameptr, codeptr);
        break;
    
      case (dneg):
        opcode_dneg(frameptr, codeptr);
        break;
    
      case (ishl):
        opcode_ishl(frameptr, codeptr);
        break;

      case (lshl):
        opcode_lshl(frameptr, codeptr);
        break;

      case (ishr):
        opcode_ishr(frameptr, codeptr);
        break;

      case (lshr):
        opcode_lshr(frameptr, codeptr);
        break;

      case (iushr):
        opcode_iushr(frameptr, codeptr);
        break;

      case (lushr):
        opcode_lushr(frameptr, codeptr);
        break;

      case (iand):
        opcode_iand(frameptr, codeptr);
        break;

      case (land):
        opcode_land(frameptr, codeptr);
        break;

      case (ior):
        opcode_ior(frameptr, codeptr);
        break;

      case (lor):
        opcode_lor(frameptr, codeptr);
        break;

      case (ixor):
        opcode_ixor(frameptr, codeptr);
        break;

      case (lxor):
        opcode_lxor(frameptr, codeptr);
        break;

      case (iinc):
        opcode_iinc(frameptr, codeptr);
        break;
  
      case (i2l):
        opcode_i2l(frameptr, codeptr);
        break;

      case (i2f):
        opcode_i2f(frameptr, codeptr);
        break;

      case (i2d):
        opcode_i2d(frameptr, codeptr);
        break;

      case (l2i):
        opcode_l2i(frameptr, codeptr);
        break;

      case (l2f):
        opcode_l2f(frameptr, codeptr);
        break;
      
      case (l2d):
        opcode_l2d(frameptr, codeptr);
        break;

      case (f2i):
        opcode_f2i(frameptr, codeptr);
        break;

      case (f2l):
        opcode_f2l(frameptr, codeptr);
        break;

      case (f2d):
        opcode_f2d(frameptr, codeptr);
        break;

      case (d2i):
        opcode_d2i(frameptr, codeptr);
        break;

      case (d2l):
        opcode_d2l(frameptr, codeptr);
        break;

      case (d2f):
        opcode_d2f(frameptr, codeptr);
        break;
 
      case (i2b):
        opcode_i2b(frameptr, codeptr);
        break;
        
      case (i2c):
        opcode_i2c(frameptr, codeptr);
        break;
        
      case (i2s):
        opcode_i2s(frameptr, codeptr);
        break;
        
      case (lcmp):
        opcode_lcmp(frameptr, codeptr);
        break;

      case (fcmpl):
        opcode_fcmpl(frameptr, codeptr);
        break;

      case (fcmpg):
        opcode_fcmpg(frameptr, codeptr);
        break;

      case (dcmpl):
        opcode_dcmpl(frameptr, codeptr);
        break;

      case (dcmpg):
        opcode_dcmpg(frameptr, codeptr);
        break;

      case (ifeq):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_ifeq(frameptr, codeptr);
        break;

      case (ifne):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_ifne(frameptr, codeptr);
        break;

      case (iflt):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_iflt(frameptr, codeptr);
        break;

      case (ifge):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_ifge(frameptr, codeptr);
        break;

      case (ifgt):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_ifgt(frameptr, codeptr);
        break;

      case (ifle):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_ifle(frameptr, codeptr);
        break;


      case (if_icmpeq):        
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_if_icmpeq(frameptr, codeptr);
        break;

      case (if_icmpne):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_if_icmpne(frameptr, codeptr);
        break;

      case (if_icmplt):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_if_icmplt(frameptr, codeptr);
        break;

      case (if_icmpge):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_if_icmpge(frameptr, codeptr);
        break;

      case (if_icmpgt):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_if_icmpgt(frameptr, codeptr);
        break;

      case (if_icmple):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_if_icmple(frameptr, codeptr);
        break;

      case (if_acmpeq):        
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_if_acmpeq(frameptr, codeptr);
        break;

      case (if_acmpne):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_if_acmpne(frameptr, codeptr);
        break;

      case (j_goto):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_goto(frameptr, codeptr);
        break;

      case (goto_w):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_goto_w(frameptr, codeptr);
        break;

      case (jsr):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_jsr(frameptr, codeptr);
        break;

      case (jsr_w):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_jsr_w(frameptr, codeptr);
        break;

      case (ret):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_ret(frameptr, codeptr);
        break;
  
      case (tableswitch): 
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_tableswitch(frameptr, codeptr);
        break;

      case (lookupswitch):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_lookupswitch(frameptr, codeptr);
        break;

      case (ireturn):
      case (freturn):
      case (areturn):
      case (lreturn):
      case (dreturn):
      case (vreturn):
        CHECK_FOR_ASYNCH_EXCEPTION
        if (*frameptr == startFrame) {
          proceed = WONKA_FALSE;
          continue;
        }
        opcode_return(frameptr, codeptr);
        if (*frameptr == NULL) {
          proceed = WONKA_FALSE;
        }
        break;

      case (getstatic):
        opcode_getstatic(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (putstatic):
        opcode_putstatic(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (getfield):
        opcode_getfield(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (putfield):
        opcode_putfield(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (invokevirtual):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_invokevirtual(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (invokespecial):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_invokespecial(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (invokestatic):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_invokestatic(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;
        
      case (invokeinterface):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_invokeinterface(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (no_code26):
        proceed = WONKA_FALSE; 
        break;
  
      case (new):
        opcode_new(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (newarray):
        opcode_newarray(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (anewarray):
        opcode_anewarray(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (arraylength):
        opcode_arraylength(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (athrow):
        opcode_athrow(frameptr, codeptr);
        if (*frameptr == NULL) {
          woempa(9, "Unhandled exception !\n");
          proceed = WONKA_FALSE;
//          continue;
        }
        break;

      case (checkcast):
        opcode_checkcast(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (instanceof):
        opcode_instanceof(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (monitorenter):
        opcode_monitorenter(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (monitorexit):
        opcode_monitorexit(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (multianewarray):
        opcode_multianewarray(frameptr, codeptr);
        CHECK_FOR_SYNCH_EXCEPTION
        break;

      case (ifnull):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_ifnull(frameptr, codeptr);
        break;

      case (ifnonnull):
        CHECK_FOR_ASYNCH_EXCEPTION
        opcode_ifnonnull(frameptr, codeptr);
        break;

      case (wide):
        proceed = perform_wide_opcode(frameptr, codeptr);
        continue;

      case (breakpoint):
        if(jpda_hooks) {

          /*
          ** Turn back the program counter 
          */
          
          (*codeptr)--;

          /*
          ** Call the event handler. This will return the original opcode that used 
          ** to be at the spot the breakpoint opcode was found.
          */
          
          opcode = jdwp_event_breakpoint(*codeptr);

          /*
          ** Put the original opcode back where it used to be
          */
          
          *(*codeptr) = opcode;

          /*
          ** Store the address of the opcode, so it can be replaced with a breakpoint 
          ** opcode as soon as it's executed.
          */
          
          break_point = *codeptr;
          
          continue;
        } else {
          wabort(ABORT_WONKA, "Well I'll be blowed! Opcode '%s' not yet implemented.\n", opc2name(opcode));
          proceed = WONKA_FALSE;
          continue;
        }

      default: 
        proceed = WONKA_FALSE;
        
    }

    if(jpda_hooks) {
      if(break_point) {

        /*
        ** If break_point is set, then a breakpoint opcode was replaced with the original opcode. That opcode 
        ** has now been executed, so it's time to overwrite it again with the breakpoint opcode.
        */ 

        *break_point = breakpoint;
        break_point = NULL;
      }
    }

#ifdef DEBUG
    if (*frameptr) {
      (*frameptr)->pc = *codeptr - (*frameptr)->method->code;
    }
#else
    if (jpda_hooks) {
      if (*frameptr) {
        (*frameptr)->pc = *codeptr - (*frameptr)->method->code;
      }
    }
#endif
  }

  woempa(1, "Returning from interpreting method '%w', flags are 0x%08x.\n", NM(startFrame->method), thread->flags);
  if (isNotSet(thread->flags, (WT_THREAD_THROWN | WT_THREAD_KILLED))) {

    return NULL;

  }
  else { 
    w_instance thrown = (w_instance)thread->Thread[F_Thread_thrown];
    setReferenceField(thread->Thread, NULL, F_Thread_thrown);
    unsetFlag(thread->flags, WT_THREAD_KILLED | WT_THREAD_THROWN);

    return thrown;
    
  }
}
