/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: locks.c,v 1.7 2002/03/20 13:04:21 gray Exp $
*/

#include <string.h>

#include "clazz.h"
#include "core-classes.h"
#include "exception.h"
#include "hashtable.h"
#include "heap.h"
#include "locks.h"
#include "ts-mem.h"
#include "realms.h"
#include "threads.h"
#include "wonka.h"

extern w_hashtable instance_hashtable;

/*
** The lock_hashtable maps instances onto x_monitor's.
*/
w_hashtable lock_hashtable;

void initLocks() {
}

/*
** allocMonitor: called to associate a monitor with an object with which no
** monitor is yet associated.
*/

x_monitor allocMonitor(w_instance instance) {
  x_monitor mon;

  ht_lock(lock_hashtable);

  mon = (x_monitor)ht_read_no_lock(lock_hashtable, (w_word)instance);
  if (mon) {
    woempa(1, "instance %p already had a monitor %p\n", instance, mon);
  }
  else {
    mon = allocMem_system(sizeof(x_Monitor));
    x_monitor_create(mon);

    if (mon) {
      woempa(1, "created new lock %p for instance %p.\n", mon, instance);
      ht_write_no_lock(lock_hashtable, (w_word)instance, (w_word)mon);
      woempa(1, "Now %d instances out of %d have locks (%d%%)\n", lock_hashtable->occupancy, instance_hashtable->occupancy, lock_hashtable->occupancy / (instance_hashtable->occupancy/100));
      setFlag(instance2flags(instance), O_HAS_LOCK);
    }
    else {
      woempa(9, "No memory to create x_Monitor\n");
      throwOutOfMemoryError(currentWonkaThread);
    }
  }

  ht_unlock(lock_hashtable);

  return mon;
}

void releaseMonitor(w_instance instance) {
  x_monitor mon;
  ht_lock(lock_hashtable);

  mon = (x_monitor)ht_erase_no_lock(lock_hashtable, (w_word)instance);
  unsetFlag(instance2flags(instance), O_HAS_LOCK);

  if (mon) {
    woempa(1, "Deleting monitor %p of instance %p\n", mon, instance);
    woempa(1, "Now %d instances out of %d have locks (%d%%)\n", lock_hashtable->occupancy, instance_hashtable->occupancy, lock_hashtable->occupancy / (instance_hashtable->occupancy/100));
    x_monitor_delete(mon);
    releaseMem(mon);
  }
  else {
    woempa(1, "Instance %p has no monitor\n", instance);
  }
  ht_unlock(lock_hashtable);
}

/*
** enterMonitor: called to acquire a lock on an object to which a w_Lock
** structure may or may not already be attached.
*/
void enterMonitor(w_instance instance) {
  x_status status;
  x_monitor mon;

  if (isSet(instance2flags(instance), O_HAS_LOCK)) {
    mon = (x_monitor)ht_read(lock_hashtable, (w_word)instance);
  }
  else {
    mon = allocMonitor(instance);
  }

  woempa(1, "Instance %k@%p : monitor is %p\n", instance2object(instance)->clazz, instance, mon);
  status = x_monitor_enter(mon, x_eternal);
  if (status) {
    woempa(9, "x_monitor_enter returned %d\n",status);
    throwException_C(currentWonkaThread, clazzInternalError, "x_monitor_enter");
  }
  else {
    woempa(1, "%w entered monitor of %p\n", currentWonkaThread->name, instance);
  }
}

/* exitMonitor: called to relinquish a lock.
*/ 
void exitMonitor(w_instance instance) {
  x_status status;
  x_monitor mon;
  w_thread  thread = currentWonkaThread;

  mon = (x_monitor)ht_read(lock_hashtable, (w_word)instance);

  if (mon) {
    woempa(1, "Instance %p : monitor is %p\n", instance, mon);
    status = x_monitor_exit(mon);
    woempa(1, "%w left monitor of %p\n", currentWonkaThread->name, instance);
    if (status == xs_not_owner) {
      throwException_C(thread, clazzIllegalMonitorStateException, "not owner");
    }
    else if (status) {
      woempa(9, "x_monitor_exit returned %d\n",status);
      throwException_C(thread, clazzInternalError, "x_monitor_exit");
    }
  }
  else {
    woempa(9, "Instance %p : monitor is %p\n", instance, mon);
    throwException_C(thread, clazzIllegalMonitorStateException, "no lock");
  }

}

/* Wait on a monitor.
*/

void waitMonitor(w_instance instance, x_sleep timeout) {
  x_status status;
  x_monitor mon;
  w_thread  thread = currentWonkaThread;

  mon = (x_monitor)ht_read(lock_hashtable, (w_word)instance);

  if (mon) {
    woempa(1, "Instance %p : monitor is %p\n", instance, mon);

    status = x_monitor_wait(mon, timeout);
    woempa(1, "%w waited on %p\n", currentWonkaThread->name, instance);
    if (status == xs_not_owner) {
      throwException_C(thread, clazzIllegalMonitorStateException, "not owner");
    }
    else if (status) {
      woempa(9, "x_monitor_wait returned %d\n",status);
      throwException_C(thread, clazzInternalError, "x_monitor_wait");
    }
  }
  else {
    woempa(9, "Instance %p : monitor is %p\n", instance, mon);
    throwException_C(thread, clazzIllegalMonitorStateException, "no lock");
  }
}

/*
*/

void notifyMonitor(w_instance instance, int notifyAll) {
  x_status status;
  x_monitor mon;
  w_thread  thread = currentWonkaThread;

  mon = (x_monitor)ht_read(lock_hashtable, (w_word)instance);

  woempa(1, "Instance %p : monitor is %p\n", instance, mon);
  if (notifyAll) {
    status = x_monitor_notify_all(mon);
    woempa(1, "%w notified all waiting on %p\n", currentWonkaThread->name, instance);
    if (status == xs_not_owner) {
      throwException_C(thread, clazzIllegalMonitorStateException, "not owner");
    }
    else if (status) {
      woempa(9, "x_monitor_notify_all returned %d\n",status);
      throwException_C(thread, clazzInternalError, "x_monitor_notify_all");
    }
  }
  else { // single notify
    status = x_monitor_notify(mon);
    woempa(1, "Notified one waiting on %p\n", instance);
    if (status == xs_not_owner) {
      throwException_C(thread, clazzIllegalMonitorStateException, "not owner");
    }
    else if (status) {
      woempa(9, "x_monitor_exit returned %d\n",status);
      throwException_C(thread, clazzInternalError, "x_monitor_exit");
    }
  }
}


