/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/**@name File wonka/vm/ps-mem.c
** @author Chris Gray
**
** Manage the block of memory used to hold persistent objects.
** This area is mirrored on (or paged out to) the persistent medium.
*/
//@{

#ifdef LINUX
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#endif

#include <stdio.h>
#include <string.h>

#include "wonka.h"
#include "list.h"
#include "locks.h"
#include "ts-mem.h"
#include "ps-mem.h"
#include "pm.h"
#include "realms.h"
#include "threads.h"
#include "oswald.h"
#include "memset.h"

/**@name PS Memory Lock
**
** Access to PS memory is controlled by a lock: a thread which
** wishes to mutate PS memory, or does not want another thread to
** mutate it, must acquire this lock by calling \texttt{lockPSmem()}.
** When finished it must call unlock\texttt{PSmem()}.
**
** \texttt{PSmemoryLock} is created by the Persistent Storage Manager 
** thread (\texttt{psMan}) after this has initialized the PS memory.
** If \texttt{lockPSmem()} is called before the lock has been created,
** the calling thread will sleep until the lock exists.
**
** When the Persistent Storage Manager has nothing better to do, it
** executes a ``wait'' on \texttt{PSmemory_monitor}. This means that it
** can be prompted to perform a write to PM by ``notifying''
** \texttt{PSmemory_monitor}.  (See the \texttt{smartMon} command
** \texttt{save}).
*/
//@{
///
x_monitor PSmemory_monitor;
/// Called via macro lockPSmem()
void _lockPSmem(const char *f, int l){
  while (!PSmemory_monitor) {
    woempa(7,"Waiting for PSmemory_lockptr to be created\n");
    x_thread_sleep(100);
  }
  woempa(7,"Locking PSmem for %s line %d\n",f,l);
  x_monitor_enter(PSmemory_monitor, x_eternal);
  woempa(7,"Locked PSmem for %s line %d\n",f,l);
}
/// Called via macro unlockPSmem()
void _unlockPSmem(const char *f, int l){
  woempa(7,"Unlocking PSmem for %s line %d\n",f,l);
  x_monitor_exit(PSmemory_monitor);
}
//@}

/**@name Persistent Storage memory area
*/
//@{
static char PS[PS_PHY_SIZE];
char *PS_start;
char *PS_end;
#define PS_first_chunk ((w_pschunk)PS_start)
#define PS_last_chunk  ((w_pschunk)PS_end)

#define BITS_PER_WORD 32

#define BYTES_PER_WODGE    (BITS_PER_WORD*PS_MIN_SIZE)
#define WORDS_PER_WODGE    (BITS_PER_WORD*PS_MIN_SIZE/sizeof(w_word))

#define DIRTY_BITMAP_WORDS (PS_PHY_SIZE/BYTES_PER_WODGE)
static w_word dirty_bitmap[DIRTY_BITMAP_WORDS];

/// Set PS_start and PS_end (must run after startMemory)
void startPSmem() {
  PS_start = PS;
  PS_end = PS_start + PS_PHY_SIZE;

  if (! openPM() || ! readPMall() || *((w_word*)PS_start) != 0x80000040) {
    clearPSmem();
    clearPM();
  }
  closePM();
  woempa(9,"Clearing %d bytes of dirty_bitmap\n",DIRTY_BITMAP_WORDS*sizeof(w_word));
  memset(dirty_bitmap,0,DIRTY_BITMAP_WORDS*sizeof(w_word));

}
//@}

/**@name Chunk sizes
*/
//@{
/// The minimum size we allocate (in bytes)
static w_size   min_PSchunk_size;
/// The maximum size we allocate (in bytes)
static w_size   max_PSchunk_size;
/// The geometric mean of min_PSchunk_size and max_PSchunk_size.
// Used to distinguish between ``small'' and ``large'' requests.
static w_size   med_PSchunk_size;
//@}

void PS_split(w_pschunk chunk0);
static w_pschunk PS_merge(w_pschunk chunk1);

/**@name _allocPSmem
** @param file File from which macro \texttt{allocPSmem} was called.
** @param line Line from which macro \texttt{allocPSmem} was called.
** @param thread Wonka thread for which the memory is to be allocated.
** @param num  The number of items for which space is needed.
** @param size The size of each item, in bytes.
** (We just allocate num*size bytes.  Alignment is no problem.)
** @doc
** Operates with PSmemory_monitor locked.
** We search from left to right (for a ``small'' allocation)
** or from right to left (for a ``large'' allocation) until
** we find a free chunk which is at least big enough.
** When we find one, we call \texttt{updatePSchunk()} to signal
** that the backing store for this chunk is no longer up to date,
** and then if necessary call \texttt{PS_split} to reduce the size
** of the block to $2^n$, where $2^{n-1} < num.size <= 2^n$.
** In other words, this is a first-fit algorithm.
**
** Note that we call \texttt{updatePSchunk()} before splitting up
** the chunk (because the unused parts of the chunk are also changing),
** but the backing store will only be updated after we leave 
** \texttt{allocPSmem} (because we own PSmemory_monitor).
** TODO: see if we can't reduce the number of backing-store writes
** by marking only the area actually allocated and the \em{boundaries}
** of the split areas.
*/
void *_allocPSmem(const char *file, const int line, w_thread thread, w_size num, w_size size) {
  w_size    req_size = num*size + sizeof(w_PSchunk);
  w_size    current_size = 0;
  w_pschunk current;
  w_word   *back_tag_ptr;

  lockPSmem();
  if (req_size < med_PSchunk_size) {
    woempa(1,"Small request (%d bytes < %d), searching from left\n",req_size,med_PSchunk_size);
    current = PS_first_chunk;
    while(1) {
      if (PS_end - (char*)current <= 0)
      {
        woempa(1,"Reached end of PS memory without finding a chunk of size %d\n",req_size);
        current = NULL;
        break;
      }
      current_size = current->front_tag & 0x7fffffff;
      if ((current->front_tag & 0x80000000) == 0 && current_size >= req_size) {
        woempa(1,"Goodee, found a chunk of size %d >= %d\n",current_size,req_size);
        break;
      }
      else {
        woempa(1,"Chunk %p is %s, has size %d (need %d), skipping\n",current,current->front_tag & 0x80000000?"busy":"free",current_size,req_size);
      }
      current = (w_pschunk)((char*)current + current_size);
    }
  }
  else {
    woempa(1,"Large request (%d bytes > %d), searching from right\n",req_size,med_PSchunk_size);
    current = PS_last_chunk;
    while(1) {
      if (current - PS_first_chunk <= 0)
      {
        woempa(1,"Reached start of PS memory without finding a chunk of size %d\n",req_size);
        current = NULL;
        break;
      }
      back_tag_ptr = PREV_BACK_TAG_PTR(current);
      current_size = *back_tag_ptr & 0x7fffffff;
      current = (w_pschunk)((char*)current - current_size);
      if ((*back_tag_ptr & 0x80000000) == 0 && current_size >= req_size) {
        woempa(1,"Goodee, found a chunk of size %d >= %d\n",current_size,req_size);
        break;
      }
      else {
        woempa(1,"Chunk %p is %s, has size %d (need %d), skipping\n",current,current->front_tag & 0x80000000?"busy":"free",current_size,req_size);
      }
    }
  }

  if (current) {
    updatePSchunk(current, current_size);
    while (current_size > req_size*2) {
      woempa(1,"Chunk %p is too big (%d bytes, need only %d): splitting\n",current,current_size,req_size);
      PS_split(current);
      current_size = current->front_tag & 0x7fffffff;
    }
    current->front_tag |= 0x80000000;
    back_tag_ptr = THIS_BACK_TAG_PTR(current);
    *back_tag_ptr |= 0x80000000;
  }
  unlockPSmem();

  return current ? (void *)current->contents : NULL;

}

/**@name _releasePSmem
** @param file File from which macro \texttt{allocPSmem} was called.
** @param line Line from which macro \texttt{allocPSmem} was called.
** @param block Block of memory to be released.
** @doc
** Operates with PSmemory_monitor locked.
** First we perform a quick check that the front and back tags of
** the memory area match, then we set the area free and repeatedly
** call \texttt{mergePSmem()} as long as a merge with a ``buddy'' is
** possible.
*/
void _releasePSmem(const char *file, const int line, void *block) {
  w_pschunk current;
  w_word   *back_tag_ptr;

  if (block == NULL) {
    wabort(ABORT_WONKA, "Freeing NULL block at %s:%d.\n", file, line);
  }

  current = PSblock2chunk(block);
  woempa(1,"Releasing block %p = chunk %p\n",block,current);
  back_tag_ptr = THIS_BACK_TAG_PTR(current);
  if (current->front_tag != *back_tag_ptr || 
     (current->front_tag & 0x80000000) == 0 ||
     (*back_tag_ptr & 0x80000000) == 0 
     ) {
    woempa(1,"front tag: %08x\n",current->front_tag);
    woempa(1,"back tag:  %08x\n",*back_tag_ptr);
    wabort(ABORT_WONKA, "Corrupt block %p.\n", block);
  }

  lockPSmem();
  current->front_tag &= 0x7fffffff;
  *back_tag_ptr &= 0x7fffffff;
  while (current) {
    updatePSchunk(current, current->front_tag &= 0x7fffffff);
    current = PS_merge(current);
  }
  unlockPSmem();
}

/**
** \texttt{PS_split} causes an existing chunk to perform binary fission.
** The free_or_busy status is replicated.
*/
void PS_split(w_pschunk chunk0) {
  w_size    new_size = (chunk0->front_tag & 0x7fffffff)/2;
  w_pschunk chunk1 = chunk0;
  w_pschunk chunk2 = (w_pschunk)((char*)chunk1 + new_size);
  w_size    free_or_busy = chunk0->front_tag & 0x80000000;
  w_word   *back_tag_ptr;

  woempa(1,"Asked to split chunk %p (0x%x)\n",chunk0,chunk0->front_tag);
  if (chunk0->front_tag == min_PSchunk_size) {
    woempa(1,"chunk %p is minimal, no split possible\n",chunk1);

    return;

  }

  chunk1->front_tag = new_size | free_or_busy;
  back_tag_ptr = THIS_BACK_TAG_PTR(chunk1);
  *back_tag_ptr = new_size | free_or_busy;
  chunk2->front_tag = new_size | free_or_busy;
  chunk2->realm = chunk1->realm;
  back_tag_ptr = THIS_BACK_TAG_PTR(chunk2);
  *back_tag_ptr = new_size | free_or_busy;
  woempa(1,"Split chunk %p into %p (0x%x) and %p (0x%x)\n",chunk0,chunk1,chunk1->front_tag,chunk2,chunk2->front_tag);

}

/**
** \texttt{PS_merge} causes a chunk to fuse with its "buddy" iff said buddy
** has the same free_or_busy status.  
** The result returned is a pointer to the start of the merged chunk
** iff a merge was performed, \texttt{NULL} otherwise.
*/
w_pschunk PS_merge(w_pschunk chunk1) {
  w_size    old_size = chunk1->front_tag & 0x7fffffff;
  w_pschunk chunk0;
  w_pschunk chunk2;
  w_word   *back_tag_ptr;
  w_size    free_or_busy = chunk1->front_tag & 0x80000000;

  woempa(1,"Asked to merge chunk %p (0x%x)\n",chunk1,old_size);
  if (old_size == max_PSchunk_size) {
    woempa(1,"chunk %p is maximal, no merge possible\n",chunk1);

    return NULL;

  }

  if ((((char*)chunk1)-PS_start) & old_size) {
    chunk2 = (w_pschunk)(((char*)chunk1) - old_size);
    woempa(1,"chunk %p is a right-hand chunk, buddy is %p (0x%x)\n",chunk1,chunk2,chunk2->front_tag);
    chunk0 = chunk2;
  }
  else {
    chunk2 = (w_pschunk)(((char*)chunk1) + old_size);
    woempa(1,"chunk %p is a left-hand chunk, buddy is %p (0x%x)\n",chunk1,chunk2,chunk2->front_tag);
    chunk0 = chunk1;
  }

  if (chunk1->front_tag == chunk2->front_tag) {
    woempa(1,"chunk %p is (%d), %p is (%d), performing merge\n",chunk1,chunk1->front_tag,chunk2,chunk2->front_tag);
    chunk0->front_tag = (old_size*2) | free_or_busy;
    back_tag_ptr = THIS_BACK_TAG_PTR(chunk0);
    *back_tag_ptr = (old_size*2) | free_or_busy;
  }
  else {
    woempa(1,"chunk %p is %x, %p is %x, no merge allowed\n",chunk1,chunk1->front_tag,chunk2,chunk2->front_tag);
    chunk0 = NULL;
  }

  return chunk0;
}

/**@name Routines to perform a memory dump
*/
//@{
void PS_memDump (void);
///
w_size dumpOnePSchunk(w_pschunk pschunk) {
  w_word  *back_tag_ptr;
  w_size  front_size;
  w_size  back_size;

  front_size = pschunk->front_tag;
  woempa(9,"chunk %p: %08x\n",pschunk,front_size);
  back_tag_ptr = THIS_BACK_TAG_PTR(pschunk);
  back_size = *back_tag_ptr;
  woempa(9,"  check back tag: %08x\n",back_size);

  if (front_size==back_size) {

    return front_size & 0x7fffffff;

  }
  else return 0;

}
///
void PS_memDump() {
  w_pschunk current;
  w_size    size;

  woempa(9,"**************** Start PS memory dump ****************\n");
  current = PS_first_chunk;
  while ((size = dumpOnePSchunk(current)) > 0) {
    current = (w_pschunk)((char*)current + size);
    if (PS_end - (char*)current <= 0)
    {
      break;
    }
  }
  woempa(9,"***************** End PS memory dump ******************\n");
}
//@}

/**@name Initialization and Miscellaneous
*/
//@{
/**
** The static variables \texttt{min/max/med_PSchunk_size} are set here.
*/
void initPSmem() {
  w_size med1, med2;

  PSmemory_monitor = NULL;

  min_PSchunk_size = PS_MIN_SIZE;
  max_PSchunk_size = PS_PHY_SIZE;
  med1 = PS_MIN_SIZE;
  med2 = PS_PHY_SIZE;
  while (med1 < med2) {
    med1 *= 2;
    med2 /= 2;
  }
  med_PSchunk_size = med1;

  return;
}

/// Make the whole PS memory area into one big free chunk.
void clearPSmem(void) {
  w_word  *back_tag_ptr;

  PS_first_chunk->front_tag = max_PSchunk_size;
  back_tag_ptr = THIS_BACK_TAG_PTR(PS_first_chunk);
  *back_tag_ptr = max_PSchunk_size;
  woempa(1,"Persistent storage area is now one big free chunk\n");
}

/// Return the address of the first block (interior of the first chunk) of PS memory.
//  Used by the FICL word `first-pobject'
void *firstPSblock(){
  w_pschunk current;
  w_word    tag;

  current = PS_first_chunk;
  if (current == NULL) {

    return NULL;

  }

  woempa(1,"first: %p\n",current);
  tag = current->front_tag;
  while (PS_end - (char*)current > 0 && !(tag & 0x80000000)) {
    woempa(1,"chunk %p: %08x\n",current,tag);
    current = (w_pschunk)((char*)current + tag);
    if (PS_end - (char*)current > 0) {
      tag = current->front_tag;
    }
    else {

      return NULL;

    }
  }
  woempa(1,"current: %p\n",current);

  if ((tag & 0x7fffffff) > (w_size)(PS_end - PS_start)) {

    return NULL;

  }

  return &current->contents;
}

/// Return the address of the next block (interior of the next chunk) of PS memory.
//  Used by the FICL word `next-pobject'
void *nextPSblock(void *block){
  w_pschunk current;
  w_word    tag;

  current = PSblock2chunk(block);
  woempa(1,"start: %p\n",current);
  tag = current->front_tag;
  do {
    woempa(1,"tag: %p\n",tag);
    current = (w_pschunk)((char*)current + (tag & 0x7fffffff));
    woempa(1,"current: %p\n",current);
    if (PS_end - (char*)current > 0) tag = current->front_tag;
  } 
  while (PS_end - (char*)current > 0 && !(tag & 0x80000000));

  if (!(PS_end - (char*)current > 0) || !(tag & 0x80000000)) {

    return NULL;

  }
  woempa(1,"current: %p\n",current);

  return &current->contents;
}

//@}

///
void updatePSchunk(w_pschunk chunk, w_size size) {
  w_size    nbits = size/PS_MIN_SIZE;
  w_size   bitnum = (((char*)chunk) - PS_start) / PS_MIN_SIZE;
  w_word   thebit = 1;
  w_word *theword = dirty_bitmap + bitnum/BITS_PER_WORD;

  woempa(1,"Asked to update chunk %p: size is %d bytes\n",chunk,size);
  woempa(1,"Chunk begins %d bytes from PS_start, PS_MIN_SIZE is %d, so first bit is bit %d\n",((char*)chunk) - PS_start,PS_MIN_SIZE,bitnum);
  if (bitnum % BITS_PER_WORD) {
    thebit <<= (bitnum % BITS_PER_WORD);
  }

  while (nbits && thebit != 1) {
    woempa(1,"setting bit %d of dirty_bitmap by ORing word %p with 0x%08x\n",bitnum,theword,thebit);
    *theword |= thebit;
    thebit <<= 1;
    if (thebit==0) {
      ++theword;
      thebit = 1;
    }
    --nbits;
    ++bitnum;
  }

  while (nbits >= BITS_PER_WORD) {
    woempa(1,"setting bits %d..%d of dirty_bitmap by setting word %p to 0xffffffff\n",bitnum,bitnum+31,theword);
    *theword++ = 0xffffffff;
    nbits -= BITS_PER_WORD;
    bitnum += BITS_PER_WORD;
  }

  while(nbits--) {
    woempa(1,"setting bit %d of dirty_bitmap by ORing word %p with 0x%08x\n",bitnum++,theword,thebit);
    *theword |= thebit;
    thebit <<= 1;
    if (thebit==0) {
      ++theword;
      thebit = 1;
    }
  }
}


///
void updatePSblock(void *block) {
  w_pschunk chunk = PSblock2chunk(block);
  w_size     size = chunk->front_tag & 0x7fffffff;

  updatePSchunk(chunk,size);
}

void updatePSword(w_word *update) {
  w_size   bitnum = (update - ((w_word*)PS_start)) / PS_MIN_SIZE;
  w_word   thebit = 1;
  w_word *theword = dirty_bitmap + bitnum/BITS_PER_WORD;

  woempa(1,"Asked to update word %p\n",update);
  woempa(1,"Word is %d words from PS_start, PS_MIN_SIZE is %d, so must set bit %d\n",update - ((w_word*)PS_start),bitnum);
  if (bitnum % BITS_PER_WORD) {
    thebit <<= bitnum;
  }
  woempa(1,"setting bit %d of dirty_bitmap by ORing word %d with 0x%08x\n",bitnum,bitnum/BITS_PER_WORD,thebit);
  *theword |= thebit;
}

void
syncPSmem() {
  int i;
  if (openPM()) {
    for (i=0;i<DIRTY_BITMAP_WORDS;++i) {
      if (dirty_bitmap[i]) {
        woempa(9,"Writing %d bytes from %p\n",BYTES_PER_WODGE,PS_start+i*BYTES_PER_WODGE);
        writePMportion(PS_start + i*BYTES_PER_WODGE, BYTES_PER_WODGE);
        dirty_bitmap[i] = 0;
      }
    }
    closePM();
  }
}

//@}

