/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: reflection.c,v 1.12 2002/03/26 10:44:15 gray Exp $
**
** Utility functions for reflection related files.
**
** The implementations for the Java conversions and promotions are to be found
** in this file. See any good JVM book or JLS book for the different promotions 
** and conversions that can take place.
*/

#include "arrays.h"
#include "checks.h"
#include "descriptor.h"
#include "exception.h"
#include "ieee754.h"
#include "heap.h"
#include "clazz.h"
#include "core-classes.h"
#include "strings.h"
#include "methods.h"
#include "stack.h"
#include "threads.h"
#include "loading.h"
#include "interpreter.h"
#include "reflection.h"


/*
** For assigning fields and variables to each other. The reflection classes need to implement
** what is called widening for some operations, or when no type conversion is required, implement
** a so called 'identity conversion' or assignment. Checking which conversions are allowed and
** what operations need to be performed by the virtual machine in order to change from one type
** to another, is done through a matrix or lookup table of conversion function pointers. A NULL
** pointer means that the conversion <from> to <to> type is not allowed, a non NULL pointer means
** that the conversion is allowed and the function pointed to will perform the assignment.
** 
** The matrix, found after the assignment functions, is addressed by means of the primitive types
** that need to be converted. Check further for 'matrix' and the 'convert_and_assign' function.
*/

/*
** Two 'identity conversion assignment' functions, for 32 bit data
** and for 64 bit data. Pretty straightforward...
*/

static void i_32(w_word S[], w_word T[]) {

  T[0] = S[0];

}

static void i_64(w_word S[], w_word T[]) {

  T[1] = S[1];
  T[0] = S[0];

}

/*
** The different conversion routine <from>__<to>. Note that after testing,
** some of these can be collapsed together for compacter code.
*/

static void C__I(w_word S[], w_word T[]) {

  /*
  ** No sign extension takes place...
  */

  T[0] = (S[0] & 0x0000ffff);

}

static void C__J(w_word S[], w_word T[]) {

  /*
  ** No sign extension takes place...
  */
  
  T[1] = 0;
  T[0] = (S[0] & 0x0000ffff);

}

static void C__F(w_word S[], w_word T[]) {

  T[0] = int32_to_float32((w_int) S[0]);

}

static void C__D(w_word S[], w_word T[]) {

  w_u64 d;
  
  d.u64 = int32_to_float64((w_int) S[0]);

  T[1] = d.words[WORD_MSW];
  T[0] = d.words[WORD_LSW];

}

static void B_SI(w_word S[], w_word T[]) {

  T[0] = (S[0] & 0x80) ? (S[0] | 0xffffff00) : (S[0] & 0x0000007f);

}

static void B__J(w_word S[], w_word T[]) {

  w_u64 l;
  
  l.s64 = (w_byte)S[0];

  T[1] = l.words[WORD_MSW];
  T[0] = l.words[WORD_LSW];

}

static void B__F(w_word S[], w_word T[]) {

  T[0] = int32_to_float32((w_int) S[0]);
  
}

static void B__D(w_word S[], w_word T[]) {

  w_u64 d;
  
  d.u64 = int32_to_float64((w_int) S[0]);

  T[1] = d.words[WORD_MSW];
  T[0] = d.words[WORD_LSW];

}

static void S__I(w_word S[], w_word T[]) {

  T[0] = (S[0] & 0x8000) ? (S[0] | 0xffff0000) : (S[0] & 0x00007fff);

}

static void S__J(w_word S[], w_word T[]) {

  w_u64 l;
  
  l.s64 = (w_short)S[0];

  T[1] = l.words[WORD_MSW];
  T[0] = l.words[WORD_LSW];

}

static void S__F(w_word S[], w_word T[]) {

  T[0] = int32_to_float32((w_int) S[0]);

}

static void S__D(w_word S[], w_word T[]) {

  w_u64 d;
  
  d.u64 = int32_to_float64((w_int) S[0]);

  T[1] = d.words[WORD_MSW];
  T[0] = d.words[WORD_LSW];

}

static void I__J(w_word S[], w_word T[]) {

  w_u64 l;
  
  l.s64 = (w_int)S[0];

  T[1] = l.words[WORD_MSW];
  T[0] = l.words[WORD_LSW];

}

static void I__F(w_word S[], w_word T[]) {

  T[0] = int32_to_float32((w_int) S[0]);
  
}

static void I__D(w_word S[], w_word T[]) {

  w_u64 d;
  
  d.u64 = int32_to_float64((w_int) S[0]);

  T[1] = d.words[WORD_MSW];
  T[0] = d.words[WORD_LSW];

}

static void J__F(w_word S[], w_word T[]) {

  w_long l = WORDS2LONG(S[1], S[0]);
  
  T[0] = int64_to_float32(l);
  
}

static void J__D(w_word S[], w_word T[]) {

  w_u64 j;

  j.words[WORD_MSW] = S[1];
  j.words[WORD_LSW] = S[0];

  j.u64 = int64_to_float64(j.s64);

  T[1] = j.words[WORD_MSW];
  T[0] = j.words[WORD_LSW];

}

static void F__D(w_word S[], w_word T[]) {

  w_u64 d;
  
  d.u64 = float32_to_float64((w_float)S[0]);

  T[1] = d.words[WORD_MSW];
  T[0] = d.words[WORD_LSW];

}

typedef void (*w_widener)(w_word S[], w_word T[]);

static w_widener matrix[8][8] = {
  /*   +--->  to  Z     C     F     D     B     S     I     J   */
  /*   |                                                        */
  /* from Z */ { i_32, NULL, NULL, NULL, NULL, NULL, NULL, NULL },
  /* from C */ { NULL, i_32, C__F, C__D, NULL, NULL, C__I, C__J },
  /* from F */ { NULL, NULL, i_32, F__D, NULL, NULL, NULL, NULL },
  /* from D */ { NULL, NULL, NULL, i_64, NULL, NULL, NULL, NULL },
  /* from B */ { NULL, NULL, B__F, B__D, i_32, B_SI, B_SI, B__J },
  /* from S */ { NULL, NULL, S__F, S__D, NULL, i_32, S__I, S__J },
  /* from I */ { NULL, NULL, I__F, I__D, NULL, NULL, i_32, I__J },
  /* from J */ { NULL, NULL, J__F, J__D, NULL, NULL, NULL, i_64 },
};

/*
** Possibly convert and assign one primitive value (F) to another (T). If the widening
** conversion is not permitted, this function returns 1, otherwise 0 indicates success.
*/

w_int convert_and_assign(w_desc F_desc, w_word F_data[], w_desc T_desc, w_word T_data[]) {

  w_widener widener;
  w_int f;
  w_int t;

  if (F_desc->tag == D_Primitive) {
    if (T_desc->tag == D_Primitive) {
      f = F_desc->body.primitive.type;
      t = T_desc->body.primitive.type;

      /*
      ** In any case, zero out the memory of the 'T_data' array.
      */

      T_data[0] = 0;
      if (desc2words[t] == 2) {
        T_data[1] = 0;
      }

      f -= 4;
      t -= 4;

      // TODO, the following if should become a runtime check
      if ((f < 8) && (t < 8)) {
        woempa(1, "Trying to widen and assign type '%s' to type '%s'...\n", primitive2name[f + 4], primitive2name[t + 4]);
        widener = matrix[f][t];
        if (widener) {
          (*widener)(F_data, T_data);
          woempa(1, "Widened and assigned type '%s' (0x%08x) to type '%s' (0x%08x) (only 32 bits shown...)\n", primitive2name[f + 4], F_data[0], primitive2name[t + 4], T_data[0]);
          return 0;
        }
      }
    }
  }

  /*
  ** An error occured. Since the 'f' and 't' indexes could be bogus, we just print the indexes and not the
  ** primitive2name element since we could segfault...
  */
  
  woempa(1, "... impossible to perform widening.\n");
  
  return 1;
  
}

/*
** Unwrap a possibly primitive wrapper instance and get the data pointer and the descriptor
** from the wrapper classes. The returned descriptor is NULL when the clazz of Value is not
** one of these primitive wrapper classes...
*/

w_desc unwrapValue(w_instance Value, w_word **data) {

  w_desc desc = NULL;
  w_clazz value = instance2clazz(Value);
  w_int i;

  woempa(1, "Trying to unwrap instance of %k.\n", value);
  for (i = 4; i <= P_long; i++) {
    if (primitive2wrapper[i] == value) {
      desc = primitive2desc[i];
      *data = Value + primitive2wrapperSlot[i];
      break;
    }
  }

  return desc;

}

/*
** Create a wrapper instance for a certain primitive type, returns NULL when the type is
** not a primitive type.
*/

w_instance createWrapperInstance(w_thread thread, P_Type type) {

  w_instance instance = NULL;
  
  if (type > 3 && type <= P_long) {
    instance = allocInstance(thread, primitive2wrapper[type]);
  }
    
  return instance;
  
}

w_frame invoke(JNIEnv *env, w_method method, w_instance This, w_instance Arguments, w_instance *exception) {
  w_thread thread = JNIEnv2w_thread(env);
  w_instance *arguments;
  w_word *F_data;
  w_desc F_desc;
  w_word T_data[2];
  w_desc T_desc;
  w_int numArgs;
  w_int i;
  w_int j;
  w_frame bottom = NULL;
  w_word *params;
  w_instance itex;

  woempa(1, "Executing %M...\n", method);

  numArgs = instance2Array_length(Arguments);
  if ((numArgs != method->nargs)) {
    woempa(7, "mag niet! numArgs %d, method nargs %d\n", numArgs, method->nargs);
    *exception = createExceptionInstance(thread,clazzIllegalArgumentException,NULL);
  }
  else {
    classMustBeInitialized(method->clazz, thread);
    bottom = allocBottomFrame(thread, method);
    if (bottom) {
      setFlag(bottom->flags, FRAME_REFLECTION);

      if (isSet(method->flags, ACC_NATIVE)) {
        woempa(1, "Native method: copying parameters to o_top = %p\n", bottom->o_top);
        params = bottom->o_top;
      }
      else {
        woempa(1, "Non-native method: copying parameters to locals = %p\n", bottom->locals);
        params = bottom->locals;
      }

      if (isSet(method->flags,ACC_STATIC)) {
        j = 0;
      }
      else {
        woempa(1,"Frame %p: setting params[0] to %p\n", bottom, This);
        params[0] = (w_word)This;
        j = 1;
      }
      arguments = instance2Array_instance(Arguments);
      for (i = 0; i < numArgs; i++) {
        method->arg_types[i] = mustBeLoaded(method->arg_types[i], exception);
        if (isSet(method->arg_types[i]->flags, CLAZZ_IS_PRIMITIVE)) {
          T_desc = method->arg_types[i]->desc;
          F_desc = unwrapValue(arguments[i], &F_data);
          if (F_desc == NULL || convert_and_assign(F_desc, F_data, T_desc, T_data)) {
            if (F_desc == NULL) {
              woempa(7, "Argument %d is not a wrapper (%k)!\n", i, instance2clazz(arguments[i]));
            }
            else {
              woempa(7, "Widening is impossible!\n");
            }
            *exception = createExceptionInstance(thread,clazzIllegalArgumentException,NULL);
            break;
          }

          if (sizeInWords(T_desc) == 1) {
            woempa(1,"Frame %p: setting params[%d] to 0x%08x\n", bottom, j, T_data[0]);
            params[j] = T_data[0];
            j++;
          }
          else {
            woempa(1,"Frame %p: setting params[%d,%d] to 0x%08x%08x\n", bottom, j, j+1, T_data[0], T_data[1]);
            params[j] = T_data[0];
            j++;
            params[j] = T_data[1];
            j++;
          }
        }
        else {
          w_clazz clazz = method->arg_types[i];
          woempa(1,"Argument %d: target class is %k, argument is %k@%p\n",i,clazz,instance2clazz(arguments[i]),arguments[i]);
          if (*exception) {

            return NULL;

          }
          if (arguments[i] == NULL || isAssignmentCompatible(instance2clazz(arguments[i]), clazz)) {
            woempa(1,"Frame %p: setting params[%d] to %p\n", bottom, j, arguments[i]);
            params[j] = (w_word)arguments[i];
            j++;
          }
          else {
            woempa(7, "clazz %k is not assignment compatible with %k.\n", instance2clazz(arguments[i]), clazz);
            *exception = createExceptionInstance(thread,clazzIllegalArgumentException,NULL);
            break;
          }
        }
      }

      if (isSet(method->flags, ACC_NATIVE)) {
        bottom->o_top += j;
        woempa(1, "Native method: adding %d to o_top => %p\n", j, bottom->o_top);
      }
    
      /*
      ** If exception is NULL, all the arguments were OK and they are in the bottom frame, let's run it now...
      ** If not, something exceptional occured.
      ** The frame is useless and should be released for the interpreter to run further.
      ** We return NULL, to signal this situation 
      ** TODO convey more info on what went wrong...
      */

      if (*exception == NULL) {
        woempa(1, "(REFLECTION) Starting frame with %M, %d arguments.\n", method, method->nargs);
        handleSynchronization(bottom, method);
        *exception = runThread(thread, bottom);
        woempa(1, "(REFLECTION) Done calling %M, %d arguments.\n", method, method->nargs);
        if (*exception) {

        /*
        ** Create an InvocationTargetException and wrap the thrown exception in there, then rethrow this
        ** exception.
        */

          classMustBeInitialized(clazzInvocationTargetException, thread);
          itex = allocInstance(thread, clazzInvocationTargetException);
          setReferenceField(itex,*exception,F_InvocationTargetException_target);
          woempa(7, "(REFLECTION) Invocation threw an exception, wrapping it into an InvocationTargetException.\n");
          *exception = itex;
        }
      }
    }
  }

  if (*exception) {
    woempa(7,"(REFLECTION) Invoked method %M returned exception %k\n",method,instance2clazz(*exception));
    if (bottom) {
      releaseBottomFrame(bottom);
    }
    bottom = NULL;
  }

  return bottom;
}
