/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: strings.c,v 1.9 2002/03/25 09:23:21 buytaert Exp $
*/

#include <string.h>
#include <stdio.h>

#include "clazz.h"
#include "core-classes.h"
#include "fifo.h"
#include "hashtable.h"
#include "heap.h"
#include "ieee754.h"
#include "ts-mem.h"
#include "realms.h"
#include "stack.h"
#include "strings.h"
#include "threads.h"
#include "descriptor.h"

#define STRING_HASHTABLE_SIZE                         719

w_hashtable string_hashtable;

x_mutex string_mutex;

/// An array that maps atypes to descriptors as w_string's 
w_string atype2string[12];

/// The string `<NULL>'
w_string string_NULL;

/// The string `<EMPTY>'
w_string string_EMPTY;

/// The string `<BOGUS>'
w_string string_BOGUS;

static w_word hashString(w_string string);

static inline w_string allocString(w_size length) {

  w_string string;
  
  string = allocMem_system(offsetof(w_String,chars) + length*sizeof(w_char));
  string->refcount = 0;
  string->instance = NULL;
  string->length = length;

  return string;

}

static inline void deallocString(w_string s) {
  releaseMem(s);
}

/*
** Unpack a w_string into an array of w_char's.  Caller must supply a
** suitably sized buffer (i.e. space for at least source->numChars). 
** Returns the number of w_char's which were found.
*/

w_size w_string2chars(w_string source, w_char *dest) {

  memcpy(dest, source->chars, sizeof(w_char) * source->length);

  return source->length;

}

/*
** Encode a unicode character into UTF-8; the resulting 1 to three bytes
** are written to **dest, and *dest is updated accordingly.
**
** Note that it it is the caller's responsiblility to ensure that sufficient
** space is available after **dest.
*/

void xputw_char(int ch, w_byte **dest) {

  if (ch == 0) {
    *(*dest)++ = 0xC0;
    *(*dest)++ = 0x80;
  }
  else if( ch < 0x80) {
    *(*dest)++ = ch;
  }
  else if(ch < 0x800) {
    *(*dest)++ = 0xC0+(ch >> 6);
    *(*dest)++ = 0x80+(ch & 63);
  }
  else {
    *(*dest)++ = 0xE0 + (ch >> 12);
    *(*dest)++ = 0x80 + ((ch >> 6) & 63);
    *(*dest)++ = 0x80 + (ch & 63);
  }

}

w_int compareStrings(w_string s1, w_string s2) {

  w_size i;
  w_int result = 0; 

  if (s1->length == 0) {
    result -= s2->length;
  }  
  else if (s2->length == 0) {
    result = s1->length;
  }
  else if (s1 != s2) {
    for (i = 0; i < s1->length; i++) {
      if (s1->chars[i] != s2->chars[i]) {
        result = s1->chars[i] - s2->chars[i];
        break;
      }
      if ((i + 1) >= s1->length || (i + 1) >= s2->length) {
        result = s1->length - s2->length;
        break;
      }
    }
  }

  return result;

}


/*
** Convert a Wonka string into an array of UTF8 w_byte's. The caller
** is responsible for releasing the memory afterwards.
** After this function returns, 'buflen' will be set to the number of 
** bytes that where allocated in the buffer returned.
**
*/

inline static int char_is_singlet(w_word c) {

  if ((w_char)c == 0) {
    return 0;
  }
  else if ((w_char)c & 0xff80) {
    return 0;
  }
  else {
    return 1;
  }

}

inline static int char_is_duplet(w_word c) {

  if ((w_char)c & 0xf800) {
    return 0;
  }
  else {
    return 1;
  }

}

inline static void char2duplet(w_byte *buffer, w_word thechar) {

  w_byte c = (w_char)thechar;
  
  buffer[0] = (w_byte)(c>>6) | 0xC0;
  buffer[1] = (w_byte)(c&63) | 0x80;
  
}

inline static void char2triplet(w_byte *buffer, w_word thechar) {

  w_char c = (w_char)thechar;
  
  buffer[0] = (w_byte)(c>>12) | 0xE0;
  buffer[1] = (w_byte)((c>>6)&63) | 0x80;
  buffer[2] = (w_byte)(c&63) | 0x80;
}

w_byte *string2UTF8(w_thread thread, w_string string, w_int *buflen) {

  w_byte *buffer;
  w_byte *handle;
  w_byte *result;
  w_size i;
  w_size length;

  /*
  ** Allocate a worst case buffer, 3 bytes for each unicode char and
  ** 2 bytes for the length.
  */

  length = 0;

  handle = allocMem(thread, 2 + (string->length * (3 * sizeof (w_byte))));
  buffer = handle + 2;
  for (i = 0; i < string->length; i++) {
    if (char_is_singlet((w_word)string->chars[i])) {
      *buffer = (w_byte)string->chars[i];
      length += 1;
      buffer += 1;
    }
    else if (char_is_duplet((w_word)string->chars[i])) {
      char2duplet(buffer, (w_word)string->chars[i]);
      length += 2;
      buffer += 2;
    }
    else {
      char2triplet(buffer, (w_word)string->chars[i]);
      length += 3;
      buffer += 3;
    }
  }
  handle[0] = (w_byte)(length << 8);
  handle[1] = (w_byte)(length);

  result = allocMem(thread, (length + 2) * sizeof ( w_byte));
  memcpy(result, handle, (w_size)length + 2);
  releaseMem(handle);
  *buflen = length + 2;
  
  return result;
    
}

/*
** Convert a UTF8 string into a Wonka string. The string returned is a registered
** string. The UTF8 string probably comes from a class file. We first create some
** small support functions.
*/

inline static w_int is_singlet(w_word a) {

  if (! (a & 0x80)) {
    return 1;
  }

  return 0;
  
}

inline static w_int is_duplet(w_word a, w_word b) {

  if (a & 0xc0) {
    if (! (a & 0x20)) {
      if (b & 0x80) {
        if (! (b & 0x40)) {
          return 1;
        }
      }
    }
  }

  return 0;
  
}

inline static w_int is_triplet(w_word a, w_word b, w_word c) {

  if (a & 0xe0) {
    if (! (a & 0x10)) {
      if (b & 0x80) {
        if (! (b & 0x40)) {
          if (c & 0x80) {
            if (! (c & 0x40)) {
              return 1;
            }
          }
        }
      }
    }
  }

  return 0;
  
}

inline static w_char duplet2char(w_word a, w_word b) {

  w_char result = (w_char)(((a & 0x1f) << 6) + (b & 0x3f));
  
  return result;
  
}

inline static w_char triplet2char(w_word a, w_word b, w_word c) {

  w_char result = (w_char)(((a & 0x0f) << 12) + ((b & 0x3f) << 6) + (c & 0x3f));

  return result;
  
}

/*
** Now the function to create the Wonka string from a UTF8 string of bytes. The
** length argument passed is the number of UTF8 bytes. Note that the UTF8 bytes
** string is not expected to be NULL terminated, so the full length should be processed.
*/

w_string utf2String(w_thread thread, const char *utf8string, w_size utf8length) {
  w_string result;
  w_realm realm = thread2realm(thread);
  w_char *chars;
  w_string string;
  w_byte x;
  w_byte y;
  w_byte z;
  w_size idx = 0;
  w_size i = 0;

  /*
  ** Allocate a temporary buffer, we allocate it too be worst case size. This
  ** could be too large, but it will be copied over later to a buffer of a 
  ** correct size in the string itself. The worst case is when all UTF8 bytes
  ** are singlets, so we allocate 1 extra as the length.
  */
  
  if (utf8length) {
    chars = allocMem_realm(realm, (utf8length + 1) * sizeof(w_char));
    while (i < utf8length) { // [CG 20000410] WAS: <=
      x = (w_byte)utf8string[i++];
      if (is_singlet(x)) {
        chars[idx++] = (w_char)x;
      }
      else {
// [CG 20000410] add:
         if (i >= utf8length) {
           woempa(9, "Incomplete UTF8 sequence 0x%02x\n", x);
           break;
         }
// [CG 20000410] end
        y = (w_byte)utf8string[i++];
        if (is_duplet(x, y)) {
          chars[idx++] = duplet2char(x, y);
        }
        else {
// [CG 20000410] add:
           if (i >= utf8length) {
             woempa(9, "Incomplete UTF8 sequence 0x%02x 0x%02x\n", x, y);
             break;
           }
// [CG 20000410] end
          z = (w_byte)utf8string[i++];
          if (is_triplet(x, y, z)) {
            chars[idx++] = triplet2char(x, y, z);
          }
          else {
            woempa(9, "Unknown UTF8 characters 0x%02x 0x%02x 0x%02x\n", x, y, z);
          }
        }
      }
    }

    string = allocString(idx);
    memcpy(string->chars, chars, sizeof(w_char) * idx);
    releaseMem(chars);
    hashString(string);
  
//    woempa(9, "Created string %w.\n", string);

    result = registerString(string);  
  }
  else {
    result = cstring2String(W_Thread_system,"",0);
  }

  return result;
  
}

/*
** Create a Wonka string from a unicode string. The unicode string is just
** and array of w_chars and the length is passed as a separate argument.
*/

w_string unicode2String(w_thread thread, w_char *chars, w_size length) {

  w_string string;
  
  if (length) {
    string = allocString(length);
    memcpy(string->chars, chars, sizeof(w_char) * length);
    hashString(string);
  }
  else {
    string = cstring2String(W_Thread_system,"",0);
  }

//  woempa(9, "Created string %w.\n", string);
  
  return registerString(string);  
  
}

/*
** Create a Wonka string from a C string. The length should NOT take into
** account the '\0' character of the C string. I.e. one can use strlen()
** to calculate the length when calling this function.
*/

w_string cstring2String(w_thread thread, const char *cstring, w_size length) {

  w_string string;
  w_size i;

  string = allocString(length);
  if (length) {
    for (i = 0; i < length; i++) {
      string->chars[i] = (w_char)cstring[i];
    }

  }
  else {
    string->length = 0;
  }

  hashString(string);
  woempa(1, "Created string %w.\n", string);
  
  return registerString(string);  

}

/*
** The function that attaches a hash value to a w_string.
** From the JDK 1.2 documentation. Notice that the hashcode calculation
** has changed from JDK 1.1 to JDK 1.2 !
*/

w_word hashString(w_string string) {

  w_int hash = 0;
  w_char *s;
  w_size i;
  s = string->chars;
   
  for (i=0;i<(w_size)(string->length);++i) hash = hash*31 + *s++;

//  string->java_hash = hash;
  string->wonka_hash = (w_word)buzHash((w_byte*)string->chars,((w_int)sizeof(w_char))*((w_int)string->length));
  
  return hash;
  
}

static void releaseString(w_string string) {
      woempa(1, "About to release (%p, chars @ %p) %w\n", string, string->chars, string);
      deallocString(string);
}    

/*
** The hash function for the string hashtable. It should return
** the hash value of the string that should be allready available.
*/

w_word ht_stringHash(w_word string_word) {

  w_string string = (w_string)string_word;

  return string->wonka_hash;
  
}

/*
** Register a Wonka string. The result returned is the string that should be
** used for subsequent operations; the argument string that is passed could
** be the same as the result string, if it was not registered yet, or the
** pointer to the allready registered string if it was a duplicate. If there
** was a duplicate, the passed argument string is released.
*/

w_string registerString(w_string string) {
 
  w_string result;

#ifdef DEBUG
//  if ((string_hashtable->occupancy%10000)==0) dumpStrings();
#endif

  woempa(1, "Registering %w at %p\n", string, string);
#ifdef NEWSTRINGREG
  ht_lock(string_hashtable);
  result = (w_string)ht_read_no_lock(string_hashtable, (w_word)string);
  if (result) {
    woempa(1, "Found `%w' at %p, refcount %d \n", result, result, result->refcount);
    if (result != string) {
      releaseString(string);
    }
  }
  else {
    woempa(1, "Didn't find it ... %w is new\n", string);
    result = string;
    ht_write_no_lock(string_hashtable, (w_word)result, (w_word)result);
  }
  ++result->refcount;
  woempa(1, "Refcount now %d\n", result->refcount);
  ht_unlock(string_hashtable);
#else
  result = (w_string)ht_register(string_hashtable, (w_word)string);
  if (result) {
    woempa(1, "Found it at %p ... %w\n", result,result);

    /*
    ** If the found string isn't the same one that was passed to us, 
    ** we can release it by making it garbage...
    */

    if (result != string) {
      releaseString(string);
    }
  }
  else {

    woempa(1,"Adding... %w\n", string);
    result = string;
  }
#endif

  return result;
  
}

/*
** Deregister a string. When we see that the registered count drops to 0, 
** we release the w_String structure.
**
** [CG 20000413]: modified to use ht_deregister
*/

void deregisterString(w_string string) {
#ifdef NEWSTRINGREG
  woempa(1, "Deregistering %w\n", string);
  ht_lock(string_hashtable);
  if ((--string->refcount) == 0) {
    ht_erase_no_lock(string_hashtable, (w_word)string);
    woempa(1,"releasing %w (%p)\n", string, string);
    releaseString(string);
  }
  ht_unlock(string_hashtable);
#else
  w_string garbage = NULL;

  woempa(1, "Deregistering %w\n", string);
  garbage = (w_string)ht_deregister(string_hashtable, (w_word)string);
  if (garbage) {
    woempa(1,"releasing %w (%p)\n",garbage,garbage);
    releaseString(garbage);
  }
#endif
}

/*
** The compare function for the string hashtable.
*/

w_boolean ht_stringCompare(w_word string1_word, w_word string2_word) {

  w_string string1 = (w_string)string1_word;
  w_string string2 = (w_string)string2_word;
  w_boolean result = WONKA_FALSE;

  /*
  ** Do a quick check first, when the two memory addresses match, the strings match.
  ** Otherwise, try some other checks, starting with the cheapest comparisons first and
  ** ending with a memory compare over the unicode strings.
  */

  if (string1 == string2) {
    result = WONKA_TRUE;
  }
  else {
    if (string1->length == string2->length) {
      if (string1->wonka_hash == string2->wonka_hash) {
        if (memcmp(string1->chars, string2->chars, sizeof(w_char) * string1->length) == 0) {
          result = WONKA_TRUE;
        }
      }
    }
  }

//  woempa(9, "Compared (%p) %w and (%p) %w: they are %sequal.\n", string1, string1, string2, string2, result ? "" : "NOT ");
  
  return result;

}

char * formatString(char * buffer, int * remain, void * s, w_int width, w_int prec, w_flags flags) {

  w_string string = s;
  w_char * chars;
  w_char * end;
  w_int padchars;
  w_int length;

  if (s) {
    length = string->length;
    padchars = 0;
    if (width) {
      if (length > width) {
        length = width;
      }
      else {
        padchars = width - length;
      }
    }

    if (flags & FMT_ALIGN_LEFT) {
      while (padchars-- && *remain > 0) {
        *buffer++ = ' ';
        *remain -= 1;
      }
    }
    
    chars = string->chars;
    end = string->chars + length;
    while (*remain > 0 && chars < end) {
      *buffer++ = *chars++;
      *remain -= 1;;
    }

    if (!(flags & FMT_ALIGN_LEFT)) {
      while (padchars-- && *remain > 0) {
        *buffer++ = ' ';
        *remain -= 1;
      }
    }

  }
  else {
    *buffer++ = '<';
    *buffer++ = 'n';
    *buffer++ = 'u';
    *buffer++ = 'l';
    *buffer++ = 'l';
    *buffer++ = '>';
    *remain -= 6;
  }
              
  return buffer;
                
}
                
void startStrings() {

  x_formatter('w', formatString);

  /*
  ** Allocate the hashtable in which the strings will be stored.
  */

  string_hashtable = ht_create(W_Realm_system, (char *)"hashtable:strings", STRING_HASHTABLE_SIZE, ht_stringHash, ht_stringCompare, 0, 0);
  woempa(1, "created string_hashtable at %p\n", string_hashtable);

  /*
  ** Create the mutex used to prevent race problems
  */
  string_mutex = (x_mutex)allocMem_system(sizeof(x_Mutex));
  x_mutex_create(string_mutex);

  /*
  ** Setup the atype2string array.
  */

  atype2string[4] = cstring2String(W_Thread_system,"[Z",2);
  atype2string[5] = cstring2String(W_Thread_system,"[C",2);
  atype2string[6] = cstring2String(W_Thread_system,"[F",2);
  atype2string[7] = cstring2String(W_Thread_system,"[D",2);
  atype2string[8] = cstring2String(W_Thread_system,"[B",2);
  atype2string[9] = cstring2String(W_Thread_system,"[S",2);
  atype2string[10] = cstring2String(W_Thread_system,"[I",2);
  atype2string[11] = cstring2String(W_Thread_system,"[J",2);

  /*
  ** Set up the special strings
  */

  string_NULL = cstring2String(W_Thread_system,"<NULL>",6);
  string_EMPTY = cstring2String(W_Thread_system,"<EMPTY>",7);
  string_BOGUS = cstring2String(W_Thread_system,"<BOGUS>",7);

  woempa(1, "Initialised string tables and static strings.\n");

}

w_int cstringAdd(w_char *buffer, const char *cstring) {

  w_byte *character = (w_byte *)cstring;
  w_int i = 0;
  
  while (*character) {
    *buffer++ = (w_char)(*character++);
    i += 1;
  }

  return i;

}

w_int unicodeAddNumber(w_char * destination, w_char * source, w_int number) {

  w_int i;
  
  for (i = 0; i < number; i++) {
    *destination++ = *source++;
  }

  return number;
  
}

w_int stringAdd(w_char *destination, w_string source) {

  memcpy(destination, source->chars, source->length * sizeof(w_char));

  return source->length;

}

w_string dots2slashes(w_thread thread, w_string string) {

  w_char *src = string->chars;
  w_string result;
  w_char *dst;
  w_size i;

  if (string->length) {
    result = allocString(string->length);
    dst = result->chars;
    for (i = 0; i < string->length; i++) {
      if (*src == '.') {
        *dst++ = '/';
      }
      else if (*src == '/') {
        *dst++ = 127;
      }
      else {
        *dst++ = *src;
      }
      src += 1;
    }
    hashString(result);
  }
  else {
    result = cstring2String(W_Thread_system,"",0);
  }

  woempa(1, "Created string %w.\n", result);
  
  result = registerString(result);  

  return result;

}

w_string slashes2dots(w_thread thread, w_string string) {

  w_char *src = string->chars;
  w_string result = string;
  w_char *dst;
  w_size i;

  if (string->length) {
    result = allocString(string->length);
    dst = result->chars;
    for (i = 0; i < string->length; i++) {
      if (*src == '/') {
        *dst++ = '.';
      }
      else if (*src == 127) {
        woempa(9,"Wow!  Found a DEL char, converted it to a slosh.\n");
        *dst++ = '/';
      }
      else {
        *dst++ = *src;
      }
      src += 1;
    }
    hashString(result);
  }
  else {
    result = cstring2String(W_Thread_system,"",0);
  }

  woempa(1, "Created string %w.\n", result);
  
  result = registerString(result);  

  return result;

}

/*
w_clazz checkSubstring(w_string s, w_int offset, w_int count) {

  woempa(1,"offset %d count %d s.length() %d\n",offset,count,s->length);

  if (offset<0 || count<0 || offset+count>(w_int)s->length) {
    return clazzStringIndexOutOfBoundsException;
  }
  else {
    return NULL;
  }
  
}
*/

w_instance string2String(w_string s) {
  w_instance instance = allocInstance(W_Thread_system, clazzString);

  instance[F_String_wotsit] = (w_word)registerString(s);

  return instance;
}

w_instance getStringInstance(w_thread thread, w_string s) {
  w_instance instance;

/*
  return string2String(s);
*/

  x_mutex_lock(string_mutex, x_eternal);
  instance = s->instance;
  if (instance == NULL || isSet(instance2object(instance)->flags, O_CONDEMNED|O_TARGET|O_GARBAGE)) {
    instance = allocInstance(thread, clazzString);
    if (instance) {
      s->instance = instance;
      woempa(1, "allocated instance %p of String for '%w'\n", instance, s);
      instance[F_String_wotsit] = (w_word)registerString(s);
    }
  }
  x_mutex_unlock(string_mutex);

  return instance;
}

#ifdef DEBUG

/*
** Routines to print out statistics about the strings and string table.
*/

void qsort(void *base, size_t nmemb, size_t size, int (*compar)(const void *, const void *));

/*
** The total number of Unicode characters we have allocated.
*/

static w_int total_length = 0;

int string_sort(const void *v_string1, const void *v_string2) {

  w_string string1 = *(w_string *)v_string1;
  w_string string2 = *(w_string *)v_string2;
  w_int result;

  /*
  ** We compare on the basis of the memory pointer of the two strings so that strings
  ** that are allocated together end up together in the list.
  */

  result = (w_int)string1 - (w_int)string2;
  
  return result;
  
}

void string_iterate(w_word string_word, w_word hash, void * v_array, void * v_indx) {

  w_string string = (w_string)string_word;
  w_string *array = (w_string *)v_array;
  w_int *indx = (w_int *)v_indx;
  
  total_length += string->length;
  array[*indx] = string;
  *indx += 1;
  
}

void dumpSingleString(w_int i, w_string string) {

  w_int desc_registered = 0;
  
  woempa(9, "|%5d | %4d |  %4d |   %3d  | 0x%08x |'%w'\n", i,
    ht_read(string_hashtable,(w_word)string), desc_registered, string->length, string->wonka_hash, 
    string);

}


void dumpStrings(void) {

  w_string *array;
  w_int indx = 0;
  w_int i;
  
  array = allocMem_system(string_hashtable->occupancy * sizeof(w_string *));

  ht_iterate(string_hashtable, string_iterate, array, &indx);  

  qsort(array, (unsigned)(indx), sizeof(w_string), string_sort);
  
  woempa(9, "+------+------+-------+--------+------------+-------------------------+\n");
  woempa(9, "|  #   |  Reg |  Desc | Length |    Hash    |    First registration   | String\n");
  woempa(9, "+------+------+-------+--------+------------+-------------------------+-------------------------------------------------------------\n");
  for (i = 0; i < indx; i++) {
    dumpSingleString(i, array[i]);
  }
  woempa(9, "+------+------+-------+--------+------------+-------------------------+-------------------------------------------------------------\n");
  woempa(9, "|       Total number of bytes for Unicode characters: %6d\n", total_length * sizeof(w_char));
  woempa(9, "| Number of bytes for w_String (%d bytes) structures: %6d\n", sizeof(w_String), indx * sizeof(w_String));
  woempa(9, "+------+------+-------+--------+------------+-------------------------+-------------------------------------------------------------\n");
  

  releaseMem(array);

}
#endif

