/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: ts-mem.c,v 1.8 2002/02/20 08:52:17 dfrederickx Exp $
*/

#include "exception.h"
#include "heap.h"
#include "ts-mem.h"
#include "realms.h"
#include "threads.h"
#include "oswald.h"
#include "ps-mem.h"

/*
** Check that a w_realm pointer really does point to a w_Realm
*/
static inline w_realm checkRealmPointer(w_realm r) {
#ifdef RUNTIME_CHECKS
  if (*((w_word*)r->label) != *((w_word*)"realm")) {
    wabort(ABORT_WONKA,"Alleged w_realm pointer %p has incorrect label\n",r);
  }
#endif
  return realm_table[r->index] == r ? r : NULL;
}

/*
** Adjust the memory usage of a realm upwards or (if size < 0) downwards.
*/
inline static void adjustRealmUsage(w_realm realm, w_int size) {

  x_mutex_lock(realm->mutex, x_eternal);
  realm->memory_usage += size;
  x_mutex_unlock(realm->mutex);

}

/*
** Derive the realm to which a chunk belongs from its memory tag.
*/
static inline w_realm get_chunk_realm(w_chunk chunk) {
  return realm_table[x_mem_tag_get(chunk) & REALM_INDEX_MASK];
}

/*
** Set the tag of a memory chunk to an index into the realm table
*/
static inline void set_chunk_realm(w_chunk chunk, w_realm realm) {
  x_mem_tag_set(chunk, (x_mem_tag_get(chunk) & (unsigned)~REALM_INDEX_MASK) | realm->index);
}

/*
** Allocate cleared (zeroed) memory in a given realm.
*/
void * _allocClearedMem_realm(w_realm realm, w_size rsize) {

  w_chunk chunk;

  checkRealmPointer(realm);

  chunk = x_mem_calloc(sizeof(w_Chunk) + rsize);
  if (chunk == NULL) {
    return NULL;
  }

  set_chunk_realm(chunk, realm);
  adjustRealmUsage(realm, + (w_int) x_mem_size(chunk));

  return chunk->data;

}

void * _allocMem_realm(w_realm realm, w_size rsize) {

  w_chunk chunk;

  checkRealmPointer(realm);

  chunk = x_mem_alloc(sizeof(w_Chunk) + rsize);
  if (chunk == NULL) {
    throwOutOfMemoryError(currentWonkaThread);

    return NULL;
  }

  set_chunk_realm(chunk, realm);
  adjustRealmUsage(realm, + (w_int) x_mem_size(chunk));

  return chunk->data;

}

void * _reallocMem(void * block, w_size newsize) {
  w_chunk oldchunk;
  w_size  oldsize;
  w_chunk newchunk;
  w_realm realm;

  oldchunk = block2chunk(block);
  realm = get_chunk_realm(oldchunk);
  oldsize = x_mem_size(oldchunk);
  newchunk = x_mem_realloc(oldchunk, sizeof(w_Chunk) + newsize);
  
  if (newchunk) {
    adjustRealmUsage(realm, (w_int)x_mem_size(newchunk) - (w_int)oldsize);
    return newchunk->data;
  }
  
  throwOutOfMemoryError(currentWonkaThread);

  return NULL;

}

void _releaseMem(void * block) {

  w_chunk chunk = block2chunk(block);

  adjustRealmUsage(get_chunk_realm(chunk), - (w_int) x_mem_size(chunk));

  x_mem_free(chunk);

}

void _discardMem(void * block) {

  w_chunk chunk = block2chunk(block);

  adjustRealmUsage(get_chunk_realm(chunk), - (w_int) x_mem_size(chunk));

  x_mem_discard(chunk);

}

void resetMemOwnerRealm(void * block, w_realm new) {

  w_chunk chunk = block2chunk(block);
  w_realm old = get_chunk_realm(chunk);
  w_int size;

  /*
  ** Only perform calculations when old realm is not the same as the new realm.
  */

  if (old != new) {
    size = (w_int) x_mem_size(chunk);
    adjustRealmUsage(new, + size);
    adjustRealmUsage(old, - size);
    set_chunk_realm(chunk, new);
  }
  
}

void resetMemOwner(void * block, w_thread new) {

  if (new == NULL) {
    wabort(ABORT_WONKA, "Thread is null\n");
  }

  if (new->group == NULL) {
    wabort(ABORT_WONKA, "Group is null.\n");
  }

  resetMemOwnerRealm(block, thread2realm(new));

}

#ifdef DEBUG // --------------------------------------------------------------

#include <stdlib.h>
#include <string.h>

#define CHUNK_MADGIC                0xbbbbbbbb
#define FRONT_FENCE                 0xcccccccc
#define BACK_BYTE                   0xdd
#define FREED_BYTE                  0xee
#define FREED_DATA                  0xaa

const int check_limit = 10000;        // Every check_limit releases/reallocs, a complete memory check is done

/*
** Our memory walker argument structure.
*/

typedef struct w_Wa * w_wa;

typedef struct w_Wa {
  w_int count;
  w_int bytes;
  w_int errors;
  w_int all_count;
  w_int all_bytes;
} w_Wa;

/*
** The memory walker callback to check for faulty chunks.
*/

static w_chunk previous;

static void checkWalk(void * mem, void * arg) {

  w_wa wa = arg;
  w_chunk chunk = mem;
  w_int tag = x_mem_tag_get(chunk);
  w_int bad;

  wa->count += 1;

  bad = 0;
  if ((tag & REALM_INDEX_MASK)) {
    if (chunk->front != FRONT_FENCE) {
      woempa(9, "chunk %p front fence at %p holds 0x%08x instead of 8x%08x.\n", chunk, &chunk->front, chunk->front, FRONT_FENCE);
      bad += 1;
    }
    if (chunk->madgic != CHUNK_MADGIC) {
      woempa(9, "chunk %p madgic at %p holds 0x%08x instead of 8x%08x.\n", chunk, &chunk->madgic, chunk->madgic, CHUNK_MADGIC);
      bad += 10;
    }    

    if ((x_size)chunk->back > chunk->size) {
      woempa(9, "chunk back index %d > chunk size %d.\n", chunk->back, chunk->size);
      bad += 100;
    }
    
    /*
    ** Only check the back byte if the previous tests turned out OK.
    ** Also, don't count the size of the chunk when it has an
    ** error, it's value is probably corrupt anyhow.
    */

    if (bad == 0 && ! (chunk->data[chunk->back] == BACK_BYTE || chunk->data[chunk->back] == FREED_BYTE)) {
      woempa(9, "bad back = 0x%08x\n", chunk->data[chunk->back]);
      bad += 1000;
    }
  }

  if (bad) {
    wa->errors += 1;
    woempa(9, "CHUNK ERROR: number %d, bad chunk 0x%08x, status = %d\n", wa->count, chunk, bad);
    woempa(9, "  alloced in %s:%d size = %d, tag = 0x%x\n", chunk->file, chunk->line, chunk->size, tag);
    if (previous) {
      woempa(9, "  previous chunk 0x%08x, alloced in %s:%d, size = %d\n", previous, previous->file, previous->line, previous->size);
    }
  }
  else {
    wa->bytes += chunk->size + sizeof(w_Chunk);
  }
  previous = chunk;
}

void _heapCheck(const char * function, const int line) {

  static w_int walks = 0;
  w_Wa Wa;
  w_wa wa = &Wa;

  wa->count = 0;
  wa->bytes = 0;
  wa->errors = 0;
  wa->all_count = 0;
  wa->all_bytes = 0;
  previous = NULL;
  x_mem_walk(x_eternal, checkWalk, wa);
  walks += 1;
  woempa(9, "%s %d : %d anon %d tagged, %d MB anon %d Mb tagged, %d errors, %d scans.\n", function, line, wa->all_count, wa->count, wa->all_bytes / (1024 * 1024), wa->bytes / (1024 * 1024), wa->errors, walks);
  
  if (wa->errors) {
    exit(1);
  }
    
}

/*
** Perform the usual checking on a chunk. When we hit the limit for full checking,
** we perform a complete memory scan for faulty chunks.
*/

static w_chunk checkChunk(void * block, w_int for_realloc, const char * file, const int line) {

  static w_int errors = 0;
  static w_int checks = 0;
  static w_int walks = 0;
  w_chunk chunk;
  w_Wa Wa;
  w_wa wa = &Wa;

  checks += 1;
  
  if (checks % check_limit == 0) {
    wa->count = 0;
    wa->bytes = 0;
    wa->errors = 0;
    wa->all_bytes = 0;
    wa->all_count = 0;
    previous = NULL;
    x_mem_walk(x_eternal, checkWalk, wa);
    walks += 1;
    errors += wa->errors;
    woempa(9, "Walked %d anon %d tagged chunks, %d MB anon %d Mb tagged, %d chunks were in error. Made %d scans.\n", wa->all_count, wa->count, wa->all_bytes / (1024 * 1024), wa->bytes / (1024 * 1024), wa->errors, walks);
    if (wa->errors) {
      wabort(ABORT_WONKA, "Found bad chunk. Stopping...\n");
    }
  }

  if (block == NULL) {
    wabort(ABORT_WONKA, "Freeing NULL block at %s:%d.\n", file, line);
  }

  /*
  ** It looks like a valid block, just proceed...
  */

  chunk = block2chunk(block);

  if (chunk->data[chunk->back] == FREED_BYTE) {
    wabort(ABORT_WONKA, "Freeing chunk %p twice (allocated %s %d, released %s %d).\n", chunk, chunk->file, chunk->line, file, line);
    return NULL;
  }

  if (chunk->data[chunk->back] != BACK_BYTE) {
    wabort(ABORT_WONKA, "Back fence overwritten of chunk %p (allocated %s %d, released %s %d).\n", chunk, chunk->file, chunk->line, file, line);
    return NULL;
  }

  if (chunk->front != FRONT_FENCE) {
    wabort(ABORT_WONKA, "Front fence of %p overwritten (allocated %s %d, released %s %d).\n", chunk, chunk->file, chunk->line, file, line);
    return NULL;
  }

  if (chunk->madgic != CHUNK_MADGIC) {
    wabort(ABORT_WONKA, "Wrong madgic of %p (allocated %s %d, released %s %d).\n", chunk, chunk->file, chunk->line, file, line);
    return NULL;
  }

  if (get_chunk_realm(chunk) == NULL) {
    wabort(ABORT_WONKA, "Chunk has no realm (allocated %s %d, released %s %d).\n", chunk->file, chunk->line, file, line);
    return NULL;
  }

  /*
  ** If this check isn't done for realloc, we overwrite the data area to quickly evoke
  ** errors when freed memory would b re-used. Also the backfence will get a new color.
  */

  if (! for_realloc) {
    memset(chunk->data, FREED_DATA, chunk->size);
    chunk->data[chunk->back] = FREED_BYTE;
  }

  return chunk;
  
}

/*
** Attach debugging information to a chunk.
*/

static void prepareChunk(w_chunk chunk, w_realm realm, w_size size, const char * file, const int line) {

  woempa(1,"Preparing chunk %p, realm %p, size %d, allocated at %s:%d\n", chunk, realm, size, file, line);
  chunk->file = file;
  chunk->line = line;
  chunk->size = size;
  chunk->madgic = CHUNK_MADGIC;
  chunk->front = FRONT_FENCE;
  chunk->back = size;
  woempa(1,"Chunk %p data starts at %p, back fence at %p\n", chunk, chunk->data, &chunk->data[size]);
  chunk->data[size] = BACK_BYTE;
  set_chunk_realm(chunk, realm);
}

void * _d_allocClearedMem_realm(w_realm realm, w_size rsize, const char * file, const int line) {

  w_chunk chunk;

  checkRealmPointer(realm);

  woempa(1,"Requested %d bytes, allocating %d bytes\n", rsize, sizeof(w_Chunk) + rsize);
  chunk = x_mem_calloc(sizeof(w_Chunk) + rsize);
  if (chunk == NULL) {
    return NULL;
  }

  prepareChunk(chunk, realm, rsize, file, line);  
  adjustRealmUsage(realm, (w_int)x_mem_size(chunk));

  return chunk->data;

}

void * _d_allocMem_realm(w_realm realm, w_size rsize, const char * file, const int line) {

  w_chunk chunk;

  checkRealmPointer(realm);

  woempa(1,"Requested %d bytes, allocating %d bytes\n", rsize, sizeof(w_Chunk) + rsize);
  chunk = x_mem_alloc(sizeof(w_Chunk) + rsize);
  if (chunk == NULL) {
    throwOutOfMemoryError(currentWonkaThread);

    return NULL;
  }

  prepareChunk(chunk, realm, rsize, file, line);  
  adjustRealmUsage(realm, (w_int)x_mem_size(chunk));
  
  return chunk->data;

}

void * _d_reallocMem(void * block, w_size newsize, const char * file, const int line) {
  w_realm realm;
  w_chunk oldchunk;
  w_chunk newchunk;
  w_size  oldsize;

  woempa(1,"Reallocating block %p, new size = %d\n", block, newsize);
  oldchunk = checkChunk(block, 1, file, line);
  if (!oldchunk) {

    return NULL;

  }


  /*
  ** Before we realloc, we overwrite the realm index with 0
  ** since otherwise we would upset the chunk checker; we could
  ** be checking the chunk in the walker before we have prepared
  ** it properly. Later, we put the realm index back in place.
  */
  
  realm = get_chunk_realm(oldchunk);
  x_mem_tag_set(oldchunk, (x_mem_tag_get(oldchunk) & (unsigned)~REALM_INDEX_MASK));

  oldsize = x_mem_size(oldchunk);
  woempa(1,"Requested %d bytes, allocating %d bytes\n", newsize, sizeof(w_Chunk) + newsize);
  newchunk = x_mem_realloc(oldchunk, sizeof(w_Chunk) + newsize);
  
  if (!newchunk) {
    throwOutOfMemoryError(currentWonkaThread);

    return NULL;

  }

  adjustRealmUsage(realm, (w_int)x_mem_size(newchunk) - (w_int)oldsize);
  prepareChunk(newchunk, realm, newsize, file, line);  

  return newchunk->data;
}

void _d_releaseMem(void * block, const char * file, const int line) {

  w_chunk chunk;

  woempa(1,"Releasing block %p\n", block);
  chunk = checkChunk(block, 0, file, line);
  if (!chunk) {

    return;

  }

  adjustRealmUsage(get_chunk_realm(chunk), - (w_int) x_mem_size(chunk));

  x_mem_free(chunk);

}

void _d_discardMem(void * block, const char * file, const int line) {

  w_chunk chunk;

  chunk = checkChunk(block, 0, file, line);

  adjustRealmUsage(get_chunk_realm(chunk), - (w_int) x_mem_size(chunk));

  x_mem_discard(chunk);

}


/*
** A chunk reporting structure.
*/

typedef struct w_Cr * w_cr;

typedef struct w_Cr {
  w_cr next;
  w_word hash;
  const char * file;
  w_int line;
  w_int hits;
  w_int total;
} w_Cr;

/*
** Our own small custom hashtable, the number of buckets it has and the total count
** of report structures that are in the hashtable.
*/

static w_cr * ht;
static const w_int num_buckets = 200;
static w_int num_cr;

/*
** Functions to pass to the qsort routine.
*/

static int cr_sort_hits(const void * v_cr1, const void * v_cr2) {

  w_cr cr1 = *(w_cr *)v_cr1;
  w_cr cr2 = *(w_cr *)v_cr2;
  int result = 0;

  if (cr1->hits > cr2->hits) {
    result = -1;
  }
  else if (cr1->hits < cr2->hits) {
    result = 1;
  }

  return result;
  
}

static int cr_sort_file(const void * v_cr1, const void * v_cr2) {

  w_cr cr1 = *(w_cr *)v_cr1;
  w_cr cr2 = *(w_cr *)v_cr2;

  return strcmp(cr1->file, cr2->file);
  
}

static w_int cr_compare(w_cr cr1, w_cr cr2) {

  if ((cr1->file == cr2->file) && (cr1->line == cr2->line)) {
    return 1;
  }
  else {
    return 0;
  }

}

/*
** See if a cr structure is allready in our hashtable. Returns the found one or NULL
** if not found.
*/

w_cr find(w_cr new) {

  w_size i = new->hash % num_buckets;
  w_cr ex;
  
  for (ex = ht[i]; ex; ex = ex->next) {
    if (cr_compare(ex, new)) {
      return ex;
    }
  }

  return NULL;

}

/*
** Add a new cr to the hashtable.
*/

void add(w_cr new) {

  w_size i = new->hash % num_buckets;
  w_cr ex;

  for (ex = ht[i]; ex; ex = ex->next) {
    if (ex->next == NULL) {
      ex->next = new;
      break;
    }
  }

  /*
  ** There was no leader yet in the hashtable, make this one the leader.
  */
  
  if (ex == NULL) {
    ht[i] = new;
  }
  
  num_cr += 1;

}

/*
** Our memory walker callback to walk over the chunks and add reporting structures to our
** hastable.
*/

static void cr_walk(void * block, void * arg) {

  w_wa wa = arg;
  w_cr cr;
  w_cr ex;
  w_chunk chunk;
  
  chunk = block;
  wa->count += 1;
  wa->bytes += chunk->size;

  cr = x_mem_alloc(sizeof(w_Cr));
  cr->file = chunk->file;
  cr->line = chunk->line;
  cr->next = NULL;
  cr->hits = 1;
  cr->total = chunk->size;
  cr->hash = (w_word)cr->file + cr->line;

  ex = find(cr);
  if (ex) {
    ex->hits += 1;
    ex->total += chunk->size;
    x_mem_free(cr);
  }
  else {
    add(cr);
  }
  
}

static const char * fno(const char * file) {

  char * location;
  
  location = rindex(file, '/');
  if (location == NULL) {
    location = (char *)file;
  }
  else {
    location += 1;
  }
  
  return location;
  
}

void reportMemStat(w_int type) {

  w_cr cr;
  w_cr *array;
  w_int indx;
  w_int i;
  w_int overhead_mb;
  w_int requested_mb;
  w_int percent;
  w_int total_percent = 0;
  const char * current_file = "";
  w_int current_total = 0;
  w_int current_percent = 0;
  w_int current_hits = 0;
  w_int was_printed = 0;
  w_wa wa;

  num_cr = 0;
  ht = x_mem_calloc(num_buckets * sizeof(w_cr));

  /*
  ** Walk over all blocks.
  */
  
  wa = x_mem_calloc(sizeof(w_Wa));
  x_mem_walk(x_eternal, cr_walk, wa);

  /*
  ** Iterate over the hashtable and put all reporting structure references
  ** in an array.
  */

  array = x_mem_alloc(num_cr * sizeof(w_cr));
  indx = 0;
  for (i = 0; i < num_buckets; i++) {
    if (ht[i]) {
      for (cr = ht[i]; cr; cr = cr->next) {
        array[indx++] = cr;
      }
    }
  }

  /*
  ** OK, now sort the array that contains our report structures
  ** according to the number of hits.
  */

  if (type == SORT_BY_HITS) {
    qsort(array, (unsigned)(indx), sizeof(w_cr), cr_sort_hits);
  }
  else {
    qsort(array, (unsigned)(indx), sizeof(w_cr), cr_sort_file);
  }

  /*
  ** Do the reporting...
  */

  x_mem_lock(x_eternal);
  overhead_mb = (wa->count * sizeof(w_Chunk)) / (1024 * 1024);
  requested_mb = wa->bytes / (1024 * 1024);

  if (indx) {
    woempa(9, "+---------------------------+--------+---------------+--------------+-------------+\n");
    woempa(9, "|                file: line |  hits  | bytes req.  %% | bytes overh. | bytes total |\n");
    woempa(9, "+---------------------------+--------+---------------+--------------+-------------+\n");
    current_file = array[0]->file;
  }
  for (i = 0; i < indx; i++) {
    cr = array[i];
    percent = ((w_int)(cr->total * 100)) / wa->bytes;
    total_percent += percent;
    if (type == SORT_BY_FILE) {
      if (current_file != (char *)cr->file) {
        if (! was_printed) {
          woempa(9, "|%20s:%5d | %6d | %7d  %3d%% |   %8d   | %9d   |\n", 
            fno(cr->file), cr->line, cr->hits, cr->total, percent, (w_int)(cr->hits * sizeof(w_Chunk)), cr->total + (w_int)(cr->hits * sizeof(w_Chunk)));
          was_printed = 1;
        }
        woempa(9, "|%20s: T => | %6d | %7d  %3d%% |   %8d   | %9d   |\n", 
          fno(current_file), current_hits, current_total, current_percent, current_hits * sizeof(w_Chunk), (current_hits * sizeof(w_Chunk)) + current_total);
        current_file = cr->file;
        current_total = 0;
        current_percent = 0;
        current_hits = 0;
        was_printed = 0;
      }
      else {
        current_total += cr->total;
        current_percent += percent;
        current_hits += cr->hits;
      }
    }
    woempa(9, "|%20s:%5d | %6d | %7d  %3d%% |   %8d   | %9d   |\n", 
      fno(cr->file), cr->line, cr->hits, cr->total, percent, (w_int)(cr->hits * sizeof(w_Chunk)), cr->total + (w_int)(cr->hits * sizeof(w_Chunk)));
    was_printed = 1;
  }
  if (indx) {
    woempa(9, "+---------------------------+--------+---------------+--------------+-------------+\n");
    woempa(9, "Total requested bytes is %d (%d Mb) %d%% in %d chunks, total overhead bytes is %d (%d Mb)\n", 
      wa->bytes, requested_mb, total_percent, wa->count, (w_int)(wa->count * sizeof(w_Chunk)), overhead_mb);
  }

  /*
  ** Now do some courtesy cleanup. For each entry in the array, 
  ** erase it from the hashtable and release the memory of the structure.
  */

  x_mem_unlock();

  for (i = 0; i < indx; i++) {
    x_mem_free(array[i]);
  }

  x_mem_free(array);
  x_mem_free(ht);
  x_mem_free(wa);

}

#endif // --- DEBUG ----------------------------------------------------------

